// Bilinear forms for linear DG FEM in 2D

#include "integration.hh"
#include "space/tmatrix.hh"
#include "toolbox/boundary.hh"
#include "basics/exceptions.hh"
#include "linDG2D.hh"

#include "bilinearForm.hh"

// debugging
#include "basics/debug.hh"

#define DGElasticFluxBf2dAppl_D 0

using concepts::Real;

// ******************************************************** DGElasticVolBf2d **

void DGElasticVolBf2d::operator() (const concepts::Element<Real>& elmV,
                                   const concepts::Element<Real>& elmU,
                                   concepts::ElementMatrix<Real>& em) {
  const uint mV = elmV.T().m();
  const uint mU = elmU.T().m();
  em.resize(mV, mU);
  em.zeros(); 
  
  if (&elmV == &elmU) {
    const linDG2D::TriangleP1* eV =
      dynamic_cast<const linDG2D::TriangleP1*>(&elmV);
    conceptsAssert(eV, concepts::Assertion());
    conceptsAssert(3 == mV, concepts::Assertion());

    Real area = eV->cell().area();
    // fetch gradients
    concepts::Array<concepts::Real2d> gradient(mV);
    for(uint i=0; i<mV; ++i) {
      gradient[i] = eV->shapeFctGradient(i);
    }
    // fill element matrix
    for(uint iV=0; iV<mV; ++iV) {
      for(uint iU=0; iU<mU; ++iU) {
        em(iV,iU) = (gradient[iV]*(submapping_*gradient[iU]))*area;
#if DGElasticVolBf2dAppl_D
        std::cout << "grad_" << iU << " = " << gradient[iU] << std::endl;
        std::cout << "grad_" << iV << " = " << gradient[iV] << std::endl;
        std::cout << "E_ij   = " << submapping_ << std::endl;
        std::cout << "E_ij * grad_" << iU << " = " << submapping_*gradient[iU] 
                  << std::endl;
        std::cout << "grad_" << iV << "^T * E_ij * grad_" << iU << " = " 
                  << gradient[iV]*(submapping_*gradient[iU]) << std::endl;
        std::cout << std::endl;
#endif
      }
    }
  }
}

// ******************************************************* DGElasticFluxBf2d **

void DGElasticFluxBf2d::operator()(const concepts::Element<Real>& elmV,
                                   const concepts::Element<Real>& elmU,
                                   concepts::ElementMatrix<Real>& em,
                                   const concepts::ElementPair<Real>& ep) {
  const uint mV = elmV.T().m();
  const uint mU = elmU.T().m();
  em.resize(mV, mU);
  em.zeros();

  const linDG2D::TriangleP1* eV =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmV);
  conceptsAssert(eV, concepts::Assertion());
  const linDG2D::TriangleP1* eU =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmU);
  conceptsAssert(eU, concepts::Assertion());
  const linDG2D::ElementPair* epc =
    dynamic_cast<const linDG2D::ElementPair*>(&ep);
  conceptsAssert(epc, concepts::Assertion());

  const linDG2D::EdgeInfo& ei = epc->edgeInfo();
  Real factor;

  if (ei.cell(1) == 0) {
    // boundary integral
    if (bc_) {
      if ((*bc_)(ei.edge().attrib()).type() == 
          concepts::Boundary::DIRICHLET) {
        DEBUGL(DGElasticFluxBf2dAppl_D,
               ei.edge() << "has dirichlet attribute");
        factor = 1;
      } else {
        DEBUGL(DGElasticFluxBf2dAppl_D,
               ei.edge() << "has non-dirichlet attribute");
        factor = 0;
      }
    } else {
      DEBUGL(DGElasticFluxBf2dAppl_D,
             ei.edge() << "is boundary assumed dirichlet");
      // assume dirichlet boundary all over
      factor = 1;
    }
  } else {
    // inner integral, <u> = 0.5*(u_left + u_right)
    DEBUGL(DGElasticFluxBf2dAppl_D,
           ei.edge() << "is inner edge");
    factor = 0.5;
  }

  if (factor != 0) {
    factor *= ei.length();
    concepts::Real2d normal = ei.normal();

    // <grad u> E_{ij} [v]
    if (&(eV->cell()) == ei.cell(1)) {
      // for jump of testfunction [v]: testfunction (on eV) is on element 1
      factor *= -1;
    }

    for(uint iU=0; iU<mU; ++iU) {
      Real f_n_E_grad_u = factor * (normal *
                                    (submapping_*eU->shapeFctGradient(iU)));
      for(uint iV=0; iV<mV; ++iV) {
        em(iV,iU) = f_n_E_grad_u * epc->shapeFct(0, iV, 0.5);
      }
    }
  }
}

// ******************************************************* DGElasticStabBf2d **

void DGElasticStabBf2d::operator()(const concepts::Element<Real>& elmV,
                                   const concepts::Element<Real>& elmU,
                                   concepts::ElementMatrix<Real>& em,
                                   const concepts::ElementPair<Real>& ep) {
  const uint mV = elmV.T().m();
  const uint mU = elmU.T().m();
  em.resize(mV, mU);
  em.zeros();

  const linDG2D::TriangleP1* eV =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmV);
  conceptsAssert(eV, concepts::Assertion());
  const linDG2D::TriangleP1* eU =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmU);
  conceptsAssert(eU, concepts::Assertion());
  const linDG2D::ElementPair* epc =
    dynamic_cast<const linDG2D::ElementPair*>(&ep);
  conceptsAssert(epc, concepts::Assertion());
  const linDG2D::EdgeInfo& ei = epc->edgeInfo();
  Real factor;

  if (ei.cell(1) == 0) {
    // boundary integral
    if (bc_) {
      if ((*bc_)(ei.edge().attrib()).type() == 
          concepts::Boundary::DIRICHLET) {
        factor = 1;
      } else {
        factor = 0;
      }
    } else {
      // assume dirichlet boundary all over
      factor = 1;
    }
  } else {
    // inner integral
    factor = 1;
  }

  if (factor != 0) {
    factor *= omega_ * ei.length() / ei.meanHeight();
    if (&(eV->cell()) != &(eU->cell())) {
      // for jumps [u][v]
      // testfunction (on eV) is on cell 1 and ansatzfct. (on eU) on cell 0
      // or
      // testfunction (on eV) is on cell 0 and ansatzfct. (on eU) on cell 1
      factor *= -1;
    }

    concepts::Quadrature<4> quad(2);
    const Real* absc = quad.abscissas();
    Real xi[2];
    xi[0] = 0.5*(1.0+absc[0]);
    xi[1] = 0.5*(1.0+absc[1]);
    factor *= 0.5; // quadrature weight
    for(uint iV=0; iV<mV; ++iV) {
      for(uint iU=0; iU<mU; ++iU) {
        for(int i=0; i<2; ++i) {
          em(iV,iU) += factor * 
            epc->shapeFct(0, iV, xi[i]) * 
            epc->shapeFct(1, iU, xi[i]);
        }
      }
    }
  }
}

// ************************************************  DGElasticNormalStabBf2d **

void DGElasticNormalStabBf2d::operator()(const concepts::Element<Real>& elmV,
                                         const concepts::Element<Real>& elmU,
                                         concepts::ElementMatrix<Real>& em,
                                         const concepts::ElementPair<Real>& ep)
{
  const uint mV = elmV.T().m();
  const uint mU = elmU.T().m();
  em.resize(mV, mU);
  em.zeros();

  const linDG2D::TriangleP1* eV =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmV);
  conceptsAssert(eV, concepts::Assertion());
  const linDG2D::TriangleP1* eU =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmU);
  conceptsAssert(eU, concepts::Assertion());
  const linDG2D::ElementPair* epc =
    dynamic_cast<const linDG2D::ElementPair*>(&ep);
  conceptsAssert(epc, concepts::Assertion());
  const linDG2D::EdgeInfo& ei = epc->edgeInfo();
  Real factor;

  if (ei.cell(1) == 0) {
    // boundary integral
    if (bc_) {
      if ((*bc_)(ei.edge().attrib()).type() == 
          concepts::Boundary::DIRICHLET) {
        factor = 1;
      } else {
        factor = 0;
      }
    } else {
      // assume dirichlet boundary all over
      factor = 1;
    }
  } else {
    // inner integral
    factor = 1;
  }

  if (factor != 0) {
    factor *= omega_ * ei.length() / ei.meanHeight();
    if (&(eV->cell()) != &(eU->cell())) {
      // for jumps [u][v]
      // testfunction (on eV) is on cell 1 and ansatzfct. (on eU) on cell 0
      // or
      // testfunction (on eV) is on cell 0 and ansatzfct. (on eU) on cell 1
      factor *= -1;
    }

    concepts::Real2d normal = ei.normal();
    concepts::Quadrature<4> quad(2);
    const Real* absc = quad.abscissas();
    Real xi[2];
    xi[0] = 0.5*(1.0+absc[0]);
    xi[1] = 0.5*(1.0+absc[1]);
    factor *= 0.5; // quadrature weight
    for(uint iV=0; iV<mV; ++iV) {
      for(uint iU=0; iU<mU; ++iU) {
        for(int i=0; i<2; ++i) {
          em(iV,iU) += factor * normal[dV_] * normal[dU_] *
            epc->shapeFct(0, iV, xi[i]) * 
            epc->shapeFct(1, iU, xi[i]);
        }
      }
    }
  }
}

// ***************************************************** DGElasticDivVolBf2d **

void DGElasticDivVolBf2d::operator()(const concepts::Element<Real>& elmV,
                                     const concepts::Element<Real>& elmP,
                                     concepts::ElementMatrix<Real>& em,
                                     const concepts::ElementPair<Real>& ep) {
  const uint mV = elmV.T().m();
  const uint mP = elmP.T().m();
  em.resize(mV, mP);
  em.zeros();

  const linDG2D::TriangleP1* eV =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmV);
  conceptsAssert(eV, concepts::Assertion());
  const linDG2D::TriangleP0* eP =
    dynamic_cast<const linDG2D::TriangleP0*>(&elmP);
  conceptsAssert(eP, concepts::Assertion());
  if (&(eV->cell()) == &(eP->cell())) {
    conceptsAssert(3 == mV, concepts::Assertion());
    conceptsAssert(1 == mP, concepts::Assertion());
    const linDG2D::ElementPair* epc =
      dynamic_cast<const linDG2D::ElementPair*>(&ep);
    conceptsAssert(epc, concepts::Assertion());
    const linDG2D::EdgeInfo& ei = epc->edgeInfo();
    if (&(ei.edge()) == eV->support().edge(0)) {
      Real factor = - scale_ * eV->cell().area();
      for(uint iV=0; iV<mV; ++iV) {
        em(iV,0) = factor * eV->shapeFctGradient(iV)[dV_];
      }
    }
  }
}

// **************************************************** DGElasticDivFluxBf2d **

void DGElasticDivFluxBf2d::operator()(const concepts::Element<Real>& elmV,
                                      const concepts::Element<Real>& elmP,
                                      concepts::ElementMatrix<Real>& em,
                                      const concepts::ElementPair<Real>& ep) {
  const uint mV = elmV.T().m();
  const uint mP = elmP.T().m();
  em.resize(mV, mP);
  em.zeros();

  const linDG2D::TriangleP1* eV =
    dynamic_cast<const linDG2D::TriangleP1*>(&elmV);
  conceptsAssert(eV, concepts::Assertion());
  conceptsAssert(mV == 3, concepts::Assertion());
  const linDG2D::TriangleP0* eP =
    dynamic_cast<const linDG2D::TriangleP0*>(&elmP);
  conceptsAssert(eP, concepts::Assertion());
  conceptsAssert(mP == 1, concepts::Assertion());
  const linDG2D::ElementPair* epc =
    dynamic_cast<const linDG2D::ElementPair*>(&ep);
  conceptsAssert(epc, concepts::Assertion());

  const linDG2D::EdgeInfo& ei = epc->edgeInfo();
  Real factor;

  if (ei.cell(1) == 0) {
    // boundary integral
    if (bc_) {
      if ((*bc_)(ei.edge().attrib()).type() == 
          concepts::Boundary::DIRICHLET) {
        DEBUGL(DGElasticFluxBf2dAppl_D,
               ei.edge() << "has dirichlet attribute");
        factor = 1;
      } else {
        DEBUGL(DGElasticFluxBf2dAppl_D,
               ei.edge() << "has non-dirichlet attribute");
        factor = 0;
      }
    } else {
      DEBUGL(DGElasticFluxBf2dAppl_D,
             ei.edge() << "is boundary assumed dirichlet");
      // assume dirichlet boundary all over
      factor = 1;
    }
  } else {
    // inner integral, <u> = 0.5*(u_left + u_right)
    DEBUGL(DGElasticFluxBf2dAppl_D,
           ei.edge() << "is inner edge");
    factor = 0.5;
  }

  if (factor != 0) {
    factor *= scale_ * ei.length() * ei.normal()[dV_];

    // <p> [v]*n
    if (&(eV->cell()) == ei.cell(1)) {
      // for jump of testfunction [v]: testfunction (on eV) is on element 1
      factor *= -1;
    }

    for(uint iV=0; iV<mV; ++iV) {
      em(iV,0) = factor * epc->shapeFct(0, iV, 0.5);
    }
  }
}

// ******************************************************* DGElasticCompBf2d **

void DGElasticCompBf2d::operator()(const concepts::Element<Real>& elmQ,
                                   const concepts::Element<Real>& elmP,
                                   concepts::ElementMatrix<Real>& em) {
  const uint mQ = elmQ.T().m();
  const uint mP = elmP.T().m();
  em.resize(mQ, mP);
  em.zeros(); 

  if (&elmQ == &elmP) {
    const linDG2D::TriangleP0* eQ =
      dynamic_cast<const linDG2D::TriangleP0*>(&elmQ);
    conceptsAssert(eQ, concepts::Assertion());
    conceptsAssert(mQ == 1, concepts::Assertion());
    em(0,0) = eQ->cell().area() / lambda_;
  }
}
