#ifndef dgNorm2d_hh
#define dgNorm2d_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "operator/compositions.hh"
#include "function/vector.hh"

using concepts::Real;

namespace concepts {
  class Formula;
}

namespace vectorial {
  template<typename F>
  class Element;
}

// **************************************************************** DGNorm2d **

/** Provides methods to calculate norms
    @author Christoph Winkelmann, 2003
*/
class DGNorm2d
{
 public:
  /** Calculates the norm defined by \c matrix of \c coeffs
      @param coeffs The coefficent vector
      @param matrix The matrix which defines the norm
  */
  static Real matrix(const concepts::Vector<Real>& coeffs,
                     concepts::Operator<Real>& matrix) {
    concepts::Vector<Real> temp(coeffs);
    matrix(coeffs, temp);
    return std::sqrt(coeffs*temp);    
  }
  /** Calculates the L2 norm of a mesh function defined by \c coeffs
      minus a general function given by an exact formula \c exactFrm
      evaluated at time \c time.
      @param coeffs The coefficent vector
      @param exactFrm The formula for the exact solution
      @param nQuadPts Number of Gauss quadrature points per dimension
      @param time Time to evaluate \c exactFrmX and \c exactFrmY at
  */
  static Real l2(const concepts::Vector<Real>& coeffs,
                 concepts::Formula& exactFrm, uint nQuadPts,
                 Real time = 0);
  /** Calculates the energy norm of a vectorial mesh function defined by \c
      coeffs minus a general function given by exact formulae for x
      (\c exactFrmX) and y (\c exactFrmY) component, both evaluated at time
      \c time. The definition of the energy norm is according to T. Wihler.
      @param coeffs The coefficent vector
      @param exactFrmX The formula for the x component of the exact solution
      @param exactFrmY The formula for the y component of the exact solution
      @param exactFrmXderX d/dx of \c exactFrmX
      @param exactFrmXderY d/dy of \c exactFrmX
      @param exactFrmYderX d/dx of \c exactFrmY
      @param exactFrmYderY d/dy of \c exactFrmY
      @param nQuadPts Number of Gauss quadrature points per dimension
      @param time Time to evaluate \c exactFrm at
  */
  static Real energyTW(const concepts::Vector<Real>& coeffs,
                       concepts::Formula& exactFrmX,
                       concepts::Formula& exactFrmY,
                       concepts::Formula& exactFrmXderX,
                       concepts::Formula& exactFrmXderY,
                       concepts::Formula& exactFrmYderX,
                       concepts::Formula& exactFrmYderY,
                       uint nQuadPts, Real time = 0);
  /** Calculates the L2 norm of a vectorial mesh function defined by \c coeffs
      minus a general function given by exact formulae for x (\c exactFrmX)
      and y (\c exactFrmY) component, both evaluated at time \c time.
      @param coeffs The coefficent vector
      @param exactFrmX The formula for the x component of the exact solution
      @param exactFrmY The formula for the y component of the exact solution
      @param nQuadPts Number of Gauss quadrature points per dimension
      @param time Time to evaluate \c exactFrm at
  */
  static Real l2(const concepts::Vector<Real>& coeffs,
                 concepts::Formula& exactFrmX, concepts::Formula& exactFrmY,
                 uint nQuadPts, Real time = 0);
  // Calculate symmetric gradient of solution on vectorial element
  static concepts::Mapping<Real, 2> elmSymmGrad
  (vectorial::Element<Real>& vElm, const concepts::Vector<Real>& coeffs);
  private:
  /// Constructor. Is private in order to prevent creating an instance.
  DGNorm2d() {};
  /// Calculate solution on vectorial element
  static concepts::Real2d elmSol(vectorial::Element<Real>& vElm,
                                 const concepts::Vector<Real>& coeffs,
                                 Real xi1, Real xi2);
  /// Calculate norm of a mapping
  template <uint dim>
  static Real mappingNorm(concepts::Mapping<Real, dim>& map);
};

#endif // dgNorm2d_hh
