#ifndef equivalentStress_hh
#define equivalentStress_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "elasticityTensor.hh"
#include "vectorial.hh"
#include "linDG2D/space.hh"
#include "operator/compositions.hh"

using concepts::Real;

// ********************************************** EquivalentStressHypothesis **

/** Interface class for equivalent stress hypotheses
    @author Christoph Winkelmann, 2004
*/
template<int dim>
class EquivalentStressHypothesis {
 public:
  /// calculate equivalent stress from stress tensor \c sigma
  virtual Real operator() (concepts::Mapping<Real, dim>& sigma) const = 0;
  /// calculate principal stresses
  virtual concepts::Point<Real, dim>
  principalStresses(concepts::Mapping<Real, dim>& sigma) const;
};

// **************************************** MaximumPrincipalStressHypothesis **

/** Maximum principal stress hypothesis
    @author Christoph Winkelmann, 2004
*/
template<int dim>
class MaximumPrincipalStressHypothesis : 
  public EquivalentStressHypothesis<dim> {
 public:
  /// calculate equivalent stress from stress tensor \c sigma
  virtual Real operator() (concepts::Mapping<Real, dim>& sigma) const;
};

// ******************************************** MaximumShearStressHypothesis **

/** Maximum shear stress hypothesis (Tresca)
    @author Christoph Winkelmann, 2004
*/
template<int dim>
class MaximumShearStressHypothesis : 
  public EquivalentStressHypothesis<dim> {
 public:
  /// calculate equivalent stress from stress tensor \c sigma
  virtual Real operator() (concepts::Mapping<Real, dim>& sigma) const;
};

// ************************************** MaximumShearStrainEnergyHypothesis **

/** Maximum shear strain energy hypothesis (von Mises)
    @author Christoph Winkelmann, 2004
*/
template<int dim>
class MaximumShearStrainEnergyHypothesis : 
  public EquivalentStressHypothesis<dim> {
 public:
  /// calculate equivalent stress from stress tensor \c sigma
  virtual Real operator() (concepts::Mapping<Real, dim>& sigma) const;
};

// ****************************************************** EquivalentStress2d **

/** Equivalent stress operator in 2d. Maps a displacement field to a scalar
    equivalent stress field calculated using a suitable equivalent stress
    hypothesis represented by a subclass of EquivalentStressHypothesis.
    @author Christoph Winkelmann, 2004
*/
class EquivalentStress2d : public concepts::Operator<Real> {
 public:
  /** Constructor.
      @param esh equivalent stress hypothesis
      @param spc vectorial (displacement, input) space
  */
  EquivalentStress2d(const EquivalentStressHypothesis<2>& esh,
                     const ElasticityTensor<2>& et,
                     const vectorial::Space<Real>& spc) :
    esh_(esh), et_(et), spc_(spc) {
      conceptsAssert(spc.vdim() == 2, concepts::Assertion());
    }
  /** Evaluation operator for the equivalent stress.
      Calculates the equivalent stress of the displacement field in \c fncX
      and stores the result in \c fncY
      @param fncX Input vector of displacements
      @param fncY Output vector of equivalent stress
  */
  void operator()
    (const concepts::Function<Real>& fncY, concepts::Function<Real>& fncX);
  /// Returns the image space
  const concepts::Space<Real>& spaceX() const { return *spc_.get(0); }
  /// Returns the source space
  const concepts::Space<Real>& spaceY() const { return spc_; }
protected:
  virtual std::ostream& info (std::ostream &os) const;
private:
  const EquivalentStressHypothesis<2>& esh_;
  const ElasticityTensor<2>& et_;
  const vectorial::Space<Real>& spc_;
};

#endif // equivalentStress_hh
