#ifndef dgElasticLinearForms_hh
#define dgElasticLinearForms_hh

#include "basics/typedefs.hh"
#include "integration.hh"
#include "timestepping/vectors.hh"
#include "toolbox/formula.hh"
#include "toolbox/boundary.hh"
#include "geometry/boundaryConditions.hh"
#include "function/linearForm.hh"
#include "linDG2D.hh"

#include "elasticityTensor.hh"

#define DEBUG_FILLDATA 0

using concepts::Real;

/** DG right hand side for time dependent elasticity problems in 2D
    @author Christoph Winkelmann, 2003
*/
class DGElasticTimeLf2d : public timestepping::TimeLinearForm {
 public:
  /** Constructor. Takes a formula object that describes the time dependent
      form of the right hand side.
      @param frm Time dependent formulae
      @param et Elasticity tensor
      @param vIdx Index of vectorial component
      @param bc Boundary conditions
      @param mi MeshInfo to get edge information from
      @param omega Weight for penalty term
      @param omegaNormal Weight for normal penalty term
      @param tau Sign for symmetric discretization of diffusion
      @param nQuadPts Number of quadrature points per dimension
  */
  DGElasticTimeLf2d(const concepts::Array<concepts::Formula*>& frm,
                    const ElasticityTensor<2>& et, uint vIdx, const
                    concepts::Array<concepts::BoundaryConditions*>& bc,
                    const linDG2D::MeshInfo& mi, Real omega,
                    Real omegaNormal, Real tau,
                    uint nQuadPts) :
    timestepping::TimeLinearForm(), frm_(frm), et_(et), vIdx_(vIdx), bc_(bc),
    mi_(mi), omega_(omega), omegaNormal_(omegaNormal), tau_(tau),
    nQuadPts_(nQuadPts) { }
  /** Application operator. Computes the element contribution.
      @param elmBase Element
      @param em Element matrix
  */
  void operator() (const concepts::Element<Real> &elmBase,
                   concepts::ElementMatrix<Real> &em);
 private:
  const concepts::Array<concepts::Formula*>& frm_;
  const ElasticityTensor<2>& et_;
  const uint vIdx_;
  const concepts::Array<concepts::BoundaryConditions*>& bc_;
  const linDG2D::MeshInfo& mi_;
  const Real omega_;
  const Real omegaNormal_;
  const Real tau_;
  const uint nQuadPts_;
};

/** pressure part of mixed DG right hand side for time dependent elasticity
    problems in 2D
    @author Christoph Winkelmann, 2004
*/
class DGElasticPressureTimeLf2d : public timestepping::TimeLinearForm {
 public:
  /** Constructor. Takes a formula object that describes the time dependent
      form of the right hand side.
      @param bc Boundary conditions
      @param mi MeshInfo to get edge information from
      @param nQuadPts Number of quadrature points per dimension
  */
  DGElasticPressureTimeLf2d(concepts::Array<concepts::BoundaryConditions*>& bc,
                            const linDG2D::MeshInfo& mi, uint nQuadPts) :
    timestepping::TimeLinearForm(), bc_(bc), mi_(mi), nQuadPts_(nQuadPts) { }
  /** Application operator. Computes the element contribution.
      @param elmBase Element
      @param em Element matrix
  */
  void operator() (const concepts::Element<Real> &elmBase,
                   concepts::ElementMatrix<Real> &em);
 private:
  const concepts::Array<concepts::BoundaryConditions*>& bc_;
  const linDG2D::MeshInfo& mi_;
  const uint nQuadPts_;
};

/** Pure L2 divergence linear form for DG in 2D. Useful for L2 projections of
    initial pressure conditions. Does numerical differentiation of formulae
    @author Christoph Winkelmann, 2004
*/
class DGDivergenceLf2d : public concepts::LinearForm<Real> {
 public:
  /** Constructor.
      @param frmX formula for x component
      @param frmY formula for y component
      @param nQuadPts number of quadrature points per dimension
  */
  DGDivergenceLf2d(concepts::Formula& frmX, concepts::Formula& frmY,
                   uint nQuadPts)
    : frmX_(frmX), frmY_(frmY), nQuadPts_(nQuadPts) { };
  /** Application operator. Computes the element contribution.
      @param elmBase Element
      @param em Element matrix
  */
  void operator() (const concepts::Element<Real> &elmBase,
                   concepts::ElementMatrix<Real> &em);
 private:
  const concepts::Formula& frmX_;
  const concepts::Formula& frmY_;
  const uint nQuadPts_;
};

#endif //dgElasticLinearForms_hh
