#include <iostream>
#include <memory>
#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"
#include "hp2D.hh"
#include "applications/square.hh"

using concepts::Real;

#define PrintMesh

// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    // Polynomgrad
    uint p = 1;
    // Anzahl Verfeinerungen
    uint l = 1;

    // Gitter
    Square msh(1.0, 1.0);
    std::cout << "Gitter: " << msh << std::endl;

    // Randbedingungen
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    bc->add(concepts::Attribute(1),
	    concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc->add(concepts::Attribute(4),
	    concepts::Boundary(concepts::Boundary::NEUMANN, "(0)"));
    std::cout << "Randbedingungen: " << *bc << std::endl;

    // Raum erzeugen
    hp2D::Space spc(msh, l, p, bc.get());
    spc.rebuild();
    std::cout << "Raum: " << spc << std::endl;

#ifdef PrintMesh
    // EPS Ausgabe des Gitters
    graphics::drawMeshEPS(spc, "mesh_Spitze.eps");
#endif

    // Laplace-Bilinearform **
    hp2D::Laplace stiff_bfl;
    concepts::DenseMatrix<Real> laplace(spc, stiff_bfl);
    std::cout << "stiff (Laplace) = " << laplace << std::endl;

    // rechte Seite (Quellterm)
    hp2D::Riesz lform(concepts::ParsedFormula("(1)"), bc.get() );
    concepts::Vector<Real> rhs(spc, lform);
    std::cout << "rhs = " << rhs << std::endl;
    
    // Lsen des Gleichungssystems
    concepts::CG<Real> solver(laplace, 1e-6, 2000);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "Lser konvergiert nicht!" << std::endl << e << std::endl;
    }
    std::cout << "Lser = " << solver << std::endl;
    std::cout << "Lsung = " << sol << std::endl;

    // Graphische Ausgabe
    std::stringstream filename;
    filename << "solution-" << l << ".gnuplot" << std::ends;
    graphics::drawDataGnuplot(spc, filename.str(), sol);

  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
