#ifndef advection_hh
#define advection_hh

#include "basics/typedefs.hh"
#include "timestepping/vectors.hh"
#include "linDG3D/element.hh"
#include "linDG3D/space.hh"
#include "linDG3D/elementPair.hh"
#include "operator/compositions.hh"
#include "toolbox/inputOutput.hh"
#include "geometry/boundaryConditions.hh"
#include "toolbox/boundary.hh"
#include "operator/diagonal.hh"

#include <map>
#include <vector>
#include <utility>
#include <numeric>

#define DEBUG_ADVECTION 0
#define DEBUG_FLUXES 0
#define DEBUG_ADAPT_LIM_ELM 0
#define DEBUG_ADAPT_LIM_INDICATOR 0
#define DEBUG_REF_ASSIGNMENT 0

using concepts::Real;


// ****************************************************************** Fluxes **
/** Burger's flux
    @author Adrian Burri, 2002
*/
class BurgerFlux {
public:
  /** Constructor.
  */
  BurgerFlux() :
    normal_(concepts::Real3d(1., 1., 1.)) {}
  /** Returns value of the Burger's flux.
      @param u Value where the flux is evaluated
  */
  concepts::Real3d flux(const Real u) const {
    return normal_*(0.5*u*u); 
  }
  /** Returns the value of the flux projected on the given direction.
      @param u Value where the flux is evaluated
      @param direction Direction
  */
  Real flux(const Real u, const concepts::Real3d& direction) const {
    return normal_*direction*(0.5*u*u);
  }
  /** Returns the derivative's value of the Burger's flux.
      @param u Value where the derivative is evaluated
  */
  concepts::Real3d deriv(const Real u) const {
    return normal_*u;
  }
  /** Returns the derivative's value of the Burger's flux projected on the
      given direction.
      @param u Value where the derivative is evaluated
      @param direction Direction
  */
  Real deriv(const Real u, const concepts::Real3d& direction) const {
    return normal_*direction*u;
  }
  /** Checks convexity of the flux function of the related 1D problem.
      For various fluxes, a local Riemann problem in the direction of the face
      normal has to be solved. Depending on the normal, the Burger flux project
      ed on this normal can be either convex or concave, which has to be taken
      into account in numerical problems that assume convexity.
      @param normal Normal vector of a face
  */
  bool isConvex(const concepts::Real3d& normal) const {
    return normal*normal_ > 0.;
  }
  /// Returns zero for Burger's flux.
  Real uMin(const concepts::Real3d& normal) const { return 0; }

private:
  const concepts::Real3d normal_;
};

/** LinearAdvectionFlux. Definition of the linear advection flux
    @author Adrian Burri, 2002
*/
class LinearAdvectionFlux {
 public:
  /** Constructor.
  */
  LinearAdvectionFlux() :
    normal_(concepts::Real3d(1., 1., 1.)) {}
  /** Returns the value of the linear advection flux.
      @param u Value where the derivative is evaluated
  */
  concepts::Real3d flux(const Real u) const { return normal_*u;}
  /** Returns the value of the linear advection flux projected on the given
      direction.
      @param u Value where the flux is evaluated
      @param direction Direction
  */
  Real flux(const Real u, const concepts::Real3d& direction) const {
    return normal_*direction*u;
  }
  /** Returns the derivative's value of the linear advection flux.
      @param u Value where the derivative is evaluated
  */
  concepts::Real3d deriv(const Real u) const { return normal_; }
  /** Returns the derivative's value of the linear advection flux projected on
      the
      direction.
      @param u Value where the derivative is evaluated
      @param direction Direction
  */
  Real deriv(const Real u, const concepts::Real3d& direction) const {
    return normal_*direction;
  }
private:
  const concepts::Real3d normal_;
};


// *************************************************************** Advection **

/** Advection operator for convective fluxes.
    @author Adrian Burri, 2002
 */
template<typename NumFlux>
class Advection : public concepts::Operator<Real> {
public:
  /** Constructor.
      @param fvFlux Flux function object
      @param spc FvdgSpace object
      @param pcb Flag indicating if boundary conditions for pure convection 
      (distinction between inflow/outflow) is considered
      @param bc Boundary condition object
  */
  Advection(const NumFlux& fvFlux,
	    const linDG3D::FvdgSpace& spc,
	    const bool pcb,
	    concepts::BoundaryConditions* bc = 0) :
    numFlux_(fvFlux),
    spc_(spc),
    innerList_(spc.innerElmPairList()),
    boundaryList_(spc.boundaryElmPairList()),
    time_(0),
    pcb_(pcb),
    bc_(bc) {};
  /** Evaluation operator for convective fluxes. Evaluates the convective
      fluxes at the cell interfaces according to the solution in \c fncX and
      stores its result in \c fncY.
      @param fncX Input vector
      @param fncY Output vector containing the contributions of the convective
      fluxes
  */
  void operator()
    (const concepts::Function<Real>& fncY, concepts::Function<Real>& fncX);
  /// Returns the image space
  const concepts::Space<Real>& spaceX() const { return spc_; }
  /// Returns the source space, which is the same as the image space
  const concepts::Space<Real>& spaceY() const { return spc_; }
protected:
  virtual std::ostream& info (std::ostream &os) const;
private:
  const NumFlux& numFlux_;
  const linDG3D::FvdgSpace& spc_;
  const concepts::ElementPairList<Real>& innerList_;
  const concepts::ElementPairList<Real>& boundaryList_;
  Real time_;
  const bool pcb_;
  concepts::BoundaryConditions* bc_;
  /** Returns value of the boundary condition at the edge center of the given 
      edge
      @param elm Element
      @param edge Edge index
      @param faceIdx Face index
  */
  Real uBoundary_(const linDG3D::FvdgElement& elm,
		  const uint faceIdx,
		  const concepts::Real3d &x) const;
};

// Numerical Fluxes

//************************************************************ LaxFriedrichs **
/** Local Lax-Friedrichs flux. This formulation of the local Lax-Friedrichs
    flux relies on convex flux functions.
    @author Adrian Burri, 2002
*/
template <typename T>
class LaxFriedrichs {
public:
  friend class Advection<LaxFriedrichs>;
  /// Constructor. Creates an instance of the flux function
  LaxFriedrichs() : fluxFct_() {}
  /** Numerical flux function. Implementation of the numerical flux function
      according to Lax-Friedrichs. This version uses the entropy fix of the
      local Lax-Friedrichs scheme.
      @param u1 Variable value of the first element
      @param u2 Variable value of the second element
      @param normal Normal vector of the common face. The vector is pointing
      from the first element to the second element
  */
  Real operator()
    (Real u1, Real u2, const concepts::Real3d& normal) const;
private:
  const T fluxFct_;
};

//******************************************************************* Upwind **
/** Upwind flux. 
    @author Adrian Burri, 2002
*/
template <typename T>
class Upwind {
  friend class Advection<Upwind>;
public:
  /// Constructor. Creates an instance of the flux function
  Upwind() : fluxFct_() {}
  /** Numerical flux function. Implementation of the numerical flux function
      according to the Upwind scheme.
      @param u1 Variable value of the first element
      @param u2 Variable value of the second element
      @param normal Normal vector of the common face. The vector is pointing
      from the first element to the second element
  */
  Real operator()
    (Real u1, Real u2, const concepts::Real3d& normal) const;
private:
  const T fluxFct_;
};

//****************************************************************** Godunov **
/** Godunov flux. The Godunov flux relies on flux functions that are strictly
    convex.
    @author Adrian Burri, 2002
*/
template <typename T>
class Godunov {
public:
  friend class Advection<Godunov>;
  /// Constructor. Creates an instance of the flux function.
  Godunov() : fluxFct_() {}
  /** Numerical flux function. Implementation of the numerical flux function
      according to the Godunov scheme.
      @param u1 Variable value of the first element
      @param u2 Variable value of the second element
      @param normal Normal vector of the common face. The vector is pointing 
      from the first element to the second element
  */
  Real operator() (Real u1, Real u2,
			 const concepts::Real3d& normal) const;
private:
  const T fluxFct_;
};

//******************************************************************** Osher **
/** Osher flux. The Osher flux relies on flux functions that are strictly
    convex.
    @author Adrian Burri, 2002
*/
template <typename T>
class Osher {
public:
  friend class Advection<Osher>;
  /// Constructor. Creates an instance of the flux function
  Osher() : fluxFct_() {}
  /** Numerical flux function. Implementation of the numerical flux function
      according to the Osher scheme.
      @param u1 Variable value of the first element
      @param u2 Variable value of the second element
      @param normal Normal vector of the common face. The vector is pointing
      from the first element to the second element
  */
  Real operator()
    (Real u1, Real u2, const concepts::Real3d& normal) const;
private:
  const T fluxFct_;
};

// Limiter

//*********************************************************** FvdgPi0Limiter **
/** pi0-projection.
    @author Adrian Burri, 2002
 */
class FvdgPi0Limiter : public concepts::Operator<Real> {
public:
  /** Constructor. Calculates the weights for the different global degrees of
      freedom. The algorithm assumes that for every element, the first degree
      of freedom corresponds to a constant valued shapefunction and the rest
      of them to linear contributions with vanishing mean value.
  */
  FvdgPi0Limiter(const linDG3D::FvdgSpace& spc);
  /// Returns the image space of the operator
  virtual const concepts::Space<Real>& spaceX() const {
    return diag_.spaceX();
  }
  /// Returns the source space of the operator
  virtual const concepts::Space<Real>& spaceY() const {
    return diag_.spaceY(); 
  }
  /// Application operator. Limits ...
  virtual void operator() (const concepts::Function<Real> &fncY,
			   concepts::Function<Real> &fncX) {
    diag_(fncY, fncX);
  }
protected:
  concepts::DiagonalMatrix<Real> diag_;
};

//*********************************************************** FvdgPi0Limiter **
/** Adaptive limiter.
    @author Adrian Burri, 2002
 */
class FvdgAdaptiveLimiter : public concepts::Operator<Real> {
public:
  /** Constructor.
      @param spc FvdgSpace
      @param alpha Weight factor for indicator
      @param lambda Weight factor for gradient damping
      @param bc Boundary conditions
  */
  FvdgAdaptiveLimiter(const linDG3D::FvdgSpace& spc,
		      const Real alpha,
		      const Real lambda,
		      concepts::BoundaryConditions* bc);
  /// Returns the image space of the operator
  virtual const concepts::Space<Real>& spaceX() const {
    return spc_;
  }
  /// Returns the source space of the operator
  virtual const concepts::Space<Real>& spaceY() const {
    return spc_; 
  }
  /** Application operator. Limits according to an indicator function.
      @param fncY Function containing unrestricted solution
      @param fncX Function for the limited solution
  */
  virtual void operator() (const concepts::Function<Real> &fncY,
			   concepts::Function<Real> &fncX);
  /** Gets values of element indicators.
      @param inidicator Vector to store element indicator values
  */
  void getIndicator(concepts::Vector<Real> &indicator);
protected:
  const linDG3D::FvdgSpace& spc_;
  concepts::BoundaryConditions* bc_;
  const Real alpha_;
  Real lambda_;
  std::map<const linDG3D::FvdgElement*, Real> elmIndicator_;
  // Private methods
  void limit_(concepts::Vector<Real>& vecX,
	      const linDG3D::FvdgElement& elm) const;
  /// Resets element indicator to zero
  void reset_();
  /** Calculates inner face contributions
   */
  void innerContrib_(const concepts::Vector<Real>& vecY, 
		     concepts::Vector<Real>& vecX);
  /** Calculates boundary face contributions
   */
  void bndryContrib_(const concepts::Vector<Real>& vecY, 
		     concepts::Vector<Real>& vecX);
  /** Calculates volume of an element
      @param elm Element of which the volume is calculated
  */
  Real elmVolume_(const linDG3D::FvdgElement& elm) const;
};

#endif // advection_hh
