#include <cmath>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <memory>
#include <fstream>
#include <unistd.h>

#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "integration.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"

#include "hp3D.hh"
#include "vectorial.hh"
#include "eigensolver.hh"

// ************************************************************* fem routine **

void fem(const concepts::InOutParameters& input,
	 concepts::InOutParameters& output) {
  const uint l = input.getInt("level");
  const uint p = input.getInt("polynomial");
  const std::string meshNameBase = input.getString("meshNameBase");

  const uint kmax = input.getInt("kmax");
  const uint solverType = input.getInt("solver");
  const Real shift = input.getDouble("shift");

  const Real rho = input.getDouble("rho");
  const uint weight = input.getInt("weight");
  const uint sNumber = input.getInt("sNumber");
  concepts::Array<Real> s(sNumber);
  for (uint i = 0; i < sNumber; ++i) s[i] = input.getArrayDouble("s", i);

  // mesh import: generate names
  std::stringstream fileCoord;
  fileCoord << meshNameBase << "Coord.dat" << std::ends;
  output.addString("coordFile", fileCoord.str().c_str());
  std::stringstream fileElements;
  fileElements << meshNameBase << "Elements.dat" << std::ends;
  output.addString("elementsFile", fileElements.str().c_str());
  std::stringstream fileBoundary;
  fileBoundary << meshNameBase << "Boundary.dat" << std::ends;
  output.addString("boundaryFile", fileBoundary.str().c_str());
  concepts::Import3dMesh msh(fileCoord.str(), fileElements.str(),
			     fileBoundary.str());

  std::cout << "Mesh: " << msh << std::endl;
  graphics::MeshDX<Real>(msh, "mesh.dx");

  concepts::BoundaryConditions bcX, bcY, bcZ;
  bcX.add(concepts::Attribute(1),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  bcY.add(concepts::Attribute(1),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));

  bcX.add(concepts::Attribute(2),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  bcZ.add(concepts::Attribute(2),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));

  bcZ.add(concepts::Attribute(3),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  bcY.add(concepts::Attribute(3),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));

  bcX.add(concepts::Attribute(4),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  bcZ.add(concepts::Attribute(4),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));

  bcZ.add(concepts::Attribute(5),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  bcY.add(concepts::Attribute(5),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));

  bcX.add(concepts::Attribute(6),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  bcY.add(concepts::Attribute(6),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));

  const uint vdim = 3;
  vectorial::Space<Real> spc(vdim);
  hp3D::Space sspcX(msh, 0, p, &bcX);    sspcX.rebuild();    spc.put(sspcX);
  hp3D::Space sspcY(msh, 0, p, &bcY);    sspcY.rebuild();    spc.put(sspcY);
  hp3D::Space sspcZ(msh, 0, p, &bcZ);    sspcZ.rebuild();    spc.put(sspcZ);

  for (uint i = 0; i < l; ++i) {
    std::cout << "i = " << i << std::endl;
    concepts::ResourceMonitor timer;
    sspcX.rebuild();
    std::cout << "  scalar space: " << std::flush;
    std::cout << sspcX << std::endl;
    sspcY.rebuild();
    std::cout << "                " << sspcY << std::endl;
    sspcZ.rebuild();
    std::cout << "                " << sspcZ << std::endl;

    spc.rebuild();
    std::cout << "  vector valued space (" << vdim << " comp.): "
	      << std::flush;
    std::cout << spc << std::endl;

    output.addArrayInt("dof", i, spc.dim());
    output.addArrayDouble("space_time", i, timer.utime());

    // graphics of the mesh
    graphics::MeshDX<Real>(sspcX, "space.dx", 3);
    graphics::drawMeshTecplot(sspcX, "space.dat", 3);

    if (spc.dim() > 20) {

      // ** stiffness matrix **
      vectorial::BilinearForm<Real> vrotrot_bf(vdim);
      hp3D::RotRot::setup(vrotrot_bf);
      concepts::SparseMatrix<Real> rotrot(spc, vrotrot_bf);
      rotrot.compress();
      std::cout << "  rot rot matrix: " << rotrot << std::endl;

      hp3D::SingularSet singularities;
      singularities.add(concepts::Attribute(input.getInt("vtxRef")), msh);
      singularities.add(concepts::Attribute(input.getInt("edgRef")), msh);

      vectorial::BilinearForm<Real> vdivdiv_bf(vdim);
      switch (weight) {
      case 0: // TrivialWeight
	hp3D::DivDiv<hp3D::TrivialWeight>::setup(vdivdiv_bf, singularities);
	break;
      case 1: // ShortestDist
	hp3D::DivDiv<hp3D::ShortestDist>::setup(vdivdiv_bf, singularities);
	break;
      case 11: // ProductOfAll
	hp3D::DivDiv<hp3D::ProductOfAll>::setup(vdivdiv_bf, singularities);
	break;
      case 21: // DaugeWeight
	hp3D::DivDiv<hp3D::DaugeWeight>::setup(vdivdiv_bf, singularities);
	break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
      std::cout << "  div div bilinear form: "
		<< *(vdivdiv_bf.get(0)) << std::endl;
      std::stringstream weightName;
      weightName << *(vdivdiv_bf.get(0));
      output.addString("weight", weightName.str().c_str());

      concepts::SparseMatrix<Real> divdiv(spc, vdivdiv_bf);
      divdiv.compress();
      std::cout << "  div div matrix: " << divdiv << std::endl;

      // ** mass matrix **
      hp3D::Identity smass_bf;
      vectorial::BilinearForm<Real> mass_bf(vdim);
      mass_bf.put(smass_bf, 0, 0);
      mass_bf.put(smass_bf, 1, 1);
      mass_bf.put(smass_bf, 2, 2);
      concepts::SparseMatrix<Real> mass(spc, mass_bf);
      std::cout << "  mass matrix: " << mass << std::endl;
      output.addArrayDouble("matrix_time", i, timer.utime());
      mass.compress();
      std::cout << "  mass matrix: " << mass << std::endl;

      // ** solver **
      for (uint sIndex = 0; sIndex < sNumber; ++sIndex) {
	std::cout << "  s = " << s[sIndex] << std::endl;
	concepts::LiCo<Real> stiff(rotrot, divdiv, 1.0, s[sIndex]);
	std::auto_ptr<eigensolver::SolverFabric> fabric(0);
#ifdef HAS_SuperLU
	concepts::SuperLUFabric linFabric;
#endif
	switch (solverType) {
#ifdef HAS_JDBSym
#  ifdef HAS_SuperLU
	case 0: // jdbsym
	  fabric.reset(new eigensolver::JdbSymFabric
		       (1e-10, 1000, shift, 1,
			kmax <= spc.dim() ? kmax : spc.dim(), &linFabric));
	  break;
#  else
	case 0: // jdbsym
	  fabric.reset(new eigensolver::JdbSymFabric
		       (1e-10, 1000, shift, 1,
			kmax <= spc.dim() ? kmax : spc.dim()));
	  break;
#  endif
#endif
#ifdef HAS_ARPACK
	case 1: // arpack, regular invert mode (2)
	  fabric.reset(new eigensolver::ArPackFabric
		       ( kmax < spc.dim() ? kmax : spc.dim()-1 , 0.0, 1000));
	  break;
	case 2: // arpack, shift-invert modes (3)
	  fabric.reset(new eigensolver::ArPackFabric
		       ( kmax < spc.dim() ? kmax : spc.dim()-1 , 0.0, 1000,
			 eigensolver::ArPack::LA,
			 eigensolver::ArPack::SHIFTINV, shift));
	  break;
#endif
#ifdef HAS_PETSC
	case 3: { // inexact inverse iteration
	  concepts::Vector<Real> init(spc);
	  fabric.reset(new eigensolver::InexactInvFabric
		       (init, 0.75, spc.dim(), 1e-11, spc.dim()*3));
	  break; }
#endif
	default:
	  throw conceptsException
	    (concepts::MissingFeature("solver not present"));
	}

	eigensolver::EigenSolver* solver = &((*fabric)(stiff, mass));
	std::cout << "  solver = " << *solver << std::endl;

	// ** results **
	const concepts::Array<Real>& sol_eigenvalue = solver->getEV();
	const concepts::Array<concepts::Vector<Real>*>& sol_eigenfunction
	  = solver->getEF();
	if (sIndex == 0)
	  output.addArrayDouble("solve_time", i, timer.utime());
	const uint nEV = sol_eigenvalue.size();
	std::cout.precision(15);
	std::cout << "  Eigenvalues: " << sol_eigenvalue << std::endl;
	std::cout << "  solver = " << *solver << std::endl;

	// computing Rotrot and Divdiv energies
	for (uint evIndex = 0; evIndex < nEV; ++evIndex) {
	  concepts::Vector<Real> tmp(spc);

	  rotrot(*(sol_eigenfunction[evIndex]), tmp);
	  Real rotEnergy = *(sol_eigenfunction[evIndex]) * tmp;

	  divdiv(*(sol_eigenfunction[evIndex]), tmp);
	  Real divEnergy = *(sol_eigenfunction[evIndex]) * tmp * s[sIndex];

	  std::string quality;
	  if ( std::fabs(rotEnergy)/std::fabs(divEnergy) >= rho) {
	    quality = "**";
	  } else {
	    if ( std::fabs(rotEnergy)/std::fabs(divEnergy) <= 1./rho)
	      quality = "  ";
	    else quality = " *";
	  }
	  std::cout << quality << " EV " << evIndex << " = ";
	  std::cout.precision(15);
	  std::cout << sol_eigenvalue[evIndex];
	  std::cout.precision(5);
	  mass(*(sol_eigenfunction[evIndex]), tmp);
	  Real L2norm = *(sol_eigenfunction[evIndex]) * tmp;
	  std::cout << " has rotEnergy = " << rotEnergy
		    << ", divEnergy*s = " << divEnergy
		    << ", L2 norm = " << L2norm
		    << std::endl;

	  std::stringstream name1, name2, name3;
	  name1 << "ev_s_" << s[sIndex] << "_k_" << evIndex << std::ends;
	  output.addArrayDouble(name1.str().c_str(), i,
				sol_eigenvalue[evIndex]);
	  name2 << "dE_s_" << s[sIndex] << "_k_" << evIndex << std::ends;
	  output.addArrayDouble(name2.str().c_str(), i, divEnergy);
	  name3 << "rE_s_" << s[sIndex] << "_k_" << evIndex << std::ends;
	  output.addArrayDouble(name3.str().c_str(), i, rotEnergy);
	} // for evIndex
      } // for sIndex
    } // if dim > 20

    // ** refinement **
    if (i < l-1) {
      int pMax[3] = { 1, 1, 1 };
      hp3D::Space* spaces[3] = { &sspcX, &sspcY, &sspcZ };
      for (uint j = 0; j < 3; ++j) {
	hp3D::APrioriRefinement refineSpace
	  (*spaces[j], input.getInt("vtxRef"), input.getInt("edgRef"),
	   input.getInt("faceRef"), pMax);
	concepts::GlobalPostprocess<Real> post(*spaces[j]);
	post(refineSpace);
      } // loop over three components
      std::cout << "--" << std::endl;
    } // if i < l-1
  } // for i
}

// ************************************************************ Main Program **

int main(int argc, char** argv) {
  try {
    concepts::InputParser inputParser;
    concepts::InOutParameters& input = inputParser.inputParameters();

    input.addInt("level", 2);
    input.addInt("polynomial", 2);
    input.addString("meshNameBase", SOURCEDIR "/applications/thickLshape");
    input.addInt("vtxRef", 66);
    input.addInt("edgRef", 77);
    input.addInt("faceRef", 0);

    input.addInt("kmax", 20);
    input.addInt("solver", 2);
    input.addDouble("shift", 9.0);

    input.addDouble("rho", 1.5);
    input.addInt("weight", 11);
    input.addInt("sNumber", 10);
    concepts::Array<Real> s(input.getInt("sNumber"));
    s[0] = 6;
    s[1] = 2;
    s[2] = 3;
    s[3] = 4;
    s[4] = 20;
    s[5] = 9;
    s[6] = 12;
    s[7] = 15;
    s[8] = 17;
    s[9] = 30;
    input.addArrayDouble("s");
    for (uint i = 0; i < (uint)input.getInt("sNumber"); ++i)
      input.addArrayDouble("s", i, s[i]);

    input.addString("title",
		    "Maxwell Eigenvalues in 3D using weighted regularization");

    std::stringstream outfile;
    outfile << argv[0] << ".out" << std::ends;
    input.addString("parameterout", outfile.str().c_str());

    std::stringstream gnuplotfile;
    gnuplotfile << argv[0] << ".gnuplot" << std::ends;
    input.addString("gnuplot", gnuplotfile.str().c_str());

    concepts::InOutParameters& output = inputParser.outputParameters();
    output.addString("version", "$Id: hpFEM3d-EV.cc,v 1.3 2004/07/20 07:14:41 pfrauenf Exp $");
    output.addArrayInt("dof");
    output.addArrayDouble("space_time");
    output.addArrayDouble("matrix_time");
    output.addArrayDouble("solve_time");
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::INT, "dof", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "space_time", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "matrix_time", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "solve_time", output);

    // ********************************************************** input data **
    std::string inputfile;
    int opt;
    while ((opt = getopt(argc, argv, "-f:e:l:p:k:r:n:s:w:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", atoi(optarg)); break;
      case 'p': input.addInt("polynomial", atoi(optarg)); break;
      case 'f': inputfile = std::string(optarg); 
	inputParser.parse(inputfile);
	break;
      case 'e': input.addInt("solver", std::atoi(optarg)); break;
      case 'k': input.addInt("kmax", std::atoi(optarg)); break;
      case 'r': input.addDouble("rho", std::atof(optarg)); break;
      case 'n': input.addString("meshNameBase", optarg); break;
      case 's': input.addDouble("shift", std::atof(optarg)); break;
      case 'w': input.addInt("weight", std::atoi(optarg)); break;
      default:
	std::cout << "Option '" << opt << "' unkown." << std::endl
		  << "Usage: " << argv[0] 
		  << " [-f FILE] [-l LEVEL] [-e SOLVER] [-k KMAX] [-p POLY] "
		  << "[-r RHO] [-n MESH_NAME] [-s SHIFT] [-w WEIGHT]"
		  << std::endl
		  << "where" << std::endl
		  << "  FILE: name of the input file" << std::endl
		  << "  LEVEL: level of refinement" << std::endl
		  << "  SOLVER: JDBSYM(0), ARPACK(1)" << std::endl
		  << "  POLY: starting polynomial degree" << std::endl
		  << "  KMAX: number of eigenvalues which should be computed"
		  << std::endl
		  << "  MESH_NAME: name of the mesh (thickLshape)" << std::endl
		  << "  SHIFT: shift for Eigenvalue solver" << std::endl
		  << "  WEIGHT:" << std::endl
		  << "    0: TrivialWeight" << std::endl
		  << "    1: ShortestDist" << std::endl
		  << "    11: ProductOfAll" << std::endl
		  << "    21: DaugeWeight" << std::endl
		  << "Options given after the input file override the values "
		  << "read from the"
		  << std::endl << "input file." << std::endl;
	exit(1);
	break;
      }

    // ********************************** create output area for eigenvalues **
    for (uint i = 0; i < (uint)input.getInt("sNumber"); ++i) {
      for (uint j = 0; j < (uint)input.getInt("kmax"); ++j) {
	std::stringstream name1, name2, name3;
	name1 << "ev_s_" << input.getArrayDouble("s", i) << "_k_" << j
	      << std::ends;
	output.addArrayDouble(name1.str().c_str());
	table.addMap(concepts::ResultsTable::DOUBLE, name1.str().c_str(),
		     output);

	name2 << "dE_s_" << input.getArrayDouble("s", i) << "_k_" << j
	      << std::ends;
	output.addArrayDouble(name2.str().c_str());
	table.addMap(concepts::ResultsTable::DOUBLE, name2.str().c_str(),
		     output);

	name3 << "rE_s_" << input.getArrayDouble("s", i) << "_k_" << j
	      << std::ends;
	output.addArrayDouble(name3.str().c_str());
	table.addMap(concepts::ResultsTable::DOUBLE, name3.str().c_str(),
		     output);
      }
    }

    // ***************************************************** show parameters **
    std::cout << '[' << argv[0] << "]" << std::endl;
    std::cout << "--" << std::endl;
    std::cout << "Parameters:" << std::endl
	      << "  input file = " << inputfile << std::endl
	      << input;
    std::cout << "--" << std::endl;

    // ******************************************************** computations **
    fem(input, output);

    // ************************************** output of input data and other **
    std::ofstream* ofs = new std::ofstream
      (input.getString("gnuplot").c_str());
    *ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(*ofs);
    delete ofs;

    std::cout << "  --" << std::endl
	      << "  Writing gathered data to disk: " 
	      << input.getString("parameterout") << std::endl;

    ofs = new std::ofstream(input.getString("parameterout").c_str());

    *ofs << "/* program:\t" << argv[0] << std::endl
	 << " * command:\t";
    for (int j = 0; j < argc; ++j)
      *ofs << argv[j] << " ";
    *ofs << std::endl
	 << " * input file:\t" << inputfile << std::endl;
    *ofs << " */" << std::endl << std::setprecision(20) << inputParser;
    delete ofs;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
