#include <iostream>
#include <memory>
#include <algorithm>
#include <sstream>
#include <unistd.h> // for command line parsing

#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
#include "graphics.hh"
#include "linearFEM.hh"
#include "linDG1D.hh"

using concepts::Real;

#include "meshes.cc"

Real exact(const Real x) {
  return  -3.0*cosh(x)/cosh(1.0)+x*x+2.0;
}

void fem(const uint level, const uint gauss_p,
	 concepts::InOutParameters& output) {
  Line msh(-1, 0, 1);

  std::auto_ptr<concepts::BoundaryConditions>
    bc(new concepts::BoundaryConditions());
  bc->add(concepts::Attribute(2),
	  concepts::Boundary(concepts::Boundary::DIRICHLET));
  std::cout << "Boundary conditions: " << *bc << std::endl;
  std::cout << "Mesh: " << msh << std::endl;

  for (uint l = 1; l < level; ++l) {
    std::cout << "l = " << l << std::endl;

    linDG1D::Linear1d spc(msh, l);
    std::cout << "Space: " << spc << std::endl;

    linearFEM::Riesz1d rhs_lf(concepts::ParsedFormula("(x^2)"), gauss_p);
    concepts::Vector<Real> rhs(spc, rhs_lf);

    linearFEM::Laplace1d stiff_bf;
    concepts::SparseMatrix<Real> stiff(spc, stiff_bf);

    linearFEM::Identity1d mass_bf;
    concepts::SparseMatrix<Real> mass(spc, mass_bf);

    linDG1D::BoundaryInt bound_int_bf(bc.get());
    concepts::SparseMatrix<Real> bound_int(spc, bound_int_bf);
    concepts::Matrix<Real>::assembly(bound_int, bound_int_bf,
				     spc.elmPairs());
    concepts::SparseMatrix<Real> bound_int_t(bound_int, true);

    linDG1D::BoundaryIntStab bound_int_stab_bf;
    concepts::SparseMatrix<Real> bound_int_stab(spc, spc);
    concepts::Matrix<Real>::assembly(bound_int_stab, bound_int_stab_bf,
				     spc.elmPairs());

    concepts::LiCo<Real> L1(stiff, mass, 1.0, 1.0);
    concepts::LiCo<Real> L2(L1, bound_int_t, 1.0, 1.0);
    const Real t = 1.0;
    concepts::LiCo<Real> L3(L2, bound_int, 1.0, t);
    concepts::LiCo<Real> L(L3, bound_int_stab, 1.0, 1.0);

    concepts::GMRes<Real> solver(L, 1e-6, 2000);
//     concepts::SparseMatrix<Real> tmp(L);
//     concepts::Umfpack solver(tmp);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
//     std::cout << "solution = " << sol << std::endl;

    const Real h = 1.0/(1<<l);
    output.addArrayDouble("h", l, h);

    Real x = -1.0;
    Real maxNodalFehler = 0.0;
    Real nodalFehler = 0.0;
    for (uint p = 0; p < spc.dim(); ++p) {
      nodalFehler = fabs(sol[p] - exact(x));
      if (nodalFehler > maxNodalFehler)
	maxNodalFehler = nodalFehler;
      if (p%2 == 0) x += h;
    }
    output.addArrayDouble("maxNodalFehler", l, maxNodalFehler);

    std::stringstream filename;
    filename << "solutionDG1d-" << l << ".gnuplot" << std::ends;
    graphics::drawDataGnuplot(spc, filename.str(), sol);
  } // for l
}

// ************************************************************ Main Program **

int main(int argc, char** argv) {
  try {
    concepts::InputParser inputParser(true);
    concepts::InOutParameters& input = inputParser.inputParameters();

    input.addInt("level", 1);
    input.addInt("gauss_p", 1);

    int opt;
    while ((opt = getopt(argc, argv, "-l:p:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P]" << std::endl
		  << "where" << std::endl
		  << "  LEVEL: maximal level of refinement" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration"
		  << std::endl;
	std::exit(1); break;
      }

    concepts::InOutParameters& output = inputParser.outputParameters();
    output.addArrayDouble("h");
    output.addArrayDouble("maxNodalFehler");

    fem(input.getInt("level"), input.getInt("gauss_p"), output);

    std::cout << inputParser << std::endl;

    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "h", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "maxNodalFehler", output);
    std::ofstream ofs("linearDG1d.gnuplot");
    ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(ofs);
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
