/* output operator for all classes in concepts
 */

#ifndef OutputOperator_h
#define OutputOperator_h

#include <iostream>
#include <memory>

namespace concepts {

  // ******************************************************** OutputOperator **

  /** Class providing an output operator.
      This class provides an output operator and a virtual member
      \c info which is called by the output operator. Every class which needs
      an output operator can derive from this class and overload the \c info
      member.

      You should not derive from this class if one of the following cases
      applies to your class:
      <ul>
      <li>the objects of your class are very frequent (e.g. in an array)
      <li>the objects of your class are created and destroyed very frequently
      (e.g. in loops)
      </ul>
      and your class does not yet have virtual members. In this case,
      the memory or performance penalty resp. would be too large.

      The format of the output of \c info should have the following form:
      <pre>ClassName(additional information)</pre>
      and it must not end with a line break. Usually, no line breaks are
      necessary inside the output, too.

      @author Philipp Frauenfelder, 2001
  */
  class OutputOperator {
    friend std::ostream& operator<<(std::ostream& os, const OutputOperator& o);
  public:
    virtual ~OutputOperator();
  protected:
    /// Returns information in an output stream
    virtual std::ostream& info(std::ostream& os) const;
  };


} // namespace concepts


// **************************************** Output operator for a auto_ptr<> **

template<class F>
std::ostream& operator<<(std::ostream& os, const std::auto_ptr<F>& p) {
  return os;
}

#endif // OutputOperator_h
