/* testing point
 */

#include "basics/testsuite.hh"
#include "basics/exceptions.hh"
#include "basics/vectorsMatrices.hh"
#include "point.hh"

namespace test {

  // ************************************************************* PointTest **

  void PointTest::run() {
    testConstr();
    testDotProduct();
    testCrossProduct();
    testSum();
  }

  void PointTest::testConstr() {
    concepts::Real2d p0(1, 2);
    _test(p0[0] == 1);
    _test(p0[1] == 2);
    concepts::Real3d p2(4, 5, 6);
    _test(p2[0] == 4);
    _test(p2[1] == 5);
    _test(p2[2] == 6);
  }

  void PointTest::testDotProduct() {
    concepts::Real2d p0(2), p1(3);
    _test(p0*p1 == 12);
    concepts::Real3d p2(4);
    _test(p0*(concepts::Real2d)p2 == 16);
    _test(p2*(concepts::Real3d)p0 == 16);
    concepts::Real3d p3(3, 4, 7);
    _test(p2 * p3 == 56);
  }

  void PointTest::testCrossProduct() {
    concepts::Real3d p2(4), p3(5);
    _test(p2 ^ p3 == concepts::Real3d(0));
    p3[1] = 1;
    _test(p2 ^ p3 == concepts::Real3d(16, 0, -16));
    concepts::Real2d p0(2), p1(3);
    _test(p0.cross(p1) == 0);
    p1[1] = 1;
    _test(p0.cross(p1) == -4);
  }

  void PointTest::testSum() {
    concepts::Real3d p2(1, 2, 4), p3(5, 7, 9), p4(p2 + p3);
    _numtest(p4[0], 6);
    _numtest(p4[1], 9);
    _numtest(p4[2], 13);
  }

  // *********************************************************** MappingTest **

  void MappingTest::run() {
    testContr();
    testDet();
    testInv();
    testTranspose();
    testMapping();
    testScaleRows();
    testScaleColumns();
    testRank1Prod();
  }

  void MappingTest::testContr() {
    concepts::MapReal2d m0(1, 2, 3, 4);
    _numtest(m0(0,0), 1);  _numtest(m0(0,1), 2);
    _numtest(m0(1,0), 3);  _numtest(m0(1,1), 4);
  }

  void MappingTest::testDet() {
    concepts::MapReal2d m0(1, 2, 3, 4);
    _test(m0.determinant() == -2);
    concepts::MapReal3d m1(1, 2, 3, 4, 5, 6, 7, 8, 10);
    _test(m1.determinant() == -3);
  }

  void MappingTest::testInv() {
    concepts::MapReal2d m0(1, 2, 3, 4), m0i(m0.inverse());
    _test(m0i(0,0) == -2);     _test(m0i(0,1) == 1);
    _test(m0i(1,0) == 1.5);    _test(m0i(1,1) == -0.5);
    concepts::MapReal3d m1(1, 2, 3, 4, 5, 6, 7, 8, 10), m1i(m1.inverse());
    _test(m1i(0,0) == -2./3.); _test(m1i(0,1) == -4./3.); _test(m1i(0,2) == 1);
    _test(m1i(1,0) == -2./3.); _test(m1i(1,1) == 11./3.); _test(m1i(1,2) ==-2);
    _test(m1i(2,0) == 1);      _test(m1i(2,1) == -2);     _test(m1i(2,2) == 1);
  }

  void MappingTest::testTranspose() {
    concepts::MapReal3d m1(1, 2, 3, 4, 5, 6, 7, 8, 10), m1i(m1.transpose());
    _test(m1i(0,0) == 1);
    _test(m1i(2,0) == 3);
    _test(m1i(0,2) == 7);
  }

  void MappingTest::testMapping() {
    concepts::MapReal2d m0(1, 2, 3, 4);
    concepts::Real2d p0(2), p1(m0*p0);
    _test(p1[0] == 6);   _test(p1[1] == 14);
    concepts::MapReal3d m1(1, 2, 3, 4, 5, 6, 7, 8, 10);
    concepts::Real3d p2(4.0), p3(m1*p2);
    _test(p3[0] == 24);  _test(p3[1] == 60); _test(p3[2] == 100);

    concepts::MapReal3d m3(1.3, -2, 3.7, 2.5, 5, 6.9, 7, 3, -10), m2(m1*m3);
    _numtest(m2(0,0), 27.3);  _numtest(m2(0,1), 17);  _numtest(m2(0,2), -12.5);
    _numtest(m2(1,0), 59.7);  _numtest(m2(1,1), 35);  _numtest(m2(1,2), -10.7);
    _numtest(m2(2,0), 99.1);  _numtest(m2(2,1), 56);  _numtest(m2(2,2), -18.9);
  }

  void MappingTest::testScaleRows() {
    concepts::MapReal2d m0(1, 2, 3, 4);
    m0.scaleRows(concepts::Real2d(3, 5));
    _numtest(m0(0,0), 3);  _numtest(m0(0,1), 6);
    _numtest(m0(1,0), 15);  _numtest(m0(1,1), 20);
  }

  void MappingTest::testScaleColumns() {
    concepts::MapReal2d m0(1, 2, 3, 4);
    m0.scaleCols(concepts::Real2d(3, 5));
    _numtest(m0(0,0), 3);  _numtest(m0(0,1), 10);
    _numtest(m0(1,0), 9);  _numtest(m0(1,1), 20);
  }

  void MappingTest::testRank1Prod() {
    concepts::MapReal3d m1(0);
    concepts::Real3d u(1, 2, 3), v(4, 5, 6);
    m1.rank1Product(u,v);
    _numtest(m1(0,0), 4);   _numtest(m1(0,1), 5);   _numtest(m1(0,2), 6);
    _numtest(m1(1,0), 8);   _numtest(m1(1,1), 10);  _numtest(m1(1,2), 12);
    _numtest(m1(2,0), 12);  _numtest(m1(2,1), 15);  _numtest(m1(2,2), 18);
  }

}

// ****************************************************************** main() **

int main(int, char**) {
  std::cout << std::setprecision(20);

  test::TestSuite t("VectorMatrices");

  test::PointTest t1;        t.addTest(&t1);
  test::MappingTest t2;      t.addTest(&t2);

  t.run();
  return t.report();

  using namespace concepts;

  try {
    Real2d p0(2.0), p1(3.0);
    std::cout << "p0 = " << p0 << std::endl;
    std::cout << "p1 = " << p1 << std::endl;
    std::cout << "p0 * p1 = " << p0*p1 << std::endl;
    Real2d p5;
    p5.lincomb(p0, p1);
    std::cout << "p5 = p0 + p1 = " << p5 << std::endl;
    p5.ortho(p0);
    std::cout << "p0 ortho = p5 = " << p5 << std::endl;

    Real3d p2(4.0);
    std::cout << "p2 = " << p2 << std::endl;
    std::cout << p0 << " * " << p2 << "  = " 
  	      << p0*(Real2d)p2 << std::endl;
    std::cout << p2 << " * " << p0 << "  = " 
  	      << p2*(Real3d)p0 << std::endl;

    Real3d p3(5.0);
    std::cout << p2 << " ^ " << p3 << " = " << (p2^p3) << std::endl;
    std::cout << p0 << " ^ " << p1 << " = " << p0.cross(p1) << std::endl;

    MapReal2d m0(1.0, 2.0, 3.0, 4.0);
    std::cout << "m0 = " << m0 << std::endl;
    std::cout << "det(m0) = " << m0.determinant() << std::endl;
    std::cout << "inv(m0) = " << m0.inverse() << std::endl;
    std::cout << "m0 * p0 = " << m0*p0 << std::endl;

    MapReal3d m1(1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 10.0);
    std::cout << "m1 = " << m1 << std::endl;
    std::cout << "m1 * p2 = " << m1*p2 << std::endl;
    std::cout << "det(m1) = " << m1.determinant() << std::endl;
    std::cout << "inv(m1) = " << m1.inverse() << std::endl;
    std::cout << "transpose(m1) = " << m1.transpose() << std::endl;

    MapReal3d m3(1.3, -2.0, 3.7, 2.5, 5.0, 6.9, 7.0, 3.0, -10.0);
    std::cout << "m3 = " << m3 << std::endl;
    for (uint i = 0; i < 3; ++i)
      for (uint j = 0; j < 3; ++j)
	std::cout << "m3(" << i << ", " << j << ") = " << m3(i,j) << std::endl;
    std::cout << "m3 * p2 = " << m3*p2 << std::endl;
    std::cout << "m1 * m3 = " << m1*m3 << std::endl;
    std::cout << "p3 = " << p3 << std::endl;
    Real3d p4;
    m3(p3, p4);
    std::cout << "p4 = m3 * p3 = " << p4 << std::endl;
    std::cout << "m3^T = " << m3.transpose() << std::endl;
    m3.transpose().operator()(p3, p4);
    std::cout << "p4 = m3^T * p3 = " << p4 << std::endl;
    m3.mapTranspose(p3, p4);
    std::cout << "p4 = m3^T * p3 = " << p4 << std::endl;
    std::cout << "|p4|^2 = " << p4.l2_2() << ", |p4| = " << p4.l2()
	      << std::endl;
    p3.max(p2, p4);
    std::cout << "p3 = max(p2, p4) = " << p3 << std::endl;
    std::cout << "det(m3) = " << m3.determinant() << std::endl;
    std::cout << "inv(m3) = " << m3.inverse() << std::endl;
    std::cout << "p2 + p3 = " << p2 + p3 << std::endl;
    std::cout << "3.0*p3 = " << p3*3.0 << std::endl;

    MapReal2d m2(3.0, 2.0, 5.0, 1.0);
    std::cout << "m2 = " << m2 << std::endl;
    for (uint i = 0; i < 2; ++i)
      for (uint j = 0; j < 2; ++j)
	std::cout << "m2(" << i << ", " << j << ") = " << m2(i,j) << std::endl;
    std::cout << "1st column of m2 = " << m2.column(0) << std::endl;
    std::cout << "3rd column of m3 = " << m3.column(2) << std::endl;
    std::cout << "m0 * m2 = " << m0*m2 << std::endl;
    std::cout << "inv(m0) = " << m0.inverse() << std::endl;
    std::cout << "inv(m2) = " << m2.inverse() << std::endl;

    m1 = MapReal3d(1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 10.0);
    std::cout << "m1 = " << m1 << ", rows scaled: ";
    std::cout << m1.scaleRows(Real3d(1., 2., 3.))  << std::endl;
    std::cout << "m1 = " << m1 << ", columns scaled: ";
    std::cout << m1.scaleCols(Real3d(1., 2., 3.))  << std::endl;

    m1 = MapReal3d(0);
    Real3d u(1.0, 2.0, 3.0), v(4.0, 5.0, 6.0);
    std::cout << "m1 = " << m1 << std::endl;
    m1.rank1Product(u,v);
    std::cout << "m1 = " << m1 << std::endl;

    std::cout << "p0 = " << p0 << std::endl;
    m1(p0, p1);
    std::cout << "m1 * p0 = " << p1 << std::endl;

    std::cout << "m2 = " << m2 << std::endl;
    std::cout << "p2 = " << p2 << std::endl;
    m2(p2, p3);
    std::cout << "m2 * p2 = " << p3 << std::endl;
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
