/* vectors and matrices for 2D and 3D
 */

#ifndef vectMatr_hh
#define vectMatr_hh

#include <iostream>
#include <cmath>
#include <cstring>

#include "functionCompiler.hh"
#include "exceptions.hh"
#include "typedefs.hh"
#include "vectorsMatricesForward.hh"
#include "debug.hh"

// debugging Mapping
#define MappingAppll_D 0

namespace concepts {

  /// Copies \c n entries from \c src to \c dest (faster than \c std::memcpy)
  template<typename F>
  void memorycpy(F *dest, const F *src, size_t n) {
    for (size_t i = 0; i < n; ++i) *dest++ = *src++;
  }

  // ***************************************************************** Point **

  /** Basic class for a Point or a vector. The template parameters
      make it possible to use this class for 2D and 3D Points and
      vectors with real or complex valued entries.

      @test test::PointTest
      @test test::MappingTest
      @author Philipp Frauenfelder, 2001
  */
  template<class F, int dim>
  class Point {
  public:
    /** Constructor.
	Initializes all elements to 0.
    */
    Point() { F* d = data_; for(uint i = dim; i--;) *d++ = 0; }
    /// Copy constructor
    Point(const Point<F, dim>& p) {
      memorycpy(data_, p.data_, dim);
    }
    /// Copy constructor
    template<int dim2>
    Point(const Point<F, dim2>& p) {
      if (dim2<dim) {
	F* d = data_+dim2; for(uint i = dim-dim2; i--;) *d++ = 0;
      }
      memorycpy(data_, (const F*)p, ((dim < dim2) ? dim : dim2));
    }
    /** Constructor.
	All elements are initialized to x.
     */
    Point(F x) { F* d = data_; for(uint i = dim; i--;) *d++ = x; }
    /// Constructor for a 2D or a 3D Point
    Point(const F e0, const F e1, const F e2 = 0) {
      data_[0] = e0;
      if (dim > 1)
	data_[1] = e1;
      if (dim > 2)
	data_[2] = e2;
    }
    /// Constructor with a processed map
    Point(const uchar *pgm, const Real x, const Real y, const Real z = 0.0);

    /// Assignement operator
    template<int dim2>
    Point<F, dim>& operator=(const Point<F, dim2>& b) {
      if (this != &b)
	memorycpy(data_, b.data_, ((dim < dim2) ? dim : dim2));
      return *this;
    }

    /// Index operator
    F operator[](const int i) const {
      conceptsAssert(i < dim, concepts::Assertion());
      return data_[i]; }
    /// Index operator
    F& operator[](const int i) {
      conceptsAssert(i < dim, concepts::Assertion());
      return data_[i]; }

    /// Returns a pointer to the data
    operator F*() { return data_; }
    /// Returns a pointer to the data
    operator const F*() const { return data_; }

    /// Addition operator
    Point<F, dim>& operator+=(const Point<F, dim>& p) {
      F* d = data_; const F* dp = p.data_;
      for(uint i = dim; i--;) *d++ += *dp++;
      return *this;
    }
    /// Addition operator
    Point<F, dim> operator+ (const Point<F, dim>& p) const {
      Point<F, dim> res(*this);
      return res += p;
    }      
    /// Subtraction operator
    Point<F, dim>& operator-=(const Point<F, dim>& p) {
      F* d = data_; const F* dp = p.data_;
      for(uint i = dim; i--;) *d++ -= *dp++;
      return *this;
    }
    /// Subtraction operator
    Point<F, dim> operator- (const Point<F, dim>& p) const {
      Point<F, dim> res(*this);
      return res -= p;
    }      

    /// Inner product
    F operator*(const Point<F, dim>& b) const;
    /// Scaling operator
    Point<F, dim>& operator*=(const F n) {
      F* d = data_; for(uint i = dim; i--;) *d++ *= n;
      return *this;
    }
    /// Scaling operator
    Point<F, dim> operator*(const F n) const {
      Point<F, dim> res(*this);
      return res *= n;
    }

    /// Outer product (for 2D)
    F cross(const Point<F, dim>& b) const {
      return data_[0]*b[1] - data_[1]*b[0];
    }
    /// Outer product (for 3D)
    Point<F, dim> operator^(const Point<F, dim>& b) const;

    /// Returns the Euclidian norm of the vector
    Real l2() const { return std::sqrt(l2_2()); }
    /// Returns the square of the Euclidian norm of the vector
    Real l2_2() const;

    /// Assign the vector the linear combination of a and b
    void lincomb(const Point<F, dim>& a, const Point<F, dim>& b,
		 const F ca = 1.0, const F cb = 1.0);

    /// Sets the vector to the elementwise maximum of a and b
    void max(const Point<F, dim>& a, const Point<F, dim>& b);

    /// Sets the vector to the elementwise minimum of a and b
    void min(const Point<F, dim>& a, const Point<F, dim>& b);

    /// Rotates the vector by 90 degrees (clockwise) (only 2D)
    void ortho(const Point<F, dim>& a);

    std::ostream& info(std::ostream& os) const;
  private:
    F data_[dim];
  };

  template<class F, int dim>
  std::ostream& operator<<(std::ostream& os, const Point<F, dim>& p) {
    return p.info(os);
  }

  // **************************************************************** UnitNd **

  /** A vector of dimension dim and length 1.

      @author Gregor Schmidlin, 2002
   */
  template <int dim>
  class UnitNd : public Point<Real, dim> {
    Real len_;

  public:
    /// Constructor. Initializes all elements to 0.
    UnitNd() : Point<Real, dim>() {len_ = 0.0;}
    /// Constructor for a point
    inline UnitNd(const Point<Real, dim>& u);
    /// Constructor for at most 3D
    inline UnitNd(Real x, Real y, Real z);

    /// Length of the initially given vector
    Real length() const {return len_;}
  };

  template <int dim>
  inline UnitNd<dim>::UnitNd(const Point<Real, dim>& u)
    : Point<Real, dim>(u) {
    len_ = this->l2();  *this *= (1.0 / len_);
  }

  template <int dim>
  inline UnitNd<dim>::UnitNd(Real x, Real y, Real z)
    : Point<Real, dim>(x, y, z) {
    len_ = this->l2();  *this *= (1.0 / len_);
  }

  // *************************************************************** Mapping **

  /** Basic class for a 2D or 3D map. The template parameters make it possible
      to use this class for real or complex values. The main use is to
      represent linear maps in 2D or 3D.

      @test test::MappingTest
      @author Philipp Frauenfelder, 2001
  */
  template<class F, int dim>
  class Mapping {
  public:
    /** Default constructor.
	The elements of the matrix are left uninitialized.
     */
    Mapping() {}
    /// Copy constructor
    Mapping(const Mapping<F, dim>& m) {
      memorycpy(data_, m.data_, dim*dim);
    }
    /// Constructor for a 2D Mapping
    Mapping(const F e0, const F e1,
	    const F e2, const F e3);
    /** Constructor for a 2D Mapping.
	@param first First column of the matrix
	@param second Second column of the matrix
    */
    Mapping(const Point<F, dim> first, const Point<F, dim> second);
    /// Constructor for a 3D Mapping
    Mapping(const F e0, const F e1, const F e2,
	    const F e3, const F e4, const F e5,
	    const F e6, const F e7, const F e8);
    /** Constructor for a 3D Mapping. The three vectors entered into the
	columns of the matrix.
    */
    Mapping(const Point<F, dim> first,
	    const Point<F, dim> second,
	    const Point<F, dim> third);
    /// Constructor for a 6D Mapping
    Mapping(const F e0, const F e1, const F e2,
	    const F e3, const F e4, const F e5,
	    const F e6, const F e7, const F e8, 
	    const F e9, const F e10, const F e11, 
	    const F e12, const F e13, const F e14, 
	    const F e15, const F e16, const F e17, 
	    const F e18, const F e19, const F e20, 
	    const F e21, const F e22, const F e23, 
	    const F e24, const F e25, const F e26, 
	    const F e27, const F e28, const F e29, 
	    const F e30, const F e31, const F e32, 
	    const F e33, const F e34, const F e35);

    /** Constructor for rotation
	@param n Normal vector of the rotation
    */
    Mapping(const UnitNd<dim>& n);

    /// Returns an entry of the matrix
    const F& operator()(uint i, uint j) const {
      conceptsAssert(i < dim, Assertion());
      conceptsAssert(j < dim, Assertion());
      DEBUGL(MappingAppll_D, '(' << i << ", " << j << ") is " << data_[i*dim+j]
	     << " at " << &(data_[i*dim+j]));
      return data_[i*dim+j];
    }
    /** Constructor.
	All elements are initialized to x.
     */
    Mapping(F x) { F* d = data_; for(uint i = dim*dim; i--;) *d++ = x; }

    /// Returns a mapped Point
    Point<F, dim> operator*(const Point<F, dim>& p) const;
    /// Multiplication operator
    Mapping<F, dim> operator*(const Mapping<F, dim>& m) const;
    /// Scaling operator
    Mapping<F, dim>& operator*=(const F n);
    /// Scaling operator
    Mapping<F, dim> operator*(const F n) const {
      Mapping<F, dim> res(*this);
      return res *= n;
    }
    /// Addition operator
    Mapping<F, dim>& operator+=(const Mapping<F, dim>& M);
    /// Scales the rows with the respective entries in \c s
    Mapping<F, dim>& scaleRows(const Point<F, dim>& s);
    /// Scales the columns with the respective entries in \c s
    Mapping<F, dim>& scaleCols(const Point<F, dim>& s);
    /// Returns the determinant of the matrix
    F determinant() const;
    /// Returns the inverse of the matrix
    Mapping<F, dim> inverse() const;
    /// Returns the transpose of the matrix
    Mapping<F, dim> transpose() const;

    /// Returns a column of the matrix
    Point<F, dim> column(const uint i) const;

    /** Application operator. Computes x = A y where A is the matrix.
	If the dimensions do not match the matrix A is extended by
	the identity matrix.
     */
    template<int dimy, int dimx>
    void operator()(const Point<F, dimy>& y, Point<F, dimx>& x) const;

    /// Fills the matrix with zeros
    void zeros() { F* d = data_;
      for (uint i = 0; i < dim*dim; ++i) *d++ = 0; }

    /// Computes x = A<sup>T</sup> y
    void mapTranspose(const Point<F, dim>& y, Point<F, dim>& x) const;

    /// Computes x y<sup>T</sup> and adds the result to the matrix
    void rank1Product(const Point<F, dim> x,
		      const Point<F, dim> y);

    std::ostream& info(std::ostream& os) const;
  private:
    F data_[dim*dim];
  };

  template<class F, int dim>
  inline void Mapping<F, dim>::mapTranspose(const Point<F, dim>& y,
					    Point<F, dim>& x) const {
    F* dr = (F*)x;
    for(uint i = 0; i < dim; i++, dr++) {
      const F* dp = (const F*)y;
      const F* dm = data_ + i;
      *dr = 0;
      for(uint j = dim; j--;) {
	*dr += (*dm)*(*dp++);
	dm += dim;
      }
    }
  }

  template<class F, int dim>
  std::ostream& operator<<(std::ostream& os, const Mapping<F, dim>& m) {
    return m.info(os);
  }

} // namespace concepts

#endif // vectMatr_hh
