// fabric for eigensolver ARPACK

// #undef HAS_PARDISO

#if __GNUC__ == 2
#  include <float.h>
#  define EPS DBL_EPSILON
#else
#  include <limits>
#  define EPS std::numeric_limits<double>::epsilon()
#endif
#include "arpackFabric.hh"
#include "basics/exceptions.hh"
#include "operator/compositions.hh"
#include "operator/cg.hh"
#include "operator/sparseMatrix.hh"
#ifdef HAS_SuperLU
#include "operator/superLU.hh"
#endif
#ifdef HAS_UMFPACK
#include "operator/umfpack.hh"
#endif
#ifdef HAS_PARDISO
#include "operator/pardiso.hh"
#endif

// debugging
#include "basics/debug.hh"

#define ARPACKDestr_D 0
#define ARPACKAppl_D 0

namespace eigensolver {

  // ********************************************************** ArPackFabric **

  ArPackFabric::~ArPackFabric() {
    DEBUGL(ARPACKDestr_D, "done.");
  }

  ArPack& ArPackFabric::operator()(concepts::Operator<Real>& stiff,
				   concepts::Operator<Real>& mass) {
    switch (mode_) {
    case ArPack::NORMAL: normalMode_(stiff, mass); break;
    case ArPack::REGINV: reginvMode_(stiff, mass); break;
    case ArPack::SHIFTINV: shiftinvMode_(stiff, mass); break;
    case ArPack::BUCKLING: bucklingMode_(stiff, mass); break;
    case ArPack::CAYLEY: cayleyMode_(stiff, mass); break;
    }
    return *solver_;
  }

  std::ostream& ArPackFabric::info(std::ostream& os) const {
    if (solver_.get())
      return os << *solver_;
    else
      return os << solver_.get();
  }

  void ArPackFabric::normalMode_(concepts::Operator<Real>& stiff,
				 concepts::Operator<Real>& mass) {
    throw conceptsException
      (concepts::MissingFeature("normal mode not implemented"));
  }

  void ArPackFabric::reginvMode_(concepts::Operator<Real>& stiff,
				 concepts::Operator<Real>& mass) {
#ifdef HAS_SuperLU
    try {
      concepts::SparseMatrix<Real> tmpMatrix(mass);
      DEBUGL(ARPACKAppl_D, "mass matrix is " << mass);
      OP_.reset(new concepts::SuperLU(tmpMatrix));
    }
    catch (concepts::ExceptionBase& e) {
#endif
      OP_.reset(new concepts::CG<Real>(mass, EPS));
#ifdef HAS_SuperLU
    }
#endif
    DEBUGL(ARPACKAppl_D, "OP = " << *OP_);
    DEBUGL(ARPACKAppl_D, "A = " << stiff);
    DEBUGL(ARPACKAppl_D, "B = " << mass);
    solver_.reset(new ArPack
		  (*OP_, stiff, mass, kmax_, tol_, maxiter_, target_,
		   mode_, sigma_, start_));
  }

  void ArPackFabric::shiftinvMode_(concepts::Operator<Real>& stiff,
				   concepts::Operator<Real>& mass) {
    tmp1_.reset(new concepts::LiCo<Real>(stiff, mass, 1.0, -sigma_));
#ifdef HAS_PARDISO
    try {
      concepts::SparseMatrix<Real> tmpMatrix
	( *(dynamic_cast<concepts::LiCo<Real>*>(tmp1_.get())) );
      DEBUGL(ARPACKAppl_D, "sparse matrix = " << tmpMatrix);
      tmpMatrix.symmetrize();
      OP_.reset(new concepts::Pardiso(tmpMatrix,
				      concepts::Pardiso::SYMM_INDEF));
    }
    catch (concepts::MissingFeature& e) {
      DEBUGL(1, "could not convert mass to SparseMatrix -> using CG instead"
	     " of Pardiso");
      OP_.reset(new concepts::CG<Real>(*tmp1_, EPS));
    }
#else
#  ifdef HAS_UMFPACK
    try {
      concepts::SparseMatrix<Real> tmpMatrix
	( *(dynamic_cast<concepts::LiCo<Real>*>(tmp1_.get())) );
      DEBUGL(ARPACKAppl_D, "sparse matrix = " << tmpMatrix);
      tmpMatrix.symmetrize();
      OP_.reset(new concepts::Umfpack(tmpMatrix, true));
    }
    catch (concepts::MissingFeature& e) {
      DEBUGL(1, "could not convert mass to SparseMatrix -> using CG instead"
	     " of Umfpack");
      OP_.reset(new concepts::CG<Real>(*tmp1_, EPS));
    }
#  else
#    ifdef HAS_SuperLU
    try {
      concepts::SparseMatrix<Real> tmpMatrix
	( *(dynamic_cast<concepts::LiCo<Real>*>(tmp1_.get())) );
      DEBUGL(ARPACKAppl_D, "sparse matrix = " << tmpMatrix);
      OP_.reset(new concepts::SuperLU(tmpMatrix));
    }
    catch (concepts::MissingFeature& e) {
      DEBUGL(1, "could not convert mass to SparseMatrix -> using CG instead"
	     " of SuperLU");
      OP_.reset(new concepts::CG<Real>(*tmp1_, EPS));
    }
#    else
    OP_.reset(new concepts::CG<Real>(*tmp1_, EPS));
#    endif
#  endif
#endif
    DEBUGL(ARPACKAppl_D, "OP = " << *OP_);
    DEBUGL(ARPACKAppl_D, "A = " << stiff);
    DEBUGL(ARPACKAppl_D, "B = " << mass);
    solver_.reset(new ArPack
		  (*OP_, stiff, mass, kmax_, tol_, maxiter_, target_,
		   mode_, sigma_, start_));
  }

  void ArPackFabric::bucklingMode_(concepts::Operator<Real>& stiff,
				   concepts::Operator<Real>& mass) {
    throw conceptsException
      (concepts::MissingFeature("Buckling mode not implemented"));
  }

  void ArPackFabric::cayleyMode_(concepts::Operator<Real>& stiff,
				 concepts::Operator<Real>& mass) {
    throw conceptsException
      (concepts::MissingFeature("Cayley mode not implemented"));
  }

} // namespace eigensolver
