// solves a constrained eigenvalue problem

#ifndef constrainedEV_hh
#define constrainedEV_hh

#include <memory>
#include <cstdlib>
#include "eigens.hh"
#include "solverFabric.hh"
#include "space/space.hh"
#include "operator/compositions.hh"
#include "operator/sparseMatrix.hh"
#include "operator/permutation.hh"
#include "sparseqr/givensRotations.hh"

namespace eigensolver {

  // forward declaration
  template<typename F, bool extend>
  class TrivExtendRestrict;

  // *********************************************************** Constrained **

  /** Solves a generalized eigenvalue problem subject to
      linear, homogeneous constraints.

      This class solve a generalized eigenvalue problem of the form
      \f[ A \vec x = \lambda M \vec x \f]
      subject to the constraints \f$ C^\top \vec x = g = 0 \f$ where
      \f$ C \in R^{n \times m} \f$ and \f$ A, M \in R^{n \times n} \f$.
      This is important for solving eigenvalue problems with essential
      boundary conditions like \f$ \vec u \times \vec n = 0 \f$.

      For the case \f$ g = 0 \f$, this is described in [1 section 12.6, 2].
      For the much more complicated case of \f$ g \neq 0 \f$, see [3, 4].
      The first case should be sufficient for most applications.

      @see [1] Gene H. Golub and Charles F. van Loan, <i>Matrix
           Computations,</i> The Johns Hopkins University Press, 1983.<br>
	   [2] Gene H. Golub, <i>Some Modified Matrix Eigenvalue Problems,</i>
	   SIAM Review, <b>15</b> (1973), No. 2, 318--334.<br>
	   [3] Walter Gander, Gene H. Golub and Urs von Matt,
	   <i>A Constrained Eigenvalue Problem,</i>
	   in Numerical Linear Algebra, Digital Signal Processing
	   and Parallel Algorithms, Gene H. Golub and P. Van Dooren (eds.),
	   NATO ASI Series, Vol. F70, 1991, 677--686.<br>
	   [4] Walter Gander, Gene H. Golub and Urs von Matt,
	   <i>A Constrained Eigenvalue Problem,</i>
	   Linear Algebra and Its Applications, <b>114/115</b> (1989),
	   815--839.

      @test test::ConstrainedEVP
      @author Philipp Frauenfelder, 2002
  */
  class Constrained : public EigenSolver {
  public:
    /** Constructor.
	@param A Matrix
	@param M Matrix
	@param C Matrix of constraints
	@param solver Fabric for the eigenvalue solver
    */
    Constrained(concepts::Operator<Real>& A, concepts::Operator<Real>& M,
		const concepts::SparseMatrix<Real>& C, SolverFabric& solver) :
      solver_(solver), computed_(false), A_(A), M_(M), C_(C),
      lambda_(0), eigenvectors_(0), ev_(0), kmax_(0),
      Pr(0), Prt(0), Qgt(0), Qg(0), ext(0), restr(0), PrQgt(0),
      QgPrt(0), restrPrQgt(0), QgPrtext(0), restrPrQgtA(0),
      restrPrQgtAQgPrtext(0), restrPrQgtM(0), restrPrQgtMQgPrtext(0),
      it_(0), k_conv_(0) {}
    virtual ~Constrained();
    virtual const concepts::Array<Real>& getEV();
    virtual const concepts::Array<concepts::Vector<Real>*>& getEF();
    virtual uint iterations() const { return it_; }
    virtual uint converged() const { return k_conv_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Fabric for the eigenvalue solver
    SolverFabric& solver_;

    /** The real compute routine. Fills \c lambda_ and \c ev_
	@see computed_
     */
    void compute_();
    /// If false, the code will have to compute, otherwise not
    bool computed_;

    /// Stiffness matrix
    concepts::Operator<Real>& A_;
    /// Mass matrix
    concepts::Operator<Real>& M_;
    /// Matrix of constraints. \f$ C^\top \vec x = g = 0 \f$ is enforced.
    const concepts::SparseMatrix<Real>& C_;

    /// Storage space for eigenvalues
    concepts::Array<Real> lambda_;
    /// Storage space for eigenvectors
    concepts::Array<Real> eigenvectors_;
    /// References into storage for eigenvectors
    concepts::Array<concepts::Vector<Real>*> ev_;

    /// Maximal number of eigenvalues which can be computed
    uint kmax_;

    concepts::Permutation<Real>* Pr;
    std::auto_ptr<concepts::Permutation<Real> > Prt;
    sparseqr::GivensRotations<Real>* Qgt, * Qg;
    concepts::TrivExtendRestrict<Real>* ext, * restr;
    std::auto_ptr<concepts::Compose<Real> > PrQgt, QgPrt, restrPrQgt,
      QgPrtext, restrPrQgtA, restrPrQgtAQgPrtext, restrPrQgtM,
      restrPrQgtMQgPrtext;

    /// Statistics
    uint it_, k_conv_;
  };

} // namespace eigensolver

#endif // constrainedEV_hh
