// jdbsym eigensolver

#include <iostream>
#include <cstring>

#include "jdbsym.hh"
#include "operator/compositions.hh"
#include "operator/matrix.hh"
#include "operator/sparseMatrix.hh"
#include "operator/solverFabric.hh"
#include "basics/exceptions.hh"

// jdbsym include files
extern "C" {
#include "include/blas.h"
#include "include/fortran.h"
#include "include/lapack.h"
#include "src/jdbsym.h"
#include "jdqrgen/jdqrgen.h"
#include "src/OrthoPack.h"
}

// debugging
#include "basics/debug.hh"

#define JDBSymDestr_D 0
#define JDBSymCompute_D 0
#define matvecb_D 0

namespace eigensolver {

  // ****************************************** matrix vector multiplication **

  concepts::Operator<Real>* stiffness_matrix, * mass_matrix, * preconditioner;
  uint* mass_matrix_counter, * stiffness_matrix_counter;

  // stiffness matrix
  void matveca(double* x, double* y) {
    const concepts::Vector<Real> X(stiffness_matrix->spaceY(), x);
    concepts::Vector<Real> Y(stiffness_matrix->spaceY(), y);
    (*stiffness_matrix)(X, Y);
    ++(*stiffness_matrix_counter);
  }

  // mass matrix
  void matvecb(double* x, double* y) {
    const concepts::Vector<Real> X(mass_matrix->spaceY(), x);
    concepts::Vector<Real> Y(mass_matrix->spaceY(), y);
    DEBUGL(matvecb_D, "x = " << X);
    DEBUGL(matvecb_D, "y = " << Y);
    (*mass_matrix)(X, Y);
    ++(*mass_matrix_counter);
    DEBUGL(matvecb_D, "y = " << Y);
  }

  // ******************************************************* preconditioners **
  Real* diag_elements;

  // preconditioner
  void precond(double* x, double* y) {
    const concepts::Vector<Real> X(preconditioner->spaceY(), x);
    concepts::Vector<Real> Y(preconditioner->spaceY(), y);
    (*preconditioner)(X, Y);
  }

  // empty preconditioner
  void identity(double* x, double* y) {
    std::memcpy(y, x, mass_matrix->spaceY().dim()*sizeof(double));
  }

  // **************************************************************** JdbSym **

  int JdbSym::jdbsymLock_ = 0;

  JdbSym::JdbSym(concepts::Operator<Real>& stiff,
		 concepts::Operator<Real>& mass,
		 Real tol, uint maxit, Real tau, uint jdtype, uint kmax,
		 concepts::SolverFabric<Real>* fabric,
		 const concepts::Array<concepts::Vector<Real>*>* start) :
    computed_(false), stiff_(stiff), mass_(mass), tol_(tol),
    maxit_(maxit), tau_(tau), jdtype_(jdtype), kmax_(kmax), it_(0), k_conv_(0),
    lambda_(0), eigenvectors_(0), ev_(0), stiffCounter_(0), massCounter_(0),
    fabric_(fabric), start_(start) {
    conceptsAssert(stiff.spaceX().dim() == stiff.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(mass.spaceX().dim() == mass.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(mass.spaceX().dim() == stiff.spaceX().dim(),
		   concepts::Assertion());
  }

  JdbSym::~JdbSym() {
    DEBUGL(JDBSymDestr_D, "start.");
    if (computed_)
      for (uint i = 0; i < kmax_; ++i)
	delete ev_[i];
    DEBUGL(JDBSymDestr_D, "done.");
  }

  const concepts::Array<Real>& JdbSym::getEV() {
    if (!computed_) {
      compute_();
    }
    return lambda_;
  }

  const concepts::Array<concepts::Vector<Real>*>& JdbSym::getEF() {
    if (!computed_) {
      compute_(); // eigenvectors_ is filled
    }
    return ev_;
  }

  void JdbSym::compute_() {
    // allocate space for eigenvalues in lambda_
    lambda_.resize(kmax_);
    Real* lambd = (Real*)lambda_; // the array to insert into jdbsym call

    eigenvectors_.resize(kmax_*stiff_.spaceX().dim());
    Real* q = (Real*)eigenvectors_; // the array to insert into jdbsym call

    int N = stiff_.spaceX().dim();   
 
    if (jdbsymLock_ == 0) {
      jdbsymLock_ = 1;

      stiffness_matrix = &stiff_;
      stiffness_matrix_counter = &stiffCounter_;
      mass_matrix = &mass_;
      mass_matrix_counter = &massCounter_;

      void (*matvecb_fun)(double *, double *);
      matvecb_fun = matvecb;

      // preconditioner
      void (*prec_fun)(double *, double *);
      concepts::LiCo<Real> toPreCond(stiff_, mass_, 1.0, -tau_);
      std::auto_ptr<concepts::Operator<Real> > precondOp(0);
      if (fabric_) {
	try {
	  concepts::SparseMatrix<Real> tmpMatrix(toPreCond);
	  precondOp.reset((*fabric_)(tmpMatrix));
	  preconditioner = precondOp.get();
	  prec_fun = precond;
	}
	catch (concepts::MissingFeature& e) {
	  DEBUGL(1, "could not convert to SparseMatrix -> no precond instead"
		 " of given solver");
	  prec_fun = 0;
	}
      } else
	prec_fun = 0;

      //default parameter values for the JDBSYM routine
      int strategy_ = 0;
      double eps_tr_ = 1e-4, toldecay_ = 2.0;
      int blksize_ = 1, blkwise_ = 0, jmax_ = (2*kmax_>(uint)N)?kmax_:2*kmax_,
	jmin_ = ((uint)jmax_>kmax_)?kmax_:jmax_/2, mvmax_ = 200; 
      int linsolver_ = 1, optype_ = OP_UNSYM1, clvl_ = 0,
	Bisempty_ = 0;

      // starting values
      concepts::Array<double> v0(0);
      int v0dim = 0;
      if (start_) {
	DEBUGL(JDBSymCompute_D, start_->size() << " starting vectors. jmax = "
	       << jmax_ << ", kmax_ = " << kmax_ << ", blksize = "
	       << blksize_);
	v0dim = (start_->size() < (uint)jmax_) ? start_->size() : jmax_;
	v0.resize(v0dim*N);
	v0.zeros();
	DEBUGL(JDBSymCompute_D, "v0dim = " << v0dim);

	double* v0Acc = v0;
	for (int j = 0; j < v0dim; ++j) {
	  DEBUGL(JDBSymCompute_D, j << "th starting vector is "
		 << (*(*start_)[j]));
	  double* vVecAcc = (*(*start_)[j]);
	  for (int k = 0; k < N; ++k)
	    *v0Acc++ = *vVecAcc++;
	}
	DEBUGL(JDBSymCompute_D, "v0 = " << v0);
      }

      DEBUGL(JDBSymCompute_D, "jdtype = " << jdtype_);
      DEBUGL(JDBSymCompute_D, "N = " << N << ", kmax = " << kmax_);
      switch (jdtype_) {
      case 1:
        jdbsym(N, tau_, tol_,
 	       kmax_, jmax_, jmin_, maxit_,
 	       blksize_, blkwise_,
 	       v0dim, v0,
 	       linsolver_, optype_,
 	       mvmax_, eps_tr_, toldecay_, strategy_,
 	       clvl_,
 	       &k_conv_, q, lambd, &it_,
 	       &matveca, 
 	       Bisempty_?NULL:matvecb_fun, 
	       prec_fun);
      break;
//       case 2:
//         jdqrgen(N, tau_, 1, V0_, N, tol_,  
//                 kmax_, jmax_, jmin_, maxit_, mvmax_,
//                 q, N, lambd, &it_,
//       	        eps_tr_, toldecay_, 0, clvl_,
//       	        &matveca, Bisempty_?NULL:matvecb_fun, Kisempty_?NULL:prec_fun);
//         for (uint i = 0; i < kmax_; ++i) {
//           s = -lambd[i];
//           matveca(q + i*N, V0_);
//           matvecb_fun(q + i*N, diag_elements);
//           F77(daxpy)(&N, &s, diag_elements, &ONE_, V0_, &ONE_);
// 	}
//       break;
      default:
	conceptsAssert(false, concepts::Assertion());
      }    

      if (ev_.size() == 0) {
	ev_.resize(kmax_);
	Real* q = (Real*)eigenvectors_;
	const uint n = stiff_.spaceX().dim();
	for (uint i = 0; i < kmax_; ++i) {
	  ev_[i] = new concepts::Vector<Real>(stiff_.spaceX(), q+i*n);
	}
      }

      jdbsymLock_ = 0;
      computed_ = true;
    }
  }

  std::ostream& JdbSym::info(std::ostream& os) const {
    return os << "JdbSym(" << (jdtype_ ? "jdbsym" : "jdqrgen") << ", tol = "
	      << tol_ << ", tau = " << tau_ << ", conv. eigenpairs = "
	      << k_conv_ << "(max = " << kmax_ << "), iter = " << it_
	      << "(max = " << maxit_ << "), " << stiffCounter_
	      << " stiff*x ops., " << massCounter_ << " mass*x ops.)";
  }
}
