// fabric for eigensolver JDBSym

#ifndef jdbsymFabric_hh
#define jdbsymFabric_hh

#include <memory>
#include "jdbsym.hh"
#include "solverFabric.hh"

namespace eigensolver {

  // ********************************************************** JdbSymFabric **

  /** Fabric class for JdbSym eigenvalue solver for a constrained problem.
      @author Philipp Frauenfelder, 2002
      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
  */
  class JdbSymFabric : public SolverFabric {
  public:
    /** Constructor.
	@param fabric Solver fabric for a linear solver (preconditioner for
	              the shifted stiffness matrix)
        @param tol Convergence tolerance for the eigenpairs. For a pair
	           \f$ (\lambda, u) \f$ convergence is defined by
		   \f$ \| A u - \lambda B u \| < tol \f$
        @param tau Target value of Jacobi-Davidson algorithm. The code will
	           find the \c kmax eigenvalues closest to \c tau.
        @param jdtype Type of solver required. An older solver and a newer
	              solver are possible to use.
        @param kmax Number of eigenpairs to be computed
	@param start Starting vectors. Used to build the initial search
	             subspace
    */
    JdbSymFabric(Real tol, uint maxit = 150, Real tau = 0.0,
		 uint jdtype = 1, uint kmax = 1,
		 concepts::SolverFabric<Real>* fabric = 0,
		 const concepts::Array<concepts::Vector<Real>*>* start = 0) :
      solver_(0), tol_(tol), tau_(tau), maxit_(maxit), jdtype_(jdtype),
      kmax_(kmax), fabric_(fabric), start_(start) {}
    virtual ~JdbSymFabric();
    virtual JdbSym& operator()(concepts::Operator<Real>& stiff,
			       concepts::Operator<Real>& mass);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Real eigenvalue solver
    std::auto_ptr<JdbSym> solver_;
    Real tol_, tau_;
    uint maxit_, jdtype_, kmax_;
    concepts::SolverFabric<Real>* fabric_;
    const concepts::Array<concepts::Vector<Real>*>* start_;
  };

} // eigensolver

#endif // jdbsymFabric_hh
