// testsuite for JDBSYM eigensolver

#include <limits>
#include <iostream>
#include <unistd.h> // for command line parsing
#include "space/space.hh"
#include "operator/sparseMatrix.hh"
#include "eigensolver/jdbsym.hh"

using concepts::Real;

int main(int argc, char** argv) {

  try {
    uint dim = 10;
    int kmax_ = 4;
    int opt;
    while ((opt = getopt(argc, argv, "-d:k:")) != EOF)
      switch(opt) {
      case 'd': dim = atoi(optarg); break;
      case 'k': kmax_ = atoi(optarg); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-d DIM] [-k KMAX]" << std::endl
		  << "where" << std::endl
		  << "  DIM: dimension of problem (default: " << dim << ')'
		  << std::endl
		  << "  KMAX: number of eigenparis to be computed (default: "
		  << kmax_ << ')' << std::endl;
	exit(1); break;
      }
    conceptsAssert((uint)kmax_ < dim, concepts::Assertion());
    conceptsAssert(kmax_ > 0, concepts::Assertion());

    concepts::DummySpace<Real> spc(dim);
    std::cout << "Space: " << spc << std::endl;

    // create and fill mass and stiff
    concepts::SparseMatrix<Real> stiff_(spc, spc), mass_(spc, spc);
    const Real h = 1.0/((Real)dim+1.0);
    stiff_(0,0) = 2.0/h; stiff_(0,1) = -1.0/h;
    for (uint i = 1; i < dim-1; ++i) {
      stiff_(i, i-1) = -1.0/h; stiff_(i,i) = 2.0/h; stiff_(i,i+1) = -1.0/h;
    }
    stiff_(dim-1,dim-2) = -1.0/h; stiff_(dim-1,dim-1) = 2.0/h;
    mass_(0,0) = 4.0*h/6.0; mass_(0,1) = h/6.0;
    for (uint i = 1; i < dim-1; ++i) {
      mass_(i, i-1) = h/6.0; mass_(i,i) = 4.0*h/6.0; mass_(i,i+1) = h/6.0;
    }
    mass_(dim-1,dim-2) = h/6.0; mass_(dim-1,dim-1) = 4.0*h/6.0;
    std::cout << "Mass matrix: " << mass_ << std::endl;
    std::cout << "Stiffness matrix: " << stiff_ << std::endl;

    eigensolver::JdbSym evsolver(stiff_, mass_, 1e-10, 300,
				 0.0, 1, kmax_);
    std::cout << "Eigenvalues: " << evsolver.getEV() << std::endl;
    std::cout << "Eigenvectors: " << std::endl;
    const concepts::Array<concepts::Vector<Real>*>& ef = evsolver.getEF();
    for (int i = 0; i < kmax_; ++i)
      std::cout << "  " << i << "  " << *(ef[i]) << std::endl;
    std::cout << evsolver << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
