/* Function derivation: Vector, stores the values in an array
 */

#ifndef vector_hh
#define vector_hh

#include <cmath>
#include <string>

#include "basics/typedefs.hh"
#include "basis.hh"
#include "toolbox/array.hh"

namespace concepts {

  // forward declaration
  template<class F>
  class LinearForm;

  template<class F>
  class Space;

  // ************************************************************** Vector **

  /** A vector. The values of the function represented by this vector are
      stored elementwise in an array.

      It is possible to write fast interfaces to other libraries since
      the array of values is under complete control of this class. An
      example is the interface to PETSc, where the vector data is not
      copied, but the data of PETSc is used.
      @ingroup bilinear
      @test test::VectorTest
  */
  template<class F>
  class Vector : public Function<F> {
  public:
    /// Copy constructor
    Vector(const Vector<F>& f);

    /** Constructor. Creates a vector with all entries set to 0.
	@param spc A space
    */
    Vector(const Space<F>& spc) : 
      Function<F>(spc), n_(spc.dim()), data_(n_, 0), v_((F*)data_) {}

    /** Constructor. The entries are filled according to \c lf.
	@param spc A space
	@param lf A linear form
    */
    Vector(const Space<F>& spc, LinearForm<F>& lf);

    /** File read constructor. The vector is read from a file.
	This should be endian proof.
	@param spc A space
	@param fname Filename
    */
    Vector(const Space<F>& spc, const std::string& fname);

    /** Constructor.
	Use this constructor to create a vector from an array of data.
	The memory in data is not freed by this class.
	@param spc Space on which the vector is defined
	@param data Data of the vector
    */
    Vector(const Space<F>& spc, F* data) :
      Function<F>(spc), n_(spc.dim()), data_(0), v_(data) {}

    virtual ~Vector();

    virtual Function<F>& operator=(const Function<F>& fnc);

    /// Assignement operator
    Vector<F>& operator=(const Vector<F>& fnc);

    /// Assignement operator
    Vector<F>& operator=(F c);

    /// Assignement operator
    Vector<F>& operator=(LinearForm<F>& lf);

    virtual F& operator()(uint i) { return v_[i]; }
    virtual F operator()(uint i) const { return v_[i]; }

    /// Addition operator
    virtual Function<F>& operator+=(const Function<F>& fnc);
    virtual Function<F>& operator+=(F c);

    /// Subtraction operator
    virtual Function<F>& operator-=(const Function<F>& fnc);
    virtual Function<F>& operator-=(F c);

    virtual Function<F>& operator*=(F sc);

    /// Inner product
    F operator*(const Vector<F>& fnc) const;

    /** Conversion operator.
	Returns the array of values.
    */
    operator F*() const { return v_; }

    virtual Function<F>& add(const Function<F>& fnc, F sc);

    /// l<sup>1</sup> norm
    Real l1() const;

    /// l<sup>2</sup> norm
    inline Real l2() const { return sqrt(l2_2()); }

    /// l<sup>2</sup> norm squared
    Real l2_2() const;

    /// Maximum of the vector, ie. \f$l^\infty\f$ norm
    Real max() const;

    /// Elements in the vector
    uint n() const { return n_; }

    /** Writes the vector to a file. This should be endian proof.
	@param fname Filename
    */
    const Vector<F>& write(const std::string& fname) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    void fillEntries_(LinearForm<F>& lf);
  private:
    /// Dimension of the vector
    const uint n_;
    /// Vector data
    Array<F> data_;
    /// Pointer to the vector data
    F* v_;
  };

} // namespace concepts

#endif // vector_hh
