/* boundary conditions
 */

#ifndef boundaryConditions_hh
#define boundaryConditions_hh

#include <map>
#include "semantics.hh"

namespace concepts {

  // forward declarations
  class Attribute;
  class Boundary;

  // **************************************************** BoundaryConditions **

  /** Boundary conditions.

      The boundary conditions itself are held in a dynamic array. The
      application operator returns a boundary condition object based
      on its attribute.

      Adding boundary conditions in the setup of the problem is done
      using the add method.

      Since the default attribute for a free element of the topology is 0,
      be carefull with using the attribute 0 in the defintion of the
      boundary conditions.

      @see Attribute for the attributes
      @see Connector for the topological elements

      @author Philipp Frauenfelder
  */
  class BoundaryConditions : public Semantics<Boundary> {

  public:
    /** Default constructor. Sets up the dynamic array.
     */
    BoundaryConditions();

    /** Destructor. Takes care of deleting all the formulae which are
	associated with the Boundary objects.

	@see Boundary
    */
    virtual ~BoundaryConditions();

    /** Application operator.

	@param attrib Attribute of a topolgical object (ie. vertex or
	edge)

	@return A reference to the boundary condition object associated to the
	given attribute.
    */
    virtual Boundary operator() (Attribute attrib) const;

    /** Adds a boundary condition with this attribute to the list of
	boundary conditions.

	Be carefull: boundary condition objects with the same attribute
	overwrite each other, the latest is valid.

	@param attrib Attribute of a topological object (ie. vertex, edge etc.)
	@param bcObject Boundary condition object associated to attrib
    */
    void add(const Attribute& attrib, const Boundary& bcObject);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /** Dynamic array of boundaries with type and formula. The index
	is the number given by the attribute of the topological element.
    */
    mutable std::map<uint, Boundary> boundaries_;
  };

} // namespace concepts

#endif // boundaryConditions_hh
