/* Cells.
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#ifndef cell_hh
#define cell_hh

#include <iostream>

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "connector.hh"

namespace concepts {

  // ****************************************************************** Cell **

  /** A cell in a mesh consist of topological information
      (neighbours, connectivity, orientation) and geometrical
      information (coordinates). This class joins the topological
      information from a Connector and the geometical information from
      an element map.

      This class prescribes the common interface for all geometrical
      elements (cells). This interface consists of methods to query the
      children and the assiociated connector.
      @see Connector
      @see Map1d
      @see Map2d
      @see Map3d
      @ingroup geometry
  */
  class Cell : public OutputOperator {
  public:
    /** Returns a pointer to the ith child. Children are created if
	they do not already exist.
    */
    virtual Cell* child(uint i) = 0;
    /** Returns a pointer to the ith child. Children are not created if
	they do not already exist, instead 0 is returned.
    */
    virtual const Cell* child(uint i) const = 0;
    /// Returns the connector
    virtual Connector& connector() const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const = 0;
  };

  // ***************************************************************** Cell1 **

  /// One dimensional cell
  class Cell1 : public Cell {
  public:
    virtual Cell1* child(uint i) = 0;
    virtual const Cell1* child(uint i) const = 0;
    virtual Connector1& connector() const = 0;
    /// Returns the coordinates of the ith vertex
    virtual Real3d vertex(uint i) const = 0;
    /// Returns the center of the cell
    virtual Real3d center() const = 0;
  };

  // ***************************************************************** Cell2 **

  /// Two dimensional cell
  class Cell2 : public Cell {
  public:
    virtual Cell2* child(uint i) = 0;
    virtual const Cell2* child(uint i) const = 0;
    virtual Connector2& connector() const = 0;
    /// Returns the coordinates of the ith vertex
    virtual Real3d vertex(uint i) const = 0;
    /// Returns the center of the cell
    virtual Real3d center() const = 0;
  };

  // ***************************************************************** Cell3 **

  /// Three dimensional cell
  class Cell3 : public Cell {
  public:
    virtual Cell3* child(uint i) = 0;
    virtual const Cell3* child(uint i) const = 0;
    virtual Connector3& connector() const = 0;
    /// Returns the coordinates of the ith vertex
    virtual Real3d vertex(uint i) const = 0;
    /// Returns the center of the cell
    virtual Real3d center() const = 0;
  };

} // namespace concepts

#endif // cell_hh
