/* 1D Cells, ie. triangles and quads
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#ifndef cell1D_hh
#define cell1D_hh

#include <iostream>
#include <memory>

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "topology.hh"
#include "cell.hh"

namespace concepts {

  // forward declaration
  class MapEdge1d;

  // **************************************************************** Edge1d **

  /** A 1D cell: edge.
      @author Philipp Frauenfelder, 2001
      @ingroup geometry
  */
  class Edge1d : public Cell1 {
  public:
    /** Constructor.
	@param cntr The edge which should be represented
	@param map Element map of the edge
    */
    Edge1d(Edge& cntr, const MapEdge1d& map);
    virtual ~Edge1d();

    /** Returns a child. If no children exist, two new children are
	created. The edge is cut in the middle.
	@param i Index of the child to be returned.
    */
    virtual Edge1d* child(uint i);
    virtual const Edge1d* child(uint i) const;

    /// Returns the triangle connector (topology)
    inline Edge& connector() const { return cntr_; }

    /** The element map. Maps a point from the unit intervall [0,1]
	onto the element.

	@return Point in 1D in physical coordinates.
	@param xi \f$\in [0,1]\f$
    */
    Real chi(Real xi) const;

    /** Returns the determinant of the Jacobian.
	@see jacobian
    */
    inline Real jacobianDeterminant(const Real x) const { return jacobian(x); }

    /** Returns the Jacobian.
	@param x \f$\in [0,1]\f$
    */
    Real jacobian(Real x) const;

    /// Returns the size of the element
    Real size() const;

    /** Returns the inverse of the Jacobian
	@see jacobian
    */
    inline Real jacobianInverse(const Real x) const { return 1.0/jacobian(x); }

    Real3d vertex(const uint i) const;
    Real3d center() const { return Real3d(chi(0.5), 0.0, 0.0); }
    inline const MapEdge1d* map() const { return map_; }

    /// Returns a copy of itself
    inline Edge1d* clone(Edge& cntr, MapEdge1d* map) const {
      return new Edge1d(cntr, map, l_, i_); }

    /// Returns the level of the cell
    uchar level() const { return l_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Reference to the edge (topology)
    Edge& cntr_;

    /// Pointer to a sibling
    Edge1d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Edge1d* chld_;

    /// Pointer to the element map
    MapEdge1d* map_;

    /// Level of the element
    uint l_;

    /// Index of the element (ie. left or right child)
    uint i_;

    /** Private copy constructor.
	@param cntr The edge which should be represented
	@param map Element map of the edge
	@param l Level
	@param i Index (0 or 1), ie. the left or right child
    */
    Edge1d(Edge& cntr, MapEdge1d* map, const uint l, const uint i);
  };

} // namespace concepts

#endif // cell1D_hh
