/* data structures and some algorithms to find the supports for
   continuous shape functions on vertices (2D and 3D) and edges (3D)
*/

#include <cmath>
#include <vector>
#include "continuityData.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

// debugging
#define Related_D 0
#define checkCellRelations_D 0

#define ConnectorDataAddCell_D 0
#define ConnectorDataReplaceCell_D 0

#define FaceRelated_D 0
#define FaceDataCheckCellRelations_D 0
#define FaceDatacheckRelatedFaces_D 0

#define EdgeRelated_D 0
#define EdgeDataCheckCellRelations_D 0
#define EdgeDataAddFace_D 0
#define EdgeDataDoCheckRel_D 0
#define EdgeDataCheckRel_D 0
#define EdgeDatacheckRelatedEdges_D 0
#define EdgeDataRecreateFaceList_D 0
#define EdgeDataCommonCell_D 0

#define VertexDataAddEdge_D 0
#define VertexDataDoCheckRel_D 0
#define VertexDataCheckRel_D 0
#define VertexDataRecreateEdgeList_D 0
#define VertexDataCommonCell_D 0

namespace concepts {

  // ****************************************************** output operators **

  template<typename T>
  std::ostream& operator<<(std::ostream& os,
			   const std::map<uint, T*>& m) {
    for (typename std::map<uint, T*>::const_iterator i = m.begin();
	 i != m.end(); ++i)
      os << "  " << *i->second << std::endl;
    return os;
  }

  template<typename T>
  std::ostream& operator<<(std::ostream& os,
			   const std::map<uint, T>& m) {
    for (typename std::map<uint, T>::const_iterator i = m.begin();
	 i != m.end(); ++i)
      os << "  " << i->second << std::endl;
    return os;
  }

  // ********************************************************* ConnectorData **

  void ConnectorData::addCell(const CellData& c) {
    for (std::map<uint, const CellData*>::iterator i = cells_.begin();
    	 i != cells_.end(); ++i) {
      int rel = i->second->related(c);
      if (rel < 0) {
    	DEBUGL(ConnectorDataAddCell_D, "refuse to add " << c
	       << ": it is smaller" << " than " << *i->second);
    	return;
      }
      if (rel > 0) {
	cells_.erase(i->second->key().key());
    	DEBUGL(ConnectorDataAddCell_D, "removed " << *i->second);
      }
    }
    cells_.insert(std::pair<const uint, const CellData*>(c.key().key(), &c));
    DEBUGL(ConnectorDataAddCell_D, "added "
    	   << *(cells_.find(c.key().key())->second));
  }

  bool ConnectorData::hasCell(const Connector& c) const {
    return (cells_.find(c.key().key()) != cells_.end());
  }

  void ConnectorData::replaceCell(const CellData& remove,
				  const CellData& add) {
    DEBUGL(ConnectorDataReplaceCell_D, "remove " << remove << ", add " << add);
    cells_.erase(remove.key().key());
    addCell(add);
  }

  // ************************************************************** CellData **

  const Connector1* CellData::edge(const uint i) const {
    const Connector2* cntr2 =
      dynamic_cast<const Connector2*>(&(cell()));
    if (cntr2) return cntr2->edge(i);
    const Connector3* cntr3 =
      dynamic_cast<const Connector3*>(&(cell()));
    if (cntr3) return cntr3->edge(i);
    return 0;
  }

  const Connector2* CellData::face(const uint i) const {
    const Connector3* cntr3 =
      dynamic_cast<const Connector3*>(&(cell()));
    if (cntr3) return cntr3->face(i);
    return 0;
  }

  bool CellData::hasFace(const Connector2& f) const {
    const Connector2* fc = 0;
    for (uint i = 0; (fc = face(i)) != 0; ++i)
      if (fc->key() == f.key()) return true;
    return false;
  }

  bool CellData::hasEdge(const Connector1& e) const {
    const Connector1* ec = 0;
    for (uint i = 0; (ec = edge(i)) != 0; ++i)
      if (ec->key() == e.key()) return true;
    return false;
  }

  void CellData::addFather(const CellData* father) {
    if (father)
      father_.addCell(*father);
  }

  const CellData* CellData::father(uint i) const {
    if (i == 0) return this;
    if (i == 1) return father();
    return father(i-1);
  }

  const ConnectorData CellData::fathers(uint i) const {
    if (i == 0) return cntrData_;
    if (father() == 0) return cntrData_;
    if (i == 1) return fathers();
    return fathers(i-1);
  }

  std::ostream& operator<<(std::ostream& os, const CellData& c) {
    return os << "CellData(" << c.cell() << ')';
  }

  // ************************************************************** FaceData **

  void FaceData::checkRelatedFaces(std::map<uint, FaceData>& faceList) {
    bool changed = true;

    while (changed) {
      changed = false;
      for (std::map<uint, FaceData>::iterator i = faceList.begin();
	   i != faceList.end(); ++i) {
	conceptsAssert3(i->second.cells_.size() > 0, Assertion(),
			"size = " << i->second.cells_.size());
	for (std::map<uint, FaceData>::iterator j = i;
	     j != faceList.end(); ++j) {
	  int rel = i->second.related(j->second);
	  if (rel != 0) {
	    DEBUGL(FaceDatacheckRelatedFaces_D, i->second
		   << " and " << j->second << " are related: " << rel);
	    changed = true;
	    FaceData *large = 0, *small = 0;
	    if (rel > 0) {
	      large = &(j->second);
	      small = &(i->second);
	    } else {
	      large = &(i->second);
	      small = &(j->second);
	    }

	    // If the larger face already has two cells, we are finished:
	    // drop the small face and continue.
	    if (large->cells_.size() == 2) {
	      faceList.erase(small->key().key());
	      break;
	    }

	    // find the cell which is the correct neighbour for large
	    conceptsAssert3(small->cells_.size() == 1, Assertion(), "size = "
			    << small->cells_.size());
	    const CellData* neigh = small->cells_.begin()->second;
	    conceptsAssert(neigh, Assertion());
	    neigh = neigh->father();
	    while (! neigh->hasFace(large->face())) {
	      if (neigh->father() == 0) {
		/* Caught bad case!
		 *
		 * Following steps seem necessary: remove large faces
		 * from list of faces and check if the parent one level up
		 * is in the list.  If not, add it and add the father cell
		 * to this face.  If the parent face is already in the
		 * list, fail an assertion.
		 */
		DEBUGL(FaceDatacheckRelatedFaces_D, "Caught bad case: "
		       << *neigh << " does not have a father. large = "
		       << *large << ", small = " << *small);

		// start with the large cell and find out which index
		// the face has in the cell
		uint largeCellFaceIndex = 0;
		const Connector2* fc = 0;
		conceptsAssert3(large->cells_.size() == 1, Assertion(),
				"size = " << large->cells_.size());
		const CellData* largeCellData = large->cells_.begin()->second;
		const Connector3* largeCell = dynamic_cast<const Connector3*>
		  (&(largeCellData->cell()));
		conceptsAssert(largeCell, Assertion());
		for (uint k = 0; (fc = largeCell->face(k)) != 0;
		     ++k) {
		  if ( fc == &(large->face()) ) {
		    DEBUGL(FaceDatacheckRelatedFaces_D, *fc << " has index "
			   << k << ": " << *fc << " = " << large->face());
		    largeCellFaceIndex = k;
		  }
		}

		// go up one level on the large side, try to add
		// the correct face to the list of faces
		conceptsAssert(largeCellData->father(), Assertion());
		const Connector3* largeCellFather =
		  dynamic_cast<const Connector3*>
		  (&(largeCellData->father()->cell()));
		conceptsAssert(largeCellFather, Assertion());
		const Connector2* newLargeFace =
		  largeCellFather->face(largeCellFaceIndex);
		conceptsAssert(newLargeFace, Assertion());
		DEBUGL(FaceDatacheckRelatedFaces_D, "father of larger cell is "
		       << *largeCellFather << " with face "
		       << largeCellFaceIndex << ": " << *newLargeFace);
		std::map<uint, concepts::FaceData>::iterator l =
		  faceList.find(newLargeFace->key().key());
		if (l == faceList.end()) {
		  l = faceList.insert(std::pair<uint, FaceData>
				      (newLargeFace->key().key(),
				       FaceData(*newLargeFace))).first;
		} else
		  DEBUGL(FaceDatacheckRelatedFaces_D, "strangely enough, "
			 << *newLargeFace
			 << " is already in the list of faces: " << l->second);
		l->second.addCell(*(largeCellData->father()));
		DEBUGL(FaceDatacheckRelatedFaces_D, "new face data: "
		       << l->second);

		if (large->key().key() != newLargeFace->key().key()) {
		  // remove large face from list of faces
		  DEBUGL(FaceDatacheckRelatedFaces_D, "removing " << *large);
		  faceList.erase(large->key().key());
		}

		// readjust large, rel and neigh
		large = &(l->second);
		DEBUGL(FaceDatacheckRelatedFaces_D, "large is now " << *large);
		neigh = small->cells_.begin()->second; // as before
		DEBUGL(FaceDatacheckRelatedFaces_D, "neigh is now " << *neigh);
		DEBUGL(FaceDatacheckRelatedFaces_D, "old rel = " << rel);
		if (rel > 0) {
		  large->related(*small);
		} else {
		  small->related(*large);
		}
		DEBUGL(FaceDatacheckRelatedFaces_D, "new rel = " << rel);
	      } else {
		neigh = neigh->father();
		conceptsAssert(neigh, Assertion());
	      }
	    }
	    DEBUGL(FaceDatacheckRelatedFaces_D, "found cell " << *neigh);
	    // add the cell to large
	    large->addCell(*neigh);

	    DEBUGL(FaceDatacheckRelatedFaces_D, "removing " << *small);
	    faceList.erase(small->key().key());
// 	    if (rel > 0)
	      break;
	  }
	}
      }    
    }
  }

  std::ostream& operator<<(std::ostream& os, const FaceData& f) {
    os << "FaceData(" << f.face();
    if (f.cells_.size() > 0) os << std::endl << f.cells_ << "  ";
    return os << ')';
  }

  // ************************************************************** EdgeData **

  void EdgeData::addFace(const FaceData& f) {
    faces_.insert(std::pair<const uint, FaceData>(f.key().key(), f));
    DEBUGL(EdgeDataAddFace_D, "added "
	   << faces_.find(f.key().key())->second << " to " << *this);
  }

  bool EdgeData::doCheckRelations(std::map<uint, EdgeData>& edgeList) {
    bool notChanged = true;
    for (std::map<uint, EdgeData>::iterator i =
	   edgeList.begin(); i != edgeList.end(); ++i) {
      try {
	notChanged &= !(i->second.checkRelations());
      }
      catch (WrongRelations& e) {
	notChanged = false;

	const EdgeData& edge = i->second;
	DEBUGL(EdgeDataDoCheckRel_D, e << " for " << edge);

	// find parent edge
	const Connector1* parentEdge = 0;
	for (std::map<uint, const CellData*>::const_iterator j =
	       edge.cells_.begin(); j != edge.cells_.end(); ++j) {
	  DEBUGL(EdgeDataDoCheckRel_D, "cell is " << *(j->second));
	  const CellData& cellD = *(j->second);
	  const CellData* newcellD = &cellD;
	  while (newcellD->hasEdge(edge.edge())) {
	    newcellD = newcellD->father();
	    conceptsAssert(newcellD, Assertion());
	  }
	  DEBUGL(EdgeDataDoCheckRel_D, "larger cell is " << *newcellD);

	  // find index of edge in cell
	  const Connector1* edgConn = 0;
	  const Connector3* cell = dynamic_cast<const Connector3*>
	    (&(cellD.cell()));
	  conceptsAssert(cell, Assertion());
	  uint idx = 0;
	  for (uint m = 0;
	       (edgConn = cell->edge(m)) != 0; ++m) {
	    if (edgConn == &(edge.edge())) {
	      idx = m;
	    }
	  } // for m
	  DEBUGL(EdgeDataDoCheckRel_D, edge.edge() << " has index " << idx
		 << " in " << *cell);
	  const Connector3* newcell = dynamic_cast<const Connector3*>
	    (&(newcellD->cell()));
	  conceptsAssert3(newcell != 0, Assertion(),
			  "dynamic_cast failed. newcellD = " << *newcellD
			  << ", newcell = " << newcellD->cell());
	  parentEdge = newcell->edge(idx);
	  DEBUGL(EdgeDataDoCheckRel_D, "larger edge is " << *parentEdge);
	  conceptsAssert3(edge.edge().key().key() != parentEdge->key().key(),
			  Assertion(), edge.edge());

	  // insert parent edge and corresponding cell
	  // if edge is really a parent of edge
	  if ((parentEdge != 0) && (edge.edge().related(*parentEdge) == 0))
	    parentEdge = 0;
	  if (parentEdge != 0) {
	    DEBUGL(EdgeDataDoCheckRel_D, "parentEdge = " << *parentEdge);
	    conceptsAssert3(newcell->edge(idx) ==
			    parentEdge, Assertion(), "newcell.edge(idx) = "
			    << *(newcell->edge(idx))
			    << ", parentEdge = " << *parentEdge);
	    std::map<uint, concepts::EdgeData>::iterator l =
	      edgeList.find(parentEdge->key().key());
	    if (l == edgeList.end())
	      l = edgeList.insert(std::pair<uint, EdgeData>
				  (parentEdge->key().key(),
				   EdgeData(*parentEdge))).first;
	    l->second.addCell(*newcellD);
	  }
	  if (parentEdge)
	    DEBUGL(EdgeDataDoCheckRel_D,
		   edgeList.find(parentEdge->key().key())->second);
	} // for j

	if (parentEdge) {
	  // recreate face list of newly added edge
	  edgeList.find(parentEdge->key().key())->second.recreateFaceList_();
	}

	// remove edge in favor of its parent
	edgeList.erase(edge.edge().key().key());
	return notChanged;

	conceptsAssert(false, Assertion());
      } // catch
    } // for i
    return notChanged;
  }

  bool EdgeData::checkRelations() {
//     int EdgeDataCheckRel_D = 0;
//     if (key() == Key(61)) EdgeDataCheckRel_D = 1;

    if (io_) return false;
    DEBUGL(EdgeDataCheckRel_D, "checking " << *this);
    bool res = false;
    for (std::map<uint, FaceData>::iterator i = faces_.begin();
	 (i != faces_.end()) && (!res); ++i) {
      for (std::map<uint, FaceData>::iterator j = i;
	   (j != faces_.end()) && (!res); ++j) {
	int k = i->second.related(j->second);
	if (k != 0) {
	  res = true;
	  DEBUGL(EdgeDataCheckRel_D, i->second << " and " << j->second
		 << " are related: " << k);
	  // replacing cell by |k| times father
	  const CellData* remove = 0;
	  const CellData* add = 0;
	  const FaceData* large = 0;
	  if (k > 0) {
	    remove = findCommonCell_(i->second);
	    large = &(j->second);
	  } else {
	    remove = findCommonCell_(j->second);
	    large = &(i->second);
	  }
	  conceptsAssert(remove, Assertion());
	  add = remove->father(std::abs(k));
	  while (! add->hasFace(large->face())) {
	    if (add->father() == 0) {
	      /* Cought bad case!
	       *
	       * Following steps seem necessary: remove edge in favor of its
	       * parent from list of edges. This has to be done outside.
	       * Therefore, throw an exception.
	       */
	      DEBUGL(EdgeDataCheckRel_D, *add << " has no father");
	      throw conceptsException(WrongRelations());
	      conceptsAssert(add->father(), Assertion());
	    } else {
	      add = add->father();
	      conceptsAssert(add, Assertion());
	    }
	  } // while
	  conceptsAssert(add, Assertion());
	  DEBUGL(EdgeDataCheckRel_D, "dropping " << *remove
		 << ", adding " << *add);
	  replaceCell(*remove, *add);
	} // relation!
      } // for faces
    } // for faces
    if (res) recreateFaceList_();
    else io_ = true;
    return res;
  }

  /* Checks if there are edges with relations in edgeList. If
     this is the case, the smaller edge is removed and the cell
     which is the ancestor (on the same level as the larger edge)
     of the cell of this smaller edge is added to the list of cells
     of the larger edge. */
  void EdgeData::checkRelatedEdges(std::map<uint, EdgeData>& edgeList) {
    bool changed = true;
    while (changed) {
    LABEL:
      changed = false;
      for (std::map<uint, EdgeData>::iterator i = edgeList.begin();
	   i != edgeList.end(); ++i) {
	for (std::map<uint, EdgeData>::iterator j = i;
	     j != edgeList.end(); ++j) {
	  int rel = i->second.related(j->second);
	  if (rel != 0) {
	    DEBUGL(EdgeDatacheckRelatedEdges_D, i->second
		   << " and " << j->second << " are related: " << rel);
	    changed = true;
	    EdgeData *large = 0, *small = 0;
	    if (rel > 0) {
	      large = &(j->second);
	      small = &(i->second);
	    } else {
	      large = &(i->second);
	      small = &(j->second);
	    }

	    // find the cell which is the correct neighbour for large
	    for (iterator k = small->cells_.begin(); k != small->cells_.end();
		 ++k) {
	      const CellData* neigh = k->second;
	      DEBUGL(EdgeDatacheckRelatedEdges_D, "cell of smaller edge : " <<
		     neigh->cell());

	      ConnectorData::iterator fathers = 0, lastfather = 0;

	      // to take all the father of the rel.th level go first on 
	      // the former level and go over all fathers
	      if (std::abs(rel)>0) {
		neigh = neigh->father(std::abs(rel)-1);

		DEBUGL(EdgeDatacheckRelatedEdges_D, std::abs(rel)-1 << 
		       ".th (grand)father of that cell : " << neigh->cell());

		fathers = neigh->fathers().begin();
		lastfather = neigh->fathers().end();
		
		if (neigh->father() != 0)
		  neigh = fathers->second;
	      }

	      while (! neigh->hasEdge(large->edge())) {
		if (fathers == lastfather) {
		  DEBUGL(EdgeDatacheckRelatedEdges_D, *neigh);

		  /* Caught bad case!
		   *
		   * Following steps seem necessary: remove large
		   * edge from list of faces and check if the parent
		   * one level up is in the list.  If not, add it and
		   * add the father cell to this edge.  If the parent
		   * edge is already in the list, fail an assertion.
		   *
		   * If the large edge does not have a father, remove
		   * both edges from the list.
		   */

		  // start with the large cell and find out which index
		  // the edge has in the cell
		  const Connector1* edg = 0;
		  std::vector<uint> largeCellEdgeIndex(large->cells_.size());
		  uint tmpIdx = 0;
		  for (iterator l = large->cells_.begin();
		       l != large->cells_.end(); ++l) {
		    const CellData* largeCellData = l->second;
		    const Connector3* largeCell =
		      dynamic_cast<const Connector3*>
		      (&(largeCellData->cell()));
		    conceptsAssert(largeCell, Assertion());
		    for (uint m = 0;
			 (edg = largeCell->edge(m)) != 0; ++m) {
		      if ( edg == &(large->edge()) ) {
			DEBUGL(EdgeDatacheckRelatedEdges_D, *edg << " = "
			       << large->edge());
			largeCellEdgeIndex[tmpIdx] = m;
			DEBUGL(EdgeDatacheckRelatedEdges_D, *edg
			       << " has index " << largeCellEdgeIndex[tmpIdx]
			       << " in " << *largeCell);
		      }
		    } // for m
		    ++tmpIdx;
		  } // for l

		  // Go up one level on the large side, check if the
		  // edges with the same index are related. If yes,
		  // replace large edge (with cells) with even larger
		  // edge in the list of edges, otherwise remove small
		  // and large edge from list of edges.
		  tmpIdx = 0;
		  bool ok = true;
		  for (iterator l = large->cells_.begin();
		       (l != large->cells_.end()) && ok; ++l) {
		    const CellData* largeCellData = l->second;
		    conceptsAssert(largeCellData->father(), Assertion());
		    const Connector3* largeCellFather =
		      dynamic_cast<const Connector3*>
		      (&(largeCellData->father()->cell()));
		    conceptsAssert(largeCellFather, Assertion());
// 		    const Connector3* largeCell = dynamic_cast<const Connector3*>
// 		      (&(largeCellData->cell()));
// 		    conceptsAssert(largeCell, Assertion());
		    const Connector1* newLargeEdge =
		      largeCellFather->edge
		      (largeCellEdgeIndex[tmpIdx]);
		    if (newLargeEdge->related(large->edge())) {
		      DEBUGL(1, "ok");
		      conceptsAssert(neigh->father(), Assertion());
		    } else {
		      ok = false;
		    }
		    ++tmpIdx;
		  } // for l
		  if (! ok) {
		    // remove both edges from list of edges
		    DEBUGL(EdgeDatacheckRelatedEdges_D, "removing " << *large);
		    edgeList.erase(large->key().key());
		    DEBUGL(EdgeDatacheckRelatedEdges_D, "removing " << *small);
		    edgeList.erase(small->key().key());
		    goto LABEL;
		  }
		} // no father
		else {
		  // try the next of the fathers of cell of former level
		  if (fathers != lastfather) {
		    fathers++;
		    DEBUGL(EdgeDatacheckRelatedEdges_D, "next "
			   << fathers->second);
		  }
		  if (fathers != lastfather) {
		    neigh = fathers->second;
		  }
		  else if (neigh->father() != 0) {  
		    // try grandfathers (fathers of the last father)
		    // its assumed that all fathers have common grandfathers
		    fathers = neigh->fathers().begin();
		    lastfather = neigh->fathers().end();
		    neigh = neigh->father();
		  }
		  conceptsAssert(neigh, Assertion());
		  DEBUGL(EdgeDatacheckRelatedEdges_D, "trying neigh "
			 << *neigh);
		}
	      } // found cell with the edge

	      DEBUGL(EdgeDatacheckRelatedEdges_D, "found cell " << *neigh);
	      // add the cell to large
	      large->addCell(*neigh);
	      large->recreateFaceList_();
	    } // for cells k

	    DEBUGL(EdgeDatacheckRelatedEdges_D, "removing " << *small);
	    edgeList.erase(small->key().key());
	    if (rel > 0)
	      break;
	  } // if rel != 0
	} // for edges j
      } // for edges i
    } // while changed
    DEBUGL(EdgeDatacheckRelatedEdges_D, "done.");
  }

  void EdgeData::recreateFaceList_() {
//     int EdgeDataRecreateFaceList_D = 0;
//     if (key() == Key(61)) EdgeDataRecreateFaceList_D = 1;

    DEBUGL(EdgeDataRecreateFaceList_D, "recreating face list of " << *this);
    faces_.clear();

    // loop over all faces in all cells and add the faces which contain
    // this edge to faces_
    for (std::map<uint, const CellData*>::iterator i = cells_.begin();
	 i != cells_.end(); ++i) {
      DEBUGL(EdgeDataRecreateFaceList_D, "faces in " << *i->second);
      const Connector2* face = 0;
      for (uint j = 0; (face = i->second->face(j)) != 0; ++j) {
	bool takeFace = false;
	const Connector1* edg = 0;
	for (uint k = 0; (!takeFace) && ( (edg = face->edge(k)) != 0 ); ++k) {
	  if (edg->key() == edge().key()) takeFace = true;
	  else if (edg->related(edge()) > 0) takeFace = true;
	}
	if (takeFace) {
	  DEBUGL(EdgeDataRecreateFaceList_D, "added " << *face
		 << " to faceList");
	  faces_.insert(std::pair<const uint, FaceData>
			(face->key(), FaceData(*face))).first->
	    second.addCell(*i->second);
	}
      } // for faces
    } // for cells
  }

  const CellData* EdgeData::findCommonCell_(const FaceData& face) const {
    DEBUGL(EdgeDataCommonCell_D, "finding common cell of " << *this
	   << " and " << face);
    for (std::map<uint, const CellData*>::const_iterator i = cells_.begin();
	 i != cells_.end(); ++i) {
      DEBUGL(EdgeDataCommonCell_D, "checking " << *i->second);
      if (face.hasCell(i->second->cell())) {
	DEBUGL(EdgeDataCommonCell_D, "it's " << *i->second);
	return i->second;
      }
    }
    DEBUGL(1, "coud not find common cell of " << *this
	   << " and " << face);
    return 0;
  }

  std::ostream& operator<<(std::ostream& os, const EdgeData& e) {
    os << "EdgeData(" << e.edge();
    if (e.cells_.size() > 0) os << std::endl << e.cells_;
    for (std::map<uint, FaceData>::const_iterator i = e.faces_.begin();
	 i != e.faces_.end(); ++i) {
      os << "  " << i->second.face() << std::endl;
    }
    return os << "  )";
  }

  // ************************************************************ VertexData **

  void VertexData::addEdge(const EdgeData& e) {
    edges_.insert(std::pair<const uint, EdgeData>(e.key().key(), e));
    DEBUGL(VertexDataAddEdge_D, "added "
	   << edges_.find(e.key().key())->second << " to " << *this);
  }

  bool VertexData::doCheckRelations(std::map<uint, VertexData>& vertexList) {
    bool notChanged = true;
    for (std::map<uint, concepts::VertexData>::iterator i =
	   vertexList.begin(); i != vertexList.end(); ++i) {
      try {
	notChanged &= !(i->second.checkRelations());
      }
      catch (WrongRelations& e) {
	DEBUGL(VertexDataDoCheckRel_D, e);
	notChanged = false;

	vertexList.erase(i->second.vertex().key().key());
	return notChanged;

	conceptsAssert(false, Assertion());
      } // catch
    } // for i
    return notChanged;
  }

  bool VertexData::checkRelations() {
//     int VertexDataCheckRel_D = 0;
//     if (key() == Key(16)) VertexDataCheckRel_D = 1;

    if (io_) return false;
    bool res = false;
    for (std::map<uint, EdgeData>::iterator i = edges_.begin();
	 (i != edges_.end()) && (!res); ++i) {
      for (std::map<uint, EdgeData>::iterator j = i;
	   (j != edges_.end()) && (!res); ++j) {
	int k = i->second.related(j->second);
	if (k != 0) {
	  res = true;
	  DEBUGL(VertexDataCheckRel_D, vertex_ << ": " << i->second
		 << " and " << j->second << " are related: " << k);

	  // replacing cell by |k| times father
	  const CellData* remove = 0;
	  const CellData* add = 0;
	  const EdgeData* large = 0;
	  if (k > 0) {
	    remove = findCommonCell_(i->second);
	    large = &(j->second);
	  } else {
	    remove = findCommonCell_(j->second);
	    large = &(i->second);
	  }
	  conceptsAssert(remove, Assertion());
	  add = remove->father(std::abs(k));
	  while (! add->hasEdge(large->edge())) {
	    if (add->father() == 0) {
	      /* Cought bad case!
	       *
	       * Following steps seem necessary: edge in favor of its
	       * parent from list of edges. This has to be done outside.
	       * Therefore, throw an exception.
	       */
	      DEBUGL(VertexDataCheckRel_D, *add << " has no father");
	      throw conceptsException(WrongRelations());
	      conceptsAssert(add->father(), Assertion());
	    } else {
	      add = add->father();
	      conceptsAssert(add, Assertion());
	    }
	  } // while
	  conceptsAssert(add, Assertion());
	  replaceCell(*remove, *add);
	}
      } // for edges
    } // for edges
    if (res) recreateEdgeList_();
    else io_ = true;
    return res;
  }

  void VertexData::recreateEdgeList_() {
//     int VertexDataRecreateEdgeList_D = 0;
//     if (key() == Key(16)) VertexDataRecreateEdgeList_D = 1;

    DEBUGL(VertexDataRecreateEdgeList_D, "recreating edge list of " << *this);
    edges_.clear();

    // loop over all edges in all cells and add the edges which contain
    // this vertex to edges_
    for (std::map<uint, const CellData*>::iterator i = cells_.begin();
	 i != cells_.end(); ++i) {
      DEBUGL(VertexDataRecreateEdgeList_D, "edges in " << *i->second);
      const Connector1* edge = 0;
      for (uint j = 0; (edge = i->second->edge(j)) != 0; ++j) {
	bool takeEdge = false;
	for (uint k = 0; (!takeEdge) && (k < 2); ++k)
	  if (edge->vertex(k)->key() == vertex().key()) takeEdge = true;
	if (takeEdge) {
	  DEBUGL(VertexDataRecreateEdgeList_D, "added " << *edge
		 << " to faceList");
	  edges_.insert(std::pair<const uint, EdgeData>
			(edge->key(), EdgeData(*edge))).first->
	    second.addCell(*i->second);
	}
      } // for edges
    } // for cells
  }

  const CellData* VertexData::findCommonCell_(const EdgeData& edge) const {
//     int VertexDataCommonCell_D = 0;
//     if (key() == Key(16)) VertexDataCommonCell_D = 1;

    DEBUGL(VertexDataCommonCell_D, "finding common cell of " << *this
	   << " and " << edge);
    for (std::map<uint, const CellData*>::const_iterator i = cells_.begin();
	 i != cells_.end(); ++i) {
      DEBUGL(VertexDataCommonCell_D, "checking " << *i->second);
      if (edge.hasCell(i->second->cell())) {
	DEBUGL(VertexDataCommonCell_D, "it's " << *i->second);
	return i->second;
      }
    }
    DEBUGL(1, "could not find common cell of " << *this
	   << " and " << edge);
    return 0;
  }

  std::ostream& operator<<(std::ostream& os, const VertexData& v) {
    os << "VertexData(" << v.vertex();
    if (v.cells_.size() > 0) os << std::endl << v.cells_ << "  ";
    return os << ')';
  }

} // namespace concepts
