/* Element maps.
 * Geometrical information (coordinates)
 */

#include <cmath>
#include "elementMaps.hh"
#include "basics/functionCompiler.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

// debugging CircleMappingEdge2d
#define CircleMappingEdge2d_D 0

// debugging MapQuad2d
#define MapQuad2dConstr_D 0

// debugging VertexQuad2d
#define VertexQuad2djacobian_D 0
#define VertexQuad2dlineElement_D 0

// debugging BlendingQuad2d
#define BlendingQuad2dConstr_D 0
#define BlendingQuad2djacobian_D 0

// debugging VertexTriangle2d
#define VertexTriangleConstr_D 0
#define VertexTriangleDestr_D 0
#define VertexTriangleClone_D 0

// debugging MapTriangle3d
#define MapTriangle3dConstr_D 0

namespace concepts {

  // ************************************************************* MapEdge1d **

  std::ostream& MapEdge1d::info(std::ostream& os) const {
    return os << "MapEdge1d(" << vtx0_ << ", " << vtx1_ << ')';
  }

  // ********************************************************** MappingEdge2d **

  MappingEdge2d::MappingEdge2d(const Real2d vtx0, const Real2d vtx1)
  {
    vtx_[0] = vtx0; vtx_[1] = vtx1;
  }

  MappingEdge2d::MappingEdge2d(const MappingEdge2d &edgemap)
  {
    vtx_[0] = edgemap.vtx(0); vtx_[1] = edgemap.vtx(1);
  }

  Real2d MappingEdge2d::operator()(const Real t) const {
    conceptsAssert(t >= 0 && t <= 1, Assertion());
    // straight line: linear interpolation
    return vtx_[0] * (1-t) + vtx_[1] * t;
  }

  Real2d MappingEdge2d::jacobian(const Real t) const {
    // derive the function above to get this result - a constant
    return vtx_[1]-vtx_[0]; 
  }

  std::ostream& MappingEdge2d::info(std::ostream& os) const {
    return os << "MappingEdge2d(" << vtx_[0] << ", " << vtx_[1] << ')';
  }

  // **************************************************** CircleMappingEdge2d **

  CircleMappingEdge2d::CircleMappingEdge2d(const Real r, const Real2d vtx0,
					   const Real2d vtx1)
    : MappingEdge2d(vtx0, vtx1), r_(std::abs(r))
  {
    DEBUGL(CircleMappingEdge2d_D, "radius = " << r << ", vertices = " << 
	   vtx0 << ", " << vtx1);

    // distance between vertices of the unit circle.
    Real dx = vtx0[x_coord]-vtx1[x_coord];
    Real dy = vtx0[y_coord]-vtx1[y_coord];
    
    // curved line
    if (r!=0) {
      Real t = r*r/(dx*dx+dy*dy) - 1./4.;
      conceptsAssert3(t >= 0, Assertion(), "Radius too small.");
      // parameter of the circle center of the perpendicular bisector
      // Geradenparameter des Kreismittelpunktes auf der Mittelsenkrechten
      t = std::sqrt(t);
      if (r > 0) t *= -1.0;
      // circle center
      m_ = (vtx0+vtx1)*0.5 + Real2d(-dy,dx) * t;
      // Angles between one vertex and the cirle center
      angle0_ = arc_(vtx0 - m_);
      angle1_ = arc_(vtx1 - m_);
      DEBUGL(CircleMappingEdge2d_D, "angles = (" << angle0_ << ", " << angle1_ << ")");
      if (r>0) 
	while(angle1_<angle0_) angle1_ += 2*M_PI;
      else
	while(angle1_>angle0_) angle1_ -= 2*M_PI;
      DEBUGL(CircleMappingEdge2d_D, "angles = (" << angle0_ << ", " << angle1_ << ")");
      DEBUGL(CircleMappingEdge2d_D, "dx = " << dx << ", dy = " <<
	     dy << ", t = " << t);
    } else {
      angle0_ = angle1_ = 0;
      m_ = Real2d(0,0);
    }
    DEBUGL(CircleMappingEdge2d_D, "center = " << m_ << ", angles = (" <<
	   angle0_ << ", " << angle1_ << ")");
  }

  Real CircleMappingEdge2d::arc_(const Real2d delta) const {
    DEBUGL(CircleMappingEdge2d_D, "delta = " << delta);
    if (delta[x_coord] != 0){
      if (delta[y_coord] != 0){
	if (delta[x_coord] > 0) 
	  return std::atan(delta[y_coord] / delta[x_coord]);
	else
	  return std::atan(delta[y_coord] / delta[x_coord]) + M_PI;
      } else if (delta[x_coord] > 0){
	return 0;
      } else {
	return M_PI;
      }
    } else if (delta[y_coord] > 0){
      return M_PI/2.;
    } else {
      return 3*M_PI/2.;
    }
  }

  Real2d CircleMappingEdge2d::operator()(const Real t) const {
    conceptsAssert(t >= 0 && t <= 1, Assertion());
    if (r_==0){
      // straight line: linear interpolation
      return vtx_[0] * (1-t) + vtx_[1] * t;
    }
    else {
      if (t==0)
	return vtx_[0];
      else if (t==1)
	return vtx_[1];
      else {
	Real phi = angle0_ * (1-t) + angle1_ * t;
	return m_ + Real2d(cos(phi),sin(phi)) * r_;
      }
    }
  }

  Real2d CircleMappingEdge2d::jacobian(const Real t) const {
    if(r_==0)
      //derive the function above to get this result - a constant
      return vtx_[1]-vtx_[0]; 
    else {
      //same as before but with a correction due to arc
      Real phi = angle0_ * (1-t) + angle1_ * t;
      return Real2d(-sin(phi),cos(phi)) * r_ * (angle1_ - angle0_);
    }
  }

  std::ostream& CircleMappingEdge2d::info(std::ostream& os) const {
    return os << "CircleMappingEdge2d(vertices = (" << vtx_[0] << ", " << 
      vtx_[1] << "), center = " << m_ << ", radius =  " << r_ << ')';
  }

  // ***************************************************** MappingTriangle2d **

  Real MappingTriangle2d::jacobianDeterminant(const Real x,
					      const Real y) const {
    return jacobian(x, y).determinant();
  }

  MapReal2d MappingTriangle2d::jacobianInverse(const Real x,
					       const Real y) const {
    return jacobian(x, y).inverse();
  }

  // ****************************************************** VertexTriangle2d **

  VertexTriangle2d::VertexTriangle2d(Real2d vtx0, Real2d vtx1, Real2d vtx2) :
    vtx0_(vtx0), vtx1_(vtx1), vtx2_(vtx2),
    jacobian_((Real2d)(vtx1_-vtx0_), (Real2d)(vtx2_-vtx1_)) {
    DEBUGL(VertexTriangleConstr_D, "done.");
  }

  VertexTriangle2d::VertexTriangle2d(const VertexTriangle2d &map) :
    vtx0_(map.vtx0_), vtx1_(map.vtx1_), vtx2_(map.vtx2_),
    jacobian_(map.jacobian_) {
    DEBUGL(VertexTriangleConstr_D, "done.");
  }

  VertexTriangle2d::~VertexTriangle2d() {
    DEBUGL(VertexTriangleDestr_D, "done.");
  }

  VertexTriangle2d* VertexTriangle2d::clone() const {
    DEBUGL(VertexTriangleClone_D, "start");
    return new VertexTriangle2d(*this);
  }

  std::ostream& VertexTriangle2d::info(std::ostream& os) const {
    return os << "VertexTriangle2d(vtx0 = " << vtx0_ << ", vtx1 = " << vtx1_ 
	      << ", vtx2 = " << vtx2_ << ")";
  }

  bool VertexTriangle2d::curved() const { 
    return false;
  }

  // ********************************************************* MapTriangle2d **

  MapTriangle2d::MapTriangle2d(char* map, const Real scx, const Real scy) :
    map_(0), sz_(0), scx_(scx), scy_(scy) {

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_];
      std::memcpy(map_, foo, sz_);
    }
  }

  MapTriangle2d::MapTriangle2d(const MapTriangle2d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_) {

    if (map.map_) {
      map_ = new uchar[sz_]; std::memcpy(map_, map.map_, sz_);
    }
  }

  MapTriangle2d::~MapTriangle2d() { delete[] map_; }

  MapTriangle2d* MapTriangle2d::clone() const {
    return new MapTriangle2d(*this);
  }

  std::ostream& MapTriangle2d::info(std::ostream& os) const {
    return os << "MapTriangle2d()";
  }

  bool MapTriangle2d::curved() const { 
    return false;
  }

  MapReal2d MapTriangle2d::jacobian(const Real, const Real) const {
    throw conceptsException(MissingFeature("Jacobian not implemented"));
  }

  // ********************************************************* MappingQuad2d **

  Real MappingQuad2d::jacobianDeterminant(const Real x, const Real y) const {
    return jacobian(x, y).determinant();
  }

  MapReal2d MappingQuad2d::jacobianInverse(const Real x, const Real y) const {
    return jacobian(x, y).inverse();
  }

  // general implementation of factor of differential element  
  // (for integrating of an edge)
  Real MappingQuad2d::lineElement(const Real x, const uint edge) const {
    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(edge < 4, Assertion());

    MapReal2d jac;
    switch(edge){
    case 0 : jac = jacobian(x,0.0);break;  // bottom edge
    case 1 : jac = jacobian(1.0,x);break;  // right
    case 2 : jac = jacobian(x,1.0);break;  // top
    case 3 : jac = jacobian(0.0,x);break;  // left
    default: throw conceptsException(concepts::ExceptionBase());
    }
    Real tmp;
    switch(edge){
    case 0 :
    case 2 :
      tmp = std::sqrt( std::pow(jac(0,0),2) + std::pow(jac(1,0),2) );
      break;
    case 1 :
    case 3 :
      tmp = std::sqrt( std::pow(jac(0,1),2) + std::pow(jac(1,1),2) );
      break;
    default: throw conceptsException(concepts::ExceptionBase());
    }
    return tmp;
  }

  // ************************************************************* MapQuad2d **

  MapQuad2d::MapQuad2d(char* map, Real scx, Real scy, 
			       Real2d org, Real stretch) :
    map_(0), sz_(0), scx_(scx), scy_(scy), org_(org),
    scaling_(Real2d(stretch, stretch)) {

    DEBUGL(MapQuad2dConstr_D, map << " parameter region (" << scx_ << ", " 
	   << scy_ << "), origin " << org_ << ", scaling " << scaling_);

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    }
  }

  MapQuad2d::MapQuad2d(char* map, Real scx, Real scy, Real2d org,
			       Real2d scaling) :
    map_(0), sz_(0), scx_(scx), scy_(scy), org_(org), scaling_(scaling) {

    DEBUGL(MapQuad2dConstr_D, map << " parameter region (" << scx_ << ", " 
	   << scy_ << "), origin " << org_ << ", scaling " << scaling_);

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    }
  }

  MapQuad2d::MapQuad2d(const MapQuad2d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_), 
    org_(map.org_), scaling_(map.scaling_) {

    map_ = new uchar[sz_]; std::memcpy(map_, map.map_, sz_);

    DEBUGL(MapQuad2dConstr_D, "parameter region (" << scx_ << ", " 
	   << scy_ << "), origin " << org_ << ", scaling " << scaling_);
  }

  MapQuad2d::~MapQuad2d() { delete[] map_; }

  MapReal2d MapQuad2d::jacobian(const Real x, const Real y) const {
    throw conceptsException(MissingFeature
			    ("Jacobian of map not implemented"));
  }

  // ********************************************************** VertexQuad2d **

  VertexQuad2d::VertexQuad2d(Real2d vtx0, Real2d vtx1,
			     Real2d vtx2, Real2d vtx3) {
    vtx_[0] = vtx0; vtx_[1] = vtx1; vtx_[1] -= vtx0;
    vtx_[3] = vtx3; vtx_[3] -= vtx0;
    vtx_[2] = vtx_[1]; vtx_[2] *= -1.0; vtx_[2] += vtx2; vtx_[2] -= vtx3;
    hasLength_ = false;
    conceptsAssert(jacobian(0.5, 0.5).determinant() > 0,
		   concepts::Assertion());
  }

  VertexQuad2d::VertexQuad2d(const VertexQuad2d& v) {
    std::memcpy(vtx_, v.vtx_, 4*sizeof(Real2d));
  }

  VertexQuad2d::~VertexQuad2d() {}

  // mapping is v0 + v1*x + v3*y + v2*x*y
  Real2d VertexQuad2d::operator()(Real x, Real y) const {
    conceptsAssert((x >= 0.0) && (x <= 1.0), concepts::Assertion());
    conceptsAssert((y >= 0.0) && (y <= 1.0), concepts::Assertion());
    Real2d res(vtx_[0]);
    res += vtx_[1]*x;    // v1 = x1-x0
    res += vtx_[3]*y;    // v3 = x3-x0
    res += vtx_[2]*x*y;  // v2 = x0+x2-x1-x3
    return res;
  }

  MapReal2d VertexQuad2d::jacobian(const Real x, const Real y) const {
    DEBUGL(VertexQuad2djacobian_D, "x = (" << x << ", " << y << ")");

    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(y >= 0.0, Assertion());
    conceptsAssert(y <= 1.0, Assertion());

    Real2d first, second;

    // first column v1 + v2*y with v1 = x1-x0, v2 = x0+x2-x1-x3
    first.lincomb(vtx_[2], vtx_[1], y);
    // second column v3 + v2*x with v3 = x3-x0
    second.lincomb(vtx_[2], vtx_[3], x);
    DEBUGL(VertexQuad2djacobian_D, "jac = "
	   << MapReal2d(first, second));
    return MapReal2d(first, second);
  }

  // differential element is just length of edge
  Real VertexQuad2d::lineElement(const Real x, const uint edge) {
    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(edge < 4, Assertion());

    if (!hasLength_){
      len_[0] = vtx_[1].l2();            // bottom: v1 = x1-x0
      len_[1] = (vtx_[2]-vtx_[3]).l2();  // right:  v2-v3 = x2-x1
      len_[2] = (vtx_[2]-vtx_[1]).l2();  // top:    v2-v1 = x2-x3
      len_[3] = vtx_[3].l2();            // left:   v3 = x3 - x0
    }

    DEBUGL(VertexQuad2dlineElement_D, "edge = " << edge <<
	   ", line element ds = " << len_[edge] << " d" << (edge%2?"x":"y"));
    return len_[edge];
  }

  VertexQuad2d* VertexQuad2d::clone() const {
    return new VertexQuad2d(*this);
  }

  // ********************************************************* BlendingQuad2d **

  BlendingQuad2d::BlendingQuad2d(const MappingEdge2d* edgemap0,
				 const MappingEdge2d* edgemap1,
				 const MappingEdge2d* edgemap2,
				 const MappingEdge2d* edgemap3)
  {
    // given mappings for the edges
    edgemap_[0] = edgemap0; edgemap_[1] = edgemap1;
    edgemap_[2] = edgemap2; edgemap_[3] = edgemap3;
#if BlendingQuad2dConstr_D
    for(uint j=0; j < 4; ++j)
      if (edgemap_[j])
	DEBUGL(1, "edgemap_[" << j << "] = " << *edgemap_[j]);
#endif
    // vertices of the quad are vertices of the edges
    for(uint j=0; j < 4; ++j) {
      // a vertex can be given by two edges, at least one has to exist
      // if both exist, the coordinates have to match
      if (existEdgemap_[j] = edgemap_[j]){
	vtx_[j] = edgemap_[j]->vtx(0);
	if (edgemap_[(j+3)%4])
	  // test matching of the vertices
	  for(uint i=0; i < 2; ++i)
	    conceptsAssert3(edgemap_[(j+3)%4]->vtx(1)[i] == vtx_[j][i],
			    Assertion(),
			    "edgemap_[" << (j+3)%4 << "]->vtx(1)[" << i << "] = "
			    << edgemap_[(j+3)%4]->vtx(1)[i] << ", vtx_[" << j <<
			    "][" << i << "] = " << vtx_[j][i]);
      } else {
	conceptsAssert3(edgemap_[(j+3)%4], concepts::Assertion(),
			"one of two adjacent edges has to be given");
	vtx_[j] = edgemap_[(j+3)%4]->vtx(1);
      }
    }

    // if there are mappings left out, take straight lines
    for(uint j=0; j < 4; ++j)
      if (!existEdgemap_[j]) {
	edgemap_[j] = new MappingEdge2d(vtx_[j], vtx_[(j+1)%4]);
      }

    // test for degenerated mappings (inverse of jacobian neccessary)
    conceptsAssert(jacobian(0.5, 0.5).determinant() > 0,
		   concepts::Assertion());
    conceptsAssert(jacobian(0.0, 0.0).determinant() > 0,
		   concepts::Assertion());
    conceptsAssert(jacobian(0.0, 1.0).determinant() > 0,
		   concepts::Assertion());
    conceptsAssert(jacobian(1.0, 0.0).determinant() > 0,
		   concepts::Assertion());
    conceptsAssert(jacobian(1.0, 1.0).determinant() > 0,
		   concepts::Assertion());
  }

  BlendingQuad2d::BlendingQuad2d(const BlendingQuad2d& v) {
    std::memcpy(vtx_, v.vtx_, 4*sizeof(Real2d));
    for(uint j=0; j<4; ++j) {
      existEdgemap_[j] = v.existEdgemap_[j];
      if (existEdgemap_[j])
	edgemap_[j] = v.edgemap_[j];
      else
	edgemap_[j] = new MappingEdge2d(vtx_[j], vtx_[(j+1)%4]);
    }
  }

  BlendingQuad2d::~BlendingQuad2d()
  {
    for(uint j=0; j < 4; ++j)
      if (!existEdgemap_[j])
	delete edgemap_[j];
  }

  // blending 
  Real2d BlendingQuad2d::operator()(Real x, Real y) const {
    conceptsAssert((x >= 0.0) && (x <= 1.0), concepts::Assertion());
    conceptsAssert((y >= 0.0) && (y <= 1.0), concepts::Assertion());
    return (*edgemap_[0])(x) * (1-y) + (*edgemap_[1])(y) * x + 
           (*edgemap_[2])(1-x) * y   + (*edgemap_[3])(1-y) * (1-x)
           - vtx_[0] * (1-x)*(1-y) - vtx_[1] * x * (1-y)
           - vtx_[2] * x*y         - vtx_[3] * (1-x) * y;
  }

  MapReal2d BlendingQuad2d::jacobian(const Real x, const Real y) const {
    DEBUGL(BlendingQuad2djacobian_D, "x = (" << x << ", " << y << ")");

    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(y >= 0.0, Assertion());
    conceptsAssert(y <= 1.0, Assertion());

    // columns
    // first derivative with respect to (local) x
    // second             "             (local) y
    Real2d first, second;

    first =  edgemap_[0]->jacobian(x) * (1-y) + (*edgemap_[1])(y) -
             edgemap_[2]->jacobian(1-x) * y   - (*edgemap_[3])(1-y) +
             (vtx_[0] - vtx_[1]) * (1-y) + (vtx_[3] - vtx_[2]) * y;
    second = (*edgemap_[0])(x)*(-1) + edgemap_[1]->jacobian(y) * x +
             (*edgemap_[2])(1-x)    - edgemap_[3]->jacobian(1-y) * (1-x) +
             (vtx_[0] - vtx_[3]) * (1-x) + (vtx_[1] - vtx_[2]) * x;

    DEBUGL(BlendingQuad2djacobian_D, "jac = " << MapReal2d(first, second));
    return MapReal2d(first, second);
  }

  Real BlendingQuad2d::lineElement(const Real x, const uint edge) const {
    conceptsAssert(edge < 4, Assertion());
    return edgemap_[edge]->jacobian(x).l2();
  }

  std::ostream& BlendingQuad2d::info(std::ostream& os) const {
    os << "BlendingQuad2d(vertices = (";
    for(uint i=0; i<4; ++i) {
      os << vtx_[i];
      if (i<3) os << ", ";
    }
    os << "), edgemappings = (";
    for(uint i=0; i<4; ++i) {
      os << *edgemap_[i];
      if (i<3) os << ", ";
    }
    return os << ")" << std::endl;
  }


  // ********************************************************* MapTriangle3d **

  MapTriangle3d::MapTriangle3d(char* map, Real scx, Real scy) :
    map_(0), sz_(0), scx_(scx), scy_(scy), B1_(0.0, 0.0, 0.0), 
    B2_(0.0, 0.0, 0.0), b_(0.0, 0.0, 0.0) {

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    }
  }

  MapTriangle3d::MapTriangle3d(Real3d vtx0, Real3d vtx1, Real3d vtx2) :
    map_(0), sz_(0), scx_(1.0), scy_(1.0), B1_(vtx1-vtx0), B2_(vtx2-vtx1), 
    b_(vtx0) {

    DEBUGL(MapTriangle3dConstr_D, "B1 = " << B1_ << ", B2 = " << B2_ 
	   << ", b = " << b_);
  }

  MapTriangle3d::MapTriangle3d(const MapTriangle3d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_), B1_(map.B1_),
    B2_(map.B2_), b_(map.b_) {

    if (map.map_) {
      map_ = new uchar[sz_]; std::memcpy(map_, map.map_, sz_);
    }
  }

} // namespace concepts
