/* Element maps.
 * Geometrical information (coordinates)
 */

#ifndef elementMaps_hh
#define elementMaps_hh

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"

namespace concepts {

  // ***************************************************************** Map1d **

  /** An abstract class for a 1d map
      @ingroup geometry
   */
  class Map1d : public OutputOperator {
  public:
    std::ostream& info(std::ostream& os) const {return os << "Map1d()";};
  };

  // ***************************************************************** Map2d **

  /** An abstract class for a 2d map
      @ingroup geometry
   */
  class Map2d : public OutputOperator {
  public:
    std::ostream& info(std::ostream& os) const {return os << "Map2d()";};
  };

  // ************************************************************* MapEdge1d **

  /** A 1D element map for an edge. The parameter domain is the unit
      intervall [0,1].

      @author Philipp Frauenfelder, 2001.
  */
  class MapEdge1d : public Map1d {
  public:
    /** Constructor.
	@param vtx0 Left vertex
	@param vtx1 Right vertex
    */
    inline MapEdge1d(const Real vtx0, const Real vtx1)  :
      vtx0_(vtx0), vtx1_(vtx1) {}

    /** Application operator. Maps the point from the parameter domain
	onto the physical space.
    */
    inline Real operator()(const Real x) const {
      return vtx0_*(1.0-x) + vtx1_*x; }

    /// Returns the jacobian
    inline Real jacobian(const Real) const { return vtx1_-vtx0_; }

    /// Returns a copy of the map
    inline MapEdge1d* clone() const {return new MapEdge1d(vtx0_, vtx1_);};
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Coordinates of the vertices
    Real vtx0_, vtx1_;
  };

  // ********************************************************** MappingEdge2d **

  class MappingEdge2d : public Map1d {

  /** Abstract class for 2D element map for an (curved) edge.

      The parameter domain is the unit intervall [0,1].

      Derive a class from this class and implement \c valuex, \c valuey and 
      \c derivativex, \c derivatey to get blending for a certain type of edge, 
      e.g. circular arcs (CircleMappingEdge2d) or similar.

      @see CircleMappingEdge2d

      @author, Kersten Schmidt, 2004
      **/
  public:
    /** Constructor

       @param vtx0,vtx1 coordinates of the vertices
     */
    MappingEdge2d(const Real2d vtx0, const Real2d vtx1);
    /// Copy Constructor
    MappingEdge2d(const MappingEdge2d &edgemap);
    /** Application operator. Maps the point from the parameter domain
	onto the physical space.
    */
    virtual Real2d operator()(const Real t) const;
    /// Returns the jacobian
    virtual Real2d jacobian(const Real t) const;
    /// Spit out one vertex of the edge.
    const Real2d vtx(uint i) const {return vtx_[i];}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    /// Coordinates of the vertices at either end of the edge.
    Real2d vtx_[2];
  };

  // **************************************************** CircleMappingEdge2d **

  /**  2D element map for an circular arc.

       The Edge is given by two vertices and a radius that determines the
       curvature of the edge. 

       This class may be used in connection with BlendingQuad2d

       @see BlendingQuad2d

       @author, Kersten Schmidt, 2004
  */
  class CircleMappingEdge2d : public MappingEdge2d {
  public:
    /** Constructor

       Use the sign of the radius to bend the edge.
       \c r = 0  No curvature
       \c r > 0  Arc is on the right side of line from 1st to 2nd vertex
       \c r < 0  Arc is on the left side of line from 1st to 2nd vertex	

       Minimal radius is half of the distance between the vertices.

       There are always two possibilites for the angle of the arc
       (\f$\phi\f$ and \f$2\pi - \phi\f$. The smaller angle is taken.
       So the maximal angle is \f$\pi\f$.
       
       @param r radius
       @param vtx0,vtx1 coordinates of the vertices
     */
    CircleMappingEdge2d(const Real r, const Real2d vtx0, const Real2d vtx1);
    virtual Real2d operator()(const Real t) const;
    virtual Real2d jacobian(const Real t) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    // zero and one correspond to x and y coordinates.
    enum coord {x_coord, y_coord};
    /// Radius of the circle, zero if straight line (no sign)
    Real r_;
    /// Mid point of the circle
    Real2d m_;
    /// Angles between one vertex and the circle center
    Real angle0_, angle1_;
    /** Calculates angle in the 2d plane between two points
	
        @param delta difference vector between the points
     */
    Real arc_(const Real2d delta) const;
  };

  // ***************************************************** MappingTriangle2d **

  /** A 2D element map for a triangle. The parameter domain is the triangle
      \f$\{(\xi,\eta) : 0 \leq \eta \leq \xi \leq 1\}\f$.

      @author Philipp Frauenfelder, 2001.
  */
  class MappingTriangle2d : public Map2d {
  public:
    /** Application operator. Maps the point from the parameter domain
	\f$\{(\xi,\eta) : 0 \leq \eta \leq \xi \leq 1\}\f$ onto a triangle
	in the 2D physical space.
    */
    virtual Real2d operator()(const Real x, const Real y) const = 0;

    /** Creates a clone of itself and returns it. This clone function
	implements a virtual constructor (cf. Stroustrup p. 424).
    */
    virtual MappingTriangle2d* clone() const = 0;

    /** Returns true if the resulting triangle has curved edges. In
	this case the integration routines have to take into account
	that the Jacobian is changing. Otherwise, the Jacobian is a
	constant.
    */
    virtual bool curved() const = 0;

    /// Returns determinant of the Jacobian
    virtual Real jacobianDeterminant(const Real x, const Real y) const;

    /// Returns the Jacobian in a 2D linear map
    virtual MapReal2d jacobian(const Real x, const Real y) const = 0;

    /// Returns the inverse of the Jacobian in a 2D linear map
    virtual MapReal2d jacobianInverse(const Real x, const Real y) const;
  };

  // ****************************************************** VertexTriangle2d **

  /** A 2D element map for a triangle. The map is specified by the three 
      vertices.

      @author Philipp Frauenfelder, 2001.
  */
  class VertexTriangle2d : public MappingTriangle2d {
  public:
    /** Constructor. Takes the three physical corners of the triangle and
	computes the element map: \f$F_K : S \rightarrow K\f$ with
	\f$\vec x = F_K(\xi) = B \cdot \vec \xi + \vec b\f$, where
	\f$\vec b = \mbox{vtx0}\f$ and 
	\f$B = [ (\mbox{vtx1} - \vec b) (\mbox{vtx2} - \mbox{vtx1}) ]\f$.
    */
    VertexTriangle2d(Real2d vtx0, Real2d vtx1, Real2d vtx2);

    /// Copy constructor
    VertexTriangle2d(const VertexTriangle2d &map);

    virtual ~VertexTriangle2d();
    virtual VertexTriangle2d* clone() const;
    virtual bool curved() const;
    virtual MapReal2d jacobian(const Real x, const Real y) const {
      return jacobian_;
    }

    Real2d operator()(const Real x, const Real y) const {
      conceptsAssert(0.0 <= y, Assertion());
      conceptsAssert(0.0 <= x, Assertion());
      conceptsAssert(y <= x, Assertion());
      conceptsAssert(x <= 1.0, Assertion());

      Real2d res(vtx1_);  res -= vtx0_;  res *= x;
      Real2d res2(vtx2_); res2 -= vtx1_; res2 *= y;
      res += res2;

      res += vtx0_;
      return res;
    }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Vertices of the triangle
    Real2d vtx0_, vtx1_, vtx2_;
    /// Jacobian
    MapReal2d jacobian_;
  };

  // ******************************************************* MapTriangle2d **

  /** A 2D element map for a triangle. The map is specified by a
      formula.  For easier construction of the element map, the
      parameter domain for the formula is a rectangle
      \f$[0,a] \times [0,b]\f$.
      The left side of the rectangle is degenerated into a
      point using a Duffy transformation.
  */
  class MapTriangle2d : public MappingTriangle2d {
  public:
    /** Constructor. The values of scX and scY are only for the map,
	they are not needed by the user for mapping a point from the
	reference element onto the real element.

	@param map The element map for this triangle as a string, 
	x and y are the allowed variables, the first and the
	second component are separated by a comma.
	@param scX The range of x is [0, scX].
	@param scY The range of y is [0, scY].
    */
    MapTriangle2d(char* map, const Real scX, const Real scY);

    /// Copy constructor
    MapTriangle2d(const MapTriangle2d &map);

    virtual ~MapTriangle2d();
    virtual MapTriangle2d* clone() const;
    virtual bool curved() const;
    virtual MapReal2d jacobian(const Real x, const Real y) const;

    Real2d operator()(const Real x, const Real y) const {
      conceptsAssert(0.0 <= y, Assertion());
      conceptsAssert(0.0 <= x, Assertion());
      conceptsAssert(y <= x, Assertion());
      conceptsAssert(x <= 1.0, Assertion());
      return Real2d(map_, scx_ * x, scy_ * (x == 0 ? 0 : y/x));
    }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Parsed formula for the map
    uchar* map_;

    /// Length of the parsed formula
    uint sz_;

    /// Right border of the x parameter domain
    const Real scx_;

    /// Right border of the y parameter domain
    const Real scy_;
  };

  // ********************************************************* MappingQuad2d **

  /** A 2D element map for a quadrilateral. The reference element is the
      unit square.
      @author Philipp Frauenfelder, 2002.
  */
  class MappingQuad2d : public Map2d {
  public:
    /** Returns a point in 2D mapped from the unit square [0,1]<sup>2</sup>
	onto the element in the original mesh.
	@return \f$F_K(x,y)\f$
    */
    virtual Real2d operator()(Real x, Real y) const = 0;

    /// Returns determinant of the Jacobian
    virtual Real jacobianDeterminant(const Real x, const Real y) const;

    /// Returns the Jacobian in a 2D linear map
    virtual MapReal2d jacobian(const Real x, const Real y) const = 0;

    /// Returns the inverse of the Jacobian in a 2D linear map
    virtual MapReal2d jacobianInverse(const Real x, const Real y) const;

    /** Returns factor of differential element integrating over an edge.

	The factor is either \f$\sqrt{J_{11}^2+J_{21}^2}\f$ or
	\f$\sqrt{J_{12}^2+J_{22}^2}\f$.

	The routine exists explicitly because it is more efficient
	to implement it than to use jacobian itself, i.e. for affine
	elements the line element is constant.

	@param x    local variable on edge (in [0,1]),
	            stands for \f$\xi_1\f$ or \f$\xi_2\f$
	@param edge number of edge, i.e.
	            0:  \f$\xi_2 = 0\f$,
		    1:  \f$\xi_1 = 1\f$,
		    2:  \f$\xi_2 = 1\f$,
		    3:  \f$\xi_1 = 0\f$
    */
    virtual Real lineElement(const Real x, const uint edge) const;

    /// Returns a copy of the map
    virtual MappingQuad2d* clone() const = 0;
  };

  // ************************************************************* MapQuad2d **

  /** A 2D element map for a quadrilateral given by a formula.  The
      reference element is the unit square. For easier construction of
      the element map, a rectangle square \f$[0,a] \times [0,b]\f$ can
      be given as the parameter domain.

      If the map gives curved boundaries, they are approxmated better
      with more refined elements.
  */
  class MapQuad2d : public MappingQuad2d {
  public:
    /** Constructor. The values of scX and scY are only for the map,
	they are not needed by the user for mapping a point from the 
	reference element onto the Real element.

	@param map The element map for this quadrilateral as a string,
	x and y are the variables, the first and the
	second component are separated by a comma.
	@param scX The range of x is [0, scX]
	@param scY The range of y is [0, scY]
	@param org The origin of the element
	@param stretch Scaling factor of the element
    */
    MapQuad2d(char* map, Real scX, Real scY, Real2d org = Real2d(0.0, 0.0),
	      Real stretch = 1.0);
    MapQuad2d(char* map, Real scX, Real scY, Real2d org,
	      Real2d scaling);

    /// Copy constructor
    MapQuad2d(const MapQuad2d &map);
    virtual ~MapQuad2d();

    virtual Real2d operator()(Real x, Real y) const { 
      Real2d v(map_, scx_ * x, scy_ * y);
      return Real2d(v[0]*scaling_[0] + org_[0], v[1]*scaling_[1] + org_[0]);
    }

    virtual MapReal2d jacobian(const Real x, const Real y) const;

    virtual MapQuad2d* clone() const { return new MapQuad2d(*this); }
  private:
    /// Parsed formula for the map
    uchar* map_;

    /// Length of the parsed formula
    uint sz_;

    /// Right border of the x parameter domain
    Real scx_;

    /// Right border of the y parameter domain
    Real scy_;

    Real2d org_;
    Real2d scaling_;

    /// Private assignement operator
    void operator=(const MapQuad2d &);
  };

  // ********************************************************** VertexQuad2d **

  /** A 2D element map for a quadrilateral given by a the four vertices.
      @author Philipp Frauenfelder, 2002.
      @test test::TestJacobian2D
      @test test::TestMatrices2D
  */
  class VertexQuad2d : public MappingQuad2d {
  public:
    /** Constructor. The four vertices have to be given counter clockwise.
     */
    VertexQuad2d(Real2d vtx0, Real2d vtx1, Real2d vtx2, Real2d vtx3);
    /// Copy constructor
    VertexQuad2d(const VertexQuad2d& v);
    virtual ~VertexQuad2d();
    virtual Real2d operator()(Real x, Real y) const;

    /** Returns the jacobian of the element map
	\f[ \left(\frac{\partial F_K}{\partial\vec\xi}\right) =
	\left\{\frac{\partial x^i}{\partial \xi_j}\right\}_{i,j=1}^2 \f]
    */
    virtual MapReal2d jacobian(const Real x, const Real y) const;

    /** Returns factor of differential element integrating over an edge.

	@see MappingQuad2d
     */
    virtual Real lineElement(const Real x, const uint edge);

    virtual VertexQuad2d* clone() const;
  private:
    /// Vertices
    Real2d vtx_[4];
    /// Length of Edges, used for 
    Real len_[4];
    /// Length computed?
    bool hasLength_;
  };

  // ********************************************************* BlendingQuad2d **

  /** A 2D element map for a curved quadrilateral.

      The mapping is defined by the Linear Blending Function Method
      (proposed by Gordon and Hall, 1971). Therefor the mapping
      of the edges has to be given.

      @test test::TestBlending2D

      @author, Kersten Schmidt, 2004
  */
  class BlendingQuad2d : public MappingQuad2d {
  public:
    /** Constructor

        The edge mappings are given, at least one of two adjacent
	edge mappings are neccessary.

	First edge is \f$\xi_2=0\f$, where \f$\xi_2\f$ is the second
	local coordinate. The edges have to be given counter clockwise.

	The last vertex of a edge has to match with the first vertex of the
	following edge. So for one edge of adjacent cells not the same
	edge mapping can be taken.
    */
    BlendingQuad2d(const MappingEdge2d* edgemap0, const MappingEdge2d* edgemap1,
		   const MappingEdge2d* edgemap2, const MappingEdge2d* edgemap3);
    /// Copy constructor
    BlendingQuad2d(const BlendingQuad2d& v);
    ~BlendingQuad2d();
    virtual Real2d operator()(Real x, Real y) const;
    virtual MapReal2d jacobian(const Real x, const Real y) const;
    virtual Real lineElement(const Real x, const uint edge) const;
    virtual BlendingQuad2d* clone() const { return new BlendingQuad2d(*this); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Vector that contains the vertices.
    Real2d vtx_[4];
    /// Map of the edges
    const MappingEdge2d* edgemap_[4];
    /// Edges, for which no mapping was given and a straight edge is taken
    bool existEdgemap_[4];
  };

  // ********************************************************* MapTriangle3d **

  /** A 3D element map for a triangle.
   */
  class MapTriangle3d : public Map2d {
  public:
    /** Constructor. The values of scX and scY are only for the map,
	they are not needed by the user for mapping a point from the 
	reference element onto the Real element.

	@param map The element map for this triangle as a string, 
	x and y are the allowed variables, the first and the
	second component are separated by a comma.
	@param scX The range of x is [0, scX].
	@param scY The range of y is [0, scY].
    */
    MapTriangle3d(char* map, Real scX, Real scY);

    /** Constructor. Takes the three physical corners of the triangle and
	computes the element map: \f$F_K : S \rightarrow K\f$ with 
	\f$\vec x = F_K(\xi) = B \cdot \vec \xi + \vec b\f$, where
	\f$\vec b = \mbox{vtx0}\f$ and \f$B = [ (\mbox{vtx1} - \vec b) 
	(\mbox{vtx2} - \mbox{vtx1}) ]\f$.
    */
    MapTriangle3d(Real3d vtx0, Real3d vtx1, Real3d vtx2);

    /// Copy constructor
    MapTriangle3d(const MapTriangle3d &map);
    ~MapTriangle3d() { delete[] map_; };

    /** Returns a point in 3D mapped from the unit square \f$[0,1]^2\f$
	onto the element in the original mesh.

	@return \f$F_K(x,y)\f$
	@param x \f$\in [0, 1]\f$
	@param y \f$\in [0, 1]\f$
    */
    Real3d operator()(Real x, Real y) const { 
      if (map_)
	return Real3d(map_, scx_ * x, scy_ * y);
      else {
	Real3d res;
	res.lincomb(B1_, B2_, x, x*y);
	res += b_;
	return res;
      }
    }

    /// Returns a copy of the map
    inline MapTriangle3d* clone() const {return new MapTriangle3d(*this);};
  private:
    /// Parsed formula for the map
    uchar* map_;

    /// Length of the parsed formula
    uint sz_;

    /// Right border of the x parameter domain
    Real scx_;

    /// Right border of the y parameter domain
    Real scy_;

    /// Computed map
    Real3d B1_, B2_, b_;

    /// Private assignement operator
    void operator=(const MapTriangle3d &);
  };

} // namespace concepts

#endif // elementMaps_hh
