/* Topology.
 * how to subdivide a hex
 */

#ifndef hexsubdiv_hh
#define hexsubdiv_hh

#include <iostream>
#include "strategyChange.hh"
#include "subdivision.hh"

namespace concepts {

  // forward declaration
  class Hexahedron;
  class Z2;
  class Z4;
  class Quad;
  class Vertex;

  // ******************************************************** HexSubdivision **

  /** Interface for topological subdivision strategies for hexahedrons.

      Every specialization of HexSubdivision should be a singleton
      whose instance can be get with a call to instance(). It
      implements the method how the children of a hexahedron are
      created. To have access to all members of the class Hexahedron, the
      specializations of HexSubdivision should be friend to Hexahedron.

      @see Hexahedron
      @test test::DeadLock1
      @test test::DeadLock2
      @test test::DeadLock3
      @author Philipp Frauenfelder, 2002
      @ingroup geometry
  */
  class HexSubdivision : public Subdivision {
  public:
    /** If no children of q exist, they are created.
     */
    virtual void createChildren(Hexahedron& q) const = 0;
    /** If there are children, they are removed. The additional faces,
	edges and vertices are removed as well.
    */
    virtual void removeChildren(Hexahedron& q) const = 0;
  protected:
    //@{
    /**	Gets the two children of \c q and sorts them.  The sorting
	rule is given in the name and parametrized by the orientation
	flags \c rho and \c tau.

	The code 0011 means the following:
	<table>
	<tr><td>\c rho \\ \c tau </td>
	    <td>\b 0 </td><td>\b 1 </td><td>\b 2 </td><td>\b 3 </td></tr>
	<tr><td>\b 0 </td><td>0</td><td>0</td><td>1</td><td>1</td></tr>
	<tr><td>\b 1 </td><td>1</td><td>1</td><td>0</td><td>0</td></tr>
	</table>
	The indices 0 and 1 denote the index in the call to Quad::child which
	has to be used to get the correct face back.

	@return \c A and \c B
    */
    void subFace0011(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B) const;
    void subFace1100(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B) const;
    void subFace0110(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B) const;
    void subFace1001(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B) const;
    //@}
    //@{
    /** Gets the four children of \c q and sorts them. The sorting
	rule is given in the name and and parametrized by the orientation
	flags \c rho and \c tau.

	The code 0123 means the following:
	<table>
	<tr><td>\c rho \\ \c tau </td>
	    <td>\b 0 </td><td>\b 1 </td><td>\b 2 </td><td>\b 3 </td></tr>
	<tr><td>\b 0 </td><td>0</td><td>1</td><td>2</td><td>3</td></tr>
	<tr><td>\b 1 </td><td>1</td><td>2</td><td>3</td><td>0</td></tr>
	</table>
	The indices 0, 1, 2 and 3 denote the index in the call to
	Quad::child which has to be used to get the correct face back.

	@return \c A, \c B, \c C and \c D
     */
    void subFace0123(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D) const;
    void subFace1230(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D) const;
    void subFace2301(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D) const;
    //@}
    /**
     * Takes a face and tries to subdivide it into 4 (by setting the
     * strategy to <tt>QuadSubdiv4</tt>).  Firstly, a normal
     * subdivision in one shot into 4 children is tried.  If this
     * fails (by catching a \c StrategyChange exception), then a
     * previous subdivision into 2 children was executed on this
     * face.  The orthogonal subdivision into 2 children is then
     * reserved (by setting the strategy accordingly) on the two
     * children of the face. This can be called a subdivision into 4
     * children in two shots.
     * @param q Face to be subdivided into 4
     * @return true if the subdivision in one shot succeeded, false otherwise
     * @throw StrategyChange if the subdivision in two shots fails too
     */
    bool faceSubdiv4(Quad& q) const throw(StrategyChange);

    /** 
     * Looks for the quad in \c q which contains the vertex
     * <tt>v</tt>.  This is done by a loop over the four entries of \c
     * q and checking the indicated vertex (given by the index
     * location <tt>vtxIdx</tt>).
     * 
     * @param v Vertex to look for
     * @param q List of quads to search through
     * @param tau Orientation flags of the quads
     * @param rho Rotation flags of the quads
     * @param vtxIdx Index location of the vertex in the quads counted
     *        from vertex 0 in the quad in \c rho = 0 direction
     * 
     * @return the matching quad
     */
    Quad* findFace(const Vertex* const v, Quad* const q[4],
		   const uint tau, const uint rho, const uint vtxIdx) const;
    //@{
    /** 
     * Subdivides a face of a hexehedron \c h into 4 children.
     * \c normal is taken into account: \c true means that the face is
     * to be subdivided in one shot, \c false means two shots.
     * 
     * @param h Hexahedron of which the face has to be subdivided
     * @param A Result
     * @param B Result
     * @param C Result
     * @param D Result
     * @param normal one or two shots
     */
    void subdivFace0(Hexahedron& h, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D, bool normal) const;
    void subdivFace1(Hexahedron& h, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D, bool normal) const;
    void subdivFace2(Hexahedron& h, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D, bool normal) const;
    void subdivFace3(Hexahedron& h, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D, bool normal) const;
    void subdivFace4(Hexahedron& h, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D, bool normal) const;
    void subdivFace5(Hexahedron& h, Quad*& A, Quad*& B,
		     Quad*& C, Quad*& D, bool normal) const;
    //@}
  };

  // ************************************************************ HexSubdiv8 **

  /** Subdivision strategy for hexahedrons which generates 8 children.
      This is the default strategy.

      The children are generated by subdividing each face into 4
      quadrilaterals and introducing 12 new faces.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv8 : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv8* instance();
    virtual ~HexSubdiv8();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv8> instance_;
  };

  // *********************************************************** HexSubdiv2x **

  /** Subdivision strategy for hexahedrons which generates 2 children
      perpendicular to the x direction.

      The children are generated by subdividing the respective faces into 2
      quadrilaterals and introducing 1 new face.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv2x : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv2x* instance();
    virtual ~HexSubdiv2x();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv2x> instance_;
  };

  // *********************************************************** HexSubdiv2y **

  /** Subdivision strategy for hexahedrons which generates 2 children
      perpendicular the y direction.

      The children are generated by subdividing the respective faces into 2
      quadrilaterals and introducing 1 new face.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv2y : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv2y* instance();
    virtual ~HexSubdiv2y();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv2y> instance_;
  };

  // *********************************************************** HexSubdiv2z **

  /** Subdivision strategy for hexahedrons which generates 2 children
      perpendicular to the z direction.

      The children are generated by subdividing the respective faces into 2
      quadrilaterals and introducing 1 new face.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv2z : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv2z* instance();
    virtual ~HexSubdiv2z();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv2z> instance_;
  };

  // *********************************************************** HexSubdiv4x **

  /** Subdivision strategy for hexahedrons which generates 4 children along
      the x direction.

      The children are generated by subdividing the respective faces into 2
      or 4 quadrilaterals and introducing one new edge and 4 new faces.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv4x : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv4x* instance();
    virtual ~HexSubdiv4x();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv4x> instance_;
  };

  // *********************************************************** HexSubdiv4y **

  /** Subdivision strategy for hexahedrons which generates 4 children along
      the y direction.

      The children are generated by subdividing the respective faces into 2
      or 4 quadrilaterals and introducing one new edge and 4 new faces.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv4y : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv4y* instance();
    virtual ~HexSubdiv4y();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv4y> instance_;
  };

  // *********************************************************** HexSubdiv4z **

  /** Subdivision strategy for hexahedrons which generates 4 children along
      the z direction.

      The children are generated by subdividing the respective faces into 2
      or 4 quadrilaterals and introducing one new edge and 4 new faces.

      @author Philipp Frauenfelder, 2002
  */
  class HexSubdiv4z : public HexSubdivision {
  public:
    virtual void createChildren(Hexahedron& q) const;
    virtual void removeChildren(Hexahedron& q) const;
    static const HexSubdiv4z* instance();
    virtual ~HexSubdiv4z();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<HexSubdiv4z> instance_;
  };

} // namespace concepts

#endif // hexsubdiv_hh
