/* Mesh converter in 2D from triangular to quadrilateral
 */

#include <memory>
#include "meshConverter.hh"

namespace concepts {

  // ****************************************************** ConvertMeshQuads **

  ConvertMeshQuads::ConvertMeshQuads(Mesh2& m, const uint level) :
    m_(m), vertexList_(0), edgeList_(0), newQuadList_(0), quadList_(0),
    hasQuads(false), hasTris(false) {
    std::auto_ptr<Scan2> sc(m_.scan());
    while (*sc) {
      Cell2& c( (*sc)++ );
      convertCell_(c, level);
      conceptsAssert(hasTris || hasQuads, concepts::Assertion());
    }
  }

  ConvertMeshQuads::~ConvertMeshQuads() {
    deleteStuff_();
  }

  std::ostream& ConvertMeshQuads::info(std::ostream& os) const {
    return os << "ConvertMeshQuads(ncell = " << ncell() << ", converted from "
	      << m_ << ")";
  }

  void ConvertMeshQuads::convertCell_(Cell2& c, const uint level) {
    if (level == 0) {
      Triangle2d* t = dynamic_cast<Triangle2d*>(&c);
      if (t && hasQuads) {
	deleteStuff_();
	throw conceptsException
	  (MissingFeature
	   ("use of Quads and Triangle in the same mesh not yet implemented"));
      }
      if (t) {
	Triangle& tconn(t->connector());
	hasTris = true;

	uint attrib;
	if (tconn.attrib().attrib() == 0) attrib = tconn.key();
	else attrib = tconn.attrib().attrib();

	// new middle vertex
	Vertex* v = new Vertex(attrib);
	vertexList_ = new Joiner<Vertex*, 1>(v, vertexList_);

	// three new edges
	Edge* e[3];
	for (uint i = 0; i < 3; ++i) {
	  e[i] = new Edge(*(tconn.edge(i)->child(0)->vertex(1)), *v, attrib);
	  edgeList_ = new Joiner<Edge*, 1>(e[i], edgeList_);
	}

	// three new connectors
	Quad* q[3];
	q[0] = new Quad(*(tconn.edge(0)->child(tconn.rho(0) ? 1 : 0)),
			*e[0], *e[2],
			*(tconn.edge(2)->child(tconn.rho(2) ? 0 : 1)), attrib);
	q[1] = new Quad(*(tconn.edge(0)->child(tconn.rho(0) ? 0 : 1)),
			*(tconn.edge(1)->child(tconn.rho(1) ? 1 : 0)),
			*e[1], *e[0], attrib);
	q[2] = new Quad(*(tconn.edge(1)->child(tconn.rho(1) ? 0 : 1)),
			*(tconn.edge(2)->child(tconn.rho(2) ? 1 : 0)),
			*e[2], *e[1], attrib);

	// three new element maps
	Real3d coord[4];
	for (uint i = 0; i < 3; ++i) {
	  coord[i] = c.vertex(i);
	  coord[i] += c.vertex( (i+1)%3 );
	  coord[i] *= 0.5;
	}
	coord[3] = t->center();
	VertexQuad2d* m[3];
	m[0] = new VertexQuad2d(c.vertex(0), coord[0], coord[3], coord[2]);
	m[1] = new VertexQuad2d(coord[0], c.vertex(1), coord[1], coord[3]);
	m[2] = new VertexQuad2d(coord[1], c.vertex(2), coord[2], coord[3]);

	// three new cells
	for (uint i = 0; i < 3; ++i) {
	  newQuadList_ = new Joiner<Quad2d*, 1>(new Quad2d(*q[i], *m[i]),
						newQuadList_);
	  delete m[i];
	  quadList_.push_back(newQuadList_->value());
	}
      }
      Quad2d* q = dynamic_cast<Quad2d*>(&c);
      if (q && hasTris) {
	deleteStuff_();
	throw conceptsException
	  (MissingFeature
	   ("use of Quads and Triangle in the same mesh not yet implemented"));
      }
      if (q) {
	hasQuads = true;
	// nothing to do
	quadList_.push_back(q);
      }
    } else {
      Cell2* chld = 0;
      for (uint i = 0; (chld = c.child(i)) != 0; ++i)
	convertCell_(*chld, level-1);
    }
  }

  void ConvertMeshQuads::deleteStuff_() {
    Joiner<Vertex*, 1>::destructor(vertexList_);
    Joiner<Edge*, 1>::destructor(edgeList_);
    Joiner<Quad2d*, 1>* tmp = newQuadList_;
    while (tmp) {
      delete &(tmp->value()->connector());
      tmp = (*tmp)[0];
    }
    Joiner<Quad2d*, 1>::destructor(newQuadList_);
  }

} // namespace concepts
