/* Meshes
 * Combining cells to a mesh to represent a computational domain
 * Importing 2D meshes from Matlab files
 */

#ifndef meshImport2D_hh
#define meshImport2D_hh

#include <vector>
#include <map>
#include "mesh.hh"
#include "toolbox/array.hh"

namespace concepts {

  // ********************************************************** Import2dMesh **

  /** Imports 2D mesh with triangles and quadrilaterals (possibly mixed).

      The notation for the meshes is used in [1] for instance.

      @author Philipp Frauenfelder, 2002
      @see [1] Alberty, Carstensen, Funken, <a href="http://www.numerik.uni-kiel.de/reports/1998/98-11.html"><em>Remarks Around 50 Lines of
      MATLAB: Finite Element Implementation</em></a>, 1998
      @test test::BEMSpaceTest
      @ingroup geometry
  */
  class Import2dMesh : public Mesh2 {
  public:
    /** Constructor. Reads the data from three files and creates a mesh
	from it. The vertices, edges and cells should be number in
	striclty increasing order, ie. 1, 2, 3 etc.
	@param coord File with coordinates of vertices. Format:
	       <pre>&lt;node&gt; &lt;x&gt; &lt;y&gt;</pre>
	       or
	       <pre>&lt;node&gt; &lt;x&gt; &lt;y&gt; &lt;z&gt;</pre>
	       For instance:
	       <pre>1 1.0 1.0
2 0.0 0.5
3 0.0 0.0
4 -1.0 2.0</pre>
               or
	       <pre>1 1.0 1.0 0.0
2 0.0 0.5 0.5
3 0.0 0.0 1.0
4 -1.0 2.0 0.5</pre>
	@param elms File with node numbers of triangles or quads. Format:
	       <pre>&lt;triangle&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt;</pre>
	       or
	       <pre>&lt;quad&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt; &lt;node3&gt;</pre>
	       For instance:
	       <pre>1 1 2 3</pre>
	       or
	       <pre>1 1 2 3 4</pre>
	@param boundary File with node numbers of edges with
	       boundary conditions. Format:
	       <pre>&lt;edge&gt; &lt;node0&gt; &lt;node1&gt; &lt;attr&gt;</pre>
	       For instance:
	       <pre>1 1 3 0</pre>
	@param idxStart Starting point of indices in the files (1 or 0)
    */
    Import2dMesh(const std::string coord, const std::string elms,
		 const std::string boundary, const uint idxStart = 1);

    /** Constructor. Reads the data from two files and creates a mesh
	from it. The vertices, edges and cells should be number in
	strictly increasing order, ie. 1, 2, 3 etc.
	@param coord File with coordinates of vertices. Format:
	       <pre>&lt;node&gt; &lt;x&gt; &lt;y&gt;</pre>
	       or
	       <pre>&lt;node&gt; &lt;x&gt; &lt;y&gt; &lt;z&gt;</pre>
	       For instance:
	       <pre>1 1.0 1.0
2 0.0 0.5
3 0.0 0.0
4 -1.0 2.0</pre>
               or
	       <pre>1 1.0 1.0 0.0
2 0.0 0.5 0.5
3 0.0 0.0 1.0
4 -1.0 2.0 0.5</pre>
	@param elms File with node numbers of triangles or quads. Format:
	       <pre>&lt;triangle&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt;</pre>
	       or
	       <pre>&lt;quad&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt; &lt;node3&gt;</pre>
	       For instance:
	       <pre>1 1 2 3</pre>
	       or
	       <pre>1 1 2 3 4</pre>
	@param idxStart Starting point of indices in the files (1 or 0)
    */
    Import2dMesh(const std::string coord, const std::string elms,
		 const uint idxStart = 1);

    virtual ~Import2dMesh();
    virtual uint ncell() const { return cell_.size(); }
    virtual Scan2* scan() { return new S(cell_); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Scanner class for Import2dMesh
    class S : public Scan<Cell2> {
    public:
      inline S(std::vector<Cell2*>& cell) :
	idx_(cell.begin()), cell_(cell) {}
      inline S(const S& scan) : idx_(scan.idx_), cell_(scan.cell_) {}
      inline bool eos() const { return idx_ == cell_.end(); }
      inline Cell2& operator++(int) { return *(*idx_++); }
      inline Scan2* clone() const { return new S(*this); }
    private:
      std::vector<Cell2*>::iterator idx_;
      std::vector<Cell2*>& cell_;
    };
    /** Stores the nodes and edges of a cell to be created.
	@author Philipp Frauenfelder, 2002
    */
    class NodeCell {
    protected:
      /// Node numbers of the cell
      Array<uint> nodes_;
      /// Edges of the cell
      Array<Edge*> edges_;
    public:
      /** Constructor. Allocates \c nodes_ and \c edges_.
	  @param size Number of nodes / edges. Ie. 3 for a triangle or
	         4 for a quadrilateral.
      */
      NodeCell(const uint size) : nodes_(size), edges_(size, 0) {}
      virtual ~NodeCell();
      /// Returns the node with index \c i
      uint operator[](const uint i) const { return nodes_[i]; }
      /// Returns the number of nodes / edges
      uint size() const { return nodes_.size(); }
      /// Returns the edge with index \c i
      Edge*& edge(const uint i) { return edges_[i]; }
      /// Returns the number of the starting vertex of edge i
      uint estart(uint i) const { return nodes_[i % size()]; }
      /// Returns the number of the stopping vertex of edge i
      uint eend(uint i) const { return nodes_[ (i+1)%size() ]; }
      /** Creates and returns the cell
	  @param vertices Coordinates of the vertices
	  @param dim Spatial dimension of the vertices
      */
      virtual Cell2* cell(const std::vector<Real3d>& vertices,
			  uint dim) const = 0;
    };
    /** Stores the nodes and edges of a triangular cell to be created.
	@author Philipp Frauenfelder, 2002
    */
    class NodeTriangle : public NodeCell {
    public:
      /** Constructor. Takes the nodes numbers of the 3 vertices and
	  stores them in \c nodes_
      */
      NodeTriangle(const uint one, const uint two, const uint three);
      virtual ~NodeTriangle();
      virtual Cell2* cell(const std::vector<Real3d>& vertices, uint dim) const;
    };
    /** Stores the nodes and edges of a quadrilateral cell to be created.
	@author Philipp Frauenfelder, 2002
    */
    class NodeQuad : public NodeCell {
    public:
      /** Constructor. Takes the nodes numbers of the 4 vertices and
	  stores them in \c nodes_
      */
      NodeQuad(const uint one, const uint two,
	       const uint three, const uint four);
      virtual ~NodeQuad();
      virtual Cell2* cell(const std::vector<Real3d>& vertices, uint dim) const;
    };
    /// List of vertices
    std::vector<Vertex*> vtx_;
    /// List of edges
    std::vector<Edge*> edg_;
    /// List of cells
    std::vector<Cell2*> cell_;

    /** Reads integers from the string
	@return Number of integers read
    */
    uint readInts_(const std::string& i, std::vector<int>& v) const;

    /** Reads an index and at most 3 Real values from the string
	@return Number of Reals read
     */
    uint readLine_(const std::string& i, uint* idx, Real3d* coord) const;

    /// Deletes the content of \c cells
    void clear_(std::vector<NodeCell*>& cells) const;

    /** Creates the edge with vertices \c ix and \c iy using \c edges as
	repository. If \c e is different from 0, the found or created edge
	is stored there.
	@param ix Index of first vertex
	@param iy Index of second vertex
	@param edges Stores the edges index by smaller and greater index
	@param attr Attribute of the edge
	@param e Return value
	@pre \c ix and \c iy are vertices which are mentioned in \c vtx_
    */
    void createEdge_(uint ix, uint iy,
		     std::map<uint, std::map<uint, Edge*> >& edges,
		     Attribute attr = 0, Edge** e = 0);
  };

} // namespace concepts

#endif // meshImport2D_hh
