/* Meshes
 * Combining cells to a mesh to represent a computational domain
 * Importing 3D meshes from Matlab files
 */

#ifndef meshImport3D_hh
#define meshImport3D_hh

#include <vector>
#include <map>

#if __ICC == 800
#  include <hash_map>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/hash_map>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <hash_map>
#    define __gnu_cxx std
#  endif
#endif

#include "mesh.hh"
#include "toolbox/array.hh"

namespace concepts {
  /** Stores a number of indices in a ordered fashion. The first index is
      the smallest, the rest of the indices keep their siblings.
      @author Philipp Frauenfelder, 2002
      @see Index::sort_
  */
  struct Index {
    /// Stores the data
    Array<uint> n_;
    /** Sorts the data somewhat. This routine is called by the constructor.
	The smallest entry in the data given to the constructor is stored
	as the first entry of \c n_. The rest of the data is order in the
	following way: the entry on the left or on the right (depending on
	which is the smaller) fixes the orientation in which the rest of
	the entries are stored into \c n_.
    */
    void sort_();

    /// Constructor for use in a triangle
    Index(const uint n0, const uint n1, const uint n2);
    /// Constructor for use in a quad
    Index(const uint n0, const uint n1, const uint n2, const uint n3);
    /// Copy constructor
    Index(const Index& i) : n_(i.n_) {}
    /// Returns number of entries in \c n_
    uint size() const { return n_.size(); }
    /// Returns the ith entry of \c n_
    uint operator[](const uint i) const { return n_[i]; }
    /// Returns the ith entry of \c n_
    uint& operator[](const uint i) { return n_[i]; }
    /// < operator for
    bool operator<(const concepts::Index& y) const;
    /// == operator
    bool operator==(const concepts::Index& y) const;
  };
}

namespace __gnu_cxx {
  /// Computes a hash for \c i
  size_t hash_value(const concepts::Index& i);

  /// Hash function class
  template<>
  struct hash<concepts::Index> {
    /// Computes the hash function value for \c x
    size_t operator()(const concepts::Index& x) const {
      return hash_value(x); }
  };

  /// Key equal function class
  struct IndexEqual {
    bool operator()(const concepts::Index& x, const concepts::Index& y) const {
      return x == y;
    }
  };
} // namespace __gnu_cxx

namespace concepts {

  // ********************************************************** Import3dMesh **

  /** Imports 3D mesh with tetrahedra and hexahedra. They can be mixed in the
      mesh---the only problem is, that tetrahedra and hexahedra cannot connect
      to each other, but this importer can be enhanced to support
      prisms or pyramids as well.

      The notation for the meshes is used in [1] for instance.

      @author Philipp Frauenfelder, 2002
      @test test::DeadLock1
      @test test::DeadLock3
      @see [1] Alberty, Carstensen, Funken, <a href="http://www.numerik.uni-kiel.de/reports/1998/98-11.html"><em>Remarks Around 50 Lines of
      MATLAB: Finite Element Implementation</em></a>, 1998
      @ingroup geometry
  */
  class Import3dMesh : public Mesh3 {
  public:
    /** Constructor. Reads the data from three files and creates a mesh
	from it. The vertices, faces and cells should be number in
	striclty increasing order, ie. 1, 2, 3 etc.
	@param coord File with coordinates of vertices. Format:
	       <pre>&lt;node&gt; &lt;x&gt; &lt;y&gt; &lt;z&gt;</pre>
	       For instance:
	       <pre>1 1.0 1.0 2.0
2 0.0 0.5 1.0
3 0.0 0.0 0.0
4 -1.0 2.0 3.0</pre>
	@param elms File with node numbers of tetrahedra or hexahedra. Format:
	       <pre>&lt;tetrahedron&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt; &lt;node3&gt;</pre>
	       or
	       <pre>&lt;hexahedron&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt; &lt;node3&gt; &lt;node4&gt; &lt;node5&gt; &lt;node6&gt; &lt;node7&gt;</pre>
	       For instance:
	       <pre>1 1 2 3 4</pre>
	       or
	       <pre>1 1 2 3 4 5 6 7 8</pre>
	@param boundary File with node numbers of edges with boundary
	       conditions and attribute number. Format for a triangle:
	       <pre>&lt;face&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt; &lt;attr&gt;</pre>
	       Format for a quad:
	       <pre>&lt;face&gt; &lt;node0&gt; &lt;node1&gt; &lt;node2&gt; &lt;node3&gt; &lt;attr&gt;</pre>
	       Format for an edge:
	       <pre>&lt;edge&gt; &lt;node0&gt; &lt;node1&gt; &lt;attr&gt;</pre>
	       Format for a vertex:
	       <pre>&lt;vtx&gt; &lt;node&gt; &lt;attr&gt;</pre>
	       For instance:
	       <pre>1 2 33
2 1 2 66
3 1 2 3 44
4 1 2 3 4 55</pre>
               Vertices should come first in the list of boundary attributes,
	       then edges and triangle or quads last.
	@param idxStart Starting point of indices in the files (1 or 0)
	@param leftHand Indicates if the order of the points in a tet
	       follow the left hand rule
	@param elmtyp only for hexahedras (0: Hexahedron3d is used,
	       1: Parallelepiped3d is used)
    */
    Import3dMesh(const std::string coord, const std::string elms,
		 const std::string boundary, const uint idxStart = 1,
		 const bool leftHand = false, const uint elmtyp = 0);
    virtual ~Import3dMesh();
    virtual uint ncell() const { return cell_.size(); }
    virtual Scan3* scan() { return new S(cell_); }

    class NodeCell;
    class NodeTetrahedron;
    class NodeHexahedron;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Scanner class for Import3dMesh
    class S : public Scan<Cell3> {
    public:
      inline S(std::vector<Cell3*>& cell) :
	idx_(cell.begin()), cell_(cell) {}
      inline S(const S& scan) : idx_(scan.idx_), cell_(scan.cell_) {}
      inline bool eos() const { return idx_ == cell_.end(); }
      inline Cell3& operator++(int) { return *(*idx_++); }
      inline Scan3* clone() const { return new S(*this); }
    private:
      std::vector<Cell3*>::iterator idx_;
      std::vector<Cell3*>& cell_;
    };
    /// List of vertices
    std::vector<Vertex*> vtx_;
    /// List of edges
    std::vector<Edge*> edg_;
    /// List of triangles
    __gnu_cxx::hash_map<Index, Triangle*, __gnu_cxx::hash<Index>,
			__gnu_cxx::IndexEqual> tri_;
    /// List of quads
    __gnu_cxx::hash_map<Index, Quad*, __gnu_cxx::hash<Index>,
			__gnu_cxx::IndexEqual> quad_;
    /// List of cells
    std::vector<Cell3*> cell_;

    /** Reads integers from the string
	@return Number of integers read
    */
    uint readInts_(const std::string& i, std::vector<int>& v) const;

    /// Deletes the content of \c cells
    void clear_(std::vector<NodeCell*>& cells) const;

    /** Creates the edge with vertices \c ix and \c iy using \c edges as
	repository. If \c e is different from 0, the found or created edge
	is stored there.
	@param ix Index of first vertex
	@param iy Index of second vertex
	@param edges Stores the edges index by smaller and greater index
	@param attr Attribute for the edge
	@param e Return value: the new (or found) edge
	@pre \c ix and \c iy are vertices which are mentioned in \c vtx_
    */
    void createEdge_(uint ix, uint iy,
		     std::map<uint, std::map<uint, Edge*> >& edges,
		     Attribute attr = 0, Edge** e = 0);
    /** Creates a triangle. The edges are taken from \c edges or created and
	stored there by calling \c createEdge_.
	@param I Indices of the vertices to create the triangle from
	@param edges Storage area for the edges
	@param attr Attribute for the triangle
	@return the new or found triangle
    */
    Triangle* createTriangle_(const Index& I,
			      std::map<uint, std::map<uint, Edge*> >& edges,
			      Attribute attr = 0);
    /** Creates a quad. The edges are taken from \c edges or created and
	stored there by calling \c createEdge_.
	@param I Indices of the vertices to create the quad from
	@param edges Storage area for the edges
	@param attr Attribute for the triangle
	@return the new or found quad
    */
    Quad* createQuad_(const Index& I,
		      std::map<uint, std::map<uint, Edge*> >& edges,
		      Attribute attr = 0);
  };

  /** Stores the nodes and edges of a cell to be created.
      @author Philipp Frauenfelder, 2002
  */
  class Import3dMesh::NodeCell {
  protected:
    /// Node numbers of the cell
    Array<uint> nodes_;
  public:
    /** Constructor. Allocates \c nodes_ and \c edges_.
	@param size Number of nodes / edges. Ie. 4 for a tetrahedron or
	8 for a hexahedron.
    */
    NodeCell(const uint size) : nodes_(size) {}
    virtual ~NodeCell();
    /// Returns the node with index \c i
    uint operator[](const uint i) const { return nodes_[i]; }
    /// Returns the number of nodes
    uint size() const { return nodes_.size(); }

    /// Returns the number of faces
    virtual uint nofaces() const = 0;
    /// Returns the indices of the nodes which define the face \c i
    virtual Index faceIndex(const uint i) = 0;
    /// Returns the face with index \c i
    virtual Quad** quad(const uint i) = 0;
    /// Returns the face with index \c i
    virtual Triangle** triangle(const uint i) = 0;
    /// Creates and returns the cell
    virtual Cell3* cell(const std::vector<Real3d>& vertices) const = 0;
  };

  /** Stores the nodes and faces of a tetrahedral cell to be created.
      @author Philipp Frauenfelder, 2002
  */
  class Import3dMesh::NodeTetrahedron : public Import3dMesh::NodeCell {
  private:
    /// Faces of the cell
    Array<Triangle*> faces_;
  public:
    /** Constructor. Takes the node numbers of the 4 vertices and
	stores them in \c nodes_
    */
    NodeTetrahedron(const uint one, const uint two, const uint three,
		    const uint four, const bool leftHand = false);
    virtual ~NodeTetrahedron();
    virtual uint nofaces() const { return 4; }
    virtual Index faceIndex(const uint i);
    virtual Tetrahedron3d* cell(const std::vector<Real3d>& vertices) const;
    virtual Quad** quad(const uint i) { return 0; }
    virtual Triangle** triangle(const uint i) { return &(faces_[i]); }
  };

  /** Stores the nodes and faces of a hexahedral cell to be created.
      @author Philipp Frauenfelder, 2002
  */
  class Import3dMesh::NodeHexahedron : public Import3dMesh::NodeCell {
  private:
    /// Faces of the cell
    Array<Quad*> faces_;
    /// Type of element to construct
    uint elmType_;
  public:
    /** Constructor. Takes the node numbers of the 8 vertices and
	stores them in \c nodes_
    */
    NodeHexahedron(const uint one, const uint two, const uint three,
		   const uint four, const uint five, const uint six,
		   const uint seven, const uint eight, const uint elmType = 0);
    virtual ~NodeHexahedron();
    virtual uint nofaces() const { return 6; }
    virtual Index faceIndex(const uint i);
    virtual Cell3* cell(const std::vector<Real3d>& vertices) const;
    virtual Quad** quad(const uint i) { return &(faces_[i]); }
    virtual Triangle** triangle(const uint i) { return 0; }
  };

} // namespace concepts

#endif // meshImport3D_hh
