/* Topology.
 * how to subdivide a quad
 */

#ifndef quadsubdiv_hh
#define quadsubdiv_hh

#include <iostream>
#include "subdivision.hh"

namespace concepts {

  // forward declaration
  class Vertex;
  class Edge;
  class Quad;

  // ******************************************************* QuadSubdivision **

  /** Interface for topological subdivision strategies for quadrilaterals.

      Every specialization of QuadSubdivision should be a singleton
      whose instance can be get with a call to instance(). It
      implements the method how the children of a quadrilateral are
      created. To have access to all members of the class Quad, the
      specializations of QuadSubdivision should be friend to Quad.

      @see Quad
      @author Philipp Frauenfelder, 2002
      @ingroup geometry
  */
  class QuadSubdivision : public Subdivision {
  public:
    /** If no children of q exist, they are created.
     */
    virtual void createChildren(Quad& q) const = 0;
    /** If there are children, they are removed. The additional edges and
	vertices are removed as well.
    */
    virtual void removeChildren(Quad& q) const = 0;
    /** Returns the ith edge which has been newly created by the subdivision.
	An edge which is the result of a refinement is not newly created.
	@param q Quad which has been refined
	@param i Index of the new edge
	@return 0 if the new edge with the index i does not exist
    */
    virtual Edge* newEdge(const Quad& q, uint i) const = 0;
    /** Return the ith vertex which has been newly created by the subdivision.
	An vertex which is the result of a refinement (ie. of an edge) is
	not newly created.
	@param q Quad which has been refined
	@param i Index of the new vertex
	@return 0 if the new vertex iwth the index i does not exist
    */
    virtual Vertex* newVertex(const Quad& q, uint i) const = 0;
  };

  // *********************************************************** QuadSubdiv4 **

  /** Subdivision strategy for quadrilaterals which generates four children.
      This is the default subdivision strategy.

      The children are generated by subdividing the four edges and introducing
      four new edges meeting in the middle of the parent quad.

      @author Philipp Frauenfelder, 2002
  */
  class QuadSubdiv4 : public QuadSubdivision {
  public:
    virtual void createChildren(Quad& q) const;
    virtual void removeChildren(Quad& q) const;
    virtual Edge* newEdge(const Quad& q, uint i) const;
    virtual Vertex* newVertex(const Quad& q, uint i) const;
    static const QuadSubdiv4* instance();
    virtual ~QuadSubdiv4();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<QuadSubdiv4> instance_;
  };

  // ********************************************************** QuadSubdiv2H **

  /** Subdivision strategy for quadrilaterals which generates two children.

      The parent quad's vertical axis is subdivided, the horizontal is not.

      @author Philipp Frauenfelder, 2002
  */
  class QuadSubdiv2H : public QuadSubdivision {
  public:
    virtual void createChildren(Quad& q) const;
    virtual void removeChildren(Quad& q) const;
    virtual Edge* newEdge(const Quad& q, uint i) const;
    virtual Vertex* newVertex(const Quad& q, uint i) const;
    static const QuadSubdiv2H* instance();
    virtual ~QuadSubdiv2H();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<QuadSubdiv2H> instance_;
  };

  // ********************************************************** QuadSubdiv2V **

  /** Subdivision strategy for quadrilaterals which generates two children.

      The parent quad's horizontal axis is subdivided, the vertical is not.

      @author Philipp Frauenfelder, 2002
  */
  class QuadSubdiv2V : public QuadSubdivision {
  public:
    virtual void createChildren(Quad& q) const;
    virtual void removeChildren(Quad& q) const;
    virtual Edge* newEdge(const Quad& q, uint i) const;
    virtual Vertex* newVertex(const Quad& q, uint i) const;
    static const QuadSubdiv2V* instance();
    virtual ~QuadSubdiv2V();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<QuadSubdiv2V> instance_;
  };

} // namespace concepts

#endif // quadsubdiv_hh
