/* graphics in OpenDX
 */

#include <sstream>
#include "dataDX.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "function/vector.hh"
#include "toolbox/elementGraphics.hh"
#include "vertexList.hh"

// debugging DataDX
#define graphDataDXConstr_D 0

// debugging DataDXCounts
#define graphDXCountsConstr_D 0
#define graphDXCountsAppl_D 0

// debugging DataDXCoordinates
#define graphDXCoordinates_D 0 // format of the file: binary or ascii
#define graphDXCoordConstr_D 0
#define graphDXCoordDestr_D 0
#define graphDXCoordAppl_D 0

// debugging DataDXConnections
#define graphDXConn_D 0 // format of the file: binary or ascii
#define graphDXConnAppl_D 0
#define graphDXConnWrite_D 0

// debugging DataDXData
#define graphDXData_D 0 // format of the file: binary or ascii
#define graphDXDataAppl_D 0

namespace graphics {

  // **************************************************************** DataDX **

  template<class F>
  DataDX<F>::DataDX(concepts::Space<F>& spc, const std::string filename, 
		    bool pc, bool boundary, const uint dim) :
    spc_(&spc), filename_(filename), postProcess_(spc), pc_(pc),
    names_(0), vtxList_(0) {

    DEBUGL(graphDataDXConstr_D, "space = " << *spc_);
    DEBUGL(graphDataDXConstr_D, "filename = " << filename_);
    DEBUGL(graphDataDXConstr_D, "dimension = " << dim);
    std::ofstream ofs(filename_.c_str());

    if (boundary) {
      vtxList_.reset(new VertexList(VertexList::COORD));
      postProcess_(*vtxList_);
    }

    std::auto_ptr<DataDXCounts> graphcount
      (new DataDXCounts(vtxList_.get()));
    postProcess_(*graphcount);
    countp_ = graphcount->counts();
    DEBUGL(graphDataDXConstr_D, "Number of points in DX file: countp = " 
	   << countp_);
    graphcount.reset();

    std::auto_ptr<DataDXCoordinates> graphcoord
      (new DataDXCoordinates(ofs, countp_, vtxList_.get(), dim));
    postProcess_(*graphcoord);
    countq_ = graphcoord->counts();
    DEBUGL(graphDataDXConstr_D, "Number of quads in DX file: countq = "
	   << countq_);
    graphcoord.reset();

    // connections one dimension lower than spatial dimension
    // for boundary elements
    std::auto_ptr<DataDXConnections> graphconn
      (new DataDXConnections(ofs, countq_, vtxList_.get(), 
			     (boundary ? dim-1 : dim)));
    postProcess_(*graphconn);
    graphconn.reset();
  }

  template<class F>
  DataDX<F>::DataDX(const DataDX<F>& dx) :
    spc_(dx.spc_), filename_(dx.filename_), postProcess_(dx.postProcess_),
    countp_(dx.countp_), countq_(dx.countq_), pc_(dx.pc_), names_(dx.names_) {

    if (dx.vtxList_.get()) {
      vtxList_.reset(new VertexList(*(dx.vtxList_)));
    }
  }

  template<class F>
  DataDX<F>::~DataDX() {

    std::ofstream ofs(filename_.c_str(), std::ios::app);
    uint j = 0;
    ofs << "# group" << std::endl;
    ofs << "object 2 class group" << std::endl;
    for (__gnu_cxx::slist<std::string>::const_iterator i = names_.begin(); 
	 i != names_.end(); ++i, ++j)
      ofs << "member " << j << " value \"" << *i << "\"" << std::endl;

    ofs << "end" << std::endl;
  }

  template<class F>
  DataDX<F>& DataDX<F>::operator=(const DataDX<F>& dx) {
    if (this != &dx) {
      spc_ = dx.spc_;
      filename_ = dx.filename_;
      postProcess_ = dx.postProcess_;
      countp_ = dx.countp_;
      countq_ = dx.countq_;
      pc_ = dx.pc_;
      names_.clear();
      names_.insert(names_.end(), dx.names_.begin(), dx.names_.end());
      if (dx.vtxList_.get()) {
	vtxList_.reset(new VertexList(*(dx.vtxList_)));
      }
    }
    return *this;
  }

  template<class F>
  void DataDX<F>::operator()(const concepts::Vector<Real>& solution,
			     const std::string& name) {

    std::ofstream ofs(filename_.c_str(), std::ios::app);
    DataDXData graphdata(ofs, solution,
			 (pc_) ? countq_ : countp_, name, vtxList_.get());
    postProcess_(graphdata);
    names_.push_front(name);
  }

  template<class F>
  void DataDX<F>::operator()(const concepts::Formula& frm,
			     const std::string& name) {

    std::ofstream ofs(filename_.c_str(), std::ios::app);
    DataDXData graphdata(ofs, frm,
			 (pc_) ? countq_ : countp_, name, vtxList_.get());
    postProcess_(graphdata);
    names_.push_front(name);
  }

  template<class F>
  void DataDX<F>::operator()(const concepts::ElementFormula<F>& frm,
			     const std::string& name) {
    std::ofstream ofs(filename_.c_str(), std::ios::app);
    DataDXData graphdata(ofs, frm,
			 (pc_) ? countq_ : countp_, name, vtxList_.get());
    postProcess_(graphdata);
    names_.push_front(name);
  }

  template class DataDX<Real>;

  // ********************************************************** DataDXCounts **

  DataDXCounts::DataDXCounts(VertexList* vtxList) : 
    counts_(0), vtxList_(vtxList) {

    // count external points
    if (vtxList_)
      counts_ = vtxList_->size();

    DEBUGL(graphDXCountsConstr_D, "counts = " << counts_);
  }

  uint DataDXCounts::counts() const {
    return counts_;
  }

  void DataDXCounts::operator() (const concepts::Element<Real>& elm) {

    DEBUGL(graphDXCountsAppl_D, "elm = " << elm);

    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<Real>::DATA_DX_COUNTS,
			       *this);
    return;
  }

  // ***************************************************** DataDXCoordinates **

  DataDXCoordinates::DataDXCoordinates(std::ostream& os, const uint count,
				       VertexList* vtxList, 
				       const uint dim) :
    os_(os), vtxList_(vtxList), dim_(dim), counts_(0), cntCoords_(0) {

    DEBUGL(graphDXCoordConstr_D, "dim = " << dim_);

    // endianness of data
#ifdef __sparc
    os_ << "data mode msb" << std::endl;
#endif
#ifdef __i386
    os_ << "data mode lsb" << std::endl;
#endif

    os_ << std::endl << "# coordinates" << std::endl;
    os_ << "object 0 class array type float rank 1 shape " << dim_
#if graphDXCoordinates_D
	<< " items " << count << " data follows" << std::endl;
#else
    << " items " << count << " ieee data follows" << std::endl;
#endif
  }

  DataDXCoordinates::~DataDXCoordinates() {

    if (vtxList_) {
      for (VertexList::iterator i = vtxList_->begin(); 
	   i != vtxList_->end(); ++i, ++cntCoords_) {
	concepts::Real3d x = i->second.coordinates();
	write(x[0]);
	write(x[1]);
	if (dim_ == 3)
	  write(x[2]);
	i->second.index() = cntCoords_;
	DEBUGL(graphDXCoordDestr_D, x << " as nr. " << cntCoords_);
      }
    }
  }

  uint DataDXCoordinates::counts() const {
    return counts_;
  }

  void DataDXCoordinates::operator() (const concepts::Element<Real>& elm) {

    DEBUGL(graphDXCoordAppl_D, "elm " << elm);

    elm.graphics()->operator()
      (elm, concepts::ElementGraphics<Real>::DATA_DX_COORDINATES, *this);
    return;
  }

#if graphDXCoordinates_D
  static int graphDXCoordinatesCnt = 0;
#endif

  void DataDXCoordinates::write(const float x) const {

#if graphDXCoordinates_D
    os_ << x << ' ';
    if ((graphDXCoordinatesCnt++ % dim_) == dim_-1) os_ << std::endl;
#else
    os_.write((char*)&x, sizeof(const float));
#endif
  }

  // ***************************************************** DataDXConnections **

  DataDXConnections::DataDXConnections(std::ostream& os, 
				       const uint count, 
				       VertexList* vtxList, 
				       const uint dim) :
    os_(os), vtxList_(vtxList), counts_(0), dim_(dim) {

    os_ << std::endl << "# connections" << std::endl;
    os_ << "object 1 class array type int rank 1 shape " 
#if graphDXConn_D
	<< (dim_ == 2 ? 4 : 8) << " items " << count << " data follows" 
#else
	<< (dim_ == 2 ? 4 : 8) << " items " << count << " ieee data follows" 
#endif
	<< std::endl;
  }

  DataDXConnections::~DataDXConnections() {

    os_ << std::endl << "attribute \"element type\" string ";
    if (dim_ == 2)
      os_ << "\"quads\"";
    else
      os_ << "\"cubes\"";
    os_ << std::endl;
    os_ << "attribute \"ref\" string \"positions\"" << std::endl;
  }

  void DataDXConnections::operator() (const concepts::Element<Real>& elm) {

    DEBUGL(graphDXConnAppl_D, "counts = " << counts_);

    elm.graphics()->operator()
      (elm, concepts::ElementGraphics<Real>::DATA_DX_CONNECTIONS, *this);
    return;
  }

#if graphDXConn_D
  static int graphDXConnCnt = 1;
#endif

  void DataDXConnections::write(const int x) const {
#if graphDXConn_D
    os_ << x << ' ';
    if (!(graphDXConnCnt++ % (1<<dim_))) os_ << std::endl;
#else
    os_.write((char*)&x, sizeof(const int));
#endif
  }

  // ************************************************************ DataDXData **

  DataDXData::DataDXData(std::ostream& os, 
			 const concepts::Vector<Real>& solution, 
			 const uint count, const std::string& name,
			 VertexList* vtxList) :
    os_(os), solution_(&solution), frm_(0), efrm_(0), coeff_(0), shpfnc_(0),
    vertexSol_(0), vertexSolIndex_(0), name_(name), pc_(0),
    vtxList_(vtxList), vtxData_(0) {

    if (vtxList) vtxData_.resize(vtxList_->size());

    os_ << "# data" << std::endl;
    os_ << "object \"data_" << name_ 
	<< "\" class array type float items " << count 
#if graphDXData_D
	<< " data follows" << std::endl;
#else
    << " ieee data follows" << std::endl;
#endif
  }

  DataDXData::DataDXData(std::ostream& os, const concepts::Formula& frm, 
			 const uint count, const std::string& name,
			 VertexList* vtxList) :
    os_(os), solution_(0), frm_(frm.clone()), efrm_(0), coeff_(0), shpfnc_(0),
    vertexSol_(0), vertexSolIndex_(0), name_(name), pc_(0),
    vtxList_(vtxList), vtxData_(0) {

    if (vtxList) vtxData_.resize(vtxList_->size());

    os_ << "# data" << std::endl;
    os_ << "object \"data_" << name_ 
	<< "\" class array type float items " << count 
#if graphDXData_D
	<< " data follows" << std::endl;
#else
    << " ieee data follows" << std::endl;
#endif
  }

  DataDXData::DataDXData(std::ostream& os,
			 const concepts::ElementFormula<Real>& frm,
			 const uint count, const std::string& name,
			 VertexList* vtxList) :
    os_(os), solution_(0), frm_(), efrm_(&frm), coeff_(0), shpfnc_(0),
    vertexSol_(0), vertexSolIndex_(0), name_(name), pc_(0),
    vtxList_(vtxList), vtxData_(0) {

    if (vtxList) vtxData_.resize(vtxList_->size());

    os_ << "# data" << std::endl;
    os_ << "object \"data_" << name_ 
	<< "\" class array type float items " << count 
#if graphDXData_D
	<< " data follows" << std::endl;
#else
    << " ieee data follows" << std::endl;
#endif
  }

  DataDXData::~DataDXData() {

    if (pc_ == 0) {
      if (vtxList_) {
	Real *d = (Real*)vtxData_;
	for(uint i = 0; i < vtxList_->size(); i++, d++)  write(d[i]);
      }
      os_ << std::endl << "attribute \"dep\" string \"positions\""
	  << std::endl;
    }
    else
      os_ << std::endl << "attribute \"dep\" string \"connections\""
	  << std::endl;

    os_ << "# field" << std::endl;
    os_ << "object \"" << name_ << "\" class field" 
	<< std::endl;
    os_ << "component \"positions\" value 0" << std::endl;
    os_ << "component \"connections\" value 1" << std::endl;
    os_ << "component \"data\" \"data_" << name_ << "\"" << std::endl;
  }

  void DataDXData::operator() (const concepts::Element<Real>& elm)
    throw(concepts::MissingFeature) {

    DEBUGL(graphDXDataAppl_D, "elm = " << elm);

    const concepts::TMatrixBase<Real>* T = &elm.T();
    if (solution_)
      T->extract(*solution_, coeff_);

    elm.graphics()->operator()
      (elm, concepts::ElementGraphics<Real>::DATA_DX_DATA, *this);
    return;

    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void DataDXData::write(const float x) const {

#if graphDXData_D
    os_ << x << std::endl;
#else
    os_.write((char*)&x, sizeof(const float));
#endif
  }

} // namespace graphics
