/* graphics of the mesh in OpenDX
 */

#include <fstream>
#include <sstream>
#include <memory>

#include "meshDX.hh"
#include "vertexList.hh"
#include "geometry/cell2D.hh"
#include "geometry/cell3D.hh"
#include "geometry/mesh.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "toolbox/elementGraphics.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

// debugging MeshDX
#define graphMeshDXConstr_D 0

// debugging MeshDXCell
#define graphMeshDXCellAppl_D 0
#define graphMeshDXCellwrite_D 0

namespace graphics {

  void MeshDX_write_(std::ostream& os, const int x) {
    os.write((char*)&x, sizeof(const int));
  }

  void MeshDX_write_(std::ostream& os, const float x) {
    os.write((char*)&x, sizeof(const float));
  }

  void MeshDX_conn_write_(std::ostream& os, const VertexList& vtxList,
			  const concepts::Edge& c) {
    __gnu_cxx::hash_map<uint, VertexInfo>::const_iterator j;
    for (uint i = 0; i < 2; ++i) {
      j = vtxList.find(c.vertex(i)->key());
      conceptsAssert(j != vtxList.end(), concepts::Assertion());
      MeshDX_write_(os, (int)j->second.index());
    }
  }

  void MeshDX_conn_write_(std::ostream& os, const VertexList& vtxList,
			  const concepts::Triangle& c) {
    __gnu_cxx::hash_map<uint, VertexInfo>::const_iterator j;
    for (uint i = 0; i < 3; ++i) {
      j = vtxList.find(c.vertex(i)->key());
      conceptsAssert(j != vtxList.end(), concepts::Assertion());
      MeshDX_write_(os, (int)j->second.index());
    }
    MeshDX_write_(os, (int)j->second.index());
  }

  void MeshDX_conn_write_(std::ostream& os, const VertexList& vtxList,
			  const concepts::Quad& c) {
    __gnu_cxx::hash_map<uint, VertexInfo>::const_iterator j;
    uint i;
    for (i = 0; i < 2; ++i) {
      j = vtxList.find(c.vertex(i)->key());
      conceptsAssert(j != vtxList.end(), concepts::Assertion());
      MeshDX_write_(os, (int)j->second.index());
    }
    i = 3;
    j = vtxList.find(c.vertex(i)->key());
    conceptsAssert(j != vtxList.end(), concepts::Assertion());
    MeshDX_write_(os, (int)j->second.index());
    i = 2;
    j = vtxList.find(c.vertex(i)->key());
    conceptsAssert(j != vtxList.end(), concepts::Assertion());
    MeshDX_write_(os, (int)j->second.index());
  }

  // **************************************************************** MeshDX **

  template<class F>
  MeshDX<F>::MeshDX(concepts::Mesh& msh, std::string filename)
    throw(concepts::Assertion) :
    spc_(0), msh_(&msh), filename_(filename) {

    DEBUGL(graphMeshDXConstr_D, "filename = " << filename_);

    concepts::GlobalPostprocess<F> postProcess(*msh_);

    // guess spatial dimension
    uint dim = 0;
    if (dynamic_cast<concepts::Mesh2*>(msh_))
      dim = 2;
    else
      if (dynamic_cast<concepts::Mesh3*>(msh_))
	dim = 3;
    conceptsAssert(dim != 0, concepts::Assertion());

    constructor_(postProcess, filename, msh_->ncell(), dim);
  }

  template<class F>
  MeshDX<F>::MeshDX(concepts::Space<F>& spc, std::string filename, uint dim)
    throw(concepts::Assertion) :
    spc_(&spc), msh_(0), filename_(filename) {

    DEBUGL(graphMeshDXConstr_D, "space = " << *spc_);
    DEBUGL(graphMeshDXConstr_D, "filename = " << filename_);

    concepts::GlobalPostprocess<F> postProcess(*spc_);

    constructor_(postProcess, filename, spc_->nelm(), dim);
  }

  template<class F>
  void MeshDX<F>::constructor_(concepts::GlobalPostprocess<F>& postProcess,
			       std::string& filename, uint nelm, uint dim) {
    VertexList vtxList(VertexList::COORD);

    DEBUGL(graphMeshDXConstr_D, "dimension = " << dim);
    postProcess(vtxList);
    DEBUGL(graphMeshDXConstr_D, "vtxList = " << vtxList);

    BoundaryDXCell<F> boundary(vtxList);
    postProcess(boundary);
    const std::list<const concepts::Connector*>& faces = boundary.faces();

    std::ofstream ofs(filename.c_str());
    MeshDXCell<F> dx(ofs, vtxList, nelm, faces.size(), dim);
    postProcess(dx);

    if (faces.size() > 0) {
      // write faces
      for (std::list<const concepts::Connector*>::const_iterator
	     i = faces.begin(); i != faces.end(); ++i) {
	const concepts::Triangle* connT =
	  dynamic_cast<const concepts::Triangle*>(*i);
	if (connT) {
	  MeshDX_conn_write_(ofs, vtxList, *connT);
	  continue;
	}
	const concepts::Quad* connQ =
	  dynamic_cast<const concepts::Quad*>(*i);
	if (connQ) {
	  MeshDX_conn_write_(ofs, vtxList, *connQ);
	  continue;
	}
	const concepts::Edge* connE =
	  dynamic_cast<const concepts::Edge*>(*i);
	if (connE) {
	  MeshDX_conn_write_(ofs, vtxList, *connE);
	  continue;
	}
	std::stringstream error;
	error << "connector " << **i << " not supported" << std::ends;
	throw conceptsException(concepts::MissingFeature(error.str()));
      }
      // write attributes of faces
      for (std::list<const concepts::Connector*>::const_iterator
	     i = faces.begin(); i != faces.end(); ++i) {
	MeshDX_write_(ofs, (int)((*i)->attrib().attrib()));
      }
    }
  }

  template class MeshDX<Real>;

  MeshDX<Real> drawMeshDX(concepts::Mesh& msh, std::string filename) {
    return MeshDX<Real>(msh, filename);
  }

  // ************************************************************ MeshDXCell **

  template<class F>
  MeshDXCell<F>::MeshDXCell(std::ostream& os, VertexList& vtxList, 
			    uint nelms, uint nfaces, uint dim) : 
    os_(os), vtxList_(vtxList), dim_(dim) {

    conceptsAssert( (dim_ == 2) || (dim_ == 3), concepts::Assertion() );

    __gnu_cxx::hash_map<uint, VertexInfo>::iterator i;

    // endianness of data
#ifdef __sparc
    os_ << "data mode msb" << std::endl;
#endif
#ifdef __i386
    os_ << "data mode lsb" << std::endl;
#endif

    os_ << "# coordinates of the vertices" << std::endl;
    os_ << "object \"vertices\" class array type float rank 1 shape " << dim_ 
	<< " items " << vtxList.size() << " ieee data 0" << std::endl;

    uint dataPointer = vtxList.size()*sizeof(float)*dim_;

    os_ << "# connections for the cells" << std::endl;
    os_ << "object \"cells\" class array type int rank 1 shape "
	<< (dim_ == 2 ? 4 : 8) 
	<< " items " << nelms 
	<< " ieee data " << dataPointer << std::endl;
    os_ << "attribute \"element type\" string " 
	<< (dim_ == 2 ? "\"quads\"" : "\"cubes\"")<< std::endl;
    os_ << "attribute \"ref\" string \"positions\"" << std::endl;

    dataPointer += (dim_ == 2 ? 4 : 8)*nelms*sizeof(int);

    os_ << "# the mesh" << std::endl;
    os_ << "object \"mesh data\" class field"
	<< std::endl;
    os_ << "component \"positions\" value \"vertices\"" << std::endl;
    os_ << "component \"connections\" value \"cells\""
	<< std::endl << std::endl;

    if (nfaces > 0) {
      os_ << "# boundary cells" << std::endl;
      os_ << "object \"faces\" class array type int rank 1 shape "
	  << (dim_ == 2 ? 2 : 4) << " items " << nfaces << " ieee data "
	  << dataPointer << std::endl;
      os_ << "attribute \"element type\" string " 
	  << (dim_ == 2 ? "\"lines\"" : "\"quads\"")<< std::endl;
      os_ << "attribute \"ref\" string \"positions\"" << std::endl;

      dataPointer += (dim_ == 2 ? 2 : 4)*nfaces*sizeof(int);

      os_ << "# boundary attributes" << std::endl;
      os_ << "object \"boundary\" class array type int rank 0 items "
	  << nfaces << " ieee data " << dataPointer << std::endl;
      os_ << "attribute \"dep\" string \"connections\"" << std::endl;    

      os_ << "# boundary data" << std::endl;
      os_ << "object \"face data\" class field" << std::endl
	  << "component \"positions\" value \"vertices\"" << std::endl
	  << "component \"connections\" value \"faces\"" << std::endl
	  << "component \"data\" value \"boundary\""
	  << std::endl << std::endl;

      os_ << "# combining the mesh and the boundary" << std::endl
	  << "object \"grid data\" class group" << std::endl
	  << "member \"cells\" value \"mesh data\"" << std::endl
	  << "member \"faces\" value \"face data\"" << std::endl;
    }
    os_ << "end" << std::endl;

    uint index = 0;
    for (i = vtxList.begin(); i != vtxList.end(); ++i) {
      MeshDX_write_(os_, (float)i->second.coordinates()[0]);
      MeshDX_write_(os_, (float)i->second.coordinates()[1]);
      if (dim_ == 3)
	MeshDX_write_(os_, (float)i->second.coordinates()[2]);
      i->second.index() = index++;
    }
  }

  template<class F>
  MeshDXCell<F>::~MeshDXCell() {}

  template<class F>
  void MeshDXCell<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<Real>::MESH_DX,
			       *this);
  }

  template<class F>
  void MeshDXCell<F>::operator() (const concepts::Cell& cell)
    throw(concepts::MissingFeature, concepts::Assertion) {

    DEBUGL(graphMeshDXCellAppl_D, "cell = " << cell);

    __gnu_cxx::hash_map<uint, VertexInfo>::iterator j;

    const concepts::Triangle2d* cellA =
      dynamic_cast<const concepts::Triangle2d*>(&cell);
    if (cellA) {
      // triangle in 2D
      conceptsAssert(dim_ == 2, concepts::Assertion());
      MeshDX_conn_write_(os_, vtxList_, cellA->connector());
      return;
    }
    const concepts::Quad2d* cellB =
      dynamic_cast<const concepts::Quad2d*>(&cell);
    if (cellB) {
      // quadrilateral in 2D
      conceptsAssert(dim_ == 2, concepts::Assertion());
      MeshDX_conn_write_(os_, vtxList_, cellB->connector());
      return;
    }
    const concepts::Triangle3d* cellT =
      dynamic_cast<const concepts::Triangle3d*>(&cell);
    if (cellT) {
      // triangle in 3D
      conceptsAssert(dim_ == 3, concepts::Assertion());
      for (uint t = 0; t < 2; ++t) {
	MeshDX_conn_write_(os_, vtxList_, cellT->connector());
      }
      return;
    }
    const concepts::Tetrahedron3d* cellC = 
      dynamic_cast<const concepts::Tetrahedron3d*>(&cell);
    if (cellC) {
      // tetrahedron
      conceptsAssert(dim_ == 3, concepts::Assertion());
      for (uint i = 0; i < 3; ++i) {
	j = vtxList_.find(cellC->connector().vertex(i)->key());
	conceptsAssert(j != vtxList_.end(), concepts::Assertion());
	MeshDX_write_(os_, (int)j->second.index());
      }
      MeshDX_write_(os_, (int)j->second.index());
      j = vtxList_.find(cellC->connector().vertex(3)->key());
      conceptsAssert(j != vtxList_.end(), concepts::Assertion());
      for (uint i = 0; i < 4; ++i)
	MeshDX_write_(os_, (int)j->second.index());
      return;
    }
    const concepts::Hexahedron3d* cellH =
      dynamic_cast<const concepts::Hexahedron3d*>(&cell);
    if (cellH) {
      // hexahedron
      conceptsAssert(dim_ == 3, concepts::Assertion());
      uint i;
      for (i = 0; i < 2; ++i) {
	j = vtxList_.find(cellH->connector().vertex(i)->key());
	conceptsAssert(j != vtxList_.end(), concepts::Assertion());
	MeshDX_write_(os_, (int)j->second.index());
      }
      i = 3;
      j = vtxList_.find(cellH->connector().vertex(i)->key());
      conceptsAssert(j != vtxList_.end(), concepts::Assertion());
      MeshDX_write_(os_, (int)j->second.index());
      i = 2;
      j = vtxList_.find(cellH->connector().vertex(i)->key());
      conceptsAssert(j != vtxList_.end(), concepts::Assertion());
      MeshDX_write_(os_, (int)j->second.index());
      for (i = 4; i < 6; ++i) {
	j = vtxList_.find(cellH->connector().vertex(i)->key());
	conceptsAssert(j != vtxList_.end(), concepts::Assertion());
	MeshDX_write_(os_, (int)j->second.index());
      }
      i = 7;
      j = vtxList_.find(cellH->connector().vertex(i)->key());
      conceptsAssert(j != vtxList_.end(), concepts::Assertion());
      MeshDX_write_(os_, (int)j->second.index());
      i = 6;
      j = vtxList_.find(cellH->connector().vertex(i)->key());
      conceptsAssert(j != vtxList_.end(), concepts::Assertion());
      MeshDX_write_(os_, (int)j->second.index());
      return;
    }
    const concepts::Cell3* cellD = dynamic_cast<const concepts::Cell3*>(&cell);
    if (cellD) {
      // default 3D cell
      conceptsAssert(dim_ == 3, concepts::Assertion());
      uint i = 0, k = 0;
      concepts::Connector0* V;
      while ( (V = cellD->connector().vertex(k)) ) {
	j = vtxList_.find(V->key());
	conceptsAssert(j != vtxList_.end(), concepts::Assertion());
	MeshDX_write_(os_, (int)j->second.index());
	++i;
	switch (i) {
	case 2: k = 3; break;
	case 3: k = 2; break;
	case 6: k = 7; break;
	case 7: k = 6; break;
	default: k = i;
	}
      }
      for (; i < 7; ++i) {
	MeshDX_write_(os_, (int)j->second.index());
      }
      return;
    }
    std::stringstream error;
    error << "cell " << cell << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  template class MeshDXCell<Real>;

  // ******************************************************** BoundaryDXCell **

  template<typename F>
  BoundaryDXCell<F>::~BoundaryDXCell() {}

  template<typename F>
  void BoundaryDXCell<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<Real>::MESH_DX,
			       *this);
  }

  template<typename F>
  void BoundaryDXCell<F>::operator() (const concepts::Cell& cell)
    throw(concepts::MissingFeature, concepts::Assertion) {

    const concepts::Cell2* cell2 =
      dynamic_cast<const concepts::Cell2*>(&cell);
    if (cell2) {
      const concepts::Connector2& cntr = cell2->connector();
      const concepts::Connector1* edge = 0;
      for (uint i = 0; (edge = cntr.edge(i)) != 0; ++i) {
	if (edge->attrib() != concepts::Attribute(0))
	  faces_.push_back(edge);
      }
      return;
    }

    const concepts::Cell3* cell3 =
      dynamic_cast<const concepts::Cell3*>(&cell);
    if (cell3) {
      const concepts::Connector3& cntr = cell3->connector();
      const concepts::Connector2* face = 0;
      for (uint i = 0; (face = cntr.face(i)) != 0; ++i) {
	if (face->attrib() != concepts::Attribute(0))
	  faces_.push_back(face);
      }
      return;
    }

    std::stringstream error;
    error << "cell " << cell << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  template class BoundaryDXCell<Real>;

} // namespace graphics
