/* graphics of the mesh in eps
 */

#ifndef graphMeshEPS_hh
#define graphMeshEPS_hh

#include <string>
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/exceptions.hh"
#include "space/postProcess.hh"

namespace concepts {
  // forward declartions
  class Mesh2;
} // namespace concepts

namespace graphics {

  using concepts::Real;

  // *************************************************************** MeshEPS **

  /** Draws a picture of a mesh in EPS format and stores the
      result in a file. This class is just an executor class for 
      MeshEPSCell, the real work to draw the picture is done there.

      The class has a template parameter which is not very handy. For this
      purpose, there are two trampoline functions drawMeshEPS which
      create this class and handle the template parameter corretctly.

      @see drawMeshEPS the trampoline function of this class
      @see MeshEPSCell the class which does the real work
      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class MeshEPS {
  public:
    /** Constructor. The coarsest grid will be depicted. To store the results,
	a new file is created.
    */
    MeshEPS(concepts::Mesh& msh, std::string filename);

    /** Constructor. The currently used grid in the space will be depicted.
	To store the results, a new file is created.
    */
    MeshEPS(concepts::Space<F>& spc, std::string filename);
  private:
    /// The space
    concepts::Space<F>* spc_;

    /// The mesh
    concepts::Mesh* msh_;
  };

  /** Trampoline function to create a MeshEPS. The purpose of this
      function is to hide the template parameters of MeshEPS.

      @see MeshEPS for the complete description
      @author Philipp Frauenfelder, 2001.
  */
//   template<class F>
  MeshEPS<Real> drawMeshEPS(concepts::Mesh& msh, std::string filename);

  /** Trampoline function to create a MeshEPS. The purpose of this
      function is to hide the template parameters of MeshEPS.

      @see MeshEPS for the complete description
      @author Philipp Frauenfelder, 2001.
  */
  template<class F>
  MeshEPS<F> drawMeshEPS(concepts::Space<F>& spc, std::string filename);

  // *********************************************************** MeshEPSCell **

  /** Draws a picture of a mesh in EPS. The application operator of this class 
      sketches the mesh in EPS format for each cell or element.

      You should not use this class directly. Instead, use the executor class
      MeshEPS or (even simpler) its trampoline function 
      drawMeshEPS (which hides the template parameter of MeshEPS).

      @see MeshEPS the executor class
      @see drawMeshEPS the trampoline function of MeshEPS
      @author Philipp Frauenfelder, 2000
  */
  template<class F>
  class MeshEPSCell : public concepts::CellPostprocess<Real> {
  public:
    /** Constructor. First scanns the mesh for the top right and 
	bottom left point and then prints the header of the EPS file.

	@param os Output stream where the data should be written to
	@param msh The mesh which should be draw
	@param scale Optional scaling parameter
    */
    MeshEPSCell(std::ostream& os, concepts::Mesh2& msh, const Real scale = 100)
      throw(concepts::MissingFeature);

    /** Constructor. First scanns the mesh for the top right and 
	bottom left point and then prints the header of the EPS file.

	@param os Output stream where the data should be written to
	@param spc The space which should be drawn
	@param scale Optional scaling parameter
    */
    MeshEPSCell(std::ostream& os, concepts::Space<F>& spc,
		const Real scale = 100) throw(concepts::MissingFeature);

    virtual void operator() (const concepts::Element<Real>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature);
    void offsetMin(const concepts::Real2d& x);
    void maxMax(const concepts::Real2d& x);
  private:
    /// Output stream
    std::ostream& os_;

    /// Scaling factor
    Real scale_;

    /// Offset of the bottom left corner from the origin
    concepts::Real2d offset_;

    /// Offset of the top right corner from the origin
    concepts::Real2d max_;

    void EPSheader_(const concepts::Real2d max, std::ostream& os) const;
  };

} // namespace graphics

#endif // graphMeshEPS_hh
