/* graphics of the mesh in Matlab
 */

#ifndef graphMeshMatlab_hh
#define graphMeshMatlab_hh

#include <string>
#include <fstream>
#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "basics/exceptions.hh"
#include "toolbox/dynArray.hh"
#include "space/postProcess.hh"

namespace concepts {
  // forward declartions
  class Mesh2;
  class Mesh3;
} // namespace concepts

namespace graphics {

  using concepts::Real;

  // ************************************************************ MeshMatlab **

  /** Creates data for Matlab to draw the mesh.
      This class is just an executor class for MeshMatlabCell, the real
      work is done there.

      The data is stored in three file by appending '_1.m', '_2.m'
      and '_3.m' to the filename. The data can be read into Matlab
      by giving the three filenames as commands (and ommmiting the '.m').
      Depending on the element type the mesh is either plotted by
      <pre>plot3(a, b, c)</pre>
      (if you want black lines in the plot, use
      <pre>plot3(a, b, c, 'k')</pre>
      ) or
      <pre>trimesh(b', a(1,:), a(2,:), a(3,:), ones(size(a,2),1))</pre>

      This class opens and closes the three files and hands the
      streams to MeshMatlabCell.

      @see drawMeshMatlab the trampoline function of this class
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class MeshMatlab {
  public:
    /** Constructor. Creates a picture of the coarsest mesh.
     */
    MeshMatlab(concepts::Mesh& msh, std::string filename);
    /** Constructor. Creates a picture of the current mesh of the space.
     */
    MeshMatlab(concepts::Space<F>& spc, std::string filename);
    ~MeshMatlab();
  private:
    /// Creates the streams and prints the header
    void createStreams_(std::string filename);
    /// Prints the footer and closes the streams
    void closeStreams_();
    /// The streams
    std::auto_ptr<std::ofstream> ofs1_, ofs2_, ofs3_;
  };

  /** Trampoline function to create an instance of MeshMatlab.
      @author Philipp Frauenfelder, 2002
  */
  void drawMeshMatlab(concepts::Mesh& msh, std::string filename);

  /** Trampoline function to create an instance of MeshMatlab.
      @author Philipp Frauenfelder, 2002
  */
  template<class F>
  void drawMeshMatlab(concepts::Space<F>& spc, std::string filename);

  // ******************************************************** MeshMatlabCell **

  /** Writes the data which describe the elements to the three streams.

      There are two different output formats depending on the cell.

      In the first output format every stream takes another component
      of the coordinates of a point. Then every line which is printed
      signifies a poly line in the final drawing.

      In the second case the first stream takes all coordinates of
      a point and the second one the mesh. The third stream is
      not used.

      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class MeshMatlabCell : public concepts::CellPostprocess<F> {
  public:
    /// Constructor.
    MeshMatlabCell(std::ostream* os1, std::ostream* os2, std::ostream* os3);
    virtual ~MeshMatlabCell();
    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature);
  private:
    /// Output streams
    std::ostream *os1_, *os2_, *os3_;
    /// Map for vertices of the elements
    std::auto_ptr<concepts::DynArray<uint> > map_;
    /// Counter
    uint cnt_;
  };

} // namespace graphics

#endif // graphMeshMatlab_hh
