/* Traits for different spaces
 */

#ifndef spaceTraits_hh
#define spaceTraits_hh

#include "basics/typedefs.hh"

namespace hp2D {
  class Space;
}

namespace hp3D {
  class Space;
}

namespace linearFEM {
  class Linear2d;
  class Linear3d;
}

namespace vectorial {
  template<typename F>
  class Space;
}

namespace graphics {

  // ********************************************************** Space traits **

  /** Traits for a space. These traits are used by trampoline
      functions to fill in certain parameters which depend on the
      space (eg. the spatial dimension of the space). Every space
      which should be drawable needs an explicit specialization of
      this class.

      To add a new spezialisation, the name of the space has to be
      known (this is best handled by adding a forward declaration) and
      an explicit template specialization of this class has to be
      created with the required functions, see the examplex for
      hp2D::Space and hp3D::Space.

      @see drawMeshDX
      @see spaceTraits<hp2D::Space> as an example for an explicit
           template specialization
      @author Philipp Frauenfelder, 2002
   */
  template<typename space>
  struct spaceTraits {};

  /** Traits for hp2D::Space. */
  template<>
  struct spaceTraits<hp2D::Space> {
    static inline uint dim() { return 2; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };

  /** Traits for hp3D::Space. */
  template<>
  struct spaceTraits<hp3D::Space> {
    static inline uint dim() { return 3; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };

  /** Traits for linearFEM::Linear2d. */
  template<>
  struct spaceTraits<linearFEM::Linear2d> {
    static inline uint dim() { return 2; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };

  /** Traits for linearFEM::Linear3d. */
  template<>
  struct spaceTraits<linearFEM::Linear3d> {
    static inline uint dim() { return 3; }
    static inline bool positionConnection() { return false; }
    static inline bool boundaryElements() { return false; }
  };

  /** Traits for vectorial::Space. */
  template<>
  struct spaceTraits<vectorial::Space<concepts::Real> > {
    static inline uint dim() { return dim_; }
    static inline bool positionConnection() { return posCon_; }
    static inline bool boundaryElements() { return bem_; }
    static uint dim_;
    static bool posCon_;
    static bool bem_;
    static inline void setDim(uint dim) { dim_ = dim; }
    static inline void setposCon(bool posCon) { posCon_ = posCon; }
    static inline void setBEM(bool bem) { bem_ = bem; }
  };

} // namespace graphics

#endif // spaceTraits_hh
