/* Bilinear forms for hp 1D FEM
 */

#include "bilinearForm.hh"
#include "element.hh"
#include "basics/debug.hh"

// debugging
#define Identity1DAppl_D 0

namespace hp1D {

  // *************************************************************** Laplace **

  Laplace::Laplace() {
  }

  Laplace::~Laplace() {
  }

  void Laplace::operator()(const concepts::Element<Real>& elmX,
			   const concepts::Element<Real>& elmY, 
			   concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const Element<Real>* eX = dynamic_cast<const Element<Real>*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const Element<Real>* eY = dynamic_cast<const Element<Real>*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size(); // element size, ie. Jacobian

      const Real* const w = eX->integration()->weights(); // quadrature weights
//       const Real* q = eX->integration()->abscissas(); // quadrature points
      const uint nq = eX->integration()->n(); // number of quadrature points

      const Real* const s = eX->shpfctD()->values(); // shape function values

      for (uint i = 0; i < n; ++i) { // loop over shape functions
	const Real* sy = s;
	for (uint j = 0; j < n; ++j) { // loop over shape functions
	  Real& m = em(i,j);
	  const Real* sx = s +i*nq;
	  const Real* wAcc = w;
	  for (uint k = 0; k < nq; ++k) { // loop over quadrature points
	    m += *wAcc++ / h * 2.0 * *sx++ * *sy++;
	  } // for k
	} // for j
      } // for i
    } // if eX == eY
  }

  std::ostream& Laplace::info(std::ostream& os) const {
    return os << "Laplace()";
  }

  // ************************************************************** Identity **

  Identity::Identity() {
  }

  Identity::~Identity() {
  }

  void Identity::operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const Element<Real>* eX = dynamic_cast<const Element<Real>*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const Element<Real>* eY = dynamic_cast<const Element<Real>*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      DEBUGL(Identity1DAppl_D, "elm = " << *eX);
      const Real h = eX->cell().size(); // element size, ie. Jacobian

      const Real* const w = eX->integration()->weights(); // quadrature weights
//       const Real* q = eX->integration()->abscissas(); // quadrature points
      const uint nq = eX->integration()->n(); // number of quadrature points

      const Real* const s = eX->shpfct()->values(); // shape function values

      for (uint i = 0; i < n; ++i) { // loop over shape functions
	const Real* sy = s;
	for (uint j = 0; j < n; ++j) { // loop over shape functions
	  Real& m = em(i,j);
	  const Real* sx = s +i*nq;
	  const Real* wAcc = w;
	  for (uint k = 0; k < nq; ++k) { // loop over quadrature points
	    m += *wAcc++ * h / 2.0 * *sx++ * *sy++;
	  } // for k
	} // for j
      } // for i
      DEBUGL(Identity1DAppl_D, "em = " << em);
    } // if eX == eY
  }

  std::ostream& Identity::info(std::ostream& os) const {
    return os << "Identity()";
  }

} // namespace hp1D
