/* hp extension for 1D FEM: ie. determine wether to refine or raise p
 * in an element
 */

#include <cmath>
#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "toolbox/array.hh"
#include "function/vector.hh"
#include "space/tmatrix.hh"
#include "space/element.hh"
#include "space/hpMethod.hh"
#include "hpExtension.hh"
#include "element.hh"
#include "basics/debug.hh"

// debugging

#define RefineRaiseAppl_D 1

namespace hp1D {
  using concepts::Real;

  // ******************************************************** RefineOrRaise  **

  template<typename F>
  RefineOrRaise<F>::RefineOrRaise(const concepts::Vector<F>& solution,
				  const Real theta, concepts::AdaptiveSpace<F,
				  concepts::AdaptiveAdjustP<1> >& spc) :
    coeff_(0), legendre_(0), solution_(solution), spc_(spc), theta_(theta) {
  }

  template<typename F>
  void RefineOrRaise<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    // check for a 1D element
    const Element<F>* e = dynamic_cast<const Element<F>*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    // get local coefficients
    elm.T().extract(solution_, coeff_);
//     DEBUGL(RefineRaiseAppl_D, "coeff = " << coeff_);

    // compute Legendre coefficients
    const uint n = coeff_.size();
    conceptsAssert(n > 1, concepts::Assertion()); // p = 1 => 2 coefficients
    legendre_.resize(n);
    legendre_.zeros();
    const Real* coeff = (const Real*)coeff_;
    legendre_[0] += *coeff/2.0;
    legendre_[1] += *coeff++/-2.0;
    legendre_[0] += *coeff/2.0;
    legendre_[1] += *coeff++/2.0;
    for (uint i = 2; i < n; ++i) {
      Real a = (1.0-i)/(2.0*(2.0*i-1.0));
      legendre_[i] -= *coeff*a;
      legendre_[i-2] += *coeff++*a;
    }
//     DEBUGL(RefineRaiseAppl_D, "legendre = " << legendre_);

    // compute regression of log(legendre)
    Real* legendre = (Real*)legendre_;
    for (uint i = 0; i < n; ++i, ++legendre)
      *legendre = std::log(std::fabs(*legendre));
    Real sum = 0.0, isum = 0.0;
    legendre = (Real*)legendre_;
    for (uint i = 0; i < n; ++i, ++legendre) {
      sum += *legendre;
      isum += i**legendre;
    }
    uint p = e->p();
    Real m = 3.0*(isum - sum/2.0*p)/((p+1)*p*(p+2));
    DEBUGL(RefineRaiseAppl_D, "m = " << m << ", e^m = " << std::exp(m));

    if (std::exp(m) < theta_) {
      // p element
      DEBUGL(RefineRaiseAppl_D, "p element");
      concepts::AdaptiveAdjustP<1> a(0, 1);
      spc_.adjust(elm, a);
    } else {
      // h element
      DEBUGL(RefineRaiseAppl_D, "h element");
      concepts::AdaptiveAdjustP<1> a(1, 0);
      spc_.adjust(elm, a);
    }
  }

  template class RefineOrRaise<Real>;

} // namespace hp1D
