/* class for the hp FEM space in 1D
 */

#include <sstream>
#include "space.hh"
#include "basics/debug.hh"

// debugging
#define Space1DDestr_D 0

namespace hp1D {

  // ***************************************************************** Space **

  Space::Space(concepts::Mesh1& msh, uint l, uint p,
	       concepts::BoundaryConditions* bc) :
    msh_(msh), bc_(bc), rebuild_(true), dim_(0), nelm_(0), elm_(0),
    ctrl0_(2, 4), ctrl1_(2, 4), adj_(2, 4) {
    std::auto_ptr<concepts::Scan1> sc(msh_.scan());
    while (*sc) {
      concepts::Key key(((*sc)++).connector().key());
      ctrl1_[key].tag_.data_ = concepts::AdaptiveControlTag::MEMBER;
      adj_[key].l_ = l;
      adj_[key].p_[0] = p;
    }
  }

  Space::~Space() {
    concepts::Joiner<Element<Real>*, 1>::destructor(elm_);
    DEBUGL(Space1DDestr_D, "done.");
  }

  void Space::adjust(const concepts::Element<Real>& elm,
		     const concepts::AdaptiveAdjustP<1>& a) {
    const Element<Real>* e = dynamic_cast<const Element<Real>*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    adj_[e->support().key()] = a;
    rebuild_ = true;
  }

  void Space::rebuild() {
    // remove old elements
    delete elm_;
    elm_ = 0;
    dim_ = 0;
    nelm_ = 0;

    // prepare for rebuild of mesh by marking level and polynomial degree
    // sort out boundary conditions (ie. mark DIRICHLET nodes as passive)
    std::auto_ptr<concepts::Scan1> sc(msh_.scan());
    while (*sc) {
      concepts::Cell1& cell = (*sc)++;
      int L = -1;
      int P;

      concepts::Edge1d* cellE =
	dynamic_cast<concepts::Edge1d*>(&cell);
      if (cellE) {
 	meshAndPoly_(*cellE, 0, L, P);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

    // rebuild elements
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell1& cell = (*sc)++;

      concepts::Edge1d* cellE =
	dynamic_cast<concepts::Edge1d*>(&cell);
      if (cellE) {
	buildElements_(*cellE);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

    // clear pre rebuild information
    rebuild_ = false;
    adj_.reset();
    // clear the control information of vertices
    ctrl0_.reset();
  }

  std::ostream& Space::info(std::ostream& os) const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return os << "Space(dim = " << dim_ << ", nelm = " << nelm_ << ')';
  }

//   void Space::deactivate_(concepts::Vertex& vtx, const uint level) {
//     ctrl0_[vtx.child(0)->key()][level].tag_.deactivate();
//   }

} // namespace hp1D
