/* a priori refinement for quads
 */

#ifndef APRIORIRef2D_hh
#define APRIORIRef2D_hh

#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "basics/level.hh"
#include "geometry/connector.hh"
#include "space/postProcess.hh"
#include "space/hpMethod.hh"

// forward declaration
namespace concepts {
  class Quad;
}

namespace hp2D {

  using concepts::Real;

  // ***************************************************** APrioriRefinement **

  /** Carries out a-priori given refinements. The mesh is geometrically
      refined towards the given vertices or edges. These can be
      determined by their attribute. If a refinement in one direction occurs,
      p is not increased in this direction. Otherwise it is increased by the
      given amount. This makes it possible to generate geometrically refined
      meshes to corners and edges with linear polynomial degree distribution.

      Currently, methods for hp2D::Quad are implemented.
      @author Philipp Frauenfelder, 2003
      @bug The parameter \c p of the constructor can be set to a pointer to
      0. This is not detected and a run-time error occurs.
  */
  class APrioriRefinement : public concepts::CellPostprocess<Real> {
  public:
    enum subdivTypes { NONE = 0, X = 1, Y = 2};
    static const short MAXSHORT = 32767;
    /** Constructor. If a given attribute for the vertices or edges is 0,
	it is not considered.
	@param spc Space to be refined
	@param attribVtx Refine towards vertices with this attribute
	@param attribEdge Refine towards edges with this attribute
	@param p If there is no refinement, p will be increased by
	       this amount (anisotropically)
	@param subdiv possibility to restrict subdivision strategy
    */
    APrioriRefinement
    (concepts::AdaptiveSpace<Real, concepts::AdaptiveAdjustP<2> >& spc,
     concepts::Attribute attribVtx, concepts::Attribute attribEdge,
     const int* const p, const uint subdiv = X | Y) :
      spc_(spc), attribVtx_(attribVtx), attribEdge_(attribEdge), p_(p),
      subdiv_(subdiv), level_(MAXSHORT) {}
    APrioriRefinement
    (concepts::AdaptiveSpace<Real, concepts::AdaptiveAdjustP<2> >& spc,
     const int* const p, const uint subdiv = X | Y, const concepts::Level<2> level = MAXSHORT) :
      spc_(spc), attribVtx_(0), attribEdge_(0), p_(p),
      subdiv_(subdiv), level_(level) {}
    virtual void operator() (const concepts::Element<Real>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature) {}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Space to be refined
    concepts::AdaptiveSpace<Real, concepts::AdaptiveAdjustP<2> >& spc_;
    /// Attribute of entity to which should be refined
    concepts::Attribute attribVtx_, attribEdge_;
    /// Maximal p in a refinement step in each subdiv
    const int* const p_;
    /// Subdivision strategy
    const uint subdiv_;
    /// This or higher level 
    const concepts::Level<2> level_;
    /// Computes the refinement solely using topological information
    concepts::AdaptiveAdjustP<2> computeRef_(const concepts::Quad& q, concepts::Level<2> level) const;
  };

} // namespace hp2D

#endif // APRIORIRef2D_hh
