/* Bilinear forms for hp 2D FEM
 */

#include <cstring>
#include <fstream>
#include "basics/vectorsMatrices.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "toolbox/array.hh"
#include "toolbox/resourceMonitor.hh"
#include "toolbox/coeffIterator.hh"
#include "toolbox/dynArray.hh"
#include "hp2D/quad.hh"
#include "hp2D/bilinearForm.hh"
#include "hp2D/sumfact.hh"
#include "hp2D/shapeFunction2d.hh"
#include "basics/debug.hh"

// debugging
#define hpLaplaceAppl_D 0
#define hpLaplaceApplTime_D 0
#define hpLaplaceApplMem_D 0
#define hpIdentityAppl_D 0
#define hpIdentityApplTime_D 0

namespace hp2D {

  // *************************************************************** Laplace **

  Laplace::Laplace(const concepts::Formula& frm,
		   const concepts::DynArray<Real>& cellwise) :
    frm_(frm.clone()), cellwise_(&cellwise) {}

  Laplace::~Laplace() {}

  Laplace* Laplace::clone() const {
    if (cellwise_)
      return new Laplace(*frm_.get(), *cellwise_);
    if (frm_.get())
      return new Laplace(*frm_.get());
    return new Laplace();
  }

  void Laplace::operator()(const concepts::Element<Real>& elmX, 
			   const concepts::Element<Real>& elmY, 
			   concepts::ElementMatrix<Real>& em) {

    const Quad* elmYH = dynamic_cast<const Quad*>(&elmY);
    if (elmYH) {
      const Quad* elmXH = dynamic_cast<const Quad*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Laplace::operator()(const Quad& elmX, const Quad& elmY, 
			   concepts::ElementMatrix<Real>& em) {

    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpLaplaceAppl_D, elmX);
      DEBUGL(hpLaplaceApplMem_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      // get cellwise value
      Real cellwiseValue = 0.0;
      if (cellwise_) {
	cellwiseValue = (*cellwise_)[elmX.support().attrib().attrib()];
      }

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();

      concepts::ResourceMonitor timer;

      bool sameJacobian = false;
      if (! sameJacobian) {
	// Jacobian
	jacobian_.resize(nqx*nqy);
	DEBUGL(hpLaplaceApplMem_D, "reserved " << (nqx*nqy)
	       << " for jacobian_");
	Real* dx = (Real*)jacobian_;
	jacobianInv_.resize(nqx*nqy);
	DEBUGL(hpLaplaceApplMem_D, "reserved " << (nqx*nqy)
	       << " for jacobianInv_");
	concepts::MapReal2d* jacInv = (concepts::MapReal2d*)jacobianInv_;

	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i, ++pxAcc) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j, ++pyAcc) {
	    const concepts::Real2d refPoint( (*pxAcc + 1.0)/2.0,
					     (*pyAcc + 1.0)/2.0 );
	    Real f = 1.0;
	    if (frm_.get())
	      f = (*frm_)(elmX.cell().chi(refPoint[0], refPoint[1]))
		+ cellwiseValue;

	    concepts::MapReal2d jacobian
	      (elmX.jacobian(refPoint[0], refPoint[1]));
	    DEBUGL(hpLaplaceAppl_D, "jacobian(" << i << ", " << j << "): "
		   << jacobian);
	    *dx++ = jacobian.determinant() * f;
	    concepts::MapReal2d jacinv(jacobian.inverse());
	    DEBUGL(hpLaplaceAppl_D, "Jacinv(" << i << ", " << j << "): "
		   << jacinv * jacinv.transpose());
	    // with a coefficient matrix:
	    // jacinv * A.transpose() * jacinv.transpose()
	    *jacInv++ = jacinv * jacinv.transpose();
	  }
	}
	DEBUGL(hpLaplaceApplTime_D, "TIME for Jacobian: " << timer);

      } // sameJacobian

      // shapefunctions
      ShapeFunction2D<Real> u;
      u.n(0, elmX.shpfctX()->n());        u.n(1, elmX.shpfctY()->n());
      u.nq(0, elmX.integrationX()->n());  u.nq(1, elmX.integrationY()->n());
      u.values(0, elmX.shpfctX()->values());
      u.values(1, elmX.shpfctY()->values());
      u.abscissas(0, elmX.integrationX()->abscissas());
      u.abscissas(1, elmX.integrationY()->abscissas());
      u.weights(0, elmX.integrationX()->weights());
      u.weights(1, elmX.integrationY()->weights());
      u.T(&(elmX.T()));

      ShapeFunction2D<Real> v;
      v.n(0, elmY.shpfctX()->n());        v.n(1, elmY.shpfctY()->n());
      v.nq(0, elmY.integrationX()->n());  v.nq(1, elmY.integrationY()->n());
      v.values(0, elmY.shpfctX()->values());
      v.values(1, elmY.shpfctY()->values());
      v.abscissas(0, elmY.integrationX()->abscissas());
      v.abscissas(1, elmY.integrationY()->abscissas());
      v.weights(0, elmY.integrationX()->weights());
      v.weights(1, elmY.integrationY()->weights());
      v.T(&(elmY.T()));

      SumFactorization sumfact;
      em.zeros();
      for (uint i = 0; i < 2; ++i) {
	switch (i) {
	case 0: u.values(i, elmX.shpfctDX()->values()); break;
	case 1: u.values(i, elmX.shpfctDY()->values()); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
	for (uint j = 0; j < 2; ++j) {
	  switch (j) {
	  case 0: v.values(j, elmY.shpfctDX()->values()); break;
	  case 1: v.values(j, elmY.shpfctDY()->values()); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }

#ifdef DEBUG
	  concepts::CoeffIterator<Real, concepts::MapReal2d>
	    c((concepts::MapReal2d*)jacobianInv_, i, j);
	  DEBUGL(hpLaplaceAppl_D, "Coeff-Matrix ( " << i << ", " << j << "): "
		 << *c++);
#endif

	  concepts::CoeffIterator<Real, concepts::MapReal2d>
	    coeff((concepts::MapReal2d*)jacobianInv_, i, j);
	  sumfact(em, jacobian_, &coeff, u, v);

	  DEBUGL(hpLaplaceAppl_D, "em = " << em);

	  switch (j) {
	  case 0: v.values(j, elmY.shpfctX()->values()); break;
	  case 1: v.values(j, elmY.shpfctY()->values()); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	}
	switch (i) {
	case 0: u.values(i, elmX.shpfctX()->values()); break;
	case 1: u.values(i, elmX.shpfctY()->values()); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
      }

      DEBUGL(hpLaplaceApplTime_D, "TIME for integration: " << timer);

      for (uint i = 1; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
      DEBUGL(hpLaplaceAppl_D, "em = " << em);
      DEBUGL(hpLaplaceApplTime_D, "TIME for symmetric em: " << timer);
    } // elmX == elmY
  }

  std::ostream& Laplace::info(std::ostream& os) const{
    return os << "Laplace";
  }

  // ************************************************************** Identity **

  Identity::Identity(const concepts::Formula* frm) : frm_(0) {
    if (frm) {
      frm_.reset(frm->clone());
    }
  }

  Identity::~Identity() {}

  Identity* Identity::clone() const {
    return new Identity(frm_.get());
  }

  void Identity::operator()(const concepts::Element<Real>& elmX, 
				 const concepts::Element<Real>& elmY, 
				 concepts::ElementMatrix<Real>& em) {

    const Quad* elmYH = dynamic_cast<const Quad*>(&elmY);
    if (elmYH) {
      const Quad* elmXH = dynamic_cast<const Quad*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Identity::operator()(const Quad& elmX, const Quad& elmY, 
				 concepts::ElementMatrix<Real>& em) {

    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpIdentityAppl_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();

      concepts::ResourceMonitor timer;

      bool sameJacobian = false;
      if (! sameJacobian) {
	// Jacobian
	jacobian_.resize(nqx*nqy);
	DEBUGL(hpIdentityAppl_D, "reserved " << (nqx*nqy)
	       << " for jacobian_");
	Real* dx = (Real*)jacobian_;
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j) {
	    const concepts::Real2d refPoint( (*pxAcc + 1.0)/2.0,
					     (*pyAcc++ + 1.0)/2.0 );
	    Real f = 1.0;
	    if (frm_.get())
	      f = (*frm_)(elmX.cell().chi(refPoint[0], refPoint[1]));
	    *dx++ = elmX.jacobianDeterminant(refPoint[0], refPoint[1]) * f;
	  }
	  ++pxAcc;
	}
//       DEBUGL(hpIdentityApplTime_D, "TIME for Jacobian: " << timer);

      } // sameJacobian

      // shapefunctions
      ShapeFunction2D<Real> u;
      u.n(0, elmX.shpfctX()->n());        u.n(1, elmX.shpfctY()->n());
      u.nq(0, elmX.integrationX()->n());  u.nq(1, elmX.integrationY()->n());
      u.values(0, elmX.shpfctX()->values());
      u.values(1, elmX.shpfctY()->values());
      u.abscissas(0, elmX.integrationX()->abscissas());
      u.abscissas(1, elmX.integrationY()->abscissas());
      u.weights(0, elmX.integrationX()->weights());
      u.weights(1, elmX.integrationY()->weights());
      u.T(&(elmX.T()));

      ShapeFunction2D<Real> v;
      v.n(0, elmY.shpfctX()->n());        v.n(1, elmY.shpfctY()->n());
      v.nq(0, elmY.integrationX()->n());  v.nq(1, elmY.integrationY()->n());
      v.values(0, elmY.shpfctX()->values());
      v.values(1, elmY.shpfctY()->values());
      v.abscissas(0, elmY.integrationX()->abscissas());
      v.abscissas(1, elmY.integrationY()->abscissas());
      v.weights(0, elmY.integrationX()->weights());
      v.weights(1, elmY.integrationY()->weights());
      v.T(&(elmY.T()));

      SumFactorization sumfact;
      em.zeros();
      sumfact(em, jacobian_, (concepts::CoeffIterator<Real, Real>*)0, u, v);

      DEBUGL(hpIdentityApplTime_D, "p = (" << elmX.p()[0] << ", "
	     << elmX.p()[1] << ", " << elmX.p()[2]
	     << "), TIME for integration: " << timer);

      for (uint i = 1; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
//       DEBUGL(hpIdentityAppl_D, "em = " << em);

//       DEBUGL(hpIdentityApplTime_D, "TIME for symmetric em: " << timer);
    } // elmX == elmY
  }

  std::ostream& Identity::info(std::ostream& os) const {
    return os << "Identity";
  }

} // namespace hp2D
