/** @file buildDofsBase.hh Base class for factored out parts of
    building elements in hp2D::Space.
    @author Philipp Frauenfelder, 2004
*/

#ifndef buildDofsBase2D_hh
#define buildDofsBase2D_hh

#include "hp2D/space.hh"

namespace hp2D {

  // forward declaration
  class Space;

  // ********************************************************* BuildDofsBase **

  /** Responsible to build the degrees of freedom in a space.  A call
      to the application operator builds the T columns for a given
      quad.

      This is the base class for the specialzations for vertices,
      edges and the interior. The access to the internal data of the
      space (which is needed) is provided through the protected member
      functions of this class (and this class is a friend of Space).

      @author Philipp Frauenfelder, 2004
  */
  class BuildDofsBase {
  public:
    /// Default constructor
    BuildDofsBase() : spc_(0) {}
    /** Constructor.
	@param spc Space to build the degrees of freedom from
    */
    BuildDofsBase(Space& spc) : spc_(&spc) {}
    /** Builds the degrees of freedom.
	@param cntr Hex which vertices should be considered
	@param T1 New T columns created for the degrees of freedom in
	the vertices
    */
    virtual void operator()(concepts::Quad& cntr,
			    concepts::TColumn<Real>*& T1) = 0;
    /** Virtual copy constructor with a twist.
	If a copy of a space is created, the strategies to build the degrees
	of freedom of the copy should als reference the copy and not the
	initial space. This is done by giving the argument \c spc to
	clone.
	@param spc Space to work on for the return value.
	If set to 0, \c spc_ is used.
    */
    virtual BuildDofsBase* clone(Space* spc = 0) const = 0;
    /// Returns the space
    Space* space() { return spc_; }
  protected:
    /// Returns vertex tag of vertex with key \c idx
    inline concepts::AdaptiveControl& ctrl0_(uint idx);
    /// Returns edge tag of edge with key \c idx
    inline concepts::AdaptiveControlP<1>& ctrl1_(uint idx);
    /// Returns face tag of face with key \c idx
    inline concepts::AdaptiveControlP<2>& ctrl2_(uint idx);
    /// Returns the iterator to the vertex data of the vertex with key \c idx
    inline std::map<uint, concepts::VertexData>::const_iterator
    vertexListFind(uint idx) const;
    /// Returns the end of the list with vertex data
    inline std::map<uint, concepts::VertexData>::const_iterator
    vertexListEnd() const;
    /// Returns the iterator to the edge data of the edge with key \c idx
    inline std::map<uint, concepts::EdgeData>::const_iterator
    edgeListFind(uint idx) const;
    /// Returns the end of the list with edge data
    inline std::map<uint, concepts::EdgeData>::const_iterator
    edgeListEnd() const;
    /// Returns a reference to the dimension of the space
    uint& dim_() { return spc_->dim_; }
    /// Deactivates the children of the vertex \c vtx
    inline void deactivate_(const concepts::Vertex& vtx);
    /** Deactivates the children of the edge \c edg
	(including the middle vertex). */
    inline void deactivate_(const concepts::Edge& edg);
    /// Computes maximal polynomial \c Pmax degree in \c cntr in tensor form
    inline void computePmax_(const concepts::Quad& cntr, ushort Pmax[2]) const;
    /// Space to build the vertex degrees of freedom from
    Space* spc_;
  };

  concepts::AdaptiveControl& BuildDofsBase::ctrl0_(uint idx) {
    conceptsAssert(spc_, concepts::Assertion());
    return spc_->ctrl0_[idx];
  }

  concepts::AdaptiveControlP<1>& BuildDofsBase::ctrl1_(uint idx) {
    conceptsAssert(spc_, concepts::Assertion());
    return (*(spc_->ctrl1_))[idx];
  }

  concepts::AdaptiveControlP<2>& BuildDofsBase::ctrl2_(uint idx) {
    conceptsAssert(spc_, concepts::Assertion());
    return spc_->ctrl2_[idx];
  }

  std::map<uint, concepts::VertexData>::const_iterator
  BuildDofsBase::vertexListFind(uint idx) const {
    conceptsAssert(spc_, concepts::Assertion());
    return spc_->vertexList_.find(idx);
  }

  std::map<uint, concepts::VertexData>::const_iterator
  BuildDofsBase::vertexListEnd() const {
    conceptsAssert(spc_, concepts::Assertion());
    return spc_->vertexList_.end();
  }

  std::map<uint, concepts::EdgeData>::const_iterator
  BuildDofsBase::edgeListFind(uint idx) const {
    conceptsAssert(spc_, concepts::Assertion());
    return spc_->edgeList_.find(idx);
  }

  std::map<uint, concepts::EdgeData>::const_iterator
  BuildDofsBase::edgeListEnd() const {
    conceptsAssert(spc_, concepts::Assertion());
    return spc_->edgeList_.end();
  }

  void BuildDofsBase::deactivate_(const concepts::Vertex& vtx) {
    conceptsAssert(spc_, concepts::Assertion());
    spc_->deactivate_(vtx);
  }

  void BuildDofsBase::deactivate_(const concepts::Edge& edg) {
    conceptsAssert(spc_, concepts::Assertion());
    spc_->deactivate_(edg);
  }

  void BuildDofsBase::computePmax_(const concepts::Quad& cntr,
				   ushort Pmax[2]) const {
    conceptsAssert(spc_, concepts::Assertion());
    spc_->computePmax_(cntr, Pmax);
  }

} // namespace hp2D

#endif // buildDofsBase2D_hh
