/** @file buildInnerDofs.hh Factored out part of building elements in
    hp2D::Space.
    @author Philipp Frauenfelder, 2004
*/

#ifndef buildInnerDofs2D_hh
#define buildInnerDofs2D_hh

#include "hp2D/space.hh"
#include "hp2D/buildDofsBase.hh"

namespace hp2D {

  // forward declaration
  class Space;

  // ******************************************************** BuildInnerDofs **

  /** Responsible to build the inner degrees of freedom in a space.
      A call to the application operator builds the T columns for the
      inner modes of a given quad.

      If you want to change the way this is done, derive from this
      class and overwrite the application operator. The access to the
      internal data of the space (which is needed) is provided through
      the protected member functions of BuildDofsBase (it is a friend
      of Space).

      @par Trunk Polynomial Spaces
      Using this class, all degrees of freedom of the face are built (full
      tensor product space). It is very simply to change this behaviour
      in a derived class as the truncation function for the degrees of
      freedom is separated into trunk(). BuildInnerDofsLinTrunk is an example
      for a truncated local polynomial space (it is the default for
      building internal degrees of freedom).
      @par
      BuildInnerDofsHypTrunk is another example. The plot below compares
      the relative energy error versus degrees of freedom for three
      different internal degrees of freedom polynomial spaces. The plot
      was done with @ref hpFEM2d.cc "the hpFEM2d tutorial program" by
      exchanging the strategy for the internal degrees of freedom right
      after creating the Space. The result at an error of 1.7e-6 needs
      the following number of degrees of freedom:
      <table><tr><td><b>Space</b></td><td><b>Degrees of Freedom</b></td>
      <td><b>Relative Energy Error</b></td></tr>
      <tr><td>Full tensor product</td><td>4955</td><td>1.76e-6</td></tr>
      <tr><td>@ref BuildInnerDofsLinTrunk "Linear trunk"</td><td>2570</td><td>1.77e-6</td></tr>
      <tr><td>@ref BuildInnerDofsHypTrunk "Hyperbolic trunk"</td><td>1841</td><td>1.79e-6</td></tr>
      </table>

      <img src="hp2D-trunk.png" height=480 width=640 alt="Different polynomial spaces for internal degrees of freedom: relative energy error vs. ndofs">

      @author Philipp Frauenfelder, 2004
  */
  class BuildInnerDofs : public BuildDofsBase {
  public:
    /// Default constructor
    BuildInnerDofs() {}
    /** Constructor.
	@param spc Space to build the inner degrees of freedom from
    */
    BuildInnerDofs(Space& spc) : BuildDofsBase(spc) {}
    virtual void operator()(concepts::Quad& cntr,
			    concepts::TColumn<Real>*& T1);
    virtual BuildInnerDofs* clone(Space* spc = 0) const;
    /** Trunk space indicator function.
	For each degree of freedom, this function is called to find out
	if it should be built or not. If the return value is \c true, the
	degree of freedom is built, otherwise it is not.

	The default behaviour of this function is to return \c true.
	Overwrite it if you want to have a trunk space (e.g. as in
	BuildInnerDofsLinTrunk).
	@param p Polynomial degree in first direction of current
	degree of freedom
	@param q Polynomial degree in second direction of current
	degree of freedom
	@param P Maximal polynomial degree in the interior
    */
    virtual bool trunk(uint p, uint q, const ushort P[2]);
  };

  // ************************************************ BuildInnerDofsLinTrunk **

  /** Responsible to build the inner degrees of freedom in a (trunk) space.

      Using this class, only a trunk of the degrees of freedom in the
      interior are built: if
      <tt>(p + q) <= ((P[0] > P[1]) ? P[0] : P[1])</tt>
      the degree of freedom with polynomial degrees \c p, \c q
      in each direction is built (<tt>P[i]</tt> are the maximal polynomial
      degrees in each direction.)

      This class is registered as the default strategy in a Space to
      build the inner degrees of freedom. Giving Space a different
      class through the use of Space::buildInnerDofs, you can change
      that behaviour.

      @author Philipp Frauenfelder, 2004
  */
  class BuildInnerDofsLinTrunk : public BuildInnerDofs {
  public:
    /// Default constructor
    BuildInnerDofsLinTrunk() {}
    /** Constructor.
	@param spc Space to build the inner degrees of freedom from
    */
    BuildInnerDofsLinTrunk(Space& spc) : BuildInnerDofs(spc) {}
    virtual BuildInnerDofsLinTrunk* clone(Space* spc = 0) const;
    virtual bool trunk(uint p, uint q, const ushort P[2]);
  };

  // ************************************************ BuildInnerDofsHypTrunk **

  /** Responsible to build the inner degrees of freedom in a (trunk) space.

      Using this class, only a trunk of the degrees of freedom in the
      interior are built: if
      <tt>p*q <= 1.1*Pmax</tt>
      the degree of freedom with polynomial degrees \c p, \c q
      in each direction is built (<tt>Pmax</tt> is the maximal polynomial
      degree in each direction.)

      This class is registered as the default strategy in a Space to
      build the inner degrees of freedom. Giving Space a different
      class through the use of Space::buildInnerDofs, you can change
      that behaviour.

      @author Philipp Frauenfelder, 2004
  */
  class BuildInnerDofsHypTrunk : public BuildInnerDofs {
  public:
    /// Default constructor
    BuildInnerDofsHypTrunk() {}
    /** Constructor.
	@param spc Space to build the inner degrees of freedom from
    */
    BuildInnerDofsHypTrunk(Space& spc) : BuildInnerDofs(spc) {}
    virtual BuildInnerDofsHypTrunk* clone(Space* spc = 0) const;
    virtual bool trunk(uint p, uint q, const ushort P[2]);
  };

} // namespace hp2D

#endif // buildInnerDofs2D_hh
