/* Functions on space with hp elements in 2D
 */

#include "basics/exceptions.hh"
#include "function.hh"
#include "quad.hh"

namespace hp2D {

  using concepts::Real;
  using concepts::Real2d;

  // ***************************************************************** Value **

  void Value::operator() (const concepts::Element<Real>& elm,
			  const concepts::Array<Real>& coeff,
			  Real& val, const uint i[2]) const {
    // Quadrilaterial
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    if (quad)
      return (*this)(*quad, coeff, val, i);

    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Value::operator() (const Quad& elm, const concepts::Array<Real>& coeff,
			  Real& val, const uint i[2]) const {
    const uint npx = elm.integrationX()->n();
    const uint npy = elm.integrationY()->n();

    conceptsAssert(i[0] < npx, concepts::Assertion());
    conceptsAssert(i[1] < npy, concepts::Assertion());

    const uint nsx = elm.shpfctX()->n();
    const uint nsy = elm.shpfctY()->n();
    const Real* sx = elm.shpfctX()->values();
    const Real* sy = elm.shpfctY()->values();

    val = 0;
    if (coeff.size() > 0) {
      conceptsAssert(coeff.size() >= nsx*nsy, concepts::Assertion());

      const Real* coeffTmp = (const Real*)coeff;
      uint i2[2] = {i[0], i[1]};
      for (uint j = 0; j < nsy; ++j, i2[1]+=npy){
	i2[0]=i[0];
	for (uint k = 0; k < nsx; ++k, i2[0]+=npx)
	  val += sx[i2[0]] * sy[i2[1]]* *coeffTmp++;
      }
    }
  }

  void Value::operator() (const concepts::Element<Real>& elm, const uint j[2],
			 Real& val, const uint i[2]) const {
    // Quadrilaterial
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    if (quad)
      return (*this)(*quad, j, val, i);

    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Value::operator() (const Quad& elm, const uint j[2],
			 Real& val, const uint i[2]) const {
      const uint npx = elm.integrationX()->n();
      const uint npy = elm.integrationY()->n();

      conceptsAssert(i[0] < npx, concepts::Assertion());
      conceptsAssert(i[1] < npy, concepts::Assertion());

      conceptsAssert(j[0] < elm.shpfctX()->n(), concepts::Assertion());
      conceptsAssert(j[1] < elm.shpfctY()->n(), concepts::Assertion());

      const Real sx = elm.shpfctX()->values()[i[0]+j[0]*npx];
      const Real sy = elm.shpfctY()->values()[i[1]+j[1]*npy];
	
      val = sx * sy;
  }

  void Value::operator() (const concepts::Element<Real>& elm,
			  const concepts::Array<Real>& coeff,
			  Real& val,
			  const Real p, const Real t) const {
    Real2d p2(p);
    operator()(elm, coeff, val, p2, t);
  }

  void Value::operator() (const concepts::Element<Real>& elm,
			  const concepts::Array<Real>& coeff,
			  Real& val,
			  const Real2d& p, const Real t) const {
    conceptsAssert(p[0] >= 0, concepts::Assertion());
    conceptsAssert(p[0] <= 1, concepts::Assertion());
    conceptsAssert(p[1] >= 0, concepts::Assertion());
    conceptsAssert(p[1] <= 1, concepts::Assertion());

    // Quadrilaterial
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    if (quad)
      return (*this)(*quad, coeff, val, p, t);

    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Value::operator() (const Quad& elm, const concepts::Array<Real>& coeff,
			  Real& val, const Real2d& p, const Real t) const {
    DEBUGL(0, "coeff = " << coeff);
    Real pTmp[2] = { (p[0]-.5)*2., (p[1]-.5)*2. };
    DEBUGL(0, "p = " << p << " -> (" << pTmp[0] << ", " << pTmp[1] << ')');
    concepts::Karniadakis<1,0> tmpX(elm.shpfctX()->n()-1, pTmp, 1,
				    0, 0, false);
    concepts::Karniadakis<1,0> tmpY(elm.shpfctY()->n()-1, pTmp+1, 1,
				    0, 0, false);
    DEBUGL(0, "x values = " << tmpX << ", y values = " << tmpY);
    const uint nsx = tmpX.n(); const Real* sx = tmpX.values();
    const uint nsy = tmpY.n(); const Real* sy = tmpY.values();
    val = 0;
    if (coeff.size() > 0) {
      conceptsAssert(coeff.size() >= nsx*nsy, concepts::Assertion());
      const Real* coeffTmp = coeff;
      for (uint i = 0; i < nsy; ++i, ++sy) {
	sx = tmpX.values();
	for (uint j = 0; j < nsx; ++j, ++sx)
	  val += *sx * *sy * *coeffTmp++;
      }
    }
  }

  void Value::operator() (const concepts::Element<Real>& elm,
			  const concepts::Array<Real>& coeff,
			  Real& val,
			  const concepts::Real3d& p,
			  const Real t) const {
    Real2d p2(p);
    operator()(elm, coeff, val, p2, t);
  }

  std::ostream& Value::info(std::ostream& os) const {
    return os << "Value()";
  }

  // ****************************************************************** Grad **

  void Grad::operator() (const concepts::Element<Real>& elm,
			 const concepts::Array<Real>& coeff,
			 concepts::Array<Real>& val, const uint i[2]) const {
    // Quadrilaterial
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    if (quad)
      return (*this)(*quad, coeff, val, i);

    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Grad::operator() (const Quad& elm, const concepts::Array<Real>& coeff,
			 concepts::Array<Real>& val, const uint i[2]) const {
      const uint npx = elm.integrationX()->n();
      const uint npy = elm.integrationY()->n();

      conceptsAssert(i[0] < npx, concepts::Assertion());
      conceptsAssert(i[1] < npy, concepts::Assertion());

      const uint nsx = elm.shpfctX()->n();
      const uint nsy = elm.shpfctY()->n();
      const Real* sx = elm.shpfctX()->values();
      const Real* sy = elm.shpfctY()->values();
      const Real* const sdx = elm.shpfctDX()->values();
      const Real* const sdy = elm.shpfctDY()->values();

      val[0] = val[1] = 0;
      if (coeff.size() > 0) {
	conceptsAssert(coeff.size() >= nsx*nsy, concepts::Assertion());

	// calculate the inverse of the jacobian
	const Real px = elm.integrationX()->abscissas()[i[0]];
	const Real py = elm.integrationY()->abscissas()[i[1]];
	concepts::MapReal2d jacinv(elm.jacobianInverse((px + 1.0)/2.0,
						       (py + 1.0)/2.0));
	
	const Real* coeffTmp = (const Real*)coeff;
	uint i2[2] = {i[0], i[1]};
	// loop over shape functions
	for (uint j = 0; j < nsy; ++j, i2[1]+=npy){
	  i2[0] = i[0];
	  for (uint k = 0; k < nsx; ++k, i2[0]+=npx){
	    // loop over two vector components
	    for (uint l = 0; l < 2; ++l)
	      val[l] += (jacinv(0,l) * sdx[i2[0]] * sy[i2[1]] + 
			 jacinv(1,l) * sx[i2[0]] * sdy[i2[1]]) * *coeffTmp;
	    coeffTmp++;
	  }
	}
      } // coeff.size()
  }

  void Grad::operator() (const concepts::Element<Real>& elm, const uint j[2],
			 concepts::Array<Real>& val, const uint i[2]) const {
    // Quadrilaterial
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    if (quad)
      return (*this)(*quad, j, val, i);

    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Grad::operator() (const Quad& elm, const uint j[2],
			 concepts::Array<Real>& val, const uint i[2]) const {
      const uint npx = elm.integrationX()->n();
      const uint npy = elm.integrationY()->n();

      conceptsAssert(i[0] < npx, concepts::Assertion());
      conceptsAssert(i[1] < npy, concepts::Assertion());

      conceptsAssert(j[0] < elm.shpfctX()->n(), concepts::Assertion());
      conceptsAssert(j[1] < elm.shpfctY()->n(), concepts::Assertion());

      const Real sx = elm.shpfctX()->values()[i[0]+j[0]*npx];
      const Real sy = elm.shpfctY()->values()[i[1]+j[1]*npy];
      const Real sdx = elm.shpfctDX()->values()[i[0]+j[0]*npx];
      const Real sdy = elm.shpfctDY()->values()[i[1]+j[1]*npy];

      // calculate the inverse of the jacobian
      const Real px = elm.integrationX()->abscissas()[i[0]];
      const Real py = elm.integrationY()->abscissas()[i[1]];
      concepts::MapReal2d jacinv(elm.jacobianInverse((px + 1.0)/2.0,
						     (py + 1.0)/2.0));
      // loop over two vector components
      for (uint l = 0; l < 2; ++l)
	val[l] = (jacinv(0,l) * sdx * sy + jacinv(1,l) * sx * sdy);
  }

  std::ostream& Grad::info(std::ostream& os) const {
    return os << "Grad()";
  }


} // namespace hp2D
