/* space class for 2D hp FEM
 * meshAndPoly_
 */

#include "hp2D/space.hh"
#include "hp2D/quadFunctions.hh"
#include "toolbox/boundary.hh"

// debugging
#include "basics/debug.hh"

#define hpSpace2DmeshAndPoly_D 0
#define hpSpace2DenforceBC_D 0

namespace hp2D {

  // ***************************************************************** Space **

  void Space::meshAndPoly_(concepts::Quad2d& cell,
			   concepts::Level<2>& L, int* P) {
    concepts::Connector2& cntr = cell.connector();
    concepts::AdaptiveControlP<2>& ctrl2(ctrl2_[cntr.key()]);
    const concepts::Level<2>& l(cell.level());
    DEBUGL(hpSpace2DmeshAndPoly_D, l << " - start: " << cntr << ", l = " << l
	   << ", L = " << L
	   << ", P = (" << P[0] << ", " << P[1] << ')');

    concepts::Quad* cntrQ = dynamic_cast<concepts::Quad*>(&cntr);
    conceptsAssert(cntrQ, concepts::Assertion());
    DEBUGL(hpSpace2DmeshAndPoly_D, l << " - " << ctrl2 << ", L = " << L);

    /* If L == -1 and the connector is member of the space: this cell was
       in the old space. Apply adjustments here!
    */
    if (L == concepts::Level<2>(-1) && ctrl2.tag_.isMember()) {
      concepts::AdaptiveAdjustP<2>& adj(adj_[cntr.key()]);
      L = l + adj.level_;
      L.maximize(concepts::Level<2>(0));
      for (uint i = 0; i < 2; ++i) {
	ctrl2.p_[i] += adj.p_[i]; // p int
	if (ctrl2.p_[i] < 1) ctrl2.p_[i] = 1;
	P[i] = ctrl2.p_[i];
      }
    } 
    DEBUGL(hpSpace2DmeshAndPoly_D, l << " - updated p and l: "
	   << ctrl2 << ", L = " << L);

    /* If the desired level L == -1 or we are not on the desired level,
       this cell is not in the new space. Therefore:
       - set refinement strategy and test if it works
    */
    if (L == concepts::Level<2>(-1) || l < L) {
      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - not on desired level");

      if (cell.getStrategy() == 0) { // setting refinement strategy
	DEBUGL(hpSpace2DmeshAndPoly_D, l << " - setting refinement strategy");
	conceptsAssert(! (L == concepts::Level<2>(-1)), concepts::Assertion());
	concepts::Level<2> levelDiff = L - l;
	conceptsAssert(! (levelDiff < concepts::Level<2>(0)),
		       concepts::Assertion());
	if (levelDiff > concepts::Level<2>(0)) {
	  enum subdivTypes { NONE = 0, X = 1, Y = 2 };
	  uint subdiv = 0;
	  if (levelDiff.l_[0] > 0) subdiv |= X;
	  if (levelDiff.l_[1] > 0) subdiv |= Y;
	  switch (subdiv) {
	  case Y | X:
	    cell.setStrategy(concepts::Quad2dSubdiv4::instance()); break;
	  case X: cell.setStrategy(concepts::Quad2dSubdiv2V::instance());
	    break;
	  case Y: cell.setStrategy(concepts::Quad2dSubdiv2H::instance());
	    break;
	  default: break;
	  }
	}
      } // end of setting refinement strategy

      // testing, if refinement is allowed
      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - testing, if refinement for "
	     << L-l << " is allowed");
      try {
	cell.child(0);
      }
      catch (concepts::StrategyChange& e) {
	DEBUGL(1, "caught " << e);
	conceptsAssert(! (L == concepts::Level<2>(-1)), concepts::Assertion());
	L = l;
      }
    }

    /* If the desired level L == -1 or we are not on the desired level,
       this cell is not in the new space. Therefore:
       - call meshAndPoly_ for all children of the cell
       - get max p for all edges from the children of the edges
       - get max p for the cell from the children of the cell
    */
    if (L == concepts::Level<2>(-1) || l < L) {
      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - not on desired level");

      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - calling children");
      // call meshAndPoly for all children
      concepts::Level<2> Lmax(-1);
      concepts::Quad2d* chld = 0;
      for(uint i = 0; (chld = cell.child(i)) != 0; ++i) {
	concepts::Level<2> LL(L);
	int PP[2] = { P[0], P[1] };

	meshAndPoly_(*chld, LL, PP);

	for (uint i = 0; i < 2; ++i) if (P[i] < PP[i]) P[i] = PP[i];
	Lmax.maximize(LL);
      }
      L = Lmax; 

      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - getting p for cell: before "
	     << ctrl2);
      for (uint i = 0; i < 2; ++i) ctrl2.p_[i] = P[i];
      ctrl2.tag_ = concepts::AdaptiveControlTag::VLD_P;
      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - ... and after " << ctrl2);

      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - getting p for edges");
      const concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
	DEBUGL(hpSpace2DmeshAndPoly_D, l << " - " << *cntr1
	       << " has " << ctrl1);

	// loop over all children of current cell
	const concepts::Connector2* cntrChild = 0;
	for (uint n = 0; (cntrChild = cntr.child(n)) != 0; ++n) {
	  // get relevant edge of the child
	  const concepts::Connector1* edgeChild = cntrChild->edge(i);
	  // check if edge is child of edge of the large cell
	  bool takeEdge = false;
	  const concepts::Connector1* cntrEdgeChild = 0;
	  for (uint m = 0;
	       (!takeEdge) && ((cntrEdgeChild = cntr1->child(m)) != 0);
	       ++m) {
	    if (cntrEdgeChild->key() == edgeChild->key())
	      takeEdge = true;
	  }

	  if (takeEdge) {
	    concepts::AdaptiveControlP<1>&
	      childCtrl1((*ctrl1_)[edgeChild->key()]);
// 	    DEBUGL(hpSpace2DmeshAndPoly_D, l << " - has child " << *edgeChild
// 		   << " with " << childCtrl1);
	    if (ctrl1.p_[0] < childCtrl1.p_[0]) ctrl1.p_[0] = childCtrl1.p_[0];
	  }
	} // for children of current cell

	ctrl1.tag_.validateP();
	DEBUGL(hpSpace2DmeshAndPoly_D, l << " - and gets " << ctrl1);
      } // for edges
    } else {
      /* cell is in the new space:
	 - take degrees of freedom into space
	 - bring polynomial degrees onto faces and edges
      */
      for (uint i = 0; i < 2; ++i) ctrl2.p_[i] = P[i]; // p int
      ctrl2.tag_ = concepts::AdaptiveControlTag::VLD_P;
      ctrl2.tag_.takeIn();
      DEBUGL(hpSpace2DmeshAndPoly_D, l << " - cell has " << ctrl2
	     << ", level = " << cell.level());

      // bring p of the cell onto edges
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	uint pIndex = 0, qIndex = 0;
	if (cntrQ)
	  QuadFunctions::edgeOrientation(i, pIndex, qIndex);
	concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
	if (ctrl1.p_[0] < ctrl2.p_[pIndex]) ctrl1.p_[0] = ctrl2.p_[pIndex];
	ctrl1.tag_.validateP();
	ctrl1.tag_.takeIn();
	DEBUGL(hpSpace2DmeshAndPoly_D, l << " - " << *cntr1
	       << " has " << ctrl1);
      } // for edges

      concepts::Connector0* cntr0 = 0;
      for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
// 	ctrl0_[cntr0->key()].tag_ = concepts::AdaptiveControlTag::MEMBER;
	concepts::AdaptiveControl& ctrl0(ctrl0_[cntr0->key()]);
	ctrl0.tag_.takeIn();
	DEBUGL(hpSpace2DmeshAndPoly_D, l << " - " << *cntr0
	       << " has " << ctrl0);
      }
    }

    /* determine boundary conditions for edges and vertices:
     * if a face has DIRICHLET bc, then all adjaccent edges and vertices
     * have them too -> such a dof is a PASSIVE member of the space
     */
    if (bc_) {
      enforceBC_(cell);
    } // if bc

    DEBUGL(hpSpace2DmeshAndPoly_D, l << " - done.");
}

  void Space::enforceBC_(concepts::Quad2d& cell) {
    DEBUGL(hpSpace2DenforceBC_D, "enforcing boundary conditions");
    conceptsAssert(bc_ != 0, concepts::Assertion());
    concepts::Connector2& cntr = cell.connector();

    // loop over edges
    concepts::Connector1* cntr1 = 0;
    for (uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      if ((*bc_)(cntr1->attrib()).type() == concepts::Boundary::DIRICHLET) {
	// mark edge as passive
	(*ctrl1_)[cntr1->key()].tag_.deactivate();

	// mark vertices as passive
	concepts::Connector0* cntr0 = 0;
	for (uint j = 0; (cntr0 = cntr1->vertex(j)) != 0; ++j) {
	  ctrl0_[cntr0->key()].tag_.deactivate();
	}
      } // DIRICHLET bc
    } // for i

    // loop over vertices
    concepts::Connector0* cntr0 = 0;
    for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
      if ((*bc_)(cntr0->attrib()).type() == concepts::Boundary::DIRICHLET) {
	// mark vertex as passive
	ctrl0_[cntr0->key()].tag_.deactivate();
      } // DIRICHLET bc
    } // for i
  }

} // namespace hp2D
