/* solving PDGL in 2D with hanging nodes
 */

#include <iostream>
#include <memory>
#include <vector>
#include <sstream>
#include <cstdlib>
#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"
#include "vectorial.hh"
#include "hp2D.hh"
#include "applications/square.hh"

using concepts::Real;

#include "applications/meshes.cc"


// ************************************************************ Main Program **

int main(int argc, char** argv) {
  concepts::Stacktrace::doit() = false;
  try {
    uint l  = 1;   // number of first refinements
    uint lv = 2;   // number of vertical refinements of first element
    uint lh = 2;   // number of horizontal refinements of first element
    uint p = 3;

    // *** Gitter ***
    Square msh(1.0, 1.0);

    // *** Randbedingung ***
    std::auto_ptr<concepts::BoundaryConditions> bc(new concepts::BoundaryConditions());
    for(uint i=1;i<5;i++)
      bc->add(concepts::Attribute(i), concepts::Boundary(concepts::Boundary::DIRICHLET) );

    // *** FEM-Raum ***
    hp2D::Space spc(msh, l, p, bc.get());
    spc.buildInnerDofs(hp2D::BuildInnerDofs());

    // Integrationsregel setzen
    hp2D::Quad::setIntegrationRule(hp2D::Quad::GAUSS_LOBATTO, true, 10);

    // vertikal halbieren
    for(uint i=0; i<lv; ++i){
      std::cout << i+1 << ".te vertikale Verfeinerung " << std::endl;
      concepts::Level<2> l(0);l.l_[0] = 1;l.l_[1] = 0;
      concepts::AdaptiveAdjustP<2> a(l, 0);
      std::auto_ptr<hp2D::Space::Scanner> sc(spc.scan());
      const concepts::Element<Real>& elm = (*sc)++;
      spc.adjust(elm, a);
    }
    // horizontal halbieren
    for(uint i=0; i<lh; ++i){
      std::cout << i+1 << ".te horizontale Verfeinerung " << std::endl;
      concepts::Level<2> l(0);l.l_[0] = 0;l.l_[1] = 1;
      concepts::AdaptiveAdjustP<2> a(l, 0);
      std::auto_ptr<hp2D::Space::Scanner> sc(spc.scan());
      const concepts::Element<Real>& elm = (*sc)++;
      spc.adjust(elm, a);
    }
    spc.rebuild();
    std::cout << "scalar Space: " << spc << std::endl;

    // *** Bilinearform ***
    hp2D::Laplace bf;
    concepts::DenseMatrix<Real> stiff(spc, bf);
    std::cout << "stiff (Laplace) = " << stiff << std::endl;

    // *** Linearform ***
    hp2D::Riesz lform(concepts::ParsedFormula("(-2*(x*(x-1)+y*(y-1)))"),  bc.get() );
    concepts::Vector<Real> rhs(spc, lform);
    std::cout << "rhs = " << rhs << std::endl;
    
    // *** Loesen ***
    concepts::CG<Real> solver(stiff, 1e-6, 2000);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;
    
    // Output of solution
    graphics::drawDataGnuplot(spc, "calc_hp2D_hang.gnuplot" , sol);
    // eps output of the mesh
    graphics::drawMeshEPS(spc, "calc_hp2D_hang.eps");

  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
