/* Tests the evaluation of functions.
 */

#include "hp2D/testsuite/functionEvaluation.hh"
#include "hp2D/space.hh"
#include "hp2D/function.hh"
#include "hp2D/buildInnerDofs.hh"
#include "geometry/meshImport2D.hh"
#include "function/vector.hh"
#include "toolbox/array.hh"

namespace test {

  // **************************************************** FunctionEvaluation **

  void FunctionEvaluation::run() {
    concepts::Import2dMesh m
      (SOURCEDIR "/geometry/testsuite/Coordinates2D.dat",
       SOURCEDIR "/geometry/testsuite/quad.dat",
       SOURCEDIR "/geometry/testsuite/empty.dat");
    _test(m.ncell() == 1);
    hp2D::Space spc(m, 0, 1);
    spc.buildInnerDofs(hp2D::BuildInnerDofs());
    spc.rebuild();
    _test(spc.nelm() == 1);
    _test(spc.dim() == 4);
    std::auto_ptr<hp2D::Space::Scan> sc(spc.scan());
    while (*sc) {
      hp2D::Element<Real>& elm = (*sc)++;
      int pMax[2] = { 1, 2 };
      spc.adjust(elm, concepts::AdaptiveAdjustP<2>(0, pMax));
    }
    spc.rebuild();
    _test(spc.nelm() == 1);
    _test(spc.dim() == 12);
    
    testValue(spc); testValueInt(spc);
  }

  void FunctionEvaluation::testValue(hp2D::Space& spc) {
    concepts::Vector<Real> f(spc);
    concepts::Array<Real> coeff;
    Real val = 0;
    hp2D::Value v;

    using concepts::Real2d;
    Real2d points[6] = { Real2d(0.1, 0.111), Real2d(0.2, 0.222),
			 Real2d(0.1, 0.222), Real2d(0.777, 0.888),
			 Real2d(0.5, 0.5), Real2d(1./3., 1./3.) };

    std::auto_ptr<hp2D::Space::Scan> sc(spc.scan());
    while (*sc) {
      hp2D::Element<Real>& elm = (*sc)++;
      for (uint i = 0; i < spc.dim(); ++i) { // loop over FE basis functions
	uint s[2] = { 0, 0 };
	switch(i) {
	case  0: s[0] = 0; s[1] = 0; break; // vtx 0
	case  1: s[0] = 1; s[1] = 0; break;
	case  2: s[0] = 1; s[1] = 1; break;
	case  3: s[0] = 0; s[1] = 1; break; // vtx 3
	case  4: s[0] = 2; s[1] = 0; break; // edg 0
	case  5: s[0] = 1; s[1] = 2; break; // edg 1
	case  6: s[0] = 1; s[1] = 3; break;
	case  7: s[0] = 2; s[1] = 1; break; // edg 2
	case  8: s[0] = 0; s[1] = 2; break; // edg 3
	case  9: s[0] = 0; s[1] = 3; break;
	case 10: s[0] = 2; s[1] = 2; break; // int 0
	case 11: s[0] = 2; s[1] = 3; break; // int 1
	}
	f[i] = 1;
	elm.T().extract(f, coeff);
	DEBUGL(0, "Shape fct. " << i << " (" << s[0] << ", " << s[1]
	       << ") f = " << f << ", coeff = " << coeff)

	hp2D::Quad* q = dynamic_cast<hp2D::Quad*>(&elm);
	_test(q != 0);

	// loop over quadrature points in x in elm
	for (uint j = 0; j < q->integrationX()->n(); ++j) {
	  // loop over quadrature points in y in elm
	  for (uint k = 0; k < q->integrationY()->n(); ++k) {
	    concepts::Real2d p((q->integrationX()->abscissas()[j] + 1.0)/2.0,
				(q->integrationY()->abscissas()[k] + 1.0)/2.0);
	    v(elm, coeff, val, p);
	    DEBUGL(0, p << " -> " << val << " ?= " << shapeFct_(s, p));
	    _numtest(val, shapeFct_(s, p));
	  } // for k
	} // for j

	for (uint j = 0; j < 6; ++j) {
	  v(elm, coeff, val, points[j]);
	  _numtest(val, shapeFct_(s, points[j]));
	}

	f[i] = 0;
      } // for i
    }
  }

  void FunctionEvaluation::testValueInt(hp2D::Space& spc) {
    concepts::Vector<Real> f(spc);
    concepts::Array<Real> coeff;
    Real val = 0;
    hp2D::Value v;

    std::auto_ptr<hp2D::Space::Scan> sc(spc.scan());
    while (*sc) {
      hp2D::Element<Real>& elm = (*sc)++;
      for (uint i = 0; i < spc.dim(); ++i) { // loop over FE basis functions
	uint s[2] = { 0, 0 };
	switch(i) {
	case  0: s[0] = 0; s[1] = 0; break; // vtx 0
	case  1: s[0] = 1; s[1] = 0; break;
	case  2: s[0] = 1; s[1] = 1; break;
	case  3: s[0] = 0; s[1] = 1; break; // vtx 3
	case  4: s[0] = 2; s[1] = 0; break; // edg 0
	case  5: s[0] = 1; s[1] = 2; break; // edg 1
	case  6: s[0] = 1; s[1] = 3; break;
	case  7: s[0] = 2; s[1] = 1; break; // edg 2
	case  8: s[0] = 0; s[1] = 2; break; // edg 3
	case  9: s[0] = 0; s[1] = 3; break;
	case 10: s[0] = 2; s[1] = 2; break; // int 0
	case 11: s[0] = 2; s[1] = 3; break; // int 1
	}
	f[i] = 1;
	elm.T().extract(f, coeff);
	DEBUGL(0, "Shape fct. " << i << " (" << s[0] << ", " << s[1]
	       << ") f = " << f << ", coeff = " << coeff)

	hp2D::Quad* q = dynamic_cast<hp2D::Quad*>(&elm);
	_test(q != 0);

	// loop over quadrature points in x in elm
	for (uint j = 0; j < q->integrationX()->n(); ++j) {
	  // loop over quadrature points in y in elm
	  for (uint k = 0; k < q->integrationY()->n(); ++k) {
	    uint l[2] = { j, k };
	    v(elm, coeff, val, l);
	    concepts::Real2d p((q->integrationX()->abscissas()[j] + 1.0)/2.0,
				(q->integrationY()->abscissas()[k] + 1.0)/2.0);
	    DEBUGL(0, p << " -> " << val << " ?= " << shapeFct_(s, p));
	    _numtest(val, shapeFct_(s, p));
	  } // for k
	} // for j
	f[i] = 0;
      } // for i
    }
  }

  Real FunctionEvaluation::shapeFct_(const uint i[2],
				     const concepts::Real2d& p) const {
    Real val = 1;
    for (uint j = 0; j < 2; ++j) {
      switch(i[j]) {
      case 0: val *= (1-p[j]); break;
      case 1: val *= p[j]; break;
      case 2: val *= p[j]*(1-p[j]); break;
      case 3: val *= p[j]*(1-p[j])*(4.*p[j]-2.); break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
    }
    return val;
  }

} // namespace test


int main(int, char**) {
  try {
    test::FunctionEvaluation t;
    t.run();
    return t.report();
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
  return 1;
}
