/* Bilinear forms for hp 3D FEM
 */

#include <cstring>
#include <fstream>
#include "basics/vectorsMatrices.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "toolbox/array.hh"
#include "toolbox/coeffIterator.hh"
#include "hp3D/hexahedron.hh"
#include "hp3D/bilinearForm.hh"
#include "hp3D/sumfact.hh"
#include "hp3D/shapeFunction3d.hh"
#include "basics/debug.hh"

// timings
#define matrixTiming

#ifdef matrixTiming
#include "toolbox/inputOutput.hh"
#include "toolbox/resourceMonitor.hh"
#endif // matrixTiming

// debugging
#define hpLaplaceAppl_D 0
#define hpLaplaceApplMem_D 0
#define hpIdentityAppl_D 0

namespace hp3D {

  // *************************************************************** Laplace **

  Laplace::Laplace() : oldElm_(0) {}

  Laplace::~Laplace() {}

  void Laplace::operator()(const concepts::Element<Real>& elmX, 
			   const concepts::Element<Real>& elmY, 
			   concepts::ElementMatrix<Real>& em) {

    const Hexahedron* elmYH = dynamic_cast<const Hexahedron*>(&elmY);
    if (elmYH) {
      const Hexahedron* elmXH = dynamic_cast<const Hexahedron*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Laplace::operator()(const Hexahedron& elmX, const Hexahedron& elmY, 
			   concepts::ElementMatrix<Real>& em) {

    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpLaplaceAppl_D, elmX);
      DEBUGL(hpLaplaceApplMem_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const uint nqz = elmX.integrationZ()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();
      const Real* const pz = elmX.integrationZ()->abscissas();

      bool sameJacobian = false;
      if (! sameJacobian) {
	// Jacobian
	jacobian_.resize(nqx*nqy*nqz);
	DEBUGL(hpLaplaceApplMem_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobian_");
	Real* dx = (Real*)jacobian_;
	jacobianInv_.resize(nqx*nqy*nqz);
	DEBUGL(hpLaplaceApplMem_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobianInv_");
	concepts::MapReal3d* jacInv = (concepts::MapReal3d*)jacobianInv_;
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i, ++pxAcc) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j, ++pyAcc) {
	    const Real* pzAcc = pz;
	    for (uint k = 0; k < nqz; ++k) {
	      concepts::MapReal3d jacobian
		(elmX.jacobian( (*pxAcc + 1.0)/2.0, 
				(*pyAcc + 1.0)/2.0,
				(*pzAcc++ + 1.0)/2.0 ));
	      *dx++ = jacobian.determinant();
	      concepts::MapReal3d jacinv(jacobian.inverse());
	      // with a coefficient matrix:
	      // jacinv * A.transpose() * jacinv.transpose()
	      *jacInv++ = jacinv * jacinv.transpose();
	    }
	  }
	}
      } // sameJacobian

      const Real* const wx = elmX.integrationX()->weights();
      const Real* const wy = elmX.integrationY()->weights();
      const Real* const wz = elmX.integrationZ()->weights();

      // shapefunctions
      const uint nsx = elmX.shpfctX()->n();
      const uint nsy = elmX.shpfctY()->n();
      const uint nsz = elmX.shpfctZ()->n();
      const Real* const sx = elmX.shpfctX()->values();
      const Real* const sy = elmX.shpfctY()->values();
      const Real* const sz = elmX.shpfctZ()->values();
      const Real* const sdx = elmX.shpfctDX()->values();
      const Real* const sdy = elmX.shpfctDY()->values();
      const Real* const sdz = elmX.shpfctDZ()->values();

      ShapeFunction3D<Real> u;
      u.n(0, nsx);        u.n(1, nsy);        u.n(2, nsz);
      u.nq(0, nqx);       u.nq(1, nqy);       u.nq(2, nqz);
      u.values(0, sx);    u.values(1, sy);    u.values(2, sz);
      u.abscissas(0, px); u.abscissas(1, py); u.abscissas(2, pz);
      u.weights(0, wx);   u.weights(1, wy);   u.weights(2, wz);
      u.T(&(elmX.T()));

      ShapeFunction3D<Real> v;
      v.n(0, nsx);        v.n(1, nsy);        v.n(2, nsz);
      v.nq(0, nqx);       v.nq(1, nqy);       v.nq(2, nqz);
      v.values(0, sx);    v.values(1, sy);    v.values(2, sz);
      v.abscissas(0, px); v.abscissas(1, py); v.abscissas(2, pz);
      v.weights(0, wx);   v.weights(1, wy);   v.weights(2, wz);
      v.T(&(elmY.T()));

      SumFactorization sumfact;
      em.zeros();
      for (uint i = 0; i < 3; ++i) {
	switch (i) {
	case 0: u.values(i, sdx); break;
	case 1: u.values(i, sdy); break;
	case 2: u.values(i, sdz); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
	for (uint j = 0; j < 3; ++j) {
	  switch (j) {
	  case 0: v.values(j, sdx); break;
	  case 1: v.values(j, sdy); break;
	  case 2: v.values(j, sdz); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  concepts::CoeffIterator<Real, concepts::MapReal3d>
	    coeff((concepts::MapReal3d*)jacobianInv_, i, j);
	  sumfact(em, jacobian_, &coeff, u, v);
	  switch (j) {
	  case 0: v.values(j, sx); break;
	  case 1: v.values(j, sy); break;
	  case 2: v.values(j, sz); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	}
	switch (i) {
	case 0: u.values(i, sx); break;
	case 1: u.values(i, sy); break;
	case 2: u.values(i, sz); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
      }

      for (uint i = 1; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
      DEBUGL(hpLaplaceAppl_D, "em = " << em);

      oldElm_ = &elmX;
    } // elmX == elmY
  }

  std::ostream& Laplace::info(std::ostream& os) const{
    return os << "Laplace";
  }

  // ************************************************************** Identity **

  Identity::Identity(const concepts::Formula* frm) : oldElm_(0), frm_(0) {
    if (frm)
      frm_.reset(frm->clone());
  }

  Identity::~Identity() {}

  void Identity::operator()(const concepts::Element<Real>& elmX, 
				 const concepts::Element<Real>& elmY, 
				 concepts::ElementMatrix<Real>& em) {

    const Hexahedron* elmYH = dynamic_cast<const Hexahedron*>(&elmY);
    if (elmYH) {
      const Hexahedron* elmXH = dynamic_cast<const Hexahedron*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void Identity::operator()(const Hexahedron& elmX, const Hexahedron& elmY, 
				 concepts::ElementMatrix<Real>& em) {

    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpIdentityAppl_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const uint nqz = elmX.integrationZ()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();
      const Real* const pz = elmX.integrationZ()->abscissas();

#ifdef matrixTiming
      concepts::ResourceMonitor timer;
#endif // matrixTiming

      bool sameJacobian = false;
      if (! sameJacobian) {
	// Jacobian
	jacobian_.resize(nqx*nqy*nqz);
	DEBUGL(hpIdentityAppl_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobian_");
	Real* dx = (Real*)jacobian_;
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j) {
	    const Real* pzAcc = pz;
	    for (uint k = 0; k < nqz; ++k) {
	      const concepts::Real3d refPoint( (*pxAcc + 1.0)/2.0,
					       (*pyAcc + 1.0)/2.0,
					       (*pzAcc++ + 1.0)/2.0 );
	      Real f = 1.0;
	      if (frm_.get())
		f = (*frm_)(elmX.cell().chi(refPoint[0],
					    refPoint[1], refPoint[2]));
	      *dx++ = elmX.jacobianDeterminant
		(refPoint[0], refPoint[1], refPoint[2]) * f;
	    }
	    ++pyAcc;
	  }
	  ++pxAcc;
	}
      } // sameJacobian
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("jacobian", timeCntr_, timer.utime());
#endif // matrixTiming

      const Real* const wx = elmX.integrationX()->weights();
      const Real* const wy = elmX.integrationY()->weights();
      const Real* const wz = elmX.integrationZ()->weights();

      // shapefunctions
      const uint nsx = elmX.shpfctX()->n();
      const uint nsy = elmX.shpfctY()->n();
      const uint nsz = elmX.shpfctZ()->n();
      const Real* const sx = elmX.shpfctX()->values();
      const Real* const sy = elmX.shpfctY()->values();
      const Real* const sz = elmX.shpfctZ()->values();

      ShapeFunction3D<Real> u;
      u.n(0, nsx);        u.n(1, nsy);        u.n(2, nsz);
      u.nq(0, nqx);       u.nq(1, nqy);       u.nq(2, nqz);
      u.values(0, sx);    u.values(1, sy);    u.values(2, sz);
      u.abscissas(0, px); u.abscissas(1, py); u.abscissas(2, pz);
      u.weights(0, wx);   u.weights(1, wy);   u.weights(2, wz);
      u.T(&(elmX.T()));

#ifdef matrixTiming
      if (timings_)
	timer.utime();
#endif // matrixTiming
      SumFactorization sumfact;
      em.zeros();
      sumfact(em, jacobian_, (concepts::CoeffIterator<Real, Real>*)0, u, u);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("whole_sumfact", timeCntr_, timer.utime());
#endif // matrixTiming

      for (uint i = 1; i < n; ++i)
	for (uint j = 0; j < i; ++j)
	  em(i,j) = em(j,i);
//       DEBUGL(hpIdentityAppl_D, "em = " << em);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("symmetrize_em", timeCntr_++, timer.utime());
#endif // matrixTiming

      oldElm_ = &elmX;
    } // elmX == elmY
  }

  concepts::InOutParameters* Identity::timings_ = 0;

  uint Identity::timeCntr_ = 0;

  void Identity::setTimings(concepts::InOutParameters* timings) {
#ifdef matrixTiming
    timings_ = timings;
    if (timings_) {
      timeCntr_ = 0;
      timings_->addArrayDouble("jacobian");
      timings_->addArrayDouble("whole_sumfact");
      timings_->addArrayDouble("symmetrize_em");
    }
#endif // matrixTiming
  }

  bool Identity::timings() {
#ifdef matrixTiming
    return true;
#else
    return false;
#endif // matrixTiming
  }

  std::ostream& Identity::info(std::ostream& os) const {
    return os << "Identity";
  }

} // namespace hp3D
