/* Bilinear forms for hp 3D FEM
 */

#ifndef hp3dbilinearform_hh
#define hp3dbilinearform_hh

#include <memory>
#include "operator/bilinearForm.hh"
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "toolbox/formula.hh"

namespace concepts {
  // forward declarations
  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  template<class F>
  class Array;

  class InOutParameters;
}

namespace hp3D {
  // forward declarations
  class Hexahedron;

  using concepts::Real;

  // *************************************************************** Laplace **

  /** A function class to calculate element matrices for the Laplacian.

      @author Philipp Frauenfelder, 2001
  */
  class Laplace : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    Laplace();
    virtual ~Laplace();

    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em); 
    void operator()(const Hexahedron& elmX, const Hexahedron& elmY, 
		    concepts::ElementMatrix<Real>& em);
    virtual Laplace* clone() const { return new Laplace(); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Intermediate data for element matrix computation
    concepts::Array<Real> jacobian_;
    concepts::Array<concepts::MapReal3d> jacobianInv_;

    const Hexahedron* oldElm_;
  };

  // ************************************************************** Identity **

  /** A function class to calculate element matrices for the mass matrix.

      This class is equiped with an interface to get timings of internal
      computations if compiled accordingly (see \c bilinearForm.cc file), see
      \c setTimings() and \c timings().

      @author Philipp Frauenfelder, 2001
  */
  class Identity : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    Identity(const concepts::Formula* frm = 0);
    virtual ~Identity();

    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em); 
    void operator()(const Hexahedron& elmX, const Hexahedron& elmY, 
		    concepts::ElementMatrix<Real>& em);

    /** @name Timing Interface
     *  These functions are used to get timings from class internal
     *  computations. The values are stored in a user defined
     *  concepts::InOutParameters structure in different arrays (see 
     *  \c setTimings). These arrays can be grouped into a table for
     *  easier postprocessing with
     *  \code
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "jacobian", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "whole_sumfact", output);
    std::ofstream ofs("table.gnuplot");
    ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(ofs);
        \endcode
     */
    //@{
    /** Sets the class to store the timing values in. Additionally,
	the \c timeCntr_ is reset to 0. This counter is used to fill
	in the values into the arrays listed below in subsequent calls.
	The following timings are taken and stored in \c timings:
	<ul>
	<li>evaluation of the Jacobian in \c jacobian
	<li>calling the sum fatorization subroutine in \c whole_sumfact
	<li>symmetrizing the element matrix in \c symmetrize_em
	</ul>
    */
    static void setTimings(concepts::InOutParameters* timings);
    /** Returns true if the class is able to do timings.  The ability
	to do timings depends on a compiler switch in \c bilinearform.cc file.
    */
    static bool timings();
    //@}
    virtual Identity* clone() const { return new Identity(frm_.get()); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Intermediate data for element matrix computation
    concepts::Array<Real> jacobian_;

    const Hexahedron* oldElm_;
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;

    /// Place to store timing values
    static concepts::InOutParameters* timings_;
    /// Counter for timing table
    static uint timeCntr_;
  };

} // namespace hp3D

#endif // hp3dbilinearform_hh
