/* hp extension for 3D FEM: ie. determine wether to refine or raise p
 * in an element
 */

#include <cmath>
#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "toolbox/array.hh"
#include "function/vector.hh"
#include "space/tmatrix.hh"
#include "space/element.hh"
#include "space/hpMethod.hh"
#include "integration/quadratureBase.hh"
#include "hpExtension.hh"
#include "element.hh"
#include "hexahedron.hh"
#include "basics/debug.hh"

// debugging

#define RefineRaise3DAppl_D 0
#define RefineRaise3DRegression_D 0

namespace hp3D {
  using concepts::Real;

  // ******************************************************** RefineOrRaise  **

  template<typename F>
  RefineOrRaise<F>::RefineOrRaise(const concepts::Vector<F>& solution,
				  const Real theta, concepts::AdaptiveSpace<F,
				  concepts::AdaptiveAdjustP<3> >& spc) :
    coeff_(0), legendre_(0), regr_(0), solution_(solution), spc_(spc),
    theta_(theta) {
  }

  template<typename F>
  void RefineOrRaise<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    DEBUGL(RefineRaise3DAppl_D, elm);
    // check for a 3D element
    const Element<F>* e = dynamic_cast<const Element<F>*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    // get local coefficients
    elm.T().extract(solution_, coeff_);
    DEBUGL(RefineRaise3DAppl_D, "coeff = " << coeff_);

    // compute Legendre coefficients
    const Hexahedron* elmH = dynamic_cast<const Hexahedron*>(&elm);
    conceptsAssert(elmH, concepts::Assertion());
    const uint nsx = elmH->shpfctX()->n();
    const uint nsy = elmH->shpfctY()->n();
    const uint nsz = elmH->shpfctZ()->n();

    const uint n = coeff_.size();
    legendre_.resize(n);
    legendre_.zeros();
    Real* legendre = (Real*)legendre_;
    for (uint r = 0; r < nsz; ++r) {
      for (uint q = 0; q < nsy; ++q) {
	for (uint p = 0; p < nsx; ++p) {
	  const Real* coeff = (Real*) coeff_;
	  for (uint k = 0; k < nsz; ++k) {
	    const Real c = transCoeff_(r, k);
	    if (c == 0.0) coeff += nsx*nsy;
	    else {
	      for (uint j = 0; j < nsy; ++j) {
		const Real d = c * transCoeff_(q, j);
		if (d == 0.0) coeff += nsx;
		else {
		  for (uint i = 0; i < nsx; ++i) {
		    *legendre += *coeff++ * d * transCoeff_(p, i);
		  } // for i
		} // d != 0
	      } // for j
	    } // c != 0
	  } // for k
	  ++legendre;
	} // for p
      } // for q
    } // for r
    DEBUGL(RefineRaise3DAppl_D, "legendre = " << legendre_);

//     // check if the transformation is correct
//     DEBUGL(RefineRaise3DAppl_D, "legendre = " << legendre_);
//     concepts::Array<Real> LegendreX(elmH->integrationX()->n()*(e->p()[0]+1));
//     concepts::JacobiPol(0.0, 0.0, e->p()[0], elmH->integrationX()->abscissas(),
// 			elmH->integrationX()->n(), (Real*)LegendreX);
//     DEBUGL(RefineRaise3DAppl_D, "Leg X = " << LegendreX);
//     concepts::Array<Real> LegendreY(elmH->integrationY()->n()*(e->p()[1]+1));
//     concepts::JacobiPol(0.0, 0.0, e->p()[1], elmH->integrationY()->abscissas(),
// 			elmH->integrationY()->n(), (Real*)LegendreY);
//     DEBUGL(RefineRaise3DAppl_D, "Leg Y = " << LegendreY);
//     concepts::Array<Real> LegendreZ(elmH->integrationZ()->n()*(e->p()[2]+1));
//     concepts::JacobiPol(0.0, 0.0, e->p()[2], elmH->integrationZ()->abscissas(),
// 			elmH->integrationZ()->n(), (Real*)LegendreZ);
//     DEBUGL(RefineRaise3DAppl_D, "Leg Z = " << LegendreZ);

//     const uint nqx = elmH->integrationX()->n();
//     const uint nqy = elmH->integrationY()->n();
//     const uint nqz = elmH->integrationZ()->n();
//     const Real * const sx = elmH->shpfctX()->values();
//     const Real * const sy = elmH->shpfctY()->values();
//     const Real * const sz = elmH->shpfctZ()->values();
//     for (uint k = 0; k < nqz; ++k) {
//       for (uint j = 0; j < nqy; ++j) {
// 	for (uint i = 0; i < nqx; ++i) {
// 	  Real resCoeff = 0.0;
// 	  Real resLegendre = 0.0;
// 	  for (uint p = 0; p < nsx; ++p) {
// 	    for (uint q = 0; q < nsy; ++q) {
// 	      for (uint r = 0; r < nsz; ++r) {
// 		resCoeff += coeff_[p + q*nsx + r*nsx*nsy] * 
// 		  ( sx[p*nqx + i] * sy[q*nqy + j] * sz[r*nqz + k] );
// 		resLegendre += legendre_[p + q*nsx + r*nsx*nsy] *
// 		  ( LegendreX[p+i*nsx] * LegendreY[q+j*nsy] *
// 		    LegendreZ[r+k*nsz] );
// 	      } // for r
// 	    } // for q
// 	  } // for p
// 	  DEBUGL(RefineRaise3DAppl_D, '('
// 		 << elmH->integrationX()->abscissas()[i] << ", "
// 		 << elmH->integrationY()->abscissas()[j] << ", "
// 		 << elmH->integrationZ()->abscissas()[k] << ") -> "
// 		 << resCoeff << " == " << resLegendre);
// 	} // for i
//       } // for j
//     } // for k

    // compute regression of log(legendre)
    concepts::Level<3> l(0);
    int P[3] = { 0, 0, 0 };
    // x
    regr_.resize(nsx);
    regr_.zeros();
    Real* regr = (Real*)regr_;
    legendre = (Real*)legendre_;
    for (uint p = 0; p < nsx; ++p, ++regr) {
      for (uint q = 0; q < nsy; ++q) {
	const Real c = 2.0/(2.0*q + 1.0);
	for (uint r = 0; r < nsz; ++r, ++legendre) {
	  *regr += (*legendre)*(*legendre) * 2.0/(2.0*r + 1.0) * c;
	} // for r
      } // for q
      *regr = std::sqrt(*regr);
    } // for p
    if (regression_(regr_, e->p()[0])) {
      P[0] = 1;
    } else {
      l.l_[0] = 1;
    }

    // y
    regr_.resize(nsy);
    regr_.zeros();
    regr = (Real*)regr_;
    for (uint q = 0; q < nsy; ++q, ++regr) {
      for (uint p = 0; p < nsx; ++p) {
	const Real c = 2.0/(2.0*p + 1.0);
	for (uint r = 0; r < nsz; ++r) {
	  const Real d = legendre_[r + q*nsz + p*nsz*nsy];
	  *regr += 2.0/(2.0*r + 1.0) * c * d*d;
	} // for r
      } // for p
      *regr = std::sqrt(*regr);
    } // for q
    if (regression_(regr_, e->p()[1])) {
      P[1] = 1;
    } else {
      l.l_[1] = 1;
    }
    DEBUGL(RefineRaise3DAppl_D, "l = " << l << ", P = ("
	   << P[0] << ", " << P[1] << ", " << P[2] << ')');

    if (l == concepts::Level<3>(0)) {
// P[0] = 1; P[1] = 1; P[2] = 1;
      concepts::AdaptiveAdjustP<3> a(0, P);
      spc_.adjust(elm, a);
    } else {
      concepts::AdaptiveAdjustP<3> a(l, P);
      spc_.adjust(elm, a);
    }
  }

  template<typename F>
  Real RefineOrRaise<F>::transCoeff_(const uint i, const uint j) const {
    if ((i == 0) && (j < 2)) return 0.5;
    if ((i == 1) && (j == 1)) return 0.5;
    if ((i == 1) && (j == 0)) return -0.5;

    if (j < i) return 0.0;
    if (j == i+1) return 0.0;
    if (j > i+2) return 0.0;

    const Real c = (1.0-j)/(2.0*(2.0*j-1.0));
    if (i == j) return c;
    if (i == j-2) return -c;

    // should not be reached!
    conceptsAssert(1 == 0, concepts::Assertion());
    return 0.0;
  }

  template<typename F>
  bool RefineOrRaise<F>::regression_(concepts::Array<F>& coeff,
				     const uint p) const {
    Real* cTmp = (Real*)coeff;
    Real sum = 0.0, isum = 0.0;
    const uint n = coeff.size();
    for (uint i = 0; i < n; ++i, ++cTmp) {
      const Real s = std::log(std::fabs(*cTmp));
      sum += s;
      isum += i*s;
    }
    Real m = 3.0*(isum - sum/2.0*p)/((p+1)*p*(p+2));
    DEBUGL(RefineRaise3DRegression_D, "m = " << m
	   << ", e^m = " << std::exp(m));

    if (std::exp(m) < theta_) {
      // p element
      DEBUGL(RefineRaise3DRegression_D, "p element");
      return true;
    } else {
      // h element
      DEBUGL(RefineRaise3DRegression_D, "h element");
      return false;
    }
  }

#ifdef __GNUC__
  template class RefineOrRaise<Real>;
#endif

} // namespace hp3D
