/* hp extension for 3D FEM: ie. determine wether to refine or raise p
 * in an element
 */

#ifndef hpExtension3D_hh
#define hpExtension3D_hh

#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "space/postProcess.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class Vector;

  template<typename F>
  class Array;
}

namespace hp3D {
  using concepts::Real;

  // ******************************************************** RefineOrRaise  **

  /** Refines element or raises its polynomial degree. The decision is based
      on a log linear regression to find out if the Legendre coefficients
      of the solution decrease exponentially.

      The parameter \c theta is used to control this decision. If \f$e^m\f$
      is less than \c theta, the element's polynomial degree is raised,
      otherwise it is refined. m is the slope of the fitted line in
      the regression. This decision is taken anisotropic!

      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class RefineOrRaise : public concepts::CellPostprocess<F> {
  public:
    /** Constructor.
	@param solution Solution vector
	@param theta Decision parameter
	@param spc Space (is used for the refinements)
    */
    RefineOrRaise(const concepts::Vector<F>& solution, const Real theta,
		  concepts::AdaptiveSpace<F,
		  concepts::AdaptiveAdjustP<3> >& spc);
    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature) {}
  private:
    /// Elementwise coefficients of the shape functions
    concepts::Array<F> coeff_;
    /// Legendre coefficients
    concepts::Array<F> legendre_;
    /// Temporary space for the log linear regression
    concepts::Array<F> regr_;
    /// Vector of the variable to plot
    const concepts::Vector<F>& solution_;
    /// Space
    concepts::AdaptiveSpace<F, concepts::AdaptiveAdjustP<3> >& spc_;
    /// Decision parameter
    const Real theta_;
    /** Returns the ith coefficient of the jth shape function in a
	Legendre expansion.
	\f[ a_{ij} = \frac{2i+1}{2} \int_{-1}^1 \psi_j(x) L_i(x) \, dx \f]
    */
    Real transCoeff_(const uint i, const uint j) const;
    /** Makes the regression test.
	The coefficients in \c coeff are one dimensional, ie. 3D coefficients
	are collapsed to 1D using
	\f[ |\hat a_i| = \left( \sum_{j=0}^{n_y} \sum_{k=0}^{n_z} |a_{ijk}|^2
	    \frac{2}{2j+1} \frac{2}{2k+1} \right)^{1/2} \f]
	before calling this function.
	@param coeff 1D coefficients, will be overwritten
	@param p Polynomial degree
	@return true if the coefficients seem to decrease exponentially, ie.
	             increasing p is recommended
    */
    bool regression_(concepts::Array<F>& coeff, const uint p) const;
  };

} // namespace hp3D

#endif // hpExtension3D_hh
