/* weighted functions for regularization, which takes
   the vertex or edge with shortest distance
*/

#if __GNUC__ == 2
#  include <float.h>
#  define MAX DBL_MAX
#else
#  include <limits>
#  define MAX std::numeric_limits<double>::max()
#endif
#include "shortestDist.hh"
#include "singularSet.hh"
#include "hexahedron.hh"
#include "basics/vectorsMatrices.hh"

// debugging
#include "basics/debug.hh"

#define hpShortestDistAppl_D 0

namespace hp3D {

  using concepts::Real;

  std::ostream& operator<<(std::ostream& os, const TrivialWeight& p) {
    return os << "TrivialWeight";
  }

  std::ostream& operator<<(std::ostream& os, const PostprocessSquare& p) {
    return os << "PostprocessSquare";
  }

  std::ostream& operator<<(std::ostream& os, const PostprocessRoot& p) {
    return os << "PostprocessRoot";
  }

  std::ostream& operator<<(std::ostream& os, const PostprocessRoot4& p) {
    return os << "PostprocessRoot4";
  }

  std::ostream& operator<<(std::ostream& os, const Postprocess34& p) {
    return os << "Postprocess34";
  }

  std::ostream& operator<<(std::ostream& os, const Postprocess14& p) {
    return os << "Postprocess14";
  }

  std::ostream& operator<<(std::ostream& os, const Postprocess4& p) {
    return os << "Postprocess4";
  }

  std::ostream& operator<<(std::ostream& os, const Postprocess3& p) {
    return os << "Postprocess3";
  }

  std::ostream& operator<<(std::ostream& os, const ShortestDistLimited& p) {
    return os << "ShortestDistLimited";
  }

  // ********************************************************* ShortestDist **

  Real ShortestDist::operator()(const SingularSet& singularities,
				concepts::Real3d point,
				const Hexahedron* elm) const {
    DEBUGL(hpShortestDistAppl_D, "elm = " << *elm);

    Real shortest = MAX;
    uint anz = 0; // number of singularities

    // scanner over all singular vertices
    {
      std::auto_ptr<concepts::Scan<SingularVertex> > sc
	(singularities.verticesScan());
      while (*sc) {
	SingularVertex& singVtx = (*sc)++;
	Real dist = singVtx.distance(point,elm);
	DEBUGL(hpShortestDistAppl_D, "dist = " << dist << " from " << point
	       << " to " << singVtx);
	if (dist < shortest) shortest = dist;
	++anz;
      }
    }
    // scanner over all singular edges
    {
      std::auto_ptr<concepts::Scan<SingularEdge> > sc
	(singularities.edgesScan());
      while (*sc) {
	SingularEdge& singEdge = (*sc)++;
	Real dist = singEdge.distance(point,elm);
	DEBUGL(hpShortestDistAppl_D, "dist = " << dist << " from " << point
	       << " to " << singEdge);
	if (dist < shortest) shortest = dist;
	++anz;
      }
    }
    // without marked singular edges or vertices is squared distance 1.0
    if (anz==0) shortest = 1.0;

    DEBUGL(hpShortestDistAppl_D, "shortest = " << shortest);

    return shortest;
  }

  std::ostream& operator<<(std::ostream& os, const ShortestDist& p) {
    return os << "ShortestDist";
  }

  // *********************************************************** DaugeWeight **

  /* Fortran code:
      XX= POINT(1)
      YY= POINT(2)
      ZZ= POINT(3)
      R1=SQRT (ZZ*ZZ+YY*YY)
      R2=SQRT (ZZ*ZZ+XX*XX)
      R3=SQRT (XX*XX+YY*YY)
      IF (XX+YY+ZZ.GE.0.) THEN
         R=SQRT (XX*XX+YY*YY+ZZ*ZZ)
      ELSE
         R=MIN(MIN(R1,R2),R3)
      ENDIF
      IF (R.GE.1.E-12) THEN
         RHO=S*EXP(ALPHA*LOG(R))
      ELSE
         RHO=0.
      ENDIF
  */
  Real DaugeWeight::operator()(const SingularSet& singularities,
			       concepts::Real3d point, const Hexahedron* elm)
    const {
    const concepts::Real3d x = elm->cell().chi(point[0], point[1], point[2]);
    Real R = 0;
    if (x[0] + x[1] + x[2] >= 0) {
      R = x.l2_2();
    } else {
      const Real R1 = x[2]*x[2] + x[1]*x[1],
	R2 = x[2]*x[2] + x[0]*x[0],
	R3 = x[0]*x[0] + x[1]*x[1];
      R = (R1 > R2) ? R2 : R1;
      R = (R > R3) ? R3 : R;
    }
    return R;
  }

  std::ostream& operator<<(std::ostream& os, const DaugeWeight& p) {
    return os << "DaugeWeight";
  }

  // ********************************************************** ProductOfAll **

  Real ProductOfAll::operator()(const SingularSet& singularities,
				concepts::Real3d point,
				const Hexahedron* elm) const {
    Real dist = 1.0;

    // scanner over all singular vertices
    {
      std::auto_ptr<concepts::Scan<SingularVertex> > sc
	(singularities.verticesScan());
      DEBUGL(hpShortestDistAppl_D, "elm = " << *elm);
      while (*sc) {
	const SingularVertex& singVtx = (*sc)++;
	dist *= singVtx.distance(point, elm);
      }
    }
    if (dist == 0) return 0;
    // scanner over all singular edges
    {
      std::auto_ptr<concepts::Scan<SingularEdge> > sc
	(singularities.edgesScan());
      while (*sc) {
	const SingularEdge& singEdge = (*sc)++;
	Real rho = singEdge.distance(point, elm);
	if (rho == 0) return 0;
	rho /= singEdge.distanceV0(point, elm);
	rho /= singEdge.distanceV1(point, elm);
	dist *= rho;
      }
    }

    return dist;
  }

  std::ostream& operator<<(std::ostream& os, const ProductOfAll& p) {
    return os << "ProductOfAll";
  }

} // namespace hp3D
