#include "singularSet.hh"
#include "hexahedron.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/exceptions.hh"
#include "toolbox/dynArray.hh"

// debugging
#include "basics/debug.hh"

#define hpSingularSetAppl_D 0
#define hpSingularSetApplArray_D 0
#define hpSingularSetApplLists_D 0
#define hpSingularEdgeAppl_D 0

// distance formula
#define SingularDistanceExact 0
#define SingularDistanceSecond 1

#define SingularDistanceFormula SingularDistanceExact

namespace hp3D {

  using concepts::Real;

  // ********************************************************** SingularEdge **

  std::ostream& operator<<(std::ostream& os, const SingularEdge& s) {
    return os << "SingularEdge(Edge = " << *(s.edge_) << ", vtx0 = "
	      << s.vtx0_ << ", vtx1 = " << s.vtx1_ << ")";
  }

  Real SingularEdge::distance(const concepts::Real3d point,
			      const Hexahedron* elm) const {
    if (SingularDistanceFormula==SingularDistanceExact){
      Real t = elm->chi(point[0],point[1],point[2]) * diffVectorNormed_
	- diffVectorTimesVtx0Normed_;
      DEBUGL(hpSingularEdgeAppl_D, "diffVectorNormed = " << diffVectorNormed_
	     << ", diffVectorTimesVtx0Normed = "
	     << diffVectorTimesVtx0Normed_);
      DEBUGL(hpSingularEdgeAppl_D, "point = " << point << ", t = " << t);
      if (t<0) {
	// first point is closest
	DEBUGL(hpSingularEdgeAppl_D, "closest point is first corner = "
	       << vtx0_);
	return distanceV0(point, elm);
      }
      else if (t>1) {
	// second point is closest
	DEBUGL(hpSingularEdgeAppl_D, "closest point is second corner = "
	       << vtx1_);
	return distanceV1(point, elm);
      }
      else {
	// closest point is on the edge, so calculate it
	concepts::Real3d vtx = diffVector_;
	vtx *= t;
	vtx += vtx0_;
	// distance from that point on the edge to the point in the element
	SingularVertex svtx(0, vtx);
	DEBUGL(hpSingularEdgeAppl_D, "closest point on the edge = " << vtx);
	return svtx.distance(point, elm);
      }
    }
    else
      throw conceptsException(concepts::MissingFeature
			      ("No known distance formula is requested."));
  }

  Real SingularEdge::distanceV0(const concepts::Real3d point,
				const Hexahedron* elm) const {
    SingularVertex svtx(0, vtx0_);
    return svtx.distance(point, elm);
  }

  Real SingularEdge::distanceV1(const concepts::Real3d point,
				const Hexahedron* elm) const {
    SingularVertex svtx(0, vtx1_);
    return svtx.distance(point, elm);
  }

  // ******************************************************** SingularVertex **

  std::ostream& operator<<(std::ostream& os, const SingularVertex& s) {
    return os << "SingularVertex(Vertex = " << *(s.vertex_)
	      << ", vtx0 = " << s.vtx0_ << ")";
  }

  // Returns the distance from the vertex to a arbitrary point
  Real SingularVertex::distance(concepts::Real3d point) const {
    if (SingularDistanceFormula==SingularDistanceExact){
      concepts::Real3d diffVector = point - vtx0_;
      return diffVector.l2_2();
    }
    else
      throw conceptsException(concepts::MissingFeature
			      ("No known distance formula is requested."));
  }

  // Returns the distance from the vertex to a point inside an Hexaedron
  // (point in reference coordinates)
  Real SingularVertex::distance(concepts::Real3d point,
				const Hexahedron* elm) const {
    return distance(elm->chi(point[0],point[1],point[2]));
  }

  // *********************************************************** SingularSet **

  std::ostream& operator<<(std::ostream& os, const SingularSet& s) {
    return os << "SingularSet()";
  }

  SingularSet::~SingularSet() {
    concepts::Joiner<SingularEdge*, 1>::destructor(edges_); 
    concepts::Joiner<SingularVertex*, 1>::destructor(vertices_); 
  }

  void SingularSet::add(const concepts::Attribute& attrib,
			concepts::Mesh3& msh) {

    // arrays for managing the marking process
    // coding for edges:      0 - already checked, 1 - not yet checked
    // coding for vertices: , 0 - checked, not to mark, 1 - not yet checked,
    //                        2 - checked, to mark
    concepts::DynArray<uint> singVertMap(2, 3, 1);
    concepts::DynArray<uint> singEdgeMap(2, 3, 1);

    // scanner over all already marked vertices or edges
    std::auto_ptr<concepts::Scan<SingularVertex> > scv(verticesScan());
    std::auto_ptr<concepts::Scan<SingularEdge> > sce(edgesScan());

    // collect already marked vertices and edges in the arrays
    while (*scv) {
      const SingularVertex& singVtx = (*scv)++;
      singVertMap[singVtx.vertex()->key()] = 0;
    }
    while (*sce) {
      SingularEdge& singEdge = (*sce)++;
      const concepts::Connector1* edge;
      const concepts::Connector0* vtx;
      singEdgeMap[(edge = singEdge.edge())->key()] = 0;
      for (uint l = 0; (vtx = edge->vertex(l)) != 0; ++l) {
	// the vertices of that edge are already marked as singular
	singVertMap[vtx->key()] = 0;
      }
    }

    DEBUGL(hpSingularSetApplArray_D, "VertMap = " << singVertMap);
    DEBUGL(hpSingularSetApplArray_D, "EdgeMap = " << singEdgeMap);

    // There are two loops over all elements.
    // The first looks for all edges and mark that which right attribute
    // as singular and mark the associated vertices in the array singVerMap
    // with flag 0, so that they won't be marked for themselves as singular.
    // In the first loop all vertices with right attribute which are not
    // associated with a proper edge are flagged with 2.
    // The two loops are necessary to avoid that a vertice, which is
    // associated with a edge of another element is marked explicitly singular.
    // The second loop looks for all vertices which are flagged with 2
    // and mark them as singular, that means attach it to the list
    std::auto_ptr<concepts::Scan3> sc(0);
    sc.reset(msh.scan());
    while (*sc) {
	concepts::Cell3& cell = (*sc)++;
	concepts::Connector1* edge;
	for (uint j = 0; (edge = cell.connector().edge(j)) != 0; ++j) {
	  concepts::Connector0* vtx;
	  // edge already checked?
	  if (singEdgeMap[edge->key()]) {
	    singEdgeMap[edge->key()] = 0;
	    // right attribute of the edge?
	    if (edge->attrib() == attrib) {
	      concepts::Real3d coord[2];
	      concepts::Connector0* vtx2;
	      for (uint k = 0; (vtx = edge->vertex(k)) != 0; ++k) {
		// the vertices of that edge are already singular marked
		singVertMap[vtx->key()] = 0;
		for (uint l = 0; (vtx2 = cell.connector().vertex(l))!=0; ++l)
		  if (vtx->key()==vtx2->key()){
		    // coordinates
		    coord[k] = cell.vertex(l);
		  }
	      }
	      // add Edge in the list of singular Edges
	      edges_ = new concepts::Joiner<SingularEdge*, 1>
		(new SingularEdge(edge, coord[0], coord[1]), edges_);
	      DEBUGL(hpSingularSetAppl_D, "singular marked: Edge("
		     << edge->key() << ", (" << *(edge->vertex(0)) << "-"
		     << coord[0] << ", " << *(edge->vertex(1)) << "-"
		     << coord[1] << ")");
	    }
	    else {
	      // edge is not to mark as singular. Are vertices to mark?
	      for (uint k = 0; (vtx = edge->vertex(k)) != 0; ++k) {
		// if not marked as part of an edge then mark singular vertices
		// vertex already checked?
		if (singVertMap[vtx->key()]==1) {
		  // right attribute of the vertex?
		  if (vtx->attrib() == attrib)
		    // vertex is to mark
		    singVertMap[vtx->key()] = 2;
		  else
		    // not mark as singular
		    singVertMap[vtx->key()] = 0;
		}
	      }
	    }
	  }
	}	
    }
    DEBUGL(hpSingularSetApplArray_D, "VertMap = " << singVertMap);
    DEBUGL(hpSingularSetApplArray_D, "EdgeMap = " << singEdgeMap);

    // mark vertices
    sc.reset(msh.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      concepts::Connector0* vtx;
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j)
	if (singVertMap[vtx->key()]==2) {
	  // add Vertex in the list of singular Vertices
	  vertices_ = new concepts::Joiner<SingularVertex*, 1>
		(new SingularVertex(vtx, cell.vertex(j)), vertices_);
	  DEBUGL(hpSingularSetAppl_D, "singular marked: " << *vtx << "-"
		 << cell.vertex(j));
	  singVertMap[vtx->key()] = 0;
      }
    }

    DEBUGL(hpSingularSetApplArray_D, "VertMap = " << singVertMap);
    DEBUGL(hpSingularSetApplArray_D, "EdgeMap = " << singEdgeMap);

#if hpSingularSetApplLists_D
    DEBUGL(hpSingularSetApplLists_D, "singular vertices:");
    {
      std::auto_ptr<concepts::Scan<SingularVertex> > sc
	(verticesScan());
      while (*sc) {
	const SingularVertex& singVtx = (*sc)++;
	DEBUGL(hpSingularSetApplLists_D, singVtx);
      }
    }
    DEBUGL(hpSingularSetApplLists_D, "singular edges:");
    {
      std::auto_ptr<concepts::Scan<SingularEdge> > sc
	(edgesScan());
      while (*sc) {
	const SingularEdge& singEdg = (*sc)++;
	DEBUGL(hpSingularSetApplLists_D, singEdg);
      }
    }
#endif
  }

} // namespace hp3D
