/* space class for 3D hp FEM
 */

#include <memory>
#include <sstream>
#include <algorithm>
#include "hp3D/space.hh"
#include "hp3D/element.hh"
#include "hp3D/hexahedron.hh"
#include "hp3D/buildVtxDofs.hh"
#include "hp3D/buildEdgeDofs.hh"
#include "hp3D/buildFaceDofs.hh"
#include "hp3D/buildInnerDofs.hh"
#include "basics/exceptions.hh"
#include "basics/Zm.hh"
#include "toolbox/boundary.hh"
#include "space/tmatrix.hh"
#include "space/smatrix.hh"
#include "geometry/quadsubdiv.hh"

// timings
#define matrixTiming

#ifdef matrixTiming
#include "toolbox/inputOutput.hh"
#include "toolbox/resourceMonitor.hh"
#endif // matrixTiming

// debugging
#include "basics/debug.hh"

#define hpSpace3DConstr_D 0
#define hpSpace3DDestr_D 0
#define hpSpace3DAdj_D 0
#define hpSpace3DRebuild_D 0
#define hpSpace3DrecomputeSmatrices_D 0

namespace hp3D {

  std::ostream& operator<<
    (std::ostream& os,
     const __gnu_cxx::hash_map<uint, concepts::AdaptiveAdjustP<3> >& adj) {
    os << "Adjustments(" << adj.size() << ": " << std::endl;
    for (__gnu_cxx::hash_map<uint,
	   concepts::AdaptiveAdjustP<3> >::const_iterator
	   i = adj.begin(); i != adj.end(); ++i)
      os << "  " << i->first << " - " << i->second << std::endl;
    return os << ')';
  }

  // ***************************************************************** Space **

  Space::Space(concepts::Mesh3& msh, uint l, uint p,
	       concepts::BoundaryConditions* bc) :
    msh_(msh), bc_(bc), rebuild_(true), dim_(0),
    nelm_(0), elm_(0), endElm_(0), S1left_(0), S1right_(0),
    buildVertexDofs_(new BuildVertexDofs(*this)),
    buildEdgeDofs_(new BuildEdgeDofs(*this)),
    buildFaceDofs_(new BuildFaceDofsLinTrunk(*this)),
    buildInnerDofs_(new BuildInnerDofsLinTrunk(*this)) {
    std::auto_ptr<concepts::Scan3> sc(msh_.scan());
    while (*sc) {
      concepts::Key key(((*sc)++).connector().key());
      ctrl3_[key].tag_.data_ = concepts::AdaptiveControlTag::MEMBER;
      adj_[key].level_ = concepts::Level<3>(l);
      for (uint i = 0; i < 3; ++i)
	adj_[key].p_[i] = p;
    }
    DEBUGL(hpSpace3DConstr_D, "done.");
  }

  Space::Space(const Space& spc) :
    msh_(spc.msh_), bc_(spc.bc_), rebuild_((spc.elm_ == 0) ? true : false),
    dim_(spc.dim_), nelm_(spc.nelm_), elm_(spc.elm_), endElm_(0),
    ctrl1_(spc.ctrl1_.get()), ctrl2_(spc.ctrl2_.get()),
    ctrl3_(spc.ctrl3_), S1left_(0), S1right_(0),
    buildVertexDofs_(spc.buildVertexDofs_->clone(this)),
    buildEdgeDofs_(spc.buildEdgeDofs_->clone(this)),
    buildFaceDofs_(spc.buildFaceDofs_->clone(this)),
    buildInnerDofs_(spc.buildInnerDofs_->clone(this)) {
    Space* space = const_cast<Space*>(&spc);
    space->elm_ = 0;
    space->ctrl1_.release();
    space->ctrl2_.release();
    conceptsAssert(this == buildVertexDofs_->space(),
		   concepts::Assertion());
    conceptsAssert(this == buildEdgeDofs_->space(), concepts::Assertion());
    conceptsAssert(this == buildFaceDofs_->space(), concepts::Assertion());
    conceptsAssert(this == buildInnerDofs_->space(), concepts::Assertion());
    DEBUGL(hpSpace3DConstr_D, "done.");
  }

  Space::~Space() {
    DEBUGL(hpSpace3DDestr_D, "elm = " << elm_);
    concepts::Joiner<Element<Real>*, 1>::destructor(elm_);
    DEBUGL(hpSpace3DDestr_D, "done.");
  }

  void Space::adjust(const concepts::Element<Real>& elm,
		     const concepts::AdaptiveAdjustP<3>& a) {
    DEBUGL(hpSpace3DAdj_D, "start. elm = " << elm << ", a = " << a);
    const Element<Real>* e = dynamic_cast<const Element<Real>*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    adj_[e->support().key()] = a;
    rebuild_ = true;
    DEBUGL(hpSpace3DAdj_D, "done.");
  }

  Space Space::rebuild() {
    DEBUGL(hpSpace3DRebuild_D, "start.");

    // prepare result
    Space result(*this);

    // remove old data
    ctrl1_.reset
      (new __gnu_cxx::hash_map<uint, concepts::AdaptiveControlP<1> >());
    ctrl2_.reset
      (new __gnu_cxx::hash_map<uint, concepts::AdaptiveControlP<2> >());
    elm_ = 0;
    dim_ = 0;
    nelm_ = 0;

    DEBUGL(hpSpace3DRebuild_D, "adj = " << adj_);

#ifdef matrixTiming
    concepts::ResourceMonitor timer;
#endif // matrixTiming

    // prepare for rebuild of mesh by marking level and polynomial degree
    // sort out boundary conditions (ie. mark DIRICHLET nodes as passive)
    DEBUGL(hpSpace3DRebuild_D, "scanning mesh with meshAndPoly_");
    std::auto_ptr<concepts::Scan3> sc(msh_.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      concepts::Level<3> L(-1);
      int P[3] = { -1, -1, -1 };

      concepts::Hexahedron3d* cellH =
	dynamic_cast<concepts::Hexahedron3d*>(&cell);
      if (cellH) {
 	meshAndPoly_(*cellH, L, P);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("meshAndPoly", timeCntr_, timer.utime());
#endif // matrixTiming

    // create list of cells, edges and so on
    DEBUGL(hpSpace3DRebuild_D, "creating cell list");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      createCellList_(cell.connector(), 0);
    } // loop over mesh
    DEBUGL(hpSpace3DRebuild_D, "registered " << cellList_.size() << " cells");
    DEBUGL(hpSpace3DRebuild_D, "creating vertex, edge and face list");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      createEdgFaceList_(cell);
      createVtxEdgList_(cell);
    } // loop over mesh
    DEBUGL(hpSpace3DRebuild_D, "registered " << faceList_.size() << " faces");
    DEBUGL(hpSpace3DRebuild_D, "registered " << edgeList_.size() << " edges");
    DEBUGL(hpSpace3DRebuild_D, "registered " << vertexList_.size()
	   << " vertices");

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("createTopologyLists", timeCntr_,
			       timer.utime());
#endif // matrixTiming

    concepts::FaceData::checkRelatedFaces(faceList_);
    DEBUGL(hpSpace3DRebuild_D, "updated faceList_");
    for (std::map<uint, concepts::FaceData>::const_iterator i =
	   faceList_.begin(); i != faceList_.end(); ++i) {
      concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[i->first]);
      ctrl2.tag_.validateP();
      ctrl2.tag_.takeIn();
    }
    DEBUGL(hpSpace3DRebuild_D, "updated ctrl2_ according to faceList_");
#ifdef DEBUG
    for (std::map<uint, concepts::FaceData>::const_iterator i =
	   faceList_.begin(); i != faceList_.end(); ++i) {
      DEBUGL(0, i->second);
      concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[i->first]);
      DEBUGL(0, ctrl2);
    }
#endif

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("checkRelatedFaces", timeCntr_, timer.utime());
#endif // matrixTiming

    concepts::EdgeData::checkRelatedEdges(edgeList_);
    bool notChanged = false;
    while (! notChanged) {
      notChanged = true;
      notChanged &= concepts::EdgeData::doCheckRelations(edgeList_);
    }
    DEBUGL(hpSpace3DRebuild_D, "updated edgeList_");
    for (std::map<uint, concepts::EdgeData>::const_iterator i =
	   edgeList_.begin(); i != edgeList_.end(); ++i) {
      concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[i->first]);
      ctrl1.tag_.validateP();
      ctrl1.tag_.takeIn();
    }
    DEBUGL(hpSpace3DRebuild_D, "updated ctrl1_ according to edgeList_");
#ifdef DEBUG
    for (std::map<uint, concepts::EdgeData>::const_iterator i =
	   edgeList_.begin(); i != edgeList_.end(); ++i) {
      DEBUGL(0, i->second);
      concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[i->first]);
      DEBUGL(0, ctrl1);
    }
#endif

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("checkRelatedEdges", timeCntr_, timer.utime());
#endif // matrixTiming

    notChanged = false;
    while (! notChanged) {
      notChanged = true;
      notChanged &= concepts::VertexData::doCheckRelations(vertexList_);
    }
    DEBUGL(hpSpace3DRebuild_D, "updated vertexList_");
#ifdef DEBUG
    for (std::map<uint, concepts::VertexData>::const_iterator i =
	   vertexList_.begin(); i != vertexList_.end(); ++i) {
      DEBUGL(0, i->second);
      concepts::AdaptiveControl& ctrl0(ctrl0_[i->first]);
      DEBUGL(0, ctrl0);
    }
#endif

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("checkRelatedVertices", timeCntr_,
			       timer.utime());
#endif // matrixTiming

    // enrich elements
    DEBUGL(hpSpace3DRebuild_D, "enriching elements with enrichElm_");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;
      int P[27];
      for (uint i = 0; i < 27; ++i) P[i] = -1;

      concepts::Hexahedron3d* cellH =
	dynamic_cast<concepts::Hexahedron3d*>(&cell);
      if (cellH) {
	enrichElm_(*cellH, P);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("enrichElements", timeCntr_, timer.utime());
#endif // matrixTiming

    // rebuild elements
    DEBUGL(hpSpace3DRebuild_D, "rebuilding elements with rebuild1_");
    sc.reset(msh_.scan());
    while (*sc) {
      concepts::Cell3& cell = (*sc)++;

      concepts::Hexahedron3d* cellH =
	dynamic_cast<concepts::Hexahedron3d*>(&cell);
      if (cellH) {
	ushort p[3] = { 0, 0, 0 };
	buildElements_(*cellH, p);
	continue;
      }

      std::stringstream error;
      error << "cell " << cell << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str()));
    }

#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("buildElements", timeCntr_++, timer.utime());
#endif // matrixTiming

    // clear pre rebuild information
    DEBUGL(hpSpace3DRebuild_D, "clear pre rebuild information");
    rebuild_ = false;
    adj_.clear();
    // clear the control information of vertices, edges and faces
    ctrl0_.clear();
    cellList_.clear();
    faceList_.clear();
    edgeList_.clear();
    vertexList_.clear();
    DEBUGL(hpSpace3DRebuild_D, "done.");
    return result;
  }

  const std::auto_ptr<concepts::SMatrixBase<Real> >*
  Space::getSmatrices(const concepts::Hex3dSubdivision* subdiv) const {
    const std::auto_ptr<concepts::SMatrixBase<Real> >* Smatrices = 0;
    if (subdiv == concepts::Hex3dSubdiv8::instance())
      Smatrices = Smatrices8_;
    if (subdiv == concepts::Hex3dSubdiv4x::instance())
      Smatrices = Smatrices4x_;
    if (subdiv == concepts::Hex3dSubdiv4y::instance())
      Smatrices = Smatrices4y_;
    if (subdiv == concepts::Hex3dSubdiv4z::instance())
      Smatrices = Smatrices4z_;
    if (subdiv == concepts::Hex3dSubdiv2x::instance())
      Smatrices = Smatrices2x_;
    if (subdiv == concepts::Hex3dSubdiv2y::instance())
      Smatrices = Smatrices2y_;
    if (subdiv == concepts::Hex3dSubdiv2z::instance())
      Smatrices = Smatrices2z_;
    conceptsAssert(Smatrices != 0, concepts::Assertion());
    return Smatrices;
  }

  concepts::InOutParameters* Space::timings_ = 0;

  uint Space::timeCntr_ = 0;

  void Space::setTimings(concepts::InOutParameters* timings) {
#ifdef matrixTiming
    timings_ = timings;
    if (timings_) {
      timeCntr_ = 0;
      timings_->addArrayDouble("meshAndPoly");
      timings_->addArrayDouble("createTopologyLists");
      timings_->addArrayDouble("checkRelatedFaces");
      timings_->addArrayDouble("checkRelatedEdges");
      timings_->addArrayDouble("checkRelatedVertices");
      timings_->addArrayDouble("enrichElements");
      timings_->addArrayDouble("buildElements");
    }
#endif // matrixTiming
  }

  bool Space::timings() {
#ifdef matrixTiming
    return true;
#else
    return false;
#endif // matrixTiming
  }

  void Space::buildVertexDofs(const BuildDofsBase& b) {
    buildVertexDofs_.reset(b.clone(this));
  }

  void Space::buildEdgeDofs(const BuildDofsBase& b) {
    buildEdgeDofs_.reset(b.clone(this));
  }

  void Space::buildFaceDofs(const BuildDofsBase& b) {
    buildFaceDofs_.reset(b.clone(this));
  }

  void Space::buildInnerDofs(const BuildDofsBase& b) {
    buildInnerDofs_.reset(b.clone(this));
  }

  std::ostream& Space::info(std::ostream& os) const {
    if (rebuild_) throw conceptsException(concepts::SpaceNotBuilt());
    return os << "Space(dim = " << dim_ << ", nelm = " << nelm_ << ')';
  }

  void Space::recomputeSmatrices(const Hexahedron* elm) const {
    bool recompute = false;
    if (S1left_.get() == 0) recompute = true;
    conceptsAssert(recompute == (S1right_.get() == 0), concepts::Assertion());

    // find highest polynomial degree
    const ushort* p = elm->p();
    uint imax = 0;
    for (uint i = 1; i < 3; ++i) if (p[i] > p[imax]) imax = i;
    // loop over the children of the cell
    const concepts::Hexahedron3d* childL = 0; // cell with the largest p
    const concepts::Hexahedron3d* child = 0;
    for (uint i = 0; (child = elm->cell().child(i)) != 0; ++i) {
      const concepts::AdaptiveControlP<3>&
	ctrli(ctrl3_.find(child->connector().key())->second);
      for (uint i = 1; i < 3; ++i)
	if (ctrli.p_[i] > p[imax]) {
	  imax = i;
	  p = ctrli.p_;
	  childL = child;
	}
    }
    // if the highest polynomial degree is in a child, create a temporary
    // element
    std::auto_ptr<const Hexahedron> elmTmp(0); // used for temporary elements
    if (childL) {
      concepts::Hexahedron3d* childLc =
	const_cast<concepts::Hexahedron3d*>(childL);
      elmTmp.reset(elm = new Hexahedron(*childLc, p, 0, 0));
    }

    if ((!recompute) && ((int)S1left_->size() < p[imax]+1)) recompute = true;
    DEBUGL(hpSpace3DrecomputeSmatrices_D, "recompute = " << recompute
	   << " for p = " << p[imax]);
    if (recompute) {
      // get the shape functions in question ...
      const concepts::Karniadakis<1,0>* small = 0;
      switch(imax) {
      case 0: small = elm->shpfctX(); break;
      case 1: small = elm->shpfctY(); break;
      default: small = elm->shpfctZ();
      }
      DEBUGL(hpSpace3DrecomputeSmatrices_D, "small = " << *small);
      // ... and the respective points
      const Real* points = 0;
      uint n = 0;
      switch(imax) {
      case 0: n = elm->integrationX()->n();
	points = elm->integrationX()->abscissas(); break;
      case 1: n = elm->integrationY()->n();
	points = elm->integrationY()->abscissas(); break;
      default: n = elm->integrationZ()->n();
	points = elm->integrationZ()->abscissas();
      }
      DEBUGL(hpSpace3DrecomputeSmatrices_D, n << " points = " << *points);
      concepts::Array<Real> leftP(n), rightP(n);
      for (uint i = 0; i < n; ++i) {
	leftP[i] = (points[i]-1.0)/2.0;
	rightP[i] = (points[i]+1.0)/2.0;
      }
      concepts::Karniadakis<1,0>
	left(p[imax], (Real*)leftP, n, 0, 0, false),
	right(p[imax], (Real*)rightP, n, 0, 0, false);
      DEBUGL(hpSpace3DrecomputeSmatrices_D, "left = " << left);
      DEBUGL(hpSpace3DrecomputeSmatrices_D, "right = " << right);
      S1left_.reset(new concepts::SMatrix1D(*small, left, right));
      S1right_.reset(new concepts::SMatrix1D(*S1left_, true));
      DEBUGL(hpSpace3DrecomputeSmatrices_D, "S1left = " << *S1left_);
      DEBUGL(hpSpace3DrecomputeSmatrices_D, "S1right = " << *S1right_);

      Smatrices2x_[0].reset
	(new concepts::SMatrixTensor<Real, 3>(*S1left_, 1));
      Smatrices2x_[1].reset
	(new concepts::SMatrixTensor<Real, 3>(*S1right_, 1));
      Smatrices2y_[0].reset
	(new concepts::SMatrixTensor<Real, 3>(*S1left_, 2));
      Smatrices2y_[1].reset
	(new concepts::SMatrixTensor<Real, 3>(*S1right_, 2));
      Smatrices2z_[0].reset
	(new concepts::SMatrixTensor<Real, 3>(*S1left_, 3));
      Smatrices2z_[1].reset
	(new concepts::SMatrixTensor<Real, 3>(*S1right_, 3));

      Smatrices8_[0].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2y_[0]),
					    Smatrices2z_[0].get()));
      Smatrices8_[1].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2y_[0]),
					    Smatrices2z_[0].get()));
      Smatrices8_[2].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2y_[1]),
					    Smatrices2z_[0].get()));
      Smatrices8_[3].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2y_[1]),
					    Smatrices2z_[0].get()));
      Smatrices8_[4].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2y_[0]),
					    Smatrices2z_[1].get()));
      Smatrices8_[5].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2y_[0]),
					    Smatrices2z_[1].get()));
      Smatrices8_[6].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2y_[1]),
					    Smatrices2z_[1].get()));
      Smatrices8_[7].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2y_[1]),
					    Smatrices2z_[1].get()));

      Smatrices4x_[0].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2y_[0]),
					    *(Smatrices2z_[0])));
      Smatrices4x_[1].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2y_[1]),
					    *(Smatrices2z_[0])));
      Smatrices4x_[2].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2y_[1]),
					    *(Smatrices2z_[1])));
      Smatrices4x_[3].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2y_[0]),
					    *(Smatrices2z_[1])));

      Smatrices4y_[0].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2z_[0])));
      Smatrices4y_[1].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2z_[0])));
      Smatrices4y_[2].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2z_[1])));
      Smatrices4y_[3].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2z_[1])));

      Smatrices4z_[0].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2y_[0])));
      Smatrices4z_[1].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2y_[0])));
      Smatrices4z_[2].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[1]),
					    *(Smatrices2y_[1])));
      Smatrices4z_[3].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2x_[0]),
					    *(Smatrices2y_[1])));
    }
  }

  void Space::deactivate_(const concepts::Connector2& face) {
// #ifdef DEBUG
//     uint d = 0;
//     if (face.key().key() == 0) d = 1;
// #endif
    const concepts::Connector2* child = 0;
    for (uint i = 0; (child = face.child(i)) != 0; ++i) {
      (*ctrl2_)[child->key()].tag_.deactivate();
      DEBUGL(0, "deactivating " << *child);
      deactivate_(*child);
    }
    const concepts::Quad* faceQ = dynamic_cast<const concepts::Quad*>(&face);
    if (faceQ) {
      const concepts::Quad* testChild = faceQ->child(0);
      const concepts::QuadSubdivision* strategy = faceQ->getStrategy();
      if (strategy && testChild) {
	const concepts::Connector1* edge = 0;
	for (uint i = 0; (edge = strategy->newEdge(*faceQ, i)) != 0; ++i) {
	  (*ctrl1_)[edge->key()].tag_.deactivate();
	  DEBUGL(0, "deactivating " << *edge);
	  deactivate_(*edge);
	}
	const concepts::Connector0* vtx = 0;
	for (uint i = 0; (vtx = strategy->newVertex(*faceQ, i)) != 0; ++i) {
	  ctrl0_[vtx->key()].tag_.deactivate();
	  DEBUGL(0, "deactivating " << *vtx);
	}
      }
      return;
    }
    const concepts::Triangle* faceT =
      dynamic_cast<const concepts::Triangle*>(&face);
    if (faceT) {
    }
    throw conceptsException(concepts::MissingFeature("face not supported"));
  }

  void Space::deactivate_(const concepts::Connector1& edg) {
// #ifdef DEBUG
//     uint d = 0;
//     if (edg.key().key() == 30) d = 1;
// #endif
    const concepts::Connector1* edge = 0;
    for (uint i = 0; (edge = edg.child(i)) != 0; ++i) {
      (*ctrl1_)[edge->key()].tag_.deactivate();
      DEBUGL(0, "deactivating " << *edge);
    }
    if ((edge = edg.child(1)) != 0) {
      ctrl0_[edge->vertex(0)->key()].tag_.deactivate();
      DEBUGL(0, "deactivating " << *(edge->vertex(0)));
    }
  }

  void Space::deactivate_(const concepts::Connector0&) {}

} // namespace hp3d
