/* space class for 3D hp FEM
 * rebuild0_
 */

#include "hexFunctions.hh"
#include "hp3D/space.hh"
#include "basics/debug.hh"

#define hpSpace3DmeshAndPoly_D 0

namespace hp3D {

  // ***************************************************************** Space **

  void Space::meshAndPoly_(concepts::Hexahedron3d& cell,
			   concepts::Level<3>& L, int* P) {
    concepts::Connector3& cntr = cell.connector();
    concepts::AdaptiveControlP<3>& ctrl3(ctrl3_[cntr.key()]);
    const concepts::Level<3>& l(cell.level());
    DEBUGL(hpSpace3DmeshAndPoly_D, l << " - start: " << cntr << ", l = " << l
	   << ", L = " << L
	   << ", P = (" << P[0] << ", " << P[1] << ", " << P[2] << ")");

    concepts::Hexahedron* cntrH = dynamic_cast<concepts::Hexahedron*>(&cntr);
    // There are different cases in the following code for different elements.
    // Right now, we only have hex available -> check here.
    conceptsAssert(cntrH, concepts::Assertion());
    DEBUGL(hpSpace3DmeshAndPoly_D, l << " - " << ctrl3 << ", L = " << L);

    /* If L == -1 and the connector is member of the space: this cell was
       in the old space. Apply adjustments here!
    */
    if (L == concepts::Level<3>(-1) && ctrl3.tag_.isMember()) {
      concepts::AdaptiveAdjustP<3>& adj(adj_[cntr.key()]);
      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - applying " << adj);
      L = cell.level() + adj.level_;
      L.maximize(concepts::Level<3>(0));
      for (uint i = 0; i < 3; ++i) {
	ctrl3.p_[i] += adj.p_[i]; // p int
	if (ctrl3.p_[i] < 1) ctrl3.p_[i] = 1;
	P[i] = ctrl3.p_[i];
      }
    }
    DEBUGL(hpSpace3DmeshAndPoly_D, l << " - updated p and l: "
	   << ctrl3 << ", L = " << L);

    /* If the desired level L == -1 or we are not on the desired level,
       this cell is not in the new space. Therefore:
       - set refinement strategy and test if it works
    */
    if (L == concepts::Level<3>(-1) || l < L) {
      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - not on desired level");

      doCorrectRefinement_(cell, L);
    }

    /* If the desired level L == -1 or we are not on the desired level,
       this cell is not in the new space. Therefore:
       - call meshAndPoly_ for all children of the cell
       - get max p for all faces from the children of the faces
       - get max p for all edges from the children of the edges
       - get max p for the cell from the children of the cell
    */
    if (L == concepts::Level<3>(-1) || l < L) {
      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - not on desired level");

      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - calling children");
      // call meshAndPoly for all children
//       for (uint i = 0; i < 3; ++i) P[i] = ctrl3.p_[i];
      concepts::Level<3> Lmax(-1);
      concepts::Hexahedron3d* chld = 0;
      for(uint i = 0; (chld = cell.child(i)) != 0; ++i) {
	concepts::Level<3> LL(L);
	int PP[3] = { P[0], P[1], P[2] };

	meshAndPoly_(*chld, LL, PP);

	for (uint i = 0; i < 3; ++i) if (P[i] < PP[i]) P[i] = PP[i];
	Lmax.maximize(LL);
      } // for chld
      L = Lmax;

      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - getting p for cell: before "
	     << ctrl3);
      for (uint i = 0; i < 3; ++i) ctrl3.p_[i] = P[i];
      ctrl3.tag_ = concepts::AdaptiveControlTag::VLD_P;
      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - ... and after " << ctrl3);

      /* get p from faces of children and set p on face of this cell
	 to the maximum
      */
      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - getting p for faces");
      const concepts::Connector2* cntr2 = 0;
      for(uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) { // loop over faces
	concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[cntr2->key()]);
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - " << *cntr2
	       << " has " << ctrl2);

	// loop over all children of current cell
	const concepts::Connector3* cntrChild = 0;
	for (uint n = 0; (cntrChild = cntr.child(n)) != 0; ++n) {
	  // get relevant face of the child
	  const concepts::Connector2* faceChild = cntrChild->face(i);
	  // check if face is child of face of the large cell
	  bool takeFace = false;
	  const concepts::Connector2* cntrFaceChild = 0;
	  for (uint m = 0;
	       (!takeFace) && ((cntrFaceChild = cntr2->child(m)) != 0);
	       ++m) {
	    if (cntrFaceChild->key() == faceChild->key())
	      takeFace = true;
	  }

	  if (takeFace) {
	    concepts::AdaptiveControlP<2>&
	      childCtrl2((*ctrl2_)[faceChild->key()]);
// 	    DEBUGL(hpSpace3DmeshAndPoly_D, l << " - has child " << *faceChild
// 		   << " with " << childCtrl2);
	    for (uint k = 0; k < 2; ++k)
	      if (ctrl2.p_[k] < childCtrl2.p_[k])
		ctrl2.p_[k] = childCtrl2.p_[k];
	  }
	} // for children of current cell

	ctrl2.tag_.validateP();
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - and gets " << ctrl2);
      } // for faces

      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - getting p for edges");
      const concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - " << *cntr1
	       << " has " << ctrl1);

	// loop over all children of current cell
	const concepts::Connector3* cntrChild = 0;
	for (uint n = 0; (cntrChild = cntr.child(n)) != 0; ++n) {
	  // get relevant edge of the child
	  const concepts::Connector1* edgeChild = cntrChild->edge(i);
	  // check if edge is child of edge of the large cell
	  bool takeEdge = false;
	  const concepts::Connector1* cntrEdgeChild = 0;
	  for (uint m = 0;
	       (!takeEdge) && ((cntrEdgeChild = cntr1->child(m)) != 0);
	       ++m) {
	    if (cntrEdgeChild->key() == edgeChild->key())
	      takeEdge = true;
	  }

	  if (takeEdge) {
	    concepts::AdaptiveControlP<1>&
	      childCtrl1((*ctrl1_)[edgeChild->key()]);
// 	    DEBUGL(hpSpace3DmeshAndPoly_D, l << " - has child " << *edgeChild
// 		   << " with " << childCtrl1);
	    if (ctrl1.p_[0] < childCtrl1.p_[0]) ctrl1.p_[0] = childCtrl1.p_[0];
	  }
	} // for children of current cell

	ctrl1.tag_.validateP();
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - and gets " << ctrl1);
      } // for edges
    } else {
      /* cell is in the new space:
	 - take degrees of freedom into space
	 - bring polynomial degrees onto faces and edges
      */
      for (uint i = 0; i < 3; ++i) ctrl3.p_[i] = P[i]; // p int
      ctrl3.tag_ = concepts::AdaptiveControlTag::VLD_P;
      ctrl3.tag_.takeIn();
      DEBUGL(hpSpace3DmeshAndPoly_D, l << " - cell has " << ctrl3
	     << ", level = " << cell.level());

      // bring p of the cell onto faces
      concepts::Connector2* cntr2 = 0;
      for(uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) {
	uint pIndex = 0, qIndex = 0, rIndex = 0;
	if (cntrH)
	  HexFunctions::faceOrientation(*cntrH, i, pIndex, qIndex, rIndex);
	concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[cntr2->key()]);
	if (ctrl2.p_[0] < ctrl3.p_[pIndex]) ctrl2.p_[0] = ctrl3.p_[pIndex];
	if (ctrl2.p_[1] < ctrl3.p_[qIndex]) ctrl2.p_[1] = ctrl3.p_[qIndex];
	ctrl2.tag_.validateP();
	ctrl2.tag_.takeIn();
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - " << *cntr2
	       << " has " << ctrl2);
      } // for faces
      // bring p of the cell onto edges
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	uint pIndex = 0, qIndex = 0, rIndex = 0;
	if (cntrH)
	  HexFunctions::edgeOrientation(i, pIndex, qIndex, rIndex);
	concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
	if (ctrl1.p_[0] < ctrl3.p_[pIndex]) ctrl1.p_[0] = ctrl3.p_[pIndex];
	ctrl1.tag_.validateP();
	ctrl1.tag_.takeIn();
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - " << *cntr1
	       << " has " << ctrl1);
      } // for edges

      concepts::Connector0* cntr0 = 0;
      for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
// 	ctrl0_[cntr0->key()].tag_ = concepts::AdaptiveControlTag::MEMBER;
	concepts::AdaptiveControl& ctrl0(ctrl0_[cntr0->key()]);
	ctrl0.tag_.takeIn();
	DEBUGL(hpSpace3DmeshAndPoly_D, l << " - " << *cntr0
	       << " has " << ctrl0);
      }
    }

    /* determine boundary conditions for edges and vertices:
     * if a face has DIRICHLET bc, then all adjaccent edges and vertices
     * have them too -> such a dof is a PASSIVE member of the space
     */
    if (bc_) {
      enforceBC_(cell);
    } // if bc

    DEBUGL(hpSpace3DmeshAndPoly_D, l << " - done.");
  }

} // namespace hp3d
