/* space class for 3D hp FEM
 * buildElements_
 */

#include <memory>
#include "hp3D/space.hh"
#include "hp3D/element.hh"
#include "hp3D/hexahedron.hh"
#include "hexFunctions.hh"

// debugging
#include "basics/debug.hh"

#define hpSpace3DBuildElements_D 0
#define hpSpace3DPmax_D 0
#define hpSpace3Ddof_D 0

// careful with those! (there are dependencies faces -> edges)
#define buildVertexModes 1
#define buildEdgeModes 1
#define buildFaceModes 1
#define buildInnerModes 1

namespace hp3D {

  // ***************************************************************** Space **

  void Space::computePmax_(const concepts::Hexahedron& cntr,
			   ushort Pmax[3]) const {
    const concepts::AdaptiveControlP<3>&
      ctrl3(ctrl3_.find(cntr.key())->second);
    DEBUGL(hpSpace3DPmax_D, "start - " << cntr << ", Pmax = ("
	   << Pmax[0] << ", " << Pmax[1] << ", " << Pmax[2] << ')');

    // start with internal polynomial degree
    for (uint j = 0; j < 3; ++j)
      if (ctrl3.p_[j] > Pmax[j]) Pmax[j] = ctrl3.p_[j];
    DEBUGL(hpSpace3DPmax_D, "int , Pmax = (" << Pmax[0] << ", " << Pmax[1]
	   << ", " << Pmax[2] << ')');

    // get maximum of edges
    const concepts::Edge* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      const concepts::AdaptiveControlP<1>&
	ctrl1(ctrl1_->find(cntr1->key())->second);
      uint pIndex = 0, qIndex = 0, rIndex = 0;
      HexFunctions::edgeOrientation(i, pIndex, qIndex, rIndex);
      if (Pmax[pIndex] < ctrl1.p_[0]) Pmax[pIndex] = ctrl1.p_[0];
    } // for edges (i)
    DEBUGL(hpSpace3DPmax_D, "edge, Pmax = (" << Pmax[0] << ", " << Pmax[1]
	   << ", " << Pmax[2] << ')');

    // get maximum of faces
    const concepts::Connector2* cntr2 = 0;
    for (uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) {
      const concepts::AdaptiveControlP<2>&
	ctrl2(ctrl2_->find(cntr2->key())->second);
      DEBUGL(hpSpace3DPmax_D, "face " << i << " has " << ctrl2);
      uint pIndex = 0, qIndex = 0, rIndex = 0;
      HexFunctions::faceOrientation(cntr, i, pIndex, qIndex, rIndex);
      if (Pmax[pIndex] < ctrl2.p_[0]) Pmax[pIndex] = ctrl2.p_[0];
      if (Pmax[qIndex] < ctrl2.p_[1]) Pmax[qIndex] = ctrl2.p_[1];
      DEBUGL(hpSpace3DPmax_D, "orientation = (" << pIndex << ", " << qIndex
	     << "), Pmax = (" << Pmax[0] << ", " << Pmax[1]
	     << ", " << Pmax[2] << ')');
    } // for faces (i)
    DEBUGL(hpSpace3DPmax_D, "face, Pmax = (" << Pmax[0] << ", " << Pmax[1]
	   << ", " << Pmax[2] << ')');
    DEBUGL(hpSpace3DPmax_D, "done.");
  }

  void Space::buildElements_(concepts::Hexahedron3d& cell,
			     ushort* Pmax,
			     concepts::TColumn<Real>* T0) {
//     uint hpSpace3DBuildElements_D = 0;
//     if ((int)cell.connector().key() == 2)
//       hpSpace3DBuildElements_D = 1;
    DEBUGL(hpSpace3DBuildElements_D, cell.connector().key() << " - dim = "
	   << dim_ << ", start.");

    concepts::Hexahedron& cntr = cell.connector();
    concepts::AdaptiveControlP<3>& ctrl3(ctrl3_[cntr.key()]);
    DEBUGL(hpSpace3Ddof_D, cntr);

    // compute p max in this element
    computePmax_(cntr, Pmax);
    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - p max = ("
	   << Pmax[0] << ", " << Pmax[1] << ", " << Pmax[2] << ')');

    // maximal number of degrees of freedom in this element
    uint nn[3] = { Pmax[0]+1, Pmax[1]+1, Pmax[2]+1 }; // p max
    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - n = "
	   << nn[0]*nn[1]*nn[2] << ", cell tag = " << ctrl3.tag_);
    concepts::TColumn<Real>* T1 = 0;

    Element<Real>* elm = 0;
    Hexahedron* elmH = 0;
    std::auto_ptr<Element<Real> > elmTmp(0); // used for temporary elements
    if (ctrl3.tag_.isMember()) {
      elm = elmH = new Hexahedron(cell, Pmax, T0, T1); // p max

      // connect polynomial degrees with element
      const concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i)
	elmH->edgeP(i, (*ctrl1_)[cntr1->key()]);
      concepts::Connector2* cntr2 = 0;
      for(uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i)
	elmH->faceP(i, (*ctrl2_)[cntr2->key()]);

      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - creating element "
	     << *elm << ", p int = (" << ctrl3.p_[0] << ", " << ctrl3.p_[1]
	     << ", " << ctrl3.p_[2] << ')');
      ++nelm_;
      if (elm_ == 0) {
	// first element in the list
	endElm_ = elm_ =
	  new concepts::Joiner<Element<Real>*, 1>(elm, elm_);
      } else {
	// list already has elements, add to the end of the list
	concepts::Joiner<Element<Real>*, 1>* newElm =
	  new concepts::Joiner<Element<Real>*, 1>(elm, 0);
	(*endElm_)[0] = newElm;
	endElm_ = newElm;
      }
    } else {
      elmTmp.reset(elmH = new Hexahedron(cell, Pmax, T0, T1)); // p max
      elm = elmTmp.get();
      DEBUGL(hpSpace3DBuildElements_D, cntr.key()
	     << " - creating temporary element " << *elm);
    }

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - vertices starting");
    conceptsAssert(this == buildVertexDofs_->space(), concepts::Assertion());
    (*buildVertexDofs_)(cntr, T1);

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - edges starting");
    conceptsAssert(this == buildEdgeDofs_->space(), concepts::Assertion());
    (*buildEdgeDofs_)(cntr, T1);

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - faces starting");
    conceptsAssert(this == buildFaceDofs_->space(), concepts::Assertion());
    (*buildFaceDofs_)(cntr, T1);

    if (ctrl3.tag_.isMember()) {
      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - interior starting");
      conceptsAssert(this == buildInnerDofs_->space(), concepts::Assertion());
      (*buildInnerDofs_)(cntr, T1);

      elm->appendT(T1);
      DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - added T columns to "
	     << *elm);
    } else {
      if (T0 || T1)
	recomputeSmatrices(elmH);
      // check subdivision strategy
      const concepts::Hex3dSubdivision* subdiv = elmH->getStrategy();
      const std::auto_ptr<concepts::SMatrixBase<Real> >* Smatrices =
	getSmatrices(subdiv);
      // loop over the children of the cell
      concepts::Hexahedron3d* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
#ifdef DEBUG
	concepts::AdaptiveControlP<3>& ctrli(ctrl3_[child->connector().key()]);
#endif
	DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - child " << i
	       << " = " << *child << " with " << ctrli);
	// number of degrees of freedom of the child
	ushort Pmaxi[3] = { 0, 0, 0 };
	if ((T0 != 0) || (T1 != 0))
	  for (uint i = 0; i < 3; ++i) Pmaxi[i] = Pmax[i];
	computePmax_(child->connector(), Pmaxi);
	uint nni[3] = { Pmaxi[0]+1, Pmaxi[1]+1, Pmaxi[2]+1 }; // p max
	DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - child has p max = ("
	       << Pmaxi[0] << ", " << Pmaxi[1] << ", " << Pmaxi[2] << ')');

	// apply S matrices to the already computed columns of the T matrix
	concepts::TColumn<Real>* T2 = 0;
	for (concepts::TColumn<Real>* TT = T0; TT != 0; TT = TT->link()) {
	  DEBUGL(hpSpace3DBuildElements_D, cntr.key()
		 << " - from T0, treating " << *TT);
	  T2 = new concepts::TColumnTensor<Real, 3>(nni, TT->index(), T2);
	  (*Smatrices[i])(*TT, *T2);
	}
	for (concepts::TColumn<Real>* TT = T1; TT != 0; TT = TT->link()) {
	  T2 = new concepts::TColumnTensor<Real, 3>(nni, TT->index(), T2);
	  (*Smatrices[i])(*TT, *T2);
	  DEBUGL(hpSpace3DBuildElements_D, cntr.key()
		 << " - from T1, treating "
		 << *TT << ", stored in " << *T2);
	}

	// buildElements_ is called with the modified columns of the T matrix
	buildElements_(*child, Pmaxi, T2);

	// clean up
	while (T2 != 0) {
	  concepts::TColumn<Real>* foo = T2; T2 = T2->link(); delete foo;
	}
      }
    }

    while (T1 != 0) {
      concepts::TColumn<Real>* foo = T1; T1 = T1->link(); delete foo;
    }

    DEBUGL(hpSpace3DBuildElements_D, cntr.key() << " - dim = " << dim_
	   << ", done.");
  }

} // namespace hp3d
