/* Functor class for sum factorization
 */

#include "sumfact.hh"
#include "shapeFunction3d.hh"
#include "space/element.hh"
#include "toolbox/coeffIterator.hh"

// timings
#define matrixTiming

#ifdef matrixTiming
#include "toolbox/inputOutput.hh"
#include "toolbox/resourceMonitor.hh"
#endif // matrixTiming

// debugging
#include "basics/debug.hh"

#define SumFactAppl_D 0

namespace hp3D {

  // ****************************************************** SumFactorization **

  template<typename T>
  void SumFactorization::operator()(concepts::ElementMatrix<Real>& em,
				    const concepts::Array<Real>& jacobian,
				    T* coeff,
				    const ShapeFunction3D<Real>& u,
				    const ShapeFunction3D<Real>& v) {
    // quadrature points
    const uint nq[3] = { u.nq(0), u.nq(1), u.nq(2) };
    const Real* const w[3] = { u.weights(0), u.weights(1), u.weights(2) };

    // assert that u and v use the same quadrature points
    conceptsAssert(nq[0] == v.nq(0), concepts::Assertion());
    conceptsAssert(nq[1] == v.nq(1), concepts::Assertion());
    conceptsAssert(nq[2] == v.nq(2), concepts::Assertion());
    conceptsAssert(w[0] == v.weights(0), concepts::Assertion());
    conceptsAssert(w[1] == v.weights(1), concepts::Assertion());
    conceptsAssert(w[2] == v.weights(2), concepts::Assertion());

    // shapefunctions for u
    const uint uns[3] = { u.n(0), u.n(1), u.n(2) };
    const Real* const us[3] = { u.values(0), u.values(1), u.values(2) };
    DEBUGL(SumFactAppl_D, "u = " << us[0][0] << ", " << us[0][1] << "; "
	   << us[1][0] << ", " << us[1][1] << "; "
	   << us[2][0] << ", " << us[2][1]);

    // shapefunctions for v
    const uint vns[3] = { v.n(0), v.n(1), v.n(2) };
    const Real* const vs[3] = { v.values(0), v.values(1), v.values(2) };
    DEBUGL(SumFactAppl_D, "v = " << vs[0][0] << ", " << vs[0][1] << "; "
	   << vs[1][0] << ", " << vs[1][1] << "; "
	   << vs[2][0] << ", " << vs[2][1]);

    // get needed indices information
    concepts::TColumnTensor<bool, 3> uneededIdx(uns, 0);
    u.T()->usedIdx(uneededIdx);
    DEBUGL(SumFactAppl_D, "needed indices for u: " << uneededIdx);
    concepts::TColumnTensor<bool, 3> vneededIdx(vns, 0);
    v.T()->usedIdx(vneededIdx);
    DEBUGL(SumFactAppl_D, "needed indices for v: " << vneededIdx);


    // swap order
    int a = 2, b = 1, c = 0; // original
    // add a method to set a, b and c to get the fastest method possible

    // prepare matrices for sum factorization
    sumfact1_.resize(nq[c]*nq[b]*uns[a]*vns[a]);
    sumfact1_.zeros();
    DEBUGL(SumFactAppl_D, "reserved " << sumfact1_.size()
	   << " for sumfact1_");
    sumfact2_.resize(nq[c]*uns[b]*vns[b]*uns[a]*vns[a]);
    sumfact2_.zeros();
    DEBUGL(SumFactAppl_D, "reserved " << sumfact2_.size()
	   << " for sumfact2_");

#ifdef matrixTiming
    concepts::ResourceMonitor timer;
#endif // matrixTiming
    // fill sumfact1_
    Real* sumfact1 = (Real*)sumfact1_;
    if (coeff != 0) {
      for (uint iz = 0; iz < uns[a]; ++iz) {
	for (uint jz = 0; jz < vns[a]; ++jz) {
	  const Real* wxAcc = w[c];
	  const Real* dx = (const Real*)jacobian;
	  T coeffTmp = *coeff;
	  for (uint I = 0; I < nq[c]; ++I) {
	    const Real* wyAcc = w[b];
	    for (uint J = 0; J < nq[b]; ++J) {
	      const Real* wzAcc = w[a];
	      const Real* siz = us[a] + iz*nq[a];
	      const Real* sjz = vs[a] + jz*nq[a];
	      for (uint K = 0; K < nq[a]; ++K) {
		*sumfact1 += *siz++ * *sjz++ * *coeffTmp++ *
		  *dx++ * *wxAcc * *wyAcc * *wzAcc++;
	      } // K
	      ++wyAcc;
	      ++sumfact1;
	    } // J
	    ++wxAcc;
	  } // I
	} // jz
      } // iz
    } else {
      for (uint iz = 0; iz < uns[a]; ++iz) {
	for (uint jz = 0; jz < vns[a]; ++jz) {
	  const Real* wxAcc = w[c];
	  const Real* dx = (const Real*)jacobian;
	  for (uint I = 0; I < nq[c]; ++I) {
	    const Real* wyAcc = w[b];
	    for (uint J = 0; J < nq[b]; ++J) {
	      const Real* wzAcc = w[a];
	      const Real* siz = us[a] + iz*nq[a];
	      const Real* sjz = vs[a] + jz*nq[a];
	      for (uint K = 0; K < nq[a]; ++K) {
		*sumfact1 += *siz++ * *sjz++ *
		  *dx++ * *wxAcc * *wyAcc * *wzAcc++;
	      } // K
	      ++wyAcc;
	      ++sumfact1;
	    } // J
	    ++wxAcc;
	  } // I
	} // jz
      } // iz
    }
#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("first_temp", timeCntr_, timer.utime());
#endif // matrixTiming

    // fill sumfact2_
    Real* sumfact2 = (Real*)sumfact2_;
    for (uint iy = 0; iy < uns[b]; ++iy) {
      for (uint jy = 0; jy < vns[b]; ++jy) {
	sumfact1 = (Real*)sumfact1_;
	for (uint iz = 0; iz < uns[a]; ++iz) {
	  for (uint jz = 0; jz < vns[a]; ++jz) {
	    for (uint I = 0; I < nq[c]; ++I) {
	      const Real* siy = us[b] + iy*nq[b];
	      const Real* sjy = vs[b] + jy*nq[b];
	      for (uint J = 0; J < nq[b]; ++J) {
		*sumfact2 += *siy++ * *sjy++ * *sumfact1++;
	      } // J
	      ++sumfact2;
	    } // I
	  } // jz
	} // iz
      } // jy
    } // iy
#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("second_temp", timeCntr_, timer.utime());
#endif // matrixTiming

    // fill em
    sumfact2 = (Real*)sumfact2_;
    Real* intermed = (Real*)em;
    for (uint iz = 0; iz < uns[2]; ++iz) {
      for (uint iy = 0; iy < uns[1]; ++iy) {
	for (uint ix = 0; ix < uns[0]; ++ix) {
	  uint i3[3] = { ix, iy, iz };
	  if (uneededIdx[i3]) {
	    for (uint jz = 0; jz < vns[2]; ++jz) {
	      for (uint jy = 0; jy < vns[1]; ++jy) {
		for (uint jx = 0; jx < vns[0]; ++jx) {
		  uint j3[3] = { jx, jy, jz };
		  if (vneededIdx[j3]) {
// 		    if (sameMatrix &&
// 			neededIdxOldElm[i3] && neededIdxOldElm[j3]) {
// 		      // no computation needed, copy of old value used
// 		    } else {
// 		      if (!((jx < ix) && (jy < iy) && (jz < iz))) {
// 			// exploiting symmetry of matrix
// 			*intermed = 0.0;
			sumfact2 = (Real*)sumfact2_
			  + i3[b]*vns[b]*uns[a]*vns[a]*nq[c]
			  + i3[a]*vns[a]*nq[c]
			  + j3[b]*uns[a]*vns[a]*nq[c]
			  + j3[a]*nq[c];
// 			Real* intermed = (Real*)em + j[0] + 
// 			  vns[0]*(j[1] + 
// 				  vns[1]*(j[2] + 
// 					  vns[2]*(i[0] + 
// 						  uns[0]*(i[1]+uns[1]*i[2]))));
			const Real* six = us[c] + i3[c]*nq[c];
			const Real* sjx = vs[c] + j3[c]*nq[c];
			for (uint I = 0; I < nq[c]; ++I) {
			  *intermed += *six++ * *sjx++ * *sumfact2++;
			} // I
// 		      } // symmetric
// 		    } // computation needed
		  } // if neededIdx[j3]
		  ++intermed;
		} // jx
	      } // jy
	    } // jz
	  }
	  else // if neededIdx[i3]
	    intermed += vns[a]*vns[b]*vns[c];
	} // ix
      } // iy
    } // iz
#ifdef matrixTiming
    if (timings_)
      timings_->addArrayDouble("fill_em", timeCntr_++, timer.utime());
#endif // matrixTiming
  }

  concepts::InOutParameters* SumFactorization::timings_ = 0;

  uint SumFactorization::timeCntr_ = 0;

  void SumFactorization::setTimings(concepts::InOutParameters* timings) {
#ifdef matrixTiming
    timings_ = timings;
    if (timings_) {
      timeCntr_ = 0;
      timings_->addArrayDouble("first_temp");
      timings_->addArrayDouble("second_temp");
      timings_->addArrayDouble("fill_em");
    }
#endif // matrixTiming
  }

  bool SumFactorization::timings() {
#ifdef matrixTiming
    return true;
#else
    return false;
#endif // matrixTiming
  }

  template
  void SumFactorization::operator()(concepts::ElementMatrix<Real>& em,
				    const concepts::Array<Real>& jacobian,
				    concepts::CoeffIterator<Real, Real>* coeff,
				    const ShapeFunction3D<Real>& u,
				    const ShapeFunction3D<Real>& v);

  template
  void SumFactorization::operator()(concepts::ElementMatrix<Real>& em,
				    const concepts::Array<Real>& jacobian,
				    concepts::CoeffIterator<Real,
				    concepts::MapReal3d>* coeff,
				    const ShapeFunction3D<Real>& u,
				    const ShapeFunction3D<Real>& v);

} // namespace hp3D
