/* zero tangential component, ie. u x n = 0
 */

#include "constraints/element.hh"
#include "zerotangential.hh"
#include "hexFunctions.hh"
#include "vectorial/vectorial.hh"

// debugging
#include "basics/debug.hh"

#define ZNVlengthT_D 0
#define ZNVassembly_D 0
#define ZNVlocalMatrix_D 0
#define ZNVfillEM_D 0

namespace hp3D {

  // *************************************************** ZeroTangentialValue **

  template<typename F>
  const concepts::Connector3& ZeroTangentialValue<F>::assertions_
  (const concepts::Element<F>& e) const {
    /* Assertions:
       - a vector element
       - 3 components
       - the same in every component
       - hp3D FEM
    */
    const vectorial::Element<F>* elm =
      dynamic_cast<const vectorial::Element<F>*>(&e);
    conceptsAssert(elm, concepts::Assertion());

#ifdef DEBUG
    const uint vdim = elm->vdim();
#endif
    conceptsAssert(vdim == 3, concepts::Assertion());

    const concepts::Element<F>& scalarE = *(elm->get(0));
    conceptsAssert(&scalarE == elm->get(1), concepts::Assertion());
    conceptsAssert(&scalarE == elm->get(2), concepts::Assertion());

    scalElm_ = dynamic_cast<const Hexahedron*>(&scalarE);
    conceptsAssert(scalElm_, concepts::Assertion());
    const concepts::Connector3& conn3 = scalElm_->support();

    return conn3;
  }

  template<typename F>
  void ZeroTangentialValue<F>::lengthT(const concepts::Element<F>& e,
				       uint& length) {
    const concepts::Connector3& conn3 = assertions_(e);
    DEBUGL(ZNVlengthT_D, attr_ << ", " << conn3);

    // store polynomial degrees of edges
    concepts::DynArray<uint> P(2, 2, 0);
    const concepts::Connector1* edge = 0;
    for (uint l = 0; (edge = conn3.edge(l)) != 0; ++l) {
      P[edge->key()] = scalElm_->edgeP(l).p_[0];
    }

    // loop over faces, check for attributes == attr_
    const concepts::Connector2* face = 0;
    for (uint i = 0; (face = conn3.face(i)) != 0; ++i) {
      if (face->attrib() == attr_) {
	const concepts::AdaptiveControlP<2>& ctrl2 = scalElm_->faceP(i);
	DEBUGL(ZNVlengthT_D, *face << " is considered, p = (" << ctrl2.p_[0]
	       << ", " << ctrl2.p_[1] << ')');
	length += (ctrl2.p_[0]-1)*(ctrl2.p_[1]-1)*2; // interior of face
	const concepts::Connector1* edge = 0;
	for (uint j = 0; (edge = face->edge(j)) != 0; ++j) {
	  const uint p = P[edge->key()];
	  length += (p-1)*2; // interior of edge
	  DEBUGL(ZNVlengthT_D, *edge << ", p = " << p);
	}
	length += 4*2; // four vertices
      }
    }
    DEBUGL(ZNVlengthT_D, "length = " << length);
  }

  template<typename F>
  void ZeroTangentialValue<F>::assembly(const concepts::Element<F>& e,
					concepts::TColumn<F>*& t,
					uint& n, uint& j, uint length) {
    const concepts::Connector3& conn3 = assertions_(e);
    DEBUGL(ZNVassembly_D, attr_ << ", " << conn3 << ", t = " << t << ", n = "
	   << n << ", j = " << j << ", length = " << length);

    // store polynomial degrees of edges
    concepts::DynArray<uint> P(2, 2, 0);
    const concepts::Connector1* edge = 0;
    for (uint l = 0; (edge = conn3.edge(l)) != 0; ++l) {
      P[edge->key()] = scalElm_->edgeP(l).p_[0];
    }

    const concepts::Connector2* face = 0;
    for (uint i = 0; (face = conn3.face(i)) != 0; ++i) {
      if (face->attrib() == attr_) {
	DEBUGL(ZNVassembly_D, "considering face " << *face << ", p = ("
	       << scalElm_->faceP(i).p_[0] << ", " << scalElm_->faceP(i).p_[1]
	       << ')');

	// vertices
	const concepts::Connector0* vtx = 0;
	for (uint l = 0; (vtx = face->vertex(l)) != 0; ++l) {
	  uint& idx = ctrl0_[vtx->key()];
	  if (idx == (uint)-1) {
	    idx = n;
	    n += 2;
	  }
	  DEBUGL(ZNVassembly_D, "vertex " << l << ": " << *vtx
		 << " has idx = " << idx);
	  for (uint m = 0; m < 2; ++m) { // every vertex has 3 dofs, 2 constr.
	    t = new concepts::TColumn<Real>(length, idx+m, t);
	    (*t)[j++] = 1.0;
	  }
	}
	DEBUGL(ZNVassembly_D, "after vertices: j = " << j);

	// edges
	const concepts::Connector1* edge = 0;
	for (uint l = 0; (edge = face->edge(l)) != 0; ++l) {
	  uint& idx = ctrl1_[edge->key()];
	  const uint p = P[edge->key()];
	  if (idx == (uint)-1) {
	    idx = n;
	    n += 2*(p-1);
	  }
	  DEBUGL(ZNVassembly_D, "edge " << l << ": " << *edge
		 << " has idx = " << idx << ", p = " << p);
	  for (uint k = 0; k < 2*(p-1); ++k) {
	    t = new concepts::TColumn<Real>(length, idx+k, t);
	    (*t)[j++] = 1.0;
	  }
	}
	DEBUGL(ZNVassembly_D, "after edges: j = " << j);

	// face
	const uint p = (scalElm_->faceP(i).p_[0]-1)*
	  (scalElm_->faceP(i).p_[1]-1)*2;
	for (uint k = 0; k < p; ++k) {
	  t = new concepts::TColumn<Real>(length, n+k, t);
	  (*t)[j++] = 1.0;
	}
	DEBUGL(ZNVassembly_D, "after face: j = " << j);
	n += p;
      } // attr_ matches
    } // for faces
  }

  template<typename F>
  constraints::AnalyticalConstraint<F>* ZeroTangentialValue<F>::clone() const {
    return new ZeroTangentialValue<F>(*this);
  }

  template<typename F>
  void ZeroTangentialValue<F>::localMatrix
  (concepts::ElementMatrix<F>& em, uint& j,
   const constraints::Element<F>& elm) const {
    if (em.size() > 0) {
      DEBUGL(ZNVlocalMatrix_D, elm.compElement() << " for " << attr_
	     << ", j = " << j);
      const concepts::Connector3& conn3 = assertions_(elm.compElement());

      const uint offset = em.m()/3;
      DEBUGL(ZNVlocalMatrix_D, "offset = " << offset);

      // store polynomial degrees of edges
      concepts::DynArray<uint> P(2, 2, 0);
      const concepts::Connector1* edge = 0;
      for (uint l = 0; (edge = conn3.edge(l)) != 0; ++l) {
	P[edge->key()] = scalElm_->edgeP(l).p_[0];
      }

      // store indices of edges and vertices
      concepts::DynArray<uint> vtxIndex(2, 2, 0);
      const concepts::Connector0* vtx = 0;
      for (uint l = 0; (vtx = conn3.vertex(l)) != 0; ++l) {
	vtxIndex[vtx->key()] = l;
      }
      concepts::DynArray<uint> edgeIndex(2, 2, 0);
      edge = 0;
      for (uint l = 0; (edge = conn3.edge(l)) != 0; ++l) {
	edgeIndex[edge->key()] = l;
      }

      const concepts::Connector2* face = 0;
      for (uint i = 0; (face = conn3.face(i)) != 0; ++i) {
	if (face->attrib() == attr_) {
	  uint ll[3] = { 0, 0, 0 };
	  const ushort* const Pmax = scalElm_->p();

	  // get normal vector of face
	  const concepts::Cell3* cell =
	    dynamic_cast<const concepts::Cell3*>(&(scalElm_->cell()));
	  concepts::Real3d e1(cell->vertex(1)-cell->vertex(0));
	  concepts::Real3d e2(cell->vertex(3)-cell->vertex(0));
	  e1 *= 1.0/e1.l2();
	  e2 *= 1.0/e2.l2();
	  // vertices
	  const concepts::Connector0* vtx = 0;
	  for (uint l = 0; (vtx = face->vertex(l)) != 0; ++l// , ++k
	       ) {
	    HexFunctions::vertexIndex(vtxIndex[vtx->key()], ll); // fill ll
	    fillEM_(ll, Pmax, offset, j, e1, em);
	    fillEM_(ll, Pmax, offset, j, e2, em);
	  }
	  DEBUGL(ZNVlocalMatrix_D, "after vertices: j = " << j);

	  // edges
	  const concepts::Connector1* edge = 0;
	  for (uint l = 0; (edge = face->edge(l)) != 0; ++l) {
	    uint pIndex = 0, qIndex = 0, rIndex = 0;
	    HexFunctions::edgeOrientation(edgeIndex[edge->key()],
					  pIndex, qIndex, rIndex);
	    HexFunctions::edgeIndex(edgeIndex[edge->key()],
				    qIndex, rIndex, ll); // fill ll
	    const uint p = P[edge->key()];
	    DEBUGL(ZNVlocalMatrix_D, *edge << ", ll = (" << ll[0] << ','
		   << ll[1] << ',' << ll[2] << "), p = " << p);
	    for (uint k = 0; k+1 < p; ++k) {
	      ll[pIndex] = k+2;
	      fillEM_(ll, Pmax, offset, j, e1, em);
	      fillEM_(ll, Pmax, offset, j, e2, em);
	    }
	  }
	  DEBUGL(ZNVlocalMatrix_D, "after edges: j = " << j);

	  // face
	  const uint p0 = scalElm_->faceP(i).p_[0],
	    p1 = scalElm_->faceP(i).p_[1];
	  uint pIndex = 0, qIndex = 0, rIndex = 0;
	  HexFunctions::faceOrientation(scalElm_->support(), i,
					pIndex, qIndex, rIndex);
	  HexFunctions::faceIndex(i, rIndex, ll); // fill ll
	  for (uint k0 = 0; k0+1 < p0; ++k0) {
	    for (uint k1 = 0; k1+1 < p1; ++k1) {
	      ll[pIndex] = k0 + 2;
	      ll[qIndex] = k1 + 2;
	      fillEM_(ll, Pmax, offset, j, e1, em);
	      fillEM_(ll, Pmax, offset, j, e2, em);
	    }
	  }
	} // attr_ matches
      } // for faces

//       DEBUGL(ZNVlocalMatrix_D, em);
    }
  }

  template<typename F>
  void ZeroTangentialValue<F>::localVector
  (concepts::ElementMatrix<F>& em, uint& j,
   const constraints::Element<F>& elm) const {}

  template<typename F>
  std::ostream& ZeroTangentialValue<F>::info(std::ostream& os) const {
    return os << "ZeroTangentialValue(" << attr_ << ')';
  }

  template<typename F>
  void ZeroTangentialValue<F>::fillEM_(const uint ll[3], const ushort p[3],
				       const uint offset, uint& j,
				       const concepts::Real3d e,
				       concepts::ElementMatrix<F>& em) const {
    DEBUGL(ZNVfillEM_D, "ll = (" << ll[0] << ',' << ll[1] << ',' << ll[2]
	   << "), p = (" << p[0] << ',' << p[1] << ',' << p[2]
	   << "), offset = " << offset << ", j = " << j << ", e = "
	   << e);
    const uint m = HexFunctions::index(ll, p);
    DEBUGL(ZNVfillEM_D, "m = " << m);
    em(m       ,   j)   = e[0];
    em(m+offset,   j)   = e[1];
    em(m+2*offset, j++) = e[2];
    DEBUGL(ZNVfillEM_D, "j = " << j);
  }

  template class ZeroTangentialValue<Real>;

} // namespace hp3D
