/* 1D quadrature rules
 */

#ifndef quadRule_hh
#define quadRule_hh

#include <map>

#include "basics/typedefs.hh"
#include "basics/outputOperator.hh"
#include "quadrature.hh"

namespace concepts {

  // ******************************************************** QuadratureRule **

  /** Quadrature rule for numerical integration.
      @author Philipp Frauenfelder, 2004.
  */
  class QuadratureRule : public OutputOperator {
  public:
    /// Returns a pointer into the array of the abscissas
    virtual const Real* abscissas() const = 0;

    /// Returns a pointer into the array of the weights
    virtual const Real* weights() const = 0;

    /// Returns the number of points
    virtual uint n() const = 0;
  };

  // ************************************************* QuadratureRuleDynamic **

  /** Base class for quadrature rules with dynamically allocated storage for
      the weights and abscissas. This is only a technical class and should
      not be used directly.
      @author Philipp Frauenfelder, 2004
  */
  class QuadratureRuleDynamic : public QuadratureRule {
  public:
    QuadratureRuleDynamic(const Real* abscissas = 0, const Real* weights = 0) :
      abscissas_(abscissas), weights_(weights) {}
    virtual const Real* abscissas() const { return abscissas_; }
    virtual const Real* weights() const { return weights_; }
  protected:
    /// Abscissas
    const Real* abscissas_;
    /// Weights
    const Real* weights_;
  };

  // ******************************************** QuadratureRuleGaussLobatto **

  /** Gauss Lobatto quadrature rule including both endpoints.
      \f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
      is exact for \f$f \in P_{2p-1}\f$ and n = p+1 points.
      n must be greater or equal to 2.

      The abscissas \f$x_i\f$ are the zeros of 
      \f$(1-x^2) P_{p-1}^{(1,1)}(x)\f$
      and the weights are 
      \f$w_i = 2/(p(p+1) (P_p^{(0,0)}(x_i))^2)\f$.

      The computations and the storage of the values are done by the
      class Quadrature with template parameter 0. The difference
      between this class and Quadrature is that it is in a class
      hierarchy of quadrature rules. This has advantages when
      dynamically switching quadrature rules is needed. On the other
      hand, this class returns the values via a virtual function call
      \c abscissas() and \c weights() should therefore not be called
      to often (inside loops etc.).
      @see Quadrature
      @author Philipp Frauenfelder, 2004
  */
  class QuadratureRuleGaussLobatto : public QuadratureRuleDynamic {
  public:
    /** Constructor. Computes the quadrature points.
	@param n Number of points to be computed
    */
    QuadratureRuleGaussLobatto(uint n) : rule_(n) {
      abscissas_ = rule_.abscissas();
      weights_ = rule_.weights();
    }
    virtual ~QuadratureRuleGaussLobatto();
    virtual uint n() const { return rule_.n(); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    Quadrature<0> rule_;
  };

  // ********************************************* QuadratureRuleGaussJacobi **

  /** Gauss Jacobi quadrature rule not including both endpoints.
      \f[ \int_{-1}^1 f(x) \, dx \approx \sum_{i=0}^p w_i f(x_i) \f]
      is exact for \f$f \in P_{2p+1}\f$ and n = p+1 points.
      n must be greater or equal to 1.

      The abscissas \f$x_i\f$ are the zeros of
      \f$P_{p+1}^{(0,0)}(x)\f$
      and the weights are
      \f[w_i = \frac{2}{1-x_i^2} \left( \frac{d}{dx} \left. P^{(0,0)}_{p+1}(x) \right|_{x=x_i} \right)^{-2}.\f]

      The computations and the storage of the values are done by the
      class Quadrature with template parameter 4. The difference
      between this class and Quadrature is that it is in a class
      hierarchy of quadrature rules. This has advantages when
      dynamically switching quadrature rules is needed. On the other
      hand, this class returns the values via a virtual function call
      \c abscissas() and \c weights() should therefore not be called
      to often (inside loops etc.).
      @see Quadrature
      @author Philipp Frauenfelder, 2004
  */
  class QuadratureRuleGaussJacobi : public QuadratureRuleDynamic {
  public:
    /** Constructor. Computes the quadrature points.
	@param n Number of points to be computed
    */
    QuadratureRuleGaussJacobi(uint n) : rule_(n) {
      abscissas_ = rule_.abscissas();
      weights_ = rule_.weights();
    }
    virtual ~QuadratureRuleGaussJacobi();
    virtual uint n() const { return rule_.n(); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    Quadrature<4> rule_;
  };

} // namespace concepts

#endif // quadRule_hh
