/* 1D quadrature rules
 */

#include "basics/debug.hh"
#include "basics/exceptions.hh"
#include "quadrature.hh"
#include "quadratureBase.hh"

// debugging
#define QuadratureConstr_D 0
#define QuadratureFind_D 0

namespace concepts {

  // ************************************************************ Quadrature **

  template<int type>
  std::map<uint, Real*> Quadrature<type>::abscissas_H
  = std::map<uint, Real*>();

  template<int type>
  std::map<uint, Real*> Quadrature<type>::weights_H
  = std::map<uint, Real*>();

  template<>
  Quadrature<3>::Quadrature(uint n) : 
    abscissas_(0), weights_(0), n_(n) {
    conceptsAssert(n >= 2, Assertion());
    if (!(find_())) {
      // compute the values
      GaussRadauAbscWght(abscissas_, weights_, n_-1, 1);
      DEBUGL(QuadratureConstr_D, "computed " << n_ << " quadrature points");
    }
  }

  template<>
  Quadrature<4>::Quadrature(uint n) : 
    abscissas_(0), weights_(0), n_(n) {
    conceptsAssert(n >= 1, Assertion());
    if (!(find_())) {
      // compute the values
      GaussJacobiAbscWght(abscissas_, weights_, n_-1);
      DEBUGL(QuadratureConstr_D, "computed " << n_ << " quadrature points");
    }
  }

  template<int type>
  Quadrature<type>::Quadrature(uint n) : 
    abscissas_(0), weights_(0), n_(n) {
    conceptsAssert(n >= 2, Assertion());
    if (!(find_())) {
      // compute the values
      GaussLobattoAbscWght(abscissas_, weights_, n_-1, type);
      DEBUGL(QuadratureConstr_D, "computed " << n_ << " quadrature points");
    }
  }

  template<int type>
  std::string Quadrature<type>::storedPoints() {
    std::stringstream res;
    for (std::map<uint, Real*>::const_iterator i = abscissas_H.begin();
	 i != abscissas_H.end(); ++i) {
      res << i->first << " ";
    }
    return res.str();
  }

  template<int type>
  std::ostream& Quadrature<type>::info(std::ostream& os) const {
    os << "Quadrature(" << std::endl << "  abscissas(";
    for (uint i = 0; i < n_; ++i) {
      os << abscissas_[i];
      if (i < n_-1) os << ", ";
    }
    os << ")" << std::endl << "  weights(";
    for (uint i = 0; i < n_; ++i) {
      os << weights_[i];
      if (i < n_-1) os << ", ";
    }
    return os << ")" << std::endl << ")";
  }

  template<int type>
  bool Quadrature<type>::find_() {
    std::map<uint, Real*>::iterator i;
    i = abscissas_H.find(n_);
    if (i != abscissas_H.end()) {
      DEBUGL(QuadratureFind_D, "found quadrature points");
      abscissas_ = i->second;
      i = weights_H.find(n_);
      assert(i != abscissas_H.end());
      weights_ = i->second;
      return true;
    } else {
      DEBUGL(QuadratureFind_D, "quadrature points must be computed");
      abscissas_ = new Real[n_];
      abscissas_H.insert(std::pair<uint, Real*>(n_, abscissas_));
      weights_ = new Real[n_];
      weights_H.insert(std::pair<uint, Real*>(n_, weights_));
      return false;
    }
  }

  template class Quadrature<0>;
  template class Quadrature<1>;
  template class Quadrature<2>;
  template class Quadrature<3>;
  template class Quadrature<4>;

} // namespace concepts
