// Boundary integrals for linear DG FEM in 1D

#include "bilinearForm.hh"
#include "elementPair.hh"
#include "space/tmatrix.hh"
#include "linearFEM/element1D.hh"
#include "toolbox/boundary.hh"

// debugging
#include "basics/debug.hh"

#define BoundaryIntAppl_D 0
#define BoundaryIntStabAppl_D 0

namespace linDG1D {

  // *********************************************************** BoundaryInt **

  void BoundaryInt::operator()(const concepts::Element<Real>& elmX,
			       const concepts::Element<Real>& elmY,
			       concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();
    
    const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
#ifdef DEBUG
    const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());
#endif

    Real bcFlag[2] = { 1.0, 1.0 };
    if (bc_) {
      for (uint i = 0; i < 2; ++i) {
	if ((*bc_)(eX->support().vertex(i)->attrib()).type() ==
	    concepts::Boundary::DIRICHLET) {
	  bcFlag[i] = 2.0;
	  DEBUGL(BoundaryIntAppl_D, "hit DIRICHLET boundary on "
		 << *(eX->support().vertex(i)));
	}
      }
    }

    if (&elmX == &elmY) {
      const Real h = eX->cell().size();
      em(0,0) = -0.5/h*bcFlag[0]; em(0,1) = 0.5/h*bcFlag[1];
      em(1,0) = 0.5/h*bcFlag[0];  em(1,1) = -0.5/h*bcFlag[1];
    }
  }

  void BoundaryInt::operator()(const concepts::Element<Real>& elmX,
			       const concepts::Element<Real>& elmY,
			       concepts::ElementMatrix<Real>& em,
			       const concepts::ElementPair<Real>& ep) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros(); 
    
    if (&elmX != &elmY) {
      const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
      conceptsAssert(eX, concepts::Assertion());
      const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
      conceptsAssert(eY, concepts::Assertion());
      
      const DGElementPair* epc = dynamic_cast<const DGElementPair*>(&ep);
      conceptsAssert(epc, concepts::Assertion());
      
      Real h = epc->elm1().cell().size();
      
      if (eX->support().key() < eY->support().key()) h *= -1.0;
      if (epc->idx1() == 1) h *= -1.0;
      bool idx1_n = !bool(epc->idx1());
      em(idx1_n, epc->idx2()) = -0.5/h;
      em(epc->idx1(), epc->idx2()) = 0.5/h;
    }
  }

  // ******************************************************* BoundaryIntStab **

  void BoundaryIntStab::operator()(const concepts::Element<Real>& elmX,
				   const concepts::Element<Real>& elmY,
				   concepts::ElementMatrix<Real>& em,
				   const concepts::ElementPair<Real>& ep) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    if (&elmX == &elmY) {
      
      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());
      
      DEBUGL(BoundaryIntStabAppl_D, elmX << ", " << elmY);
     
#ifdef DEBUG
      const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
      conceptsAssert(eX, concepts::Assertion());
      const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
      conceptsAssert(eY, concepts::Assertion());
#endif
      
      const DGElementPair* epc = dynamic_cast<const DGElementPair*>(&ep);
      conceptsAssert(epc, concepts::Assertion());
      
      const Real w_stab = 1.0; 
      
      DEBUGL(BoundaryIntStabAppl_D, epc->size1() << ", " << epc->size2());

      if ((epc->size1() == 0) || (epc->size2() == 0)) {
	Real s = std::max(epc->size1(), epc->size2());
	em(0,0) = w_stab/s;
	em(1,1) = w_stab/s;
      }
      else {
	em(0,0) = w_stab/epc->size1();
	em(1,1) = w_stab/epc->size2();
      }
    }
    if (&elmX != &elmY) {
      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());
      
      const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
      conceptsAssert(eX, concepts::Assertion());
      const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
      conceptsAssert(eY, concepts::Assertion());
      
      const DGElementPair* epc = dynamic_cast<const DGElementPair*>(&ep);
      conceptsAssert(epc, concepts::Assertion());
      
      const Real hmin = std::min(eX->cell().size(), eY->cell().size());

      const Real w_stab = 1.0;
      em(epc->idx1(), epc->idx2()) = -1.0/hmin*w_stab;
    }
  }

} // namespace linDG1D
