// element pair for linear DG FEM in 1D

#ifndef elementPairDG1D_hh
#define elementPairDG1D_hh

#include "basics/typedefs.hh"
#include "space/elementPairs.hh"
#include "linearFEM/element1D.hh"

// debugging
#include "basics/debug.hh"

#define DGElementPairConstr_D 0

namespace linDG1D {

  using concepts::Real;

  // ********************************************************* DGElementPair **

  /** Element pair for DG FEM. Combines two elements for computing the
      off-diagonal blocks of the system matrices.
      @author Thomas Wihler, 2003
  */
  class DGElementPair : public concepts::ElementPair<Real> {
  public:
    /** Constructor. Used for two different elements
	@param elm1 First element
	@param elm2 Second element
	@param idx1 Index of the common vertex in the first element
	@param idx2 Index of the common vertex in the second element
    */
    DGElementPair(const linearFEM::Line& elm1, const uint idx1,
		  const linearFEM::Line& elm2, const uint idx2) : 
      elm1_(elm1), elm2_(elm2), idx1_(idx1), idx2_(idx2),
      size1_(0), size2_(0) {
      DEBUGL(DGElementPairConstr_D, elm1 << ", " << elm2);
    }
    /** Constructor. Used for two identical elements
	@param elm Element
	@param idx Index of the vertex in the element
	@param size1 Size of the element on one side of the vertex
	@param size2 Size of the element on the other side of the vertex
    */
    DGElementPair(const linearFEM::Line& elm, const uint idx,
		  const Real size1, const Real size2) : 
      elm1_(elm), elm2_(elm), idx1_(idx), idx2_(idx),
      size1_(size1), size2_(size2) {
      DEBUGL(DGElementPairConstr_D, "Element 1:" << size1 << ", Element2:"
	     << size2);
    }
    virtual const linearFEM::Line& elm1() const { return elm1_; }
    virtual const linearFEM::Line& elm2() const { return elm2_; }
    /// Returns index of the vertex in the first element
    uint idx1() const { return idx1_; }
    /// Returns index of the vertex in the second element
    uint idx2() const { return idx2_; }
    /// Returns size of the first element
    Real size1() const { return size1_; }
    /// Returns size of the second element
    Real size2() const { return size2_; }
  private:
    const linearFEM::Line& elm1_;
    const linearFEM::Line& elm2_;
    const uint idx1_;
    const uint idx2_;
    const Real size1_, size2_;
  };

} // namespace linDG1D

#endif // elementPairDG1D_hh
