// Bilinear forms for linear DG FEM in 2D

#include "integration.hh"
#include "space/tmatrix.hh"
#include "toolbox/boundary.hh"
#include "basics/exceptions.hh"

#include "triangle.hh"
#include "edgeInfo.hh"
#include "elementPair.hh"

#include "bilinearForm.hh"

// debugging
#include "basics/debug.hh"

#define DGElasticFluxBf2dAppl_D 0

using concepts::Real;

namespace linDG2D {

  // ************************************************************ IdentityBf **

  void IdentityBf::operator()(const concepts::Element<Real>& elmV,
                              const concepts::Element<Real>& elmU,
                              concepts::ElementMatrix<Real>& em) {
  
    const uint mV = elmV.T().m();
    const uint mU = elmU.T().m();
    em.resize(mV, mU);
    em.zeros(); 

    if (&elmV == &elmU) {
      const TriangleP1* eV = dynamic_cast<const TriangleP1*>(&elmV);
      conceptsAssert(eV, concepts::Assertion());
      Real val = eV->cell().area() / 3.;
      for (uint i=0; i<mU; ++i) {
        em(i,i) = val;
      }
    }
  }

  // ********************************************************** LaplaceVolBf **

  void LaplaceVolBf::operator()(const concepts::Element<Real>& elmV,
                                const concepts::Element<Real>& elmU,
                                concepts::ElementMatrix<Real>& em) {
  
    const uint mV = elmV.T().m();
    const uint mU = elmU.T().m();
    em.resize(mV, mU);
    em.zeros(); 

    if (&elmV == &elmU) {
      const TriangleP1* eV = dynamic_cast<const TriangleP1*>(&elmV);
      conceptsAssert(eV, concepts::Assertion());
      conceptsAssert(3 == mV, concepts::Assertion());

      Real area = eV->cell().area();
      // fetch gradients
      concepts::Array<concepts::Real2d> gradient(mV);
      for(uint iShapeFct=0; iShapeFct<mV; ++iShapeFct) {
        gradient[iShapeFct] = eV->shapeFctGradient(iShapeFct);
      }
      // fill element matrix
      for(uint p=0; p<mV; ++p) {
        em(p,p) = (gradient[p]*gradient[p])*area;
        for(uint q=p+1; q<mV; ++q) {
          em(p,q) = em(q,p) = (gradient[p]*gradient[q])*area;
        }
      }
    }
  }

  // ********************************************************* LaplaceFluxBf **

  void LaplaceFluxBf::operator()(const concepts::Element<Real>& elmV,
                                 const concepts::Element<Real>& elmU,
                                 concepts::ElementMatrix<Real>& em,
                                 const concepts::ElementPair<Real>& ep) {
    const uint mV = elmV.T().m();
    const uint mU = elmU.T().m();
    em.resize(mV, mU);
    em.zeros();

    const TriangleP1* eV = dynamic_cast<const TriangleP1*>(&elmV);
    conceptsAssert(eV, concepts::Assertion());
    const TriangleP1* eU = dynamic_cast<const TriangleP1*>(&elmU);
    conceptsAssert(eU, concepts::Assertion());
    const ElementPair* epc = dynamic_cast<const ElementPair*>(&ep);
    conceptsAssert(epc, concepts::Assertion());

    const EdgeInfo& ei = epc->edgeInfo();
    Real factor;

    if (ei.cell(1) == 0) {
      // boundary integral
      if (bc_) {
        if ((*bc_)(ei.edge().attrib()).type() == 
            concepts::Boundary::DIRICHLET) {
          factor = 1;
        } else {
          factor = 0;
        }
      } else {
        // assume dirichlet boundary all over
        factor = 1;
      }
    } else {
      // inner integral, <u> = 0.5*(u_left + u_right)
      factor = 0.5;
    }

    if (factor != 0) {
      factor *= ei.length();
      concepts::Real2d normal = ei.normal();

      // <grad u> [v]
      if (&(eV->cell()) == ei.cell(1)) {
        // for jump of testfunction [v]: testfunction (on eV) is on cell 1
        factor *= -1;
      }

      for(uint iU=0; iU<mU; ++iU) {
        Real f_n_grad_u = factor * (normal * eU->shapeFctGradient(iU));
        for(uint iV=0; iV<mV; ++iV) {
          em(iV,iU) = f_n_grad_u * epc->shapeFct(0, iV, 0.5);
        }
      }
    }
  }

  // ********************************************************* LaplaceStabBf **

  void LaplaceStabBf::operator()(const concepts::Element<Real>& elmV,
                                 const concepts::Element<Real>& elmU,
                                 concepts::ElementMatrix<Real>& em,
                                 const concepts::ElementPair<Real>& ep) {
    const uint mV = elmV.T().m();
    const uint mU = elmU.T().m();
    em.resize(mV, mU);
    em.zeros();

    const TriangleP1* eV = dynamic_cast<const TriangleP1*>(&elmV);
    conceptsAssert(eV, concepts::Assertion());
    const TriangleP1* eU = dynamic_cast<const TriangleP1*>(&elmU);
    conceptsAssert(eU, concepts::Assertion());
    const ElementPair* epc = dynamic_cast<const ElementPair*>(&ep);
    conceptsAssert(epc, concepts::Assertion());
    const EdgeInfo& ei = epc->edgeInfo();
    Real factor;

    if (ei.cell(1) == 0) {
      // boundary integral
      if (bc_) {
        if ((*bc_)(ei.edge().attrib()).type() == 
            concepts::Boundary::DIRICHLET) {
          factor = 1;
        } else {
          factor = 0;
        }
      } else {
        // assume dirichlet boundary all over
        factor = 1;
      }
    } else {
      // inner integral
      factor = 0.5;
    }

    if (factor != 0) {
      factor *= omega_;
      if (&(eV->cell()) != &(eU->cell())) {
        // for jumps [u][v]
        // testfunction (on eV) is on cell 1 and ansatzfct. (on eU) on cell 0
        // or
        // testfunction (on eV) is on cell 0 and ansatzfct. (on eU) on cell 1
        factor *= -1;
      }

      concepts::Quadrature<4> quad(2);
      const Real* absc = quad.abscissas();
      Real xi[2];
      xi[0] = 0.5*(1.0+absc[0]);
      xi[1] = 0.5*(1.0+absc[1]);
      factor *= 0.5; // quadrature weight
      for(uint iV=0; iV<mV; ++iV) {
        for(uint iU=0; iU<mU; ++iU) {
          for(int i=0; i<2; ++i) {
            em(iV,iU) += factor * 
              epc->shapeFct(0, iV, xi[i]) * 
              epc->shapeFct(1, iU, xi[i]);
          }
        }
      }
    }
  }

} // namespace linDG2D
