// edge information for linear DG FEM in 2D

#ifndef EdgeInfo_hh
#define EdgeInfo_hh

#include <basics.hh>
#include <geometry.hh>

#include "triangle.hh"

namespace test {
  class DGEdgeInfoTest;
  class DGSpace2dP1Test;
}

namespace linDG2D {

  using concepts::Real;

  // ************************************************************** EdgeInfo **

  /** Edge information for DG FEM. Provides information about an edge and its
      adjacent cell(s) used for DG face integrals. EdgeInfo objects for a
      mesh can be built up efficently using a MeshInfo object.
      @author Christoph Winkelmann, 2003
  */

  class EdgeInfo {
  public:
    friend class test::DGEdgeInfoTest;
    friend class test::DGSpace2dP1Test;
    /// Default Constructor. Needed to get new EdgeInfo from a hash map access
    EdgeInfo();
    /// Add a cell adjacent to the edge
    void addCell(const concepts::Triangle2d* cell, const uint idx);
    /** Return cell adjacent to the edge
        @param i which cell, left(0) or right(1)
    */
    const concepts::Triangle2d* cell(uint i) const { return cells_[i]; }
    /** Return index of edge represented in cell i
        @param i which cell, left(0) or right(1)
    */
    uint idx(uint i) const { return idxs_[i]; }
    /// Return topological edge represented
    const concepts::Edge& edge() const {
      return *(cells_[0]->connector().edge(idxs_[0]));
    }
    /// Return the length of the edge
    Real length() const;
    /// Return the mean height of the triangle(s) adjacent to the edge
    Real meanHeight() const;
    /// Return a vector normal to the edge, pointing out of first triangle
    const concepts::Real2d normal() const;
    /** Maps a point on the reference edge (0,1) onto the reference triangle
        of cell \c i. The reference edge is oriented anticlockwise in the left
        triangle and clockwise in the right triangle. The reference triangle
        has the corners (0,0), (0,1), (1,1).
        @param xi coordinate of the point on the reference edge
        @param i which cell, left(0) or right(1)
    */
    concepts::Real2d mapRefTri(Real xi, uint i) const;
    /** Maps a point on the reference edge (0,1) onto the physical edge. The
        reference edge is oriented anticlockwise in the left triangle and
        clockwise in the right triangle. 
        @param xi coordinate of the point on the reference edge
    */
    concepts::Real2d map(Real xi) const;
  private:
    const concepts::Triangle2d* cells_[2];
    uint idxs_[2];
  };

} // namespace linDG2D

#endif // EdgeInfo_hh
