#include "linearForm.hh"

// debugging
#include "basics/debug.hh"

#define DGElasticTimeLf2dAppl_D 0

using concepts::Real;
using concepts::Real2d;

namespace linDG2D {

  void LaplaceTimeLf::operator() (const concepts::Element<Real> &elmBase,
                                  concepts::ElementMatrix<Real> &em) {
      // downcast element
      const TriangleP1* elm = dynamic_cast<const TriangleP1*>(&elmBase);
      conceptsAssert(elm, concepts::Assertion());

      // resize and initialize element matrix
      const uint nDof = elm->T().m();
      em.resize(nDof, 1);
      em.zeros();

      const Real area = elm->cell().area();
    
      // quadrature points and weights
      concepts::Quadrature<4> quad(nQuadPts_);
      const Real* absc = quad.abscissas();
      const Real* weig = quad.weights();
      concepts::Array<Real> xi(nQuadPts_);
      concepts::Array<Real> w(nQuadPts_);
      for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
        xi[iQuad] = 0.5*(1.+absc[iQuad]);
        w[iQuad] = 0.5*weig[iQuad];
      }
    
      // volume integral
      // loop over all shape functions
      for (uint iShapeFct = 0; iShapeFct < nDof; ++iShapeFct) {
        Real integral = 0;
        // loop over quadrature points
        for (uint iQuad1 = 0; iQuad1 < nQuadPts_; ++iQuad1) {
          for (uint iQuad2 = 0; iQuad2 < nQuadPts_; ++iQuad2) {
            integral += frm_(elm->cell().chi(xi[iQuad1], xi[iQuad2]), time_)*
              elm->shapeFct(iShapeFct, xi[iQuad1], xi[iQuad1]*xi[iQuad2]) * 
              w[iQuad1] * w[iQuad2] * xi[iQuad1];
          }
        }
        // add to element matrix
        em(iShapeFct, 0) = integral * 2 * area;
      }
    
      // diffusive and penalty boundary fluxes
      // lacks neumann boundary contribution!
      if (bc_) {
        // Loop over edges
        for (uint iEdge = 0; iEdge < 3; ++iEdge) {
          const concepts::Edge& edge = *(elm->cell().connector().edge(iEdge));
          if ( (*bc_)(edge.attrib()).type() == concepts::Boundary::DIRICHLET ){
            const EdgeInfo& ei = mi_.edgeInfo(edge);
            Real2d normal = ei.normal();
            const Real prefactorDiff = tau_ * ei.length();
            const Real prefactorPen = omega_;
            concepts::Array<Real2d> xirt(nQuadPts_);
            for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
              xirt[iQuad] = ei.mapRefTri(xi[iQuad], 0);
            }

            for (uint iShapeFct=0; iShapeFct<nDof; ++iShapeFct) {
              Real sumPen = 0;
              Real sumDiff = 0;
              for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
                Real w_uBndQp = w[iQuad] *
                  (*bc_)(edge.attrib())(ei.map(xi[iQuad]), time_);
                sumPen += w_uBndQp *
                  elm->shapeFct(iShapeFct, xirt[iQuad][0], xirt[iQuad][1]);
                sumDiff += w_uBndQp;
              } // end loop (quadrature points)
              em(iShapeFct, 0) += sumPen * prefactorPen;
              em(iShapeFct, 0) += sumDiff * prefactorDiff *
                (elm->shapeFctGradient(iShapeFct) * normal);
            } // end loop (shape functions)
          }
        }
      }
  }

  void ScalarProductLf::operator() (const concepts::Element<Real> &elmBase,
                                    concepts::ElementMatrix<Real> &em) {
      // downcast element
      const TriangleP1* elm = dynamic_cast<const TriangleP1*>(&elmBase);
      conceptsAssert(elm, concepts::Assertion());

      // resize and initialize element matrix
      const uint nDof = elm->T().m();
      em.resize(nDof, 1);
      em.zeros();

      const Real area = elm->cell().area();
    
      // quadrature points and weights
      concepts::Quadrature<4> quad(nQuadPts_);
      const Real* absc = quad.abscissas();
      const Real* weig = quad.weights();
      concepts::Array<Real> xi(nQuadPts_);
      concepts::Array<Real> w(nQuadPts_);
      for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
        xi[iQuad] = 0.5*(1.+absc[iQuad]);
        w[iQuad] = 0.5*weig[iQuad];
      }

      // loop over all shape functions
      for (uint iShapeFct = 0; iShapeFct < nDof; ++iShapeFct) {
        Real integral = 0;
        // loop over quadrature points
        for (uint iQuad1 = 0; iQuad1 < nQuadPts_; ++iQuad1) {
          for (uint iQuad2 = 0; iQuad2 < nQuadPts_; ++iQuad2) {
            integral += frm_(elm->cell().chi(xi[iQuad1], xi[iQuad2])) *
              elm->shapeFct(iShapeFct, xi[iQuad1], xi[iQuad1]*xi[iQuad2]) *
              w[iQuad1] * w[iQuad2] * xi[iQuad1];
          }
        }
        // add to element matrix
        em(iShapeFct, 0) = integral * 2 * area;
      }
  }

} // namespace linDG2D
