#ifndef dgLinearForms_hh
#define dgLinearForms_hh

#include "basics/typedefs.hh"
#include "integration.hh"
#include "timestepping/vectors.hh"
#include "toolbox/formula.hh"
#include "toolbox/boundary.hh"
#include "geometry/boundaryConditions.hh"
#include "function/linearForm.hh"

#include "triangle.hh"
#include "meshInfo.hh"

#define DEBUG_FILLDATA 0

using concepts::Real;

namespace linDG2D {

  /** DG right hand side for time dependent problems with diffusion in 2D
      @author Christoph Winkelmann, 2003
  */
  class LaplaceTimeLf : public timestepping::TimeLinearForm {
  public:
    /** Constructor. Takes a formula object that describes the time dependent
        form of the right hand side.
        @param frm Time dependent formula
        @param bc Boundary conditions
        @param mi MeshInfo to get edge information from
        @param omega Weight for penalty term
        @param tau Sign for symmetric discretization of diffusion
        @param nQuadPts Number of quadrature points per dimension
    */
    LaplaceTimeLf(concepts::Formula& frm, 
                      concepts::BoundaryConditions* bc,
                      MeshInfo& mi,
                      Real omega, Real tau,
                      uint nQuadPts) :
      timestepping::TimeLinearForm(), frm_(frm), bc_(bc) , mi_(mi), 
      omega_(omega), tau_(tau), nQuadPts_(nQuadPts) { }
    /** Application operator. Computes the element contribution.
        @param elmBase Element
        @param em Element matrix
    */
    void operator() (const concepts::Element<Real> &elmBase,
                     concepts::ElementMatrix<Real> &em);
  private:
    const concepts::Formula& frm_;
    const concepts::BoundaryConditions* bc_;
    const MeshInfo& mi_;
    const Real omega_;
    const Real tau_;
    const uint nQuadPts_;
  };

  /** Pure L2 scalar product linear form for DG in 2D. Useful e. g. for
      L2 projections of initial conditions
      @author Christoph Winkelmann, 2003
  */
  class ScalarProductLf : public concepts::LinearForm<Real> {
  public:
    /** Constructor.
        @param frm formula
        @param nQuadPts number of quadrature points per dimension
    */
    ScalarProductLf(concepts::Formula& frm, uint nQuadPts) 
      : frm_(frm), nQuadPts_(nQuadPts) { };
    /** Application operator. Computes the element contribution.
        @param elmBase Element
        @param em Element matrix
    */
    void operator() (const concepts::Element<Real> &elmBase,
                     concepts::ElementMatrix<Real> &em);
  private:
    const concepts::Formula& frm_;
    const uint nQuadPts_;
  };

} // namespace linDG2D

#endif //dgLinearForms_hh
