// triangle element for linear DGFEM

#ifndef DGTriangle_hh
#define DGTriangle_hh

#include "linearFEM/element.hh"
#include "geometry/cell2D.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class ElementGraphics;
}

namespace linDG2D {

  // forward declaration
  class TriangleGraphics;

  using concepts::Real;

  // ************************************************************** Triangle **

  /** Abstract base class for triangular elements in 2D DGFEM.
      @author Christoph Winkelmann, 2003
  */
  class Triangle : public linearFEM::Element {
  public:
    /** Constructor. Creates a T matrix out of \c idx.
        @see concepts::TIndex
        @param cell Cell of the element
        @param nDoF Number of degrees of freedom per element
        @param idx Describes the map local / global dof
    */
    Triangle(const concepts::Triangle2d& cell, uint nDoF, uint idx[])
      : linearFEM::Element(nDoF, idx), cell_(cell) {}
    /// Returns the topological triangle supporting the element
    virtual const concepts::Triangle& support() const {
      return cell_.connector();
    }
    /// Returns the underlying geometrical triangle
    virtual const concepts::Triangle2d& cell() const { return cell_; }
    /** Computes the value of the ith shape function in the point
        (<tt>xi1, xi2</tt>).
        @pre <tt>xi1, xi2 >= 0</tt> and <tt>xi2 <= xi1 <= 1</tt>
        @param i Index of the shape fct. (0, 1, 2)
        @param xi1 First coordinate
        @param xi2 Second coordinate
    */
    virtual Real shapeFct(const uint i, const Real xi1, const Real xi2) const = 0;
    /** Returns the value of a mesh function within this element
        @param coeffs Vector of coefficients representing the mesh function
        @param xi1 first reference element coordinate, in [0,1]
        @param xi2 second reference element coordinate, in [0,xi1]
    */
    virtual Real solution(const concepts::Vector<Real>& coeffs,
                          const Real xi1, const Real xi2) const = 0;
    /// Returns the element graphics object
    virtual const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Cell under this element
    const concepts::Triangle2d& cell_;
    /// Graphics object of this element
    static std::auto_ptr<TriangleGraphics> graphics_;
  };

  // ************************************************************ TriangleP1 **

  /** Triangular element with orthogonal linear shape functions in 2D.
      @author Christoph Winkelmann, 2003
  */
  class TriangleP1 : public Triangle {
  public:
    /** Constructor. Creates a T matrix out of \c idx.
        @see concepts::TIndex
        @param cell Cell of the element
        @param idx Describes the map local / global dof
    */
    TriangleP1(const concepts::Triangle2d& cell, uint idx[])
      : Triangle(cell, 3, idx) {}
    /** Computes the value of the ith shape function in the point
        (<tt>xi1, xi2</tt>).
        @pre <tt>xi1, xi2 >= 0</tt> and <tt>xi2 <= xi1 <= 1</tt>
        @param i Index of the shape fct. (0, 1, 2)
        @param xi1 First coordinate
        @param xi2 Second coordinate
    */
    virtual Real shapeFct(const uint i, const Real xi1, const Real xi2) const;
    /** Returns the gradient of the ith shape function, transformed to real
        world coordinates. The shape functions are linear and the gradient is
        constant. Therefore, no coordinates are necessary.
        @param i Index of the shape fct. (0, 1, 2)
    */
    concepts::Real2d shapeFctGradient(const uint i) const;
    /** Returns the value of a mesh function within this element
        @param coeffs Vector of coefficients representing the mesh function
        @param xi1 first reference element coordinate, in [0,1]
        @param xi2 second reference element coordinate, in [0,xi1]
    */
    virtual Real solution(const concepts::Vector<Real>& coeffs,
                          const Real xi1, const Real xi2) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // ************************************************************ TriangleP0 **

  /** Triangular element with constant shape function in 2D.
      @author Christoph Winkelmann, 2004
  */
  class TriangleP0 : public Triangle {
  public:
    /** Constructor. Creates a T matrix out of \c idx.
        @see concepts::TIndex
        @param cell Cell of the element
        @param idx Describes the map local / global dof
    */
    TriangleP0(const concepts::Triangle2d& cell, uint* idx)
      : Triangle(cell, 1, idx) {}
    /** Computes the value of the ith shape function in the point
        (<tt>xi1, xi2</tt>).
        @pre <tt>xi1, xi2 >= 0</tt> and <tt>xi2 <= xi1 <= 1</tt>
        @param i Index of the shape fct. (0, 1, 2)
        @param xi1 First coordinate
        @param xi2 Second coordinate
    */
    virtual Real shapeFct(const uint i, const Real xi1, const Real xi2) const {
      return 1;
    }
    virtual Real solution(const concepts::Vector<Real>& coeffs,
                          const Real xi1, const Real xi2) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

} // namespace linDG2D

#endif // DGTriangle_hh

