// bilinear forms for intervalls for the Laplacian and the mass matrix

#ifndef bilinearLinFEM1d_hh
#define bilinearLinFEM1d_hh

#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "operator/bilinearForm.hh"
#include "toolbox/formula.hh"
#include "integration/quadrature.hh"

namespace linearFEM {

  using concepts::Real;

  class Line;

  // ************************************************************* Laplace1d **

  /** Discrete equivalent of the Laplacian in 1D for linear FEM. his
      bilinear form computes the stiffness matrix resulting from the
      discretization of the Laplacian (with integration by parts):
      \f[ \int_K \nabla \phi_i \cdot \nabla \phi_j \, dx \f]
      for the element shape functions \f$\phi_i\f$.
      @author Philipp Frauenfelder, 2002
  */
  class Laplace1d : public concepts::BilinearForm<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    virtual Laplace1d* clone() const { return new Laplace1d(); }
  };

  // ************************************************************ Identity1d **

  /** Discrete equivalent for a reaction term in 1D for linear FEM. This
      bilinear form computes the mass matrix resulting from the discretization
      of a reaction term with reaction coefficient 1:
      \f[ \int_K \phi_i \phi_j \, dx \f]
      for the element shape functions \f$\phi_i\f$.
      @author Philipp Frauenfelder, 2002
  */
  class Identity1d : public concepts::BilinearForm<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    virtual Identity1d* clone() const { return new Identity1d(); }
  };

  // *********************************************************** CIdentity1d **

  /** Discrete equivalent for a reaction term in 1D for linear FEM. This
      bilinear form computes the mass matrix resulting from the discretization
      of a reaction term with a varying reaction coefficient c given by
      a formula \c frm (to be specified in the constructor):
      \f[ \int_K c(x) \phi_i \phi_j \, dx \f]
      for the element shape functions \f$\phi_i\f$.
      @author Philipp Frauenfelder, 2002
  */
  class CIdentity1d : public concepts::BilinearForm<Real> {
  public:
    /** Constructor.
	@param frm Formula for the coefficients
	@param gauss_p Number of points for the Gaussian quadrature rule
    */
    CIdentity1d(const concepts::Formula& frm,
		const uint gauss_p = 1) :
      frm_(frm.clone()), quad_(gauss_p) {}
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    virtual CIdentity1d* clone() const {
      return new CIdentity1d(*frm_.get(), quad_.n()); }
  private:
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
    /// Quadrature rule
    const concepts::Quadrature<4> quad_;
  };

} // namespace linearFEM

#endif // bilinearLinFEM1d_hh
