// elements for linear FEM in 2D

#ifndef element2DLin_hh
#define element2DLin_hh

#include "element.hh"
#include "geometry/cell2D.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class ElementGraphics;
}

namespace linearFEM {
  // forward declaration
  class TriangleGraphics;
  class QuadGraphics;

  // ************************************************************** Triangle **

  /** Triangular element with linear shape functions in 2D.
      In 2D, two different elements are possible: triangles and quadrilaterals.
      @author Philipp Frauenfelder, 2002
  */
  class Triangle : public Element {
  public:
    /** Constructor. Creates a T matrix out of \c idx.
	@see concepts::TIndex
	@param cell Cell of the element
	@param idx Describes the map local / global dof
    */
    Triangle(const concepts::Triangle2d& cell, uint idx[]) :
      Element(3, idx), cell_(cell) {}
    /** Constructor. Creates a T matrix out of \c T0 and \c T1
	@see concepts::TMatrix
	@param cell Cell of the element
	@param T0 Contains T columns
	@param T1 Contains T columns
    */
    Triangle(const concepts::Triangle2d& cell, concepts::TColumn<Real>* T0,
	     concepts::TColumn<Real>* T1 = 0) :
      Element(T0, T1), cell_(cell) {}
    virtual const concepts::Triangle& support() const {
      return cell_.connector(); }
    virtual const concepts::Triangle2d& cell() const { return cell_; }
    /** Computes the value of the ith shape function in the point
	(<tt>xi1, xi2</tt>).
	@pre <tt>xi1, xi2 >= 0</tt> and <tt>xi2 <= xi1 <= 1</tt>
	@param i Index of the shape fct. (0, 1, 2)
	@param xi1 First coordinate
	@param xi2 Second coordinate
    */
    Real shapefct(const uint i, const Real xi1, const Real xi2) const;
    /** Returns the gradient of the ith shape function. The shape functions
	are linear and the gradient is constant. Therefore, no coordinates
	are necessary.
	@param i Index of the shape fct. (0, 1, 2)
    */
    concepts::Real2d shapefctD(const uint i) const;
    virtual const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Cell of the element
    const concepts::Triangle2d& cell_;
    static std::auto_ptr<TriangleGraphics> graphics_;
  };

  // ****************************************************************** Quad **

  /** Quadrilateral element with bilinear shape functions in 2D.
      In 2D, two different elements are possible: triangles and quadrilaterals.
      @author Philipp Frauenfelder, 2002
  */
  class Quad : public Element {
  public:
    /** Constructor. Creates a T matrix out of \c idx.
	@see concepts::TIndex
	@param cell Cell of the element
	@param idx Describes the map local / global dof
    */
    Quad(const concepts::Quad2d& cell, uint idx[]) :
      Element(4, idx), cell_(cell) {}
    /** Constructor. Creates a T matrix out of \c T0 and \c T1
	@see concepts::TMatrix
	@param cell Cell of the element
	@param T0 Contains T columns
	@param T1 Contains T columns
    */
    Quad(const concepts::Quad2d& cell, concepts::TColumn<Real>* T0,
	 concepts::TColumn<Real>* T1 = 0) : Element(T0, T1), cell_(cell) {}
    virtual const concepts::Quad& support() const { return cell_.connector(); }
    virtual const concepts::Quad2d& cell() const { return cell_; }
    /** Evaluates the shape functions in the given one dimensional
	points. The points in \c x have to be between -1 and 1 and are
	tensorised to get two dimensional points.
	@pre res has to allocated, size: 4*n*n
	@param res Values of the shapefunctions with index k in the points
	       (x_i, x_j) with j running fastest and k running slowest
	@param x Points in 1D where the shape functions have to be evaluated.
	       Size: \c n
        @param n Number of points in \c x
    */
    void evaluate(Real* res, const Real* x, const uint n) const;
    /** Evaluates the derivatives of the shape functions in the given
	one dimensional points. The points in \c x have to be between
	-1 and 1 and are tensorised to get two dimensional points.
	@pre res has to allocated, size: 4*n*n
	@param res Values of the shapefunctions with index k in the points
	       (x_i, x_j) with j running fastest and k running slowest
	@param x Points in 1D where the shape functions have to be evaluated.
	       Size: \c n
        @param n Number of points in \c x
    */
    void evaluateD(concepts::Real2d* res, const Real* x, const uint n) const;
    virtual const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Cell of the element
    const concepts::Quad2d& cell_;
    static std::auto_ptr<QuadGraphics> graphics_;
  };

} // namespace linearFEM

#endif // element2DLin_hh
