// linear form for 2D linear FEM (quads and triangles)

#include <cmath>
#include <sstream>
#include "linearForm2D.hh"
#include "element2D.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "integration/quadrature.hh"

namespace linearFEM {

  // *************************************************************** Riesz2d **

  const Real Riesz2d::weights_[4] = { -9./32., 25./96., 25./96., 25./96. };

  const concepts::Real2d Riesz2d::coord_[4] =
    { concepts::Real2d(2./3., 1./3.), concepts::Real2d(4./5., 1./5.),
      concepts::Real2d(2./5., 1./5.), concepts::Real2d(4./5., 3./5.) };

  void Riesz2d::operator()(const concepts::Element<Real>& elm,
			   concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
    const linearFEM::Quad* eQ = dynamic_cast<const linearFEM::Quad*>(&elm);
    if (eQ) { this->operator()(*eQ, em); return; }
    const linearFEM::Triangle* eT =
      dynamic_cast<const linearFEM::Triangle*>(&elm);
    if (eT) { this->operator()(*eT, em); return; }
    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void Riesz2d::operator()(const linearFEM::Quad& elm,
			   concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    /* Get the vertices of the element. They are needed to compute the
       Jacobian of the element map.
      */
    concepts::Real2d y0(elm.cell().vertex(0));
    concepts::Real2d y1(elm.cell().vertex(1));
    concepts::Real2d y2(elm.cell().vertex(2));
    concepts::Real2d y3(elm.cell().vertex(3));

    concepts::Real2d p(y1 - y0);
    concepts::Real2d q(y3 - y0);
    concepts::Real2d r(y1 - y0 + y3 - y2);

    concepts::Real2d s, t;
    /* Get the quadrature points.
     */
    const concepts::Quadrature<4> quadrature(3);
    const Real* abscissas = quadrature.abscissas();
    const Real* weights = quadrature.weights();
    const uint qn = quadrature.n();
    /* Get the values of the shape functions in the quadrature points.
     */
    shpfct_.resize(n*qn*qn);
    elm.evaluate((concepts::Real*)shpfct_, abscissas, qn);
    /* Numerical integration: loop over all quadrature points and compute
       the Jacobian. Then, loop over all reference element shape functions
       and add the contributions up.
      */
    for (uint qi = 0; qi < qn; ++qi) {
      for (uint qj = 0; qj < qn; ++qj) {
	s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	const Real dx = std::fabs(s.cross(t)) * weights[qi] * weights[qj] /
	  4.0;
	const Real f = (*frm_)(elm.cell().chi( (1.0+abscissas[qi])/2.0,
					       (1.0+abscissas[qj])/2.0 ));
	for (uint i = 0; i < n; ++i)
	  em(i, 0) += f * dx * shpfct_[i*qn*qn+qi*qn+qj];
      }
    }
  }

  void Riesz2d::operator()(const linearFEM::Triangle& elm,
			   concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    const Real jac = elm.cell().jacobianDeterminant(0.0, 0.0);

    for (uint k = 0; k < 4; ++k) { // integration
      concepts::Real2d x(concepts::Triangle2d::duffyInv
			 (concepts::Real2d(coord_[k][0], coord_[k][1])));
      const Real w = weights_[k]*jac*
	(*frm_)(elm.cell().chi(x[0], x[1]));
      for (uint i = 0; i < n; ++i) { // Ansatz fcts
	em(i, 0) += w * elm.shapefct(i,coord_[k][0], coord_[k][1]);
      } // for i
    } // for k
  }

} // namespace linearFEM
