// graphics for tetrahedra

#include <sstream>
#include "tetGraphics.hh"
#include "element3D.hh"
#include "space/postProcess.hh"
#include "graphics/meshGnuplot.hh"
#include "graphics/dataGnuplot.hh"
#include "graphics/dataDX.hh"

namespace linearFEM {

  // *************************************************** TetrahedronGraphics **

  void TetrahedronGraphics::operator()(const concepts::Element<Real>& elm,
				enum graphicsType type,
				concepts::CellPostprocess<Real>& post) const {
    const Tetrahedron* e = dynamic_cast<const Tetrahedron*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    switch(type) {

    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_MATLAB:
    case MESH_DX:
      post(e->cell()); break;

    case DATA_DX_COUNTS: {
      graphics::DataDXCounts* counter =
	dynamic_cast<graphics::DataDXCounts*>(&post);
      conceptsAssert(counter, concepts::Assertion());
      counter->add(4);
      break;
    }

    case DATA_DX_COORDINATES: {
      graphics::DataDXCoordinates* coord = 
	dynamic_cast<graphics::DataDXCoordinates*>(&post);
      conceptsAssert(coord, concepts::Assertion());

      for (uint i = 0; i < 4; ++i) {
	concepts::Real3d x(e->cell().vertex(i));
	coord->write(x[0]);
	coord->write(x[1]);
	coord->write(x[2]);
      }
      coord->add(1);

      break;
    }

    case DATA_DX_CONNECTIONS: {
      graphics::DataDXConnections* conn = 
	dynamic_cast<graphics::DataDXConnections*>(&post);
      conceptsAssert(conn, concepts::Assertion());
      uint l = conn->counts();
      uint I[8] = { 0, 1, 2, 2, 3, 3, 3, 3 };
      for (uint i = 0; i < 8; ++i)
	conn->write(l+I[i]);
      conn->add(4);
      break;
    }

    case DATA_DX_DATA: {
      graphics::DataDXData* data = 
	dynamic_cast<graphics::DataDXData*>(&post);
      conceptsAssert(data, concepts::Assertion());

      const concepts::Formula* formula = data->formula();
      Real res = 0;
      const concepts::Array<Real>& coeff = data->coeff();
      for (uint i = 0; i < 4; ++i) {
	if (! formula) {
	  res = coeff[i];
	} else {
	  concepts::Real3d x(e->cell().vertex(i));
	  res = (*formula)(x);
	}
	data->write(res);
      }

      break;
    }

    default:
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    }
  }

  std::ostream& TetrahedronGraphics::info(std::ostream& os) const {
    return os << "TetrahedronGraphics()";
  }

} // namespace linearFEM
