// fabric class for PETSc

#ifndef petscFabric_hh
#define petscFabric_hh

#include <list>
#include "operator/PETSc.hh"
#include "operator/solverFabric.hh"

namespace concepts {

  // *********************************************************** PETScFabric **

  /** Fabric class for \c PETSc solvers. Creates a \c PETScMat
      if Operator is not a \c PETScMat and a preconditioner is desired.
      This may have severe disadvantages:
      <ul>
      <li>It is always memory inefficent.</li>
      <li>It may be O(n^2) slow, if Operator is not Matrix, LiCo, LiCoI or
      Compose</li>
      <li>It may cause a segmentation fault if the fabric goes out of scope
      before any of the solvers created like this. Make sure this does not
      happen.</li>
      </ul>
      @author Christoph Winkelmann, 2004
  */
  class PETScFabric : public SolverFabric<Real> {
  public:
    /** Constructor.
        @param maxeps Maximal residual
        @param ksptype Type of Krylov Subspace Solver
        @param pctype Type of preconditioner
    */
    PETScFabric(const Real maxeps, const std::string ksptype, 
                const std::string pctype = std::string("none"))
      : maxeps_(maxeps), ksptype_(ksptype), pctype_(pctype) {}
    virtual Operator<Real>* operator()(Operator<Real>& matrix) {
      if (pctype_ == "none" || dynamic_cast<concepts::PETScMat*>(&matrix)) {
        return new PETSc(matrix, maxeps_, ksptype_, pctype_);
      } else {
        concepts::SparseMatrix<Real>* sparsemat =
          dynamic_cast<concepts::SparseMatrix<Real>*>(&matrix);
        if (sparsemat) {
          petscmat_.push_front(new concepts::PETScMat(*sparsemat));
        } else {
          sparsemat = new concepts::SparseMatrix<Real>(matrix);
          petscmat_.push_front(new concepts::PETScMat(*sparsemat));
          delete sparsemat;
        }
        return new PETSc(*(petscmat_.front()), maxeps_, ksptype_, pctype_);
      }
    }
    ~PETScFabric() {
      while (!petscmat_.empty()) {
        delete petscmat_.front();
        petscmat_.pop_front();
      }
    }
  protected:
    virtual std::ostream& info(std::ostream& os) const {
      return os << "PETScFabric(maxeps = " << maxeps_ << ", ksptype = "
                << ksptype_ << ", pctype = " << pctype_ << ")";
    }
  private:
    /// Maximal residual
    const Real maxeps_;
    /// Type of Krylov Subspace Solver
    const std::string ksptype_;
    /// Type of preconditioner
    const std::string pctype_;
    /// Created auxiliary PETSc Matrices
    std::list<concepts::PETScMat*> petscmat_;
  };

} // concepts

#endif // petscFabric_hh
