/* Bilinear Form
 */

#ifndef BilinearForm_hh
#define BilinearForm_hh

#include <memory>
#include "basics/outputOperator.hh"
#include "basics/cloneable.hh"
#include "space/elementPairs.hh"

namespace concepts {

  // forward declaration
  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  // ********************************************************** BilinearForm **

  /** Abstract function class to evaluate a bilinear form. The bilinear form
      is evaluated for all the shape functions on two elements, the resulting
      matrix is returned.
      @ingroup bilinear
  */
  template<class F>
  class BilinearForm : public Cloneable, virtual public OutputOperator {
  public:
    /** Evaluates the bilinear form for all shape functions on \c elmX
	and \c elmY and stores the result in the matrix \c em.

	@post The returned matrix \c em has the correct size.
	@param elmX Left element
	@param elmY Right element
	@param em Return element matrix
    */
    virtual void operator()(const Element<F>& elmX, const Element<F>& elmY,
			    ElementMatrix<F>& em) = 0;
    /** Evaluates the bilinear form for all shape functions on \c elmX
	and \c elmY and stores the result in the matrix \c em. If this
	method is not reimplemented in a derived class, the default
	behaviour is to call the application operator without \c ep.

	@post The returned matrix \c em has the correct size.
	@param elmX Left element
	@param elmY Right element
	@param em Return element matrix
	@param ep Element pair holding more information on the pair
	       \c elmX and \c elmY
    */
    virtual void operator()(const Element<F>& elmX, const Element<F>& elmY,
			    ElementMatrix<F>& em, const ElementPair<F>& ep) {
      operator()(elmX, elmY, em); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // ****************************************************** BilinearFormLiCo **

  /** A linear combination of bilinear forms. The resulting element matrices
      of the bilinear forms are added into each other with the correct
      coefficient and can then be assembled.

      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class BilinearFormLiCo : public BilinearForm<F> {
  public:
    /** Constructor
	@param bfA First bilinear form
	@param bfB Second bilinear form
	@param cA Coefficient of bfA
	@param cB Coefficient of bfB
    */
    BilinearFormLiCo(BilinearForm<F>& bfA,
		     BilinearForm<F>& bfB,
		     const F cA = 1.0, const F cB = 1.0) :
      bfAptr_(0), bfBptr_(0),
      bfA_(&bfA), bfB_(&bfB), cA_(cA), cB_(cB), em_() {}
    /// Copy constructor. This copy constructor implements a deep copy.
    BilinearFormLiCo(const BilinearFormLiCo& b);
    virtual ~BilinearFormLiCo() {}
    virtual void operator()(const Element<F>& elmX,
			    const Element<F>& elmY, 
			    ElementMatrix<F>& em);
    virtual BilinearFormLiCo<F>* clone() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    std::auto_ptr<BilinearForm<F> > bfAptr_, bfBptr_;
    /// Bilinear form
    BilinearForm<F>* bfA_, * bfB_;
    /// Coefficient of \c bfA
    const F cA_;
    /// Coefficient of \c bfB
    const F cB_;
    /// Local element matrix used as temporary storage
    ElementMatrix<F> em_;
  };

} // namespace concepts

#endif // BilinearForm_hh
