/* A dense matrix
 */

#include <memory>
#include <cstring>
#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "toolbox/scannerConnectors.hh"
#include "bilinearForm.hh"
#include "denseMatrix.hh"

#include "basics/debug.hh"

// debugging
#define DenseMatrixConstr_D 0

namespace concepts {

  // *********************************************************** DenseMatrix **

  template<class F>
  DenseMatrix<F>::DenseMatrix(const Space<F>& spcX, const Space<F>& spcY,
			      BilinearForm<F>& bf) :
    ElementMatrix<F>(spcX.dim(), spcY.dim()), spcX_(spcX), spcY_(spcY) {
    zeros();
    Matrix<F>::assembly(*this, spcX, spcY, bf);
  }

  template<class F>
  DenseMatrix<F>::DenseMatrix(const Space<F>& spc, BilinearForm<F>& bf) :
    ElementMatrix<F>(spc.dim(), spc.dim()), spcX_(spc), spcY_(spc) {
    zeros();
    Matrix<F>::assembly(*this, spc, bf);
  }

  template<class F>
  DenseMatrix<F>::DenseMatrix(const Space<F>& spcX, const Space<F>& spcY) : 
    ElementMatrix<F>(spcX.dim(), spcY.dim()), spcX_(spcX), spcY_(spcY) {
    zeros();
    DEBUGL(DenseMatrixConstr_D, "done.");
  }

  template<class F>
  DenseMatrix<F>::DenseMatrix(const LiCo<F>& L) :
    ElementMatrix<F>(L.spaceX().dim(), L.spaceY().dim()),
    spcX_(L.spaceX()), spcY_(L.spaceY()) {
    zeros();
#ifdef DEBUG
    bool test =
#endif
      L.collapse(*this);
    conceptsAssert(test, Assertion());
  }

  template<class F>
  DenseMatrix<F>::DenseMatrix(const Compose<F>& L) :
    ElementMatrix<F>(L.spaceX().dim(), L.spaceY().dim()),
    spcX_(L.spaceX()), spcY_(L.spaceY()) {
    DEBUGL(DenseMatrixConstr_D, "L = " << L);
    zeros();
#ifdef DEBUG
    bool test =
#endif
      L.collapse(*this);
    conceptsAssert(test, Assertion());
  }

  template<class F>
  DenseMatrix<F>::DenseMatrix(const DenseMatrix<F>& m) :
    ElementMatrix<F>(m.spaceX().dim(), m.spaceY().dim()),
    spcX_(m.spaceX()), spcY_(m.spaceY()) {
    std::memcpy((F*)(*this), (const F*)m,
		spaceX().dim()*spaceY().dim()*sizeof(F));
  }

  template<class F>
  DenseMatrix<F>::DenseMatrix(const Matrix<F>& m, bool t) :
    ElementMatrix<F>(t ? m.spaceY().dim() : m.spaceX().dim(),
		     t ? m.spaceX().dim() : m.spaceY().dim()),
    spcX_(t ? m.spaceY() : m.spaceX()), spcY_(t ? m.spaceX() : m.spaceY()) {
    if (t == false) {
      F* d = (F*)(*this);
      for (uint i = 0; i < m.spaceX().dim(); ++i)
	for (uint j = 0; j < m.spaceY().dim(); ++j)
	  *d++ = m(i,j);
    } else {
      F* d = (F*)(*this);
      for (uint j = 0; j < m.spaceY().dim(); ++j)
	for (uint i = 0; i < m.spaceX().dim(); ++i)
	  *d++ = m(i,j);
    }
  }

  template<class F>
  void DenseMatrix<F>::operator()(const Function<F>& fncY, Function<F>& fncX) {
    conceptsAssert(ElementMatrix<F>::m() == fncX.space().dim(), Assertion());
    conceptsAssert(ElementMatrix<F>::n() == fncY.space().dim(), Assertion());
    const Vector<F>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<F>*>(&fncY))) {
      Vector<F>* fncXA;
      if ((fncXA = dynamic_cast<Vector<F>*>(&fncX))) {
	(*this)(*fncYA, *fncXA);
	return;
      }
    }
    const F* m = (const F*)(*this);
    for (uint i = 0; i < ElementMatrix<F>::m(); ++i) {
      F fi = 0.0;
      for (uint j = 0; j < ElementMatrix<F>::n(); ++j)
	fi += *m++ * fncY(j);
      fncX(i) = fi;
    }
  }

  template<class F>
  void DenseMatrix<F>::operator()(const Vector<F>& fncY, Vector<F>& fncX) {
    conceptsAssert(ElementMatrix<F>::m() == fncX.space().dim(), Assertion());
    conceptsAssert(ElementMatrix<F>::n() == fncY.space().dim(), Assertion());
    const F* m = (const F*)(*this);
    F* x(fncX);	
    for (uint i = 0; i < ElementMatrix<F>::m(); ++i) {
      F* y(fncY);	
      F fi = 0.0;
      for (uint j = 0; j < ElementMatrix<F>::n(); ++j)
	fi += *m++ * *y++;
      *x++ = fi;
    }
  }

  template<class F>
  void DenseMatrix<F>::transpMult(const Vector<F>& fncY, Vector<F>& fncX) {
    conceptsAssert(ElementMatrix<F>::m() == fncY.space().dim(), Assertion());
    conceptsAssert(ElementMatrix<F>::n() == fncX.space().dim(), Assertion());
    const F* m = (const F*)(*this);
    const F* y(fncY);
    fncX = 0.0;
    for (uint i = 0; i < ElementMatrix<F>::m(); ++i, ++y) {
      F* x(fncX);
      for (uint j = 0; j < ElementMatrix<F>::n(); ++j)
	*x++ += *m++ * *y;
    }
  }

  template<class F>
  void DenseMatrix<F>::addInto(Matrix<F>& dest, const F fact) const {
    if (fact != 0.0) {
      for (uint i = 0; i < spaceX().dim(); ++i)
	for (uint j = 0; j < spaceY().dim(); ++j) {
	  const F a = operator()(i,j);
	  if (a != 0.0) dest(i,j) += a*fact;
	}
    }
  }

  template<class F>
  std::ostream& DenseMatrix<F>::info(std::ostream& os) const {
    const int N = n();
    const int M = m();
    os << "DenseMatrix(" << M << 'x' << N << ", [" << std::endl;
    for (int i = 0; i < M ; ++i) {
      for (int j = 0; j < N; ++j) {
  	os << (*this)(i,j) << ((j == N-1) ? "" : " ");
      } // for j
      os << ((i == M-1) ? "" : "\n");
    } // for i
    return os << "])";
  }

} // namespace concepts
