/* A dense matrix
 */

#ifndef denseMatrix_hh
#define denseMatrix_hh

#include "space/element.hh"
#include "compositions.hh"
#include "matrix.hh"

namespace concepts {

  // forward declaration
  template<class F>
  class Space;

  template<class F>
  class BilinearForm;

  // *********************************************************** DenseMatrix **

  /** Dense matrix. The matrix has the size \f$m \times n\f$ where m is the
      dimension of the image space (<tt>spaceX</tt>) and n is the dimension
      of the source space (<tt>spaceY</tt>).

      The matrix is setup and assembled in the constructor. It calls the
      bilinear form on every element of the space and uses the T matrices
      of the elements to assemble the element matrices into the global
      matrix.

      There are quite a few solver which can be used to solve the system.

      @see TMatrixBase
      @see CG
      @see GMRes
  */
  template<class F>
  class DenseMatrix : public Matrix<F>, public ElementMatrix<F> {
  public:
    /** Constructor. Computes the global matrix by assembling the
	element matrices.

	This constructor features a double loop over the elements of the
	image and the source space. On each combination, the bilinear form is
	called.

	Using this constructor, if \c spcX != \c spcY or if you have
	local matrices which express the interaction of the two
	elements. You can force this constructor to execute the double
	loop in such a way that only for diagonal combinations of the
	elements in both space the integration and assembling is executed.
	Use \c single and set it to \c true.

	In non-symmetric FEM (eg. DGFEM), one has to solve
	A<sup>T</sup> u = f. This constructor computes A and
	not A<sup>T</sup>.

	@param spcX Image space
	@param spcY Source space
	@param bf Bilinear form
    */
    DenseMatrix(const Space<F>& spcX, const Space<F>& spcY,
		BilinearForm<F>& bf);
    /** Constructor. Computes the global matrix by assembling the
	element matrices.

	This constructor features a single loop over the elements of the
	space. On each element, the bilinear form is called.

	Use the constructor, if spcX == spcY and there is no special element
	interaction. This is the case for standard FEM.

	@param spc Image and source space
	@param bf Bilinear form
    */
    DenseMatrix(const Space<F>& spc, BilinearForm<F>& bf);
    /** Constructor. Computes the global matrix by assembling the
	element matrices.

	This creates an emtpy matrix.
	@param spcX Image space
	@param spcY Source space
    */
    DenseMatrix(const Space<F>& spcX, const Space<F>& spcY);
    /** Constructor. Reads the matrix from a file.
	@param spcX Image space
	@param spcY Source space
	@param fname Filename
    */
    DenseMatrix(const Space<F>& spcX, const Space<F>& spcY, const char* fname);
    /** Constructor. Evaluates a linear combination and stores the result
	in the matrix.
    */
    DenseMatrix(const LiCo<F>& L);
    /** Constructor. Evaluates a composition and stores the result
	in the matrix.
    */
    DenseMatrix(const Compose<F>& L);
    /// Copy constructor
    DenseMatrix(const DenseMatrix<F>& m);
    /** Copy constructor. If \c t is set to true, the matrix is
	transposed during the copy process
    */
    DenseMatrix(const Matrix<F>& m, bool t = false);

    virtual void operator()(const Function<F>& fncY, Function<F>& fncX);
    void operator()(const Vector<F>& fncY, Vector<F>& fncX);

    virtual void transpMult(const Vector<F>& fncY, Vector<F>& fncX);

    virtual F operator()(const uint i, const uint j) const {
      return ElementMatrix<F>::operator()(i,j);
    }

    virtual F& operator()(const uint i, const uint j) {
      return ElementMatrix<F>::operator()(i,j);
    }

    virtual const Space<F>& spaceX() const { return spcX_; }
    virtual const Space<F>& spaceY() const { return spcY_; }

    /** This matrix is added into to the given matrix.
	@param dest Matrix into which this matrix should be added.
	@param fact Factor by which this matrix should be multiplied.
    */
    void addInto(Matrix<F>& dest, const F fact) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Image space
    const Space<F>& spcX_;
    /// Source space
    const Space<F>& spcY_;
  };

} // namespace concepts

#endif // senseMatrix_hh
