// diagonal matrix and solver for it

#include "diagonal.hh"

// debugging
#include "basics/debug.hh"

#define DEBUG_DIAGONAL_SOLVER 0

namespace concepts {

  // ******************************************************** DiagonalMatrix **

  template<typename F>
  DiagonalMatrix<F>::DiagonalMatrix(const Space<F>& spc) :
    spc_(spc), entries_(spc.dim(), 0), dummy_(0) {}

  template<typename F>
  DiagonalMatrix<F>::DiagonalMatrix(const Space<F>& spc,
				    const Array<F> entries) :
    spc_(spc), entries_(entries), dummy_(0) {
    conceptsAssert(spc.dim() == entries.size(), Assertion());
  }

  template<typename F>
  DiagonalMatrix<F>::DiagonalMatrix(const Matrix<F>& matrix) :
    spc_(matrix.spaceX()), entries_(spc_.dim()), dummy_(0) {
    conceptsAssert(matrix.spaceX().dim() == matrix.spaceY().dim(),
		   Assertion());
    for (uint j = 0; j < spc_.dim(); ++j)
      entries_[j] = matrix(j, j);
  }

  template<typename F>
  DiagonalMatrix<F>::~DiagonalMatrix() {}

  template<typename F>
  void DiagonalMatrix<F>::operator()(const Function<F>& fncY,
				     Function<F>& fncX) {
    conceptsAssert(fncY.space().dim() == spc_.dim(), Assertion());
    conceptsAssert(fncX.space().dim() == spc_.dim(), Assertion());
    const Vector<F>* vecY = dynamic_cast<const Vector<F>*>(&fncY);
    Vector<F>* vecX = dynamic_cast<Vector<F>*>(&fncX);
    if ((vecY != 0) && (vecY != 0)) {
      transpMult(*vecY, *vecX);
      return;
    }

    // dynamic_casts did not work
    for (uint i = 0; i < spc_.dim(); ++i)
      fncX(i) = fncY(i) * entries_[i];
  }

  template<typename F>
  F DiagonalMatrix<F>::operator()(const uint i, const uint j) const {
    return i == j ? entries_[i] : 0.0;
  }

  template<typename F>
  F& DiagonalMatrix<F>::operator()(const uint i, const uint j) {
    return i == j ? entries_[i] : dummy_;
  }

  template<typename F>
  void DiagonalMatrix<F>::transpMult(const Vector<F>& fncY, Vector<F>& fncX) {
    const F* vecYdata(fncY);
    F* vecXdata(fncX);
    const F* entryData(entries_);
    for (uint i = 0; i < spc_.dim(); ++i)
      *vecXdata++ = *vecYdata++ * *entryData++;
  }

  template<typename F>
  std::ostream& DiagonalMatrix<F>::info(std::ostream& os) const {
    return os << "DiagonalMatrix(" << spc_ << ", entries = " << entries_
	      << ')';
  }

  // ******************************************************** DiagonalSolver **

  template <typename F>
  DiagonalSolver<F>::DiagonalSolver(const concepts::DiagonalMatrix<F>& m) :
    inverse_(m.spaceX()) {
    for (uint i=0; i<m.spaceX().dim(); ++i) {
      if (m(i,i) == 0.0) inverse_(i,i) = 1;
      else inverse_(i,i) = 1.0 / m(i,i);
    }
    DEBUGL(DEBUG_DIAGONAL_SOLVER, "inverse_ = " << inverse_);
  }

  template <typename F>
  DiagonalSolver<F>::~DiagonalSolver() {}

  template <typename F>
  std::ostream& DiagonalSolver<F>::info(std::ostream& os) const {
    return os << "DiagonalSolver(entries = " << inverse_ << ')';
  }

} // namespace concepts
