// diagonal matrix and solver for it

#ifndef diagonal_hh
#define diagonal_hh

#include "matrix.hh"

namespace concepts {

  // ******************************************************** DiagonalMatrix **

  /** Diagonal matrix. A diagonal matrix must be square. Any access to off
      diagonal entries does not lead to an error but is simply discarded.
      @author Philipp Frauenfelder, 2002
      @see DiagonalSolver
  */
  template<typename F>
  class DiagonalMatrix : public Matrix<F> {
  public:
    /// Constructor. Creates a diagonal matrix with all entries set to 0.
    DiagonalMatrix(const Space<F>& spc);
    /** Constructor. Creates a diagonal matrix with \c entries on the
	diagonal.
    */
    DiagonalMatrix(const Space<F>& spc, const Array<F> entries);
    /** Constructor. Creates a diagonal matrix by taking the diagonal of
	<tt>matrix</tt>.
    */
    DiagonalMatrix(const Matrix<F>& matrix);
    virtual ~DiagonalMatrix();

    virtual void operator()(const Function<F>& fncY, Function<F>& fncX);

    virtual const Space<F>& spaceX() const { return spc_; }
    virtual const Space<F>& spaceY() const { return spc_; }

    virtual F operator()(const uint i, const uint j) const;
    virtual F& operator()(const uint i, const uint j);

    virtual void transpMult(const Vector<F>& fncY, Vector<F>& fncX);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    const Space<F>& spc_;
    /// Diagonal entries of the matrix
    Array<F> entries_;
    /// Dummy entry for the access to the off diagonal entries
    F dummy_;
  };

  // ******************************************************** DiagonalSolver **

  /** A solver for diagonal matrices. Inverts a diagonal matrix and stores this
      inverse in order to apply it when applying the solver.
      @author Christoph Winkelmann, 2002
      @ingroup solver
  */
  template <typename F>
  class DiagonalSolver : public concepts::Operator<F> {
  public:
    /** Constructor.
	@param m Diagonal matrix to be inverted
    */
    DiagonalSolver(const concepts::DiagonalMatrix<F>& m);
    virtual ~DiagonalSolver();

    virtual void operator() (const concepts::Function<F>& rhs,
			     concepts::Function<F>& sol) {
      inverse_(rhs, sol); };
    const concepts::Space<F>& spaceX() const { return inverse_.spaceY(); }
    const concepts::Space<F>& spaceY() const { return inverse_.spaceX(); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    concepts::DiagonalMatrix<F> inverse_;
  };

} // namespace concepts

#endif // diagonal_hh
