/* Iterative linear solver: GMRes
   Implementation for Cmplx
 */

#include "operator/gmres.cc"

// debugging
#define GMResAppl_D 0

namespace concepts {

  // ***************************************************************** GMRes **

  template<>
  void GMRes<Cmplx>::operator()(const Vector<Cmplx>& fncY, Vector<Cmplx>& fncX)
    throw(NoConvergence, MissingFeature) {

    DEBUGL(GMResAppl_D, "solving " << A_ << " with RHS " << fncY);
    if (W_ == 0) {
      Real           l2Y = (relres_) ? fncY.l2() : 1.0;
      Real           maxepsY = maxeps_ * l2Y;
      Cmplx          eps = maxepsY + 1.0;
      R*             r0 = new R(2);
      Vector<Cmplx>* v = new Vector<Cmplx>(fncY.space());

      eps_ = std::abs(eps);
      it_ = 0;

      while (!(eps_ < maxepsY  ||  (maxit_ > 0 && it_ > maxit_))) {
	A_(fncX, *v);
	*v -= fncY;

	eps = eps_ = v->l2(); *v *= -1.0 / eps_;

	R*   r = r0;
	uint step = 1;

	while (1) {
	  it_++;

	  if (r->v) {
	    Vector<Cmplx>* vv = r->v; r->v = v; v = vv;
	  } else {
	    r->v = v; v = new Vector<Cmplx>(fncY.space());
	  }

	  A_(*r->v, *v);

          Cmplx* h = r->h;
	  for (R* rr = r; rr; rr = rr->left) {
	    h++;
	    Cmplx* rrv = (Cmplx*)(*rr->v);
	    Cmplx* vv  = (Cmplx*)(*v);
	    for (uint i = 0; i < rr->v->n(); i++) {
	      *h += std::conj(*rrv++) * *vv++;
	    }
	  }

	  h = r->h;
	  for (R* rr = r; rr; rr = rr->left) v->add(*rr->v, -*++h);
	  h = r->h + step;
	  for (R* rr = r0; rr != r; rr = rr->right) {
	    Real h1r = h->real(), h1i = h->imag();  h--;
	    Real h0r = h->real(), h0i = h->imag();
	    Real sinr = rr->sin.real(), sini = rr->sin.imag();
	    h[0] = Cmplx(rr->cos * h0r - sinr * h1r + sini * h1i,
			 rr->cos * h0i - sinr * h1i - sini * h1r);
	    h[1] = Cmplx(sinr * h0r + sini * h0i + rr->cos * h1r,
			 sinr * h0i - sini * h0r + rr->cos * h1i);
	  }

	  Real  h0 = v->l2();
	  Cmplx h1(r->h[1]);
	  Real  h2 = 1.0 / sqrt(std::norm(h1) + h0 * h0);

	  r->cos = std::abs(h1) * h2;
	  r->sin = h0 / h1 * r->cos;
	  r->h[0] = eps * r->cos;
	  r->h[1] = Cmplx(r->sin.real() * h0 + r->cos * h1.real(),
			  -r->sin.imag() * h0  + r->cos * h1.imag());

	  eps *= -r->sin;
	  eps_ = std::abs(eps);
	  DEBUGL(GMResAppl_D, "residual = " << eps_);
	  if (eps_ < maxepsY || step == rs_ || it_ > maxit_)
	    break;

	  *v *= 1.0 / h0;

	  ++step;

	  r = r->right ? r->right : new R(step + 1, r);
	}

	Cmplx* h0 = r->h;
	Cmplx  a = h0[0] / h0[1];

	for(uint j = 2; !(j > step); j++)  h0[j] *= a;
	fncX.add(*r->v, a);

	++h0;
	while ((r = r->left)) {
	  Cmplx* h1 = r->h;
	  Cmplx  a = (h1[0] - h0[1]) / h1[1];

	  for(uint j = 2; j < step; j++)  h1[j] = a * h1[j] + h0[j];
	  fncX.add(*r->v, a);

	  h0 = r->h + 1;  --step;
	}
      }

      delete v;
      while (r0) { R* foo = r0; r0 = r0->right; delete foo; }

      eps_ /= l2Y;
      if (fabs(eps_) > maxeps_) {
	std::stringstream errMsg;
	errMsg << "it: " << it_ << ", eps: " << eps_
	       << ", wanted: " << maxeps_ << std::ends;
	throw conceptsException(NoConvergence(errMsg.str()));
      }
    } else
      throw conceptsException(MissingFeature("operator W not supported"));
  }

  template class GMRes<Cmplx>;

} // namespace concepts
