/* SparseMatrix
 * sparse matrices
 */

#include <iomanip>
#include <sstream>
#include <cstring>
#include "hashedSMatrix.hh"
#include "matrix.hh"
#include "denseMatrix.hh"
#include "toolbox/array.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"

// debugging
#define HashedSparseMatrixInfo_D 0
#define HSMpermute_D 0
#define HSMdropEntry_D 0

namespace concepts {

  // **************************************************** HashedSparseMatrix **

  template<class T>
  HashedSparseMatrix<T>::HashedSparseMatrix(uint r, uint c, uint h) :
    pool((c >> 1) + 1) {
    nofRows = r; nofCols = c;
    hashMsk = (1 << (hashBits = h)) - 1;
    matrix = new Value*[r << h];
    std::memset(matrix, 0, (r << h) * sizeof(Value*));
    _used = 0;
  }

  template<class T>
  HashedSparseMatrix<T>::~HashedSparseMatrix() {
    delete[] matrix;
  }

  template<class T>
  T& HashedSparseMatrix<T>::operator ()(const uint r, const uint c) {
    conceptsAssert(r < nofRows, Assertion());
    conceptsAssert(c < nofCols, Assertion());

    Value** m = matrix + (r << hashBits | c & hashMsk);

    Value* v = *m; 
    for (; v != NULL && v->idx != c; v = v->lnk);

    if (v == NULL) {
      (v = *m = new(pool) Value(*m, c))->val = 0.0;
      ++_used;
    }
    return v->val;
  }

  template<class T>
  T HashedSparseMatrix<T>::operator ()(const uint r, const uint c) const {
    conceptsAssert(r < nofRows, Assertion());
    conceptsAssert(c < nofCols, Assertion());

    Value** m = matrix + (r << hashBits | c & hashMsk);

    Value* v = *m; 
    for (; v != NULL && v->idx != c; v = v->lnk);

    if (v == NULL) {
      return 0.0;
    }
    return v->val;
  }

  template<class T>
  void HashedSparseMatrix<T>::operator()(const T f[], T g[]) const {
    T sum = 0.0;
    uint r = nofRows << hashBits;

    while (r--) {
      for(Value* v = matrix[r]; v != NULL; v = v->lnk) 
	sum += v->val * f[v->idx];
      if (!(r & hashMsk)) { 
	g[r >> hashBits] = sum;
	sum = 0.0;
      }
    }
  }

  template<class T>
  void HashedSparseMatrix<T>::transpMult(const T f[], T g[]) const {
    uint r = nofRows << hashBits;

    while (r--) {
      for(Value* v = matrix[r]; v != NULL; v = v->lnk) 
	g[v->idx] += v->val * f[r >> hashBits];
    }
  }

  template<class T>
  void HashedSparseMatrix<T>::write(std::ofstream& ofs) const {
    Array<T> foo(nofCols);
    foo[0] = nofRows; foo[1] = nofCols;
    ofs.write((const char*)((T*)foo), 2 * sizeof(T));

    for (uint i = 0; i < nofRows; i++) {
      for (uint j = 0; j < nofCols; j++) {
	Value* v = matrix[i << hashBits | j & hashMsk];
	for (; v != 0 && v->idx != j; v = v->lnk);
	foo[j] = v == 0 ? 0.0 : v->val;
      }
      ofs.write((const char*)((T*)foo), nofCols * sizeof(T));
    }
  }

  template<class T>
  void HashedSparseMatrix<T>::outputMatlab(std::ofstream& ofs) const {
    int p = 1 << hashBits;
    ofs << nofRows << ' ' << nofCols << " 0" << std::endl;
    for (uint r = 0; r < nofRows; ++r) {
      for (int c_mod = 0; c_mod < p; ++c_mod) {
	for (HashedSparseMatrix::Value* v = matrix[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk) {
	  if (v->val != (T)0.0) {
	    ofs << r+1 << ' ' << v->idx+1 << ' ' << v->val << std::endl;
	  }
	}
      }
    }
  }

  template<class T>
  void HashedSparseMatrix<T>::outputSparseQR(std::ofstream& ofs) const {
    int p = 1 << hashBits;
    // header with size of matrix
    ofs << "# Number of rows #" << std::endl
	<< nofRows << std::endl
	<< "# Number of columns #" << std::endl
	<< nofCols << std::endl
	<< "# Data #" << std::endl
	<< "# Number of entries in row:  column ... #" << std::endl
	<< "# Value ... Right hand side #" << std::endl;
    ofs.setf(std::ios::scientific);
    ofs.setf(std::ios::uppercase);
    ofs.precision(10);
    // data of matrix with right hand size = 0
    for (uint r = 0; r < nofRows; ++r) {
      std::stringstream columnIndices, entries;
      uint number = 0;
      for (int c_mod = 0; c_mod < p; ++c_mod) {
	for (HashedSparseMatrix::Value* v = matrix[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk) {
	  if (v->val != (T)0.0) {
	    entries << '\t' << v->val;
	    columnIndices << '\t' << v->idx+1;
	    ++number;
	  }
	}
      }
      if (number == 0) {
	ofs << '\t' << 1 << '\t' << 1 << std::endl
	    << '\t' << 0.0 << '\t' << 0.0 << std::endl;
      } else {
	//	columnIndices << std::ends;
	//	entries << std::ends;
	ofs << '\t' << number << columnIndices.str() << std::endl
	    << '\t' << entries.str() << '\t' << 0.0 << std::endl;
      }
    } // for r
    ofs << '\t' << 1 << '\t' << 1 << std::endl
	<< '\t' << 0.0 << '\t' << 0.0 << std::endl;
  }

  template<class T>
  void HashedSparseMatrix<T>::multiply(const HashedSparseMatrix<T>* const fact,
				       concepts::Matrix<T>& dest) const {
    conceptsAssert(nofCols == fact->nofRows, Assertion());
    conceptsAssert(nofRows == dest.spaceX().dim(), Assertion());
    conceptsAssert(fact->nofCols == dest.spaceY().dim(), Assertion());
    uint r = nofRows << hashBits;

    while (r--) {
      for (Value* v= matrix[r]; v != 0; v = v->lnk) {
	for (uint k = 0; k < fact->nofCols; ++k) {
	  Value* w = fact->matrix[v->idx << fact->hashBits |
				  k & fact->hashMsk];
	  for (; w != 0 && w->idx != k; w = w->lnk);
	  if (w != 0) {
	    T sum = v->val * w->val;
	    if (sum != (T)0)
	      dest(r >> hashBits, k) += sum;
	  }
	} // for k
      } // for v
    } // while
  }

  template<class T>
  void HashedSparseMatrix<T>::multiply(const concepts::Matrix<T>* const fact,
				       concepts::Matrix<T>& dest) const {
    conceptsAssert(nofCols == fact->spaceX().dim(), Assertion());
    conceptsAssert(nofRows == dest.spaceX().dim(), Assertion());
    conceptsAssert(fact->spaceY().dim() == dest.spaceY().dim(), Assertion());
    uint r = nofRows << hashBits;
    const uint BnofCols = fact->spaceY().dim();
    while (r--) {
      for (Value* v= matrix[r]; v != 0; v = v->lnk) {
	for (uint k = 0; k < BnofCols; ++k) {
 	  T sum = v->val * (*fact)(v->idx,k);
	  if (sum != (T)0)
	    dest(r >> hashBits, k) += sum;
	} // for k
      } // for v
    } // while
  }

  template<class T>
  std::ostream& HashedSparseMatrix<T>::info(std::ostream& os) const {
    os << "HashedSparseMatrix: " << _used << " (";
#if HashedSparseMatrixInfo_D
    os << std::endl;
    Value* v = 0;
    for (uint i = 0; i < nofRows; i++) {
      for (uint j = 0; j < nofCols; j++) {
	for (v = matrix[i << hashBits | j & hashMsk];
	     v != 0 && v->idx != j; v = v->lnk);
	if (v == 0)
	  os << 0 << ' ';
	else 
	  os << v->val << ' ';
      }
      os << std::endl;
    }
    os << ')' << std::endl;
#else
    os << (Real)_used / ((Real)nofRows * (Real)nofCols) * (Real)100
       << "%) entries bound.";
#endif
    return os;
  }

  template<class T>
  void HashedSparseMatrix<T>::dropEntry_(Value*& v, Value*& vPrev,
					 const uint r, const uint c) {
    conceptsAssert(r < nofRows, Assertion());
    conceptsAssert(c < nofCols, Assertion());
    
    Value** m = matrix + (r << hashBits | c & hashMsk);
    if (v == 0) {
      DEBUGL(HSMdropEntry_D, "entry (" << r << ", " << c << ") not found");
    } else {
      DEBUGL(HSMdropEntry_D, "entry (" << r << ", " << c << ") found: "
	     << v->val << ", v = " << v << ", vPrev = " << vPrev);
      if (vPrev == 0) {
	*m = v->lnk;
	--_used;
      } else {
	vPrev->lnk = v->lnk;
	--_used;
      }
    }
#if HSMdropEntry_D
    const HashedSparseMatrix<T>* tmp = this;
    DEBUGL(HSMdropEntry_D, "entry is now = " << tmp->operator()(r,c));
#endif
  }

} // namespace concepts
