/* interface for a matrix
 */

#include "matrix.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "toolbox/scannerConnectors.hh"
#include "bilinearForm.hh"

// timings
#define matrixTiming

#ifdef matrixTiming
#include "toolbox/inputOutput.hh"
#include "toolbox/resourceMonitor.hh"
#endif // matrixTiming

// Boost.Thread
#undef HAS_BOOST // turn off threading support by default
#ifdef HAS_BOOST
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#define nTHREADS 2
#endif

// debugging
#include "basics/debug.hh"

#define asseblyop_D 0
#define assemblyopThreads_D 0

namespace concepts {

#ifdef HAS_BOOST
  template<typename F>
  class dofast {
  public:
    dofast(typename Space<F>::Scanner* scanner, Matrix<F>& m,
	   boost::mutex& lock1, boost::mutex& lock2,
	   BilinearForm<F>& bilinear) :
      A(3,3), B(3,3), sc(scanner), dest(m),
      scan_lock(lock1), matrix_lock(lock2),
      bf((BilinearForm<F>*)bilinear.clone()), threadID_(thread_number++) {
      DEBUGL(assemblyopThreads_D, "created thread " << threadID_);
    }
    dofast(const dofast& d) :
      A(3,3), B(3,3), sc(d.sc), dest(d.dest), scan_lock(d.scan_lock),
      matrix_lock(d.matrix_lock), bf((BilinearForm<F>*)d.bf->clone()),
      threadID_(d.threadID_) {
      DEBUGL(assemblyopThreads_D, "copied thread " << threadID_);
    }
    void operator()() {
      boost::mutex::scoped_lock lock_scanner(scan_lock);
      while (*sc) {
	const Element<F>& elm = (*sc)++;
	lock_scanner.unlock();
	DEBUGL(assemblyopThreads_D, "thread " << threadID_ << ": " << elm);
	const TMatrixBase<F>& T = elm.T();

	(*bf)(elm, elm, A);
	T(A, B); B.transpose(); T(B, A); A.transpose();
	DEBUGL(asseblyop_D, "assembling " << A << " into " << dest);

	boost::mutex::scoped_lock lock_matrix(matrix_lock);
	for(uint i = T.n(); i--;)
	  for(uint j = T.n(); j--;)
	    dest(T.index(i), T.index(j)) += A(i, j);
	lock_scanner.lock();
      } // while sc
    }
  private:
    ElementMatrix<F> A, B;
    typename Space<F>::Scanner* sc;
    Matrix<F>& dest;
    boost::mutex& scan_lock, & matrix_lock;
    std::auto_ptr<BilinearForm<F> > bf;
    const uint threadID_;
    static uint thread_number;
  };

  template<typename F>
  uint dofast<F>::thread_number = 0;
#endif

  // **************************************************************** Matrix **

  template<class F>
  void Matrix<F>::assembly(Matrix<F>& dest,
			   const Space<F>& spc, BilinearForm<F>& bf) {
    conceptsAssert(dest.spaceX().dim() == spc.dim(), Assertion());
    conceptsAssert(dest.spaceY().dim() == spc.dim(), Assertion());

#ifdef matrixTiming
    ResourceMonitor timer;
#endif // matrixTiming

#ifndef HAS_BOOST
    ElementMatrix<F> A(3, 3);
    ElementMatrix<F> B(3, 3);
#endif

    std::auto_ptr<typename Space<F>::Scanner> sc(spc.scan());
#ifdef HAS_BOOST
    boost::thread_group threads;
    boost::mutex lock1, lock2;
    for (int i = 0; i < nTHREADS; ++i)
      threads.create_thread(dofast<F>(sc.get(), dest, lock1, lock2, bf));
    threads.join_all();
#else
    while (*sc) {
      const Element<F>& elm = (*sc)++;
      const TMatrixBase<F>& T = elm.T();

#ifdef matrixTiming
      if (timings_)
	timer.utime();
#endif // matrixTiming
      bf(elm, elm, A);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("bilinear_form", timeCntr_, timer.utime());
#endif // matrixTiming
      T(A, B); B.transpose(); T(B, A); A.transpose();
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("tmatrix_apply", timeCntr_, timer.utime());
#endif // matrixTiming
      DEBUGL(asseblyop_D, "assembling " << A << " into " << dest);

      for(uint i = T.n(); i--;)
	for(uint j = T.n(); j--;)
	  dest(T.index(i), T.index(j)) += A(i, j);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("global_assembly", timeCntr_++,
				 timer.utime());
#endif // matrixTiming
    }
#endif // HAS_BOOST
    DEBUGL(asseblyop_D, "done.");
  }

  template<class F>
  void Matrix<F>::assembly(Matrix<F>& dest,
			   const Space<F>& spcX, const Space<F>& spcY,
			   BilinearForm<F>& bf) {
    conceptsAssert(dest.spaceX().dim() == spcX.dim(), Assertion());
    conceptsAssert(dest.spaceY().dim() == spcY.dim(), Assertion());

#ifdef matrixTiming
    ResourceMonitor timer;
#endif // matrixTiming

    ElementMatrix<F> A(3, 3);
    ElementMatrix<F> B(3, 3);

    std::auto_ptr<typename Space<F>::Scanner> scX(spcX.scan());
    while ( *(scX.get()) ) {
      const Element<F>& elmX = (*scX)++;
      const TMatrixBase<F>& TX = elmX.T();

      std::auto_ptr<typename Space<F>::Scanner> scY(spcY.scan());
      while ( !scY->eos() ) {
	const Element<F>& elmY = (*scY)++;
	const TMatrixBase<F>& TY = elmY.T();

#ifdef matrixTiming
	if (timings_)
	  timer.utime();
#endif // matrixTiming
	bf(elmX, elmY, A);
	conceptsAssert(A.m() == TX.m(), Assertion());
	conceptsAssert(A.n() == TY.m(), Assertion());
#ifdef matrixTiming
	if (timings_)
	  timings_->addArrayDouble("bilinear_form", timeCntr_, timer.utime());
#endif // matrixTiming
	TY(A, B); B.transpose(); TX(B, A); A.transpose();
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("tmatrix_apply", timeCntr_, timer.utime());
#endif // matrixTiming

	DEBUGL(asseblyop_D, "assembling " << A << " into " << dest);
	for(uint i = TX.n(); i--;) {
	  for(uint j = TY.n(); j--;) {
	    dest(TX.index(i), TY.index(j)) += A(i, j);
	  }
	}
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("global_assembly", timeCntr_++,
				 timer.utime());
#endif // matrixTiming
      } // while scY
    } // while scX
    DEBUGL(asseblyop_D, "done.");
  }

  template<class F>
  void Matrix<F>::assembly(Matrix<F>& dest,
			   BilinearForm<F>& bf,
			   const ElementPairList<F>& pairs) {
    ElementMatrix<F> A(3, 3);
    ElementMatrix<F> B(3, 3);

#ifdef matrixTiming
    ResourceMonitor timer;
#endif // matrixTiming

    std::auto_ptr<typename ElementPairList<F>::Scanner> sc(pairs.scan());
    while (*sc) {
      const ElementPair<F>& elmPair = (*sc)++;
      const Element<F>& elmX = elmPair.elm1();
      const TMatrixBase<F>& TX = elmX.T();
  
      const Element<F>& elmY = elmPair.elm2();
      const TMatrixBase<F>& TY = elmY.T();

#ifdef matrixTiming
      if (timings_)
	timer.utime();
#endif // matrixTiming
      bf(elmX, elmY, A, elmPair);
#ifdef matrixTiming
      if (timings_)
	timings_->addArrayDouble("bilinear_form", timeCntr_, timer.utime());
#endif // matrixTiming
      if (A.size() > 0) {
	TY(A, B); B.transpose(); TX(B, A); A.transpose();
#ifdef matrixTiming
	if (timings_)
	  timings_->addArrayDouble("tmatrix_apply", timeCntr_, timer.utime());
#endif // matrixTiming

	for(uint i = TX.n(); i--;) {
	  for(uint j = TY.n(); j--;) {
	    dest(TX.index(i), TY.index(j)) += A(i, j);
	  } // for x
	} // for y
#ifdef matrixTiming
	if (timings_)
	  timings_->addArrayDouble("global_assembly", timeCntr_++,
				   timer.utime());
#endif // matrixTiming
      } // if size
    } // while scX
    DEBUGL(asseblyop_D, "done.");
  }

  template<class F>
  InOutParameters* Matrix<F>::timings_ = 0;

  template<class F>
  uint Matrix<F>::timeCntr_ = 0;

  template<class F>
  void Matrix<F>::setTimings(InOutParameters* timings) {
#ifdef matrixTiming
    timings_ = timings;
    if (timings_) {
      timeCntr_ = 0;
      timings_->addArrayDouble("bilinear_form");
      timings_->addArrayDouble("tmatrix_apply");
      timings_->addArrayDouble("global_assembly");
    }
#endif // matrixTiming
  }

  template<class F>
  bool Matrix<F>::timings() {
#ifdef matrixTiming
    return true;
#else
    return false;
#endif // matrixTiming
  }

} // namespace concepts
