/* interface for a matrix
 */

#ifndef matrix_hh
#define matrix_hh

#include "compositions.hh"
#include "space/elementPairs.hh"

namespace concepts {

  // forward declarations
  template<typename F>
  class BilinearForm;

  class InOutParameters;

  // **************************************************************** Matrix **

  /** Abstract class for an operator.

      This class is equiped with an interface to get timings of internal
      computations if compiled accordingly (see \c bilinearForm.cc file), see
      \c setTimings() and \c timings().

      @author Philipp Frauenfelder, 2002
      @ingroup bilinear
  */
  template<class F>
  class Matrix : public Operator<F> {
  public:
    /** Assembly operator for \c dest using the bilinear form \c bf. This
	assembly operator does not compute element matrices for two different
	elements. The elements are taken from the space \c spc.
    */
    static void assembly(Matrix<F>& dest,
			 const Space<F>& spc, BilinearForm<F>& bf);
    /** Assembly operator for \c dest using the bilinear form \c bf. This
	assembly operator computes also the element matrices for two different
	elements (coming from \c spcX and \c spcY).
     */
    static void assembly(Matrix<F>& dest,
			 const Space<F>& spcX, const Space<F>& spcY,
			 BilinearForm<F>& bf);
    /**  Assembly operator for \c dest using the bilinear form \c bf. This
	 assembly operator uses the element pairs taken from \c pairs. For
	 every two elements found in a ElementPair in \c pairs, the bilinear
	 form is evaluated and the result assembled into \c dest.
    */
    static void assembly(Matrix<F>& dest,
			 BilinearForm<F>& bf, const ElementPairList<F>& pairs);

    /// Returns entry with indices \c i and \c j
    virtual F operator()(const uint i, const uint j) const = 0;
    /// Returns and allows access to entry with indices \c i and \c j
    virtual F& operator()(const uint i, const uint j) = 0;
    /// Computes \c fncX = A(\c fncY) where A is this matrix.
    virtual void operator()(const Function<F>& fncY, Function<F>& fncX) = 0;
    /// Computes \c fncX = A<sup>T</sup> \c fncY where A is this matrix
    virtual void transpMult(const Vector<F>& fncY, Vector<F>& fncX) = 0;

    /** @name Timing Interface
     *  These functions are used to get timings from class internal
     *  computations. The values are stored in a user defined
     *  concepts::InOutParameters structure in different arrays (see 
     *  \c setTimings). These arrays can be grouped into a table for
     *  easier postprocessing with
     *  \code
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "jacobian", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "whole_sumfact", output);
    std::ofstream ofs("table.gnuplot");
    ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(ofs);
        \endcode
     */
    //@{
    /** Sets the class to store the timing values in. Additionally,
	the \c timeCntr_ is reset to 0. This counter is used to fill
	in the values into the arrays listed below in subsequent calls.
	The following timings are taken and stored in \c timings:
	<ul>
	<li>evaluation of bilinear form in \c bilinear_form
	<li>application of T matrix in \c tmatrix_apply
	<li>assembling into global matrix in \c global_assembly
	</ul>
    */
    static void setTimings(InOutParameters* timings);
    /** Returns true if the class is able to do timings.  The ability
	to do timings depends on a compiler switch in \c matrix.cc file.
    */
    static bool timings();
    //@}
  private:
    /// Place to store timing values
    static InOutParameters* timings_;
    /// Counter for timing table
    static uint timeCntr_;
  };

} // namespace concepts

#endif // matrix_hh
