// permutation and restriction/extension operator

#ifndef permutation_hh
#define permutation_hh

#include "matrix.hh"
#include "basics/exceptions.hh"

// debugging
#include "basics/debug.hh"

#define PermutationAppl_D 0

#define trivExtRestrConstr_D 0
#define trivExtRestrAppl_D 0

namespace concepts {

  // forward declaration
  template<typename F>
  class TrivExtendRestrict;

  // *********************************************************** Permutation **

  /** Permutation operator. This operator permutes a vector according to
      a given permutation (given as an array of indices).

      The forward permutation acts like \c x[i] = \c y[perm[i]] on the vectors.
      The inverse acts like \c x[perm[i]] = \c y[i].

      @test test::MoreCompositionsTest
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class Permutation : public Matrix<F> {
  public:
    /** Constructor.
	@param space Source and image space
	@param perm Array defining the permutation
	@param transpose Applies the permutation (\c false) or its transpose
               (ie. its inverse, \c true)
    */
    Permutation(const Space<F>& space,
		const Array<int>& perm, bool transpose = false);
    /** Copy constructor.
	@param perm Permutation
	@param transpose Transposes the permutation while copying
    */
    Permutation(const Permutation<F>& perm, bool transpose = false);
    virtual void operator()(const Function<F>& fncY,
			    Function<F>& fncX);
    /// Returns the permutation of index \c i
    int operator[](const uint i) const { return p_[i]; }
    virtual const Space<F>& spaceX() const { return spc_; }
    virtual const Space<F>& spaceY() const { return spc_; }

    virtual F operator()(const uint i, const uint j) const;
    virtual F& operator()(const uint i, const uint j);
    virtual void transpMult(const Vector<F>& fncY, Vector<F>& fncX);

    /** Computes the composition of a \c Permutation and a
	\c TrivExtendRestrict and stores the result in \c dest.
    */
    void composeRestr(const TrivExtendRestrict<F>& restr,
		      Matrix<F>& dest) const;

    /** Convertes this permutation to the matrix \c dest
	@pre \c dest is empty
    */
    void convertToMatrix(Matrix<F>& dest) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Source and image space
    const Space<F>& spc_;
    /// Array defining the permutation
    Array<int> p_;
  };

  template<typename F>
  void Permutation<F>::operator()(const Function<F>& fncY,
				  Function<F>& fncX) {
    conceptsAssert(fncY.space().dim() == spc_.dim(), Assertion());
    conceptsAssert(fncX.space().dim() == spc_.dim(), Assertion());
    DEBUGL(PermutationAppl_D, "y = " << fncY);
    for (uint j = 0; j < spc_.dim(); ++j) fncX(j) = fncY(p_[j]);
    DEBUGL(PermutationAppl_D, "x = " << fncX);
  }

  template<typename F>
  F Permutation<F>::operator()(const uint i, const uint j) const {
    if (j == (uint)p_[i]) return 1;
    else return 0;
  }

  template<typename F>
  F& Permutation<F>::operator()(const uint i, const uint j) {
    throw conceptsException(MissingFeature
			    ("reference access to matrix not allowed"));
    // a never used return value, just to avoid warnings
    return *(F*)0;
  }

  template<typename F>
  void Permutation<F>::transpMult(const Vector<F>& fncY, Vector<F>& fncX) {
    conceptsAssert(fncY.space().dim() == spc_.dim(), Assertion());
    conceptsAssert(fncX.space().dim() == spc_.dim(), Assertion());
    DEBUGL(PermutationAppl_D, "y = " << fncY);
    for (uint j = 0; j < spc_.dim(); ++j) fncX(p_[j]) = fncY(j);
    DEBUGL(PermutationAppl_D, "x = " << fncX);
  }

  // **************************************************** TrivExtendRestrict **

  /** Trivial extension and restriction operator. Extends a vector by adding
      zeros at the beginning or restricts a vector by stripping off the
      beginning of the vector.
      @author Philipp Frauenfelder, 2002
      @test test::MoreCompositionsTest
  */
  template<typename F>
  class TrivExtendRestrict : public concepts::Operator<F> {
  public:
    /** Constructor.
	@param spcX Image space
	@param spcY Source space
	@param extend Applies the extension (\c true) or the restriction
               (\c false)
    */
    TrivExtendRestrict(const concepts::Space<F>& spcX,
		       const concepts::Space<F>& spcY, bool extend = false) :
      spcX_(spcX), spcY_(spcY), extend_(extend) {
      DEBUGL(trivExtRestrConstr_D, "extend = " << extend
	     << ", spcX = " << spcX_ << ", spcY = " << spcY_);
#ifdef DEBUG
      if (extend_) {
	conceptsAssert(spcX.dim() >= spcY.dim(), concepts::Assertion());
      } else {
	conceptsAssert(spcX.dim() <= spcY.dim(), concepts::Assertion());
      }
#endif
    }
    virtual const concepts::Space<F>& spaceX() const { return spcX_; }
    virtual const concepts::Space<F>& spaceY() const { return spcY_; }
    virtual void operator()(const concepts::Function<F>& fncY,
			    concepts::Function<F>& fncX);
    /// Returns the extension flag
    bool extend() const { return extend_; }

    /** Convertes this operator to the matrix \c dest
	@pre \c dest is empty
    */
    void convertToMatrix(Matrix<F>& dest) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Image space
    const concepts::Space<F>& spcX_;
    /// Source space
    const concepts::Space<F>& spcY_;
    const bool extend_;
  };

  template<typename F>
  void TrivExtendRestrict<F>::operator()
    (const concepts::Function<F>& fncY, concepts::Function<F>& fncX) {
    conceptsAssert(spcX_.dim() == fncX.space().dim(), concepts::Assertion());
    conceptsAssert(spcY_.dim() == fncY.space().dim(), concepts::Assertion());
    DEBUGL(trivExtRestrAppl_D, "y = " << fncY);
    const uint diff = spcX_.dim() > spcY_.dim() ? spcX_.dim()-spcY_.dim() :
      spcY_.dim()-spcX_.dim();
    DEBUGL(trivExtRestrAppl_D, "diff = " << diff << ", extend = " << extend_);
    if (extend_) {
      for (uint i = 0; i < diff; ++i) fncX(i) = 0;
      for (uint i = 0; i < spcY_.dim(); ++i) fncX(i+diff) = fncY(i);
    } else {
      for (uint i = 0; i < spcX_.dim(); ++i) fncX(i) = fncY(i+diff);
    }
    DEBUGL(trivExtRestrAppl_D, "x = " << fncX);
  }

} // namespace concepts

#endif // permutation_hh
