// testing linear combination and composition

#include <cmath>
#include "compositions.hh"
#include "basics/exceptions.hh"
#include "basics/testsuite.hh"
#include "operator/compositions.hh"
#include "operator/denseMatrix.hh"
#include "operator/permutation.hh"

using namespace concepts;

namespace test {

  // *************************************************** CompositionTestBase **

  template<typename F>
  void CompositionTestBase::compare(const Matrix<F>& a, Operator<F>& b) {
    const uint m = a.spaceX().dim();
    const uint n = a.spaceY().dim();
    _test(m == b.spaceX().dim());
    _test(n == b.spaceY().dim());

    concepts::Vector<F> tmp(b.spaceY()), res(b.spaceX());
    for (uint j = 0; j < n; ++j) {
      tmp[j] = 1;
      b(tmp, res);
      for (uint i = 0; i < m; ++i) {
	_numtest(a(i,j), res[i]);
// 	if (std::fabs(a(i,j) - res[i]) > 1e-3)
// 	  DEBUGL(1, "a(" << i << ", " << j << ") = " << a(i,j) << " =? "
// 		 << res[i]);
      }
      tmp[j] = 0;
    }
  }

  // ****************************************************** CompositionsTest **

  CompositionsTest::CompositionsTest() :
    spc(5), a(spc, spc), b(spc, spc), c(spc, spc) {
    a(0,0) = 2.0; a(0,1) = a(1,0) = 0.3; a(0,3) = a(3,0) = 0.5;
    a(1,1) = 3.0;
    a(2,2) = 7.0; a(2,4) = a(4,2) = 2.0;
    a(3,3) = 8.0; a(3,4) = a(4,3) = 6.0;
    a(4,4) = 7.0;
    DEBUGL(0, a << " = " << DenseMatrix<Real>(a));

    b(0,0) = 1.0; b(0,1) = 0.5; b(3,0) = 1.5;
    b(1,1) = b(3,1) = 13.0;
    b(2,2) = 7.0; b(4,2) = 12.0;
    b(3,3) = 8.0; b(3,4) = b(4,3) = b(1,3) = 6.0;
    b(4,4) = 0.7;
    DEBUGL(0, c << " = " << DenseMatrix<Real>(c));

    c(0,1) = c(1,0) = 5.3; c(0,3) = c(3,0) = 2.5;
    c(1,1) = c(1,3) = 13.0;
    c(2,4) = c(4,2) = 2.0;
    c(3,3) = 8.0; c(4,3) = 6.0;
    c(4,4) = 17.0;
    DEBUGL(0, c << " = " << DenseMatrix<Real>(c));
  }

  void CompositionsTest::run() {
    DEBUGL(0, a << " = " << DenseMatrix<Real>(a));
    test1(); test2(); test3(); test4(); test5();
  }

  void CompositionsTest::test1() {
    DEBUGL(0, a << " = " << DenseMatrix<Real>(a));
    SparseMatrix<Real> r(spc, spc); // (a+b)*c
    r(0,0) = 5.49; r(0,1) = 26.3; r(0,3) = 21.9;
    r(1,0) = 99.8; r(1,1) = 209.59; r(1,3) = 256.75;
    r(2,2) = 4; r(2,3) = 12; r(2,4) = 62;
    r(3,0) = 108.9; r(3,1) = 179.6; r(3,2) = 24; r(3,3) = 374;
    r(3,4) = 204;
    r(4,0) = 30; r(4,2) = 15.4; r(4,3) = 142.2; r(4,4) = 158.9;

    LiCo<Real> s(a, b);
    Compose<Real> C(s, c);
    compare(r, C);

    SparseMatrix<Real> R(C);
    compare(R, r);
  }

  void CompositionsTest::test2() {
    SparseMatrix<Real> r(spc, spc); // a*(b+c)
    r(0,0) = 5.59; r(0,1) = 25.9; r(0,3) = 18.7; r(0,4) = 3;
    r(1,0) = 16.2; r(1,1) = 79.74; r(1,3) = 57.75;
    r(2,2) = 77; r(2,3) = 24; r(2,4) = 49.4;
    r(3,0) = 32.5; r(3,1) = 106.9; r(3,2) = 84; r(3,3) = 201.25;
    r(3,4) = 154.2;
    r(4,0) = 24; r(4,1) = 78; r(4,2) = 112; r(4,3) = 180; r(4,4) = 163.9;

    LiCo<Real> s(b, c);
    Compose<Real> C(a, s);
    compare(r, C);

    SparseMatrix<Real> R(C);
    compare(R, r);
  }

  void CompositionsTest::test3() {
    SparseMatrix<Real> r(spc, spc); // a+b*c
    r(0,0) = 4.65; r(0,1) = 12.1; r(0,3) = 9.5;
    r(1,0) = 84.2; r(1,1) = 172; r(1,3) = 217;
    r(2,2) = 7; r(2,4) = 16;
    r(3,0) = 89.4; r(3,1) = 176.95; r(3,2) = 12; r(3,3) = 280.75;
    r(3,4) = 108;
    r(4,0) = 15; r(4,2) = 3.4; r(4,3) = 58.2; r(4,4) = 42.9;

    Compose<Real> C(b, c);
    LiCo<Real> s(a, C);
    compare(r, s);

    SparseMatrix<Real> R(s);
    compare(R, r);
  }

  void CompositionsTest::test4() {
    SparseMatrix<Real> r(spc, spc); // a*b+c
    r(0,0) = 2.75; r(0,1) = 16.7; r(0,3) = 8.3; r(0,4) = 3;
    r(1,0) = 5.6; r(1,1) = 52.15; r(1,3) = 31;
    r(2,2) = 73; r(2,3) = 12; r(2,4) = 3.4;
    r(3,0) = 15; r(3,1) = 104.25; r(3,2) = 72; r(3,3) = 108;
    r(3,4) = 52.2;
    r(4,0) = 9; r(4,1) = 78; r(4,2) = 100; r(4,3) = 96; r(4,4) = 57.9;

    Compose<Real> C(a, b);
    LiCo<Real> s(c, C);
    compare(r, s);

    SparseMatrix<Real> R(s);
    compare(R, r);
  }

  void CompositionsTest::test5() { // a*b-b*c
    Compose<Real> C(a, b);
    Compose<Real> D(b, c);
    LiCo<Real> s(C, D, 1, -1);
    compare(SparseMatrix<Real>(s, true), s);
    compare(SparseMatrix<Real>(s), s);
  }

  // ************************************************** MoreCompositionsTest **

  MoreCompositionsTest::MoreCompositionsTest() :
    spc(5), a(spc, spc) {
    a(0,0) = 2.0; a(0,1) = a(1,0) = 0.3; a(0,3) = a(3,0) = 0.5;
    a(1,1) = 3.0;
    a(2,2) = 7.0; a(2,4) = a(4,2) = 2.0;
    a(3,3) = 8.0; a(3,4) = a(4,3) = 6.0;
    a(4,4) = 7.0;
  }

  void MoreCompositionsTest::run() {
    compare(a, a);
    testPermutationConsistency();
    testTrivExtendRestrictConsistency();
    testCompositions();
  }

  void MoreCompositionsTest::testPermutationConsistency() {
    Array<int> perm(5);
    perm[0] = 1; perm[1] = 2; perm[2] = 0; perm[3] = 4; perm[4] = 3;

    Permutation<Real> p(spc, perm);
    SparseMatrix<Real> ps(p);
    compare(ps, p);

    Permutation<Real> p2(spc, perm, true);
    SparseMatrix<Real> ps2(p2);
    compare(ps2, p2);

    perm[0] = 1; perm[1] = 3; perm[2] = 0; perm[3] = 4; perm[4] = 2;

    Permutation<Real> p3(spc, perm);
    SparseMatrix<Real> ps3(p3);
    compare(ps3, p3);

    Permutation<Real> p4(spc, perm, true);
    SparseMatrix<Real> ps4(p4);
    compare(ps4, p4);
  }

  void MoreCompositionsTest::testTrivExtendRestrictConsistency() {
    DummySpace<Real> spcSmall(3);
    TrivExtendRestrict<Real> restrict(spcSmall, spc);
    SparseMatrix<Real> rs(restrict);
    compare(rs, restrict);

    TrivExtendRestrict<Real> extend(spc, spcSmall, true);
    SparseMatrix<Real> es(extend);
    compare(es, extend);

    DummySpace<Real> spcLarge(7);
    TrivExtendRestrict<Real> extendL(spcLarge, spc, true);
    SparseMatrix<Real> esL(extendL);
    compare(esL, extendL);

    TrivExtendRestrict<Real> restrictL(spc, spcLarge);
    SparseMatrix<Real> rsL(restrictL);
    compare(rsL, restrictL);
  }

  void MoreCompositionsTest::testCompositions() {
    Array<int> perm(5);
    perm[0] = 1; perm[1] = 2; perm[2] = 0; perm[3] = 4; perm[4] = 3;

    Permutation<Real> p(spc, perm);
    Compose<Real> ap(a, p);
    SparseMatrix<Real> aps(ap);
    compare(aps, ap);

    Compose<Real> pa(p, a);
    SparseMatrix<Real> pas(pa);
    compare(pas, pa);

    Permutation<Real> p2(spc, perm, true);
    Compose<Real> pa2(p2, a);
    SparseMatrix<Real> pa2s(pa2);
    compare(pa2s, pa2);

    DummySpace<Real> spcSmall(2);
    TrivExtendRestrict<Real> restrict(spcSmall, spc);
    Compose<Real> ra(restrict, a);
    SparseMatrix<Real> ras(ra);
    compare(ras, ra);

    DummySpace<Real> spcLarge(8);
    TrivExtendRestrict<Real> extend(spcLarge, spc, true);
    Compose<Real> ae(extend, a);
    SparseMatrix<Real> aes(ae);
    compare(aes, ae);

    Compose<Real> ep(extend, p);
    SparseMatrix<Real> eps(ep);
    compare(eps, ep);
  }

  // ************************************************** DeepCompositionsTest **

  DeepCompositionsTest::DeepCompositionsTest() :
    spc(6), cspc(4), a(spc, spc), b(spc, spc), c(spc, cspc) {
    a(0,0) = 1;  a(0,1) = -1;
    a(1,0) = -1; a(1,1) = 2; a(1,2) = -1;
    a(2,1) = -1; a(2,2) = 2; a(2,3) = -1;
    a(3,2) = -1; a(3,3) = 2; a(3,4) = -1;
    a(4,3) = -1; a(4,4) = 2; a(4,5) = -1;
    a(5,4) = -1; a(5,5) = 2;

    b(0,0) = 6;
    b(0,1) = b(1,0) = b(1,1) = 5;
    b(0,2) = b(2,0) = b(2,2) = b(1,2) = b(2,1) = 4;
    b(3,3) = b(3,2) = b(3,1) = b(3,0) = b(2,3) = b(1,3) = b(0,3) = 3;
    b(4,4) = b(4,3) = b(4,2) = b(4,1) = b(4,0) = b(3,4) = b(2,4) = b(1,4)
      = b(0,4) = 2;
    b(5,5) = b(5,4) = b(5,3) = b(5,2) = b(5,1) = b(5,0) = b(4,5) = b(3,5)
      = b(2,5) = b(1,5) = b(0,5) = 1;

    for (uint i = 0; i < spc.dim(); i += 2) {
      c(i,0) = c(i,1) = 1; c(i,2) = 8; c(i,3) = 5;
      c(i+1,0) = c(i+1,3) = 1; c(i+1,1) = -1; c(i+1,2) = 2;
    }
  }

  void DeepCompositionsTest::run() {
    createSparseQR();
    testCompositions();
  }

  void DeepCompositionsTest::createSparseQR() {
    qr.reset(new sparseqr::Driver(c));
    _test(qr->rank() == 2);
    Pr = qr->Prow();
    Qgt = qr->Qt();
    Prt.reset(new concepts::Permutation<Real>(*Pr, true));
    Qg = qr->Q();
    ext = qr->extension();
    restr = qr->restriction();

    PrQgt.reset(new concepts::Compose<Real>(*Pr, *Qgt));
    QgPrt.reset(new concepts::Compose<Real>(*Qg, *Prt));
    restrPrQgt.reset(new concepts::Compose<Real>(*restr, *PrQgt));
    QgPrtext.reset(new concepts::Compose<Real>(*QgPrt, *ext));

    restrPrQgtA.reset(new concepts::Compose<Real>(*restrPrQgt, a));
    restrPrQgtAQgPrtext.reset(new concepts::Compose<Real>(*restrPrQgtA,
							  *QgPrtext));

    restrPrQgtM.reset(new concepts::Compose<Real>(*restrPrQgt, b));
    restrPrQgtMQgPrtext.reset(new concepts::Compose<Real>(*restrPrQgtM,
							  *QgPrtext));

    _test(restrPrQgtMQgPrtext->spaceX().dim() == 4);
    _test(restrPrQgtMQgPrtext->spaceY().dim() == 4);
    _test(restrPrQgtAQgPrtext->spaceX().dim() == 4);
    _test(restrPrQgtAQgPrtext->spaceY().dim() == 4);
  }

  void DeepCompositionsTest::testCompositions() {
    compare(concepts::SparseMatrix<Real>(*restrPrQgtMQgPrtext, true),
	    *restrPrQgtMQgPrtext);
    compare(concepts::SparseMatrix<Real>(*restrPrQgtMQgPrtext),
	    *restrPrQgtMQgPrtext);

    compare(concepts::SparseMatrix<Real>(*restrPrQgtAQgPrtext, true),
	    *restrPrQgtAQgPrtext);
    compare(concepts::SparseMatrix<Real>(*restrPrQgtAQgPrtext),
	    *restrPrQgtAQgPrtext);

    concepts::LiCo<Real> l1(*restrPrQgtAQgPrtext, *restrPrQgtMQgPrtext);
    compare(concepts::SparseMatrix<Real>(l1, true), l1);
    compare(concepts::SparseMatrix<Real>(l1), l1);

    concepts::LiCo<Real> l(*restrPrQgtAQgPrtext, *restrPrQgtMQgPrtext, 1, -1);
    compare(concepts::SparseMatrix<Real>(l, true), l);
    compare(concepts::SparseMatrix<Real>(l), l);
  }

}

// ******************************************************************** main **

int main(int, char**) {
  try {
    concepts::Stacktrace::doit() = false;

    test::TestSuite t("Compositions");

    test::CompositionsTest t1;     t.addTest(&t1);
    test::MoreCompositionsTest t2; t.addTest(&t2);
    test::DeepCompositionsTest t3; t.addTest(&t3);

    t.run();
    return t.report();
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
}
