/* post processing, ie. looping over all cells / elements
 * and doing something
 */

#ifndef Postprocess_hh
#define Postprocess_hh

#include "basics/outputOperator.hh"

namespace concepts {

  // forward declarations
  template<class F>
  class Space;

  template<class F>
  class Element;

  class Cell;

  class Mesh;

  // ******************************************************* CellPostprocess **

  /** Abstract class for per cell postprocessing. An instance of a 
      derived class is called on each cell for the per cell postprocessing.
      You don't have to do this by hand, only pass an instance of a derived
      class to the application operator of GlobalPostprocess.

      @see GlobalPostprocess to see how this class can be used.

      @author Philipp Frauenfelder, 2000
  */
  template<class F>
  class CellPostprocess : public OutputOperator {
  public:
    /** Application operator. This application operator has to
	be overloaded to post process a space.
	@param elm Current element
    */
    virtual void operator() (const Element<F>& elm) = 0;

    /** Application operator. This application operator has to
	be overloaded to post process a mesh.
	@param cell Current cell
    */
    virtual void operator() (const Cell& cell) = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // ***************************************************** GlobalPostprocess **

  /** Global Postprocessing. The application operator calls the 
      application operator of the object for the per cell postprocessing
      on each cell.

      Data which is needed for the postprocessing (eg. the solution, 
      an output stream etc.) should be included in the class which does
      the real work, ie. a derived class from CellPostprocess.

      @see CellPostprocess

      @author Philipp Frauenfelder, 2000
  */
  template<class F>
  class GlobalPostprocess : public OutputOperator {
  public:
    /** Constructor.
	@param spc Space over which the application operator should loop.
    */
    GlobalPostprocess(Space<F>& spc);

    /** Constructor.
	@param msh Mesh over which the application operator should loop.
    */
    GlobalPostprocess(Mesh& msh);

    /** Application operator. It calls the application operator of perCell
	on each cell.

	@param perCell The application operator of this class is called on
	each element.
    */
    void operator() (CellPostprocess<F>& perCell) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// The space
    Space<F>* spc_;

    /// The mesh
    Mesh* msh_;
  };

} // namespace concepts

#endif // Postprocess_hh
