/* driver class for the sparseqr factorization
 */

#include "driver.hh"
#include "basics/exceptions.hh"

// debugging
#include "basics/debug.hh"

#define compute_D 0
#define GivensAppl_D 1

namespace sparseqr {

  // **************************************************************** Driver **

  Driver::~Driver() {}

  concepts::Permutation<Real>* Driver::Prow() {
    if (!computed_) compute_();
    if (Prow_.get() == 0)
      Prow_.reset(new concepts::Permutation<Real>(A_.spaceX(), prow_, false));
    return Prow_.get();
  }

  concepts::Permutation<Real>* Driver::Pcol() {
    if (!computed_) compute_();
    if (Pcol_.get() == 0)
      Pcol_.reset(new concepts::Permutation<Real>(A_.spaceY(), pcol_, false));
    return Pcol_.get();
  }

  sparseqr::GivensRotations<Real>* Driver::Q() {
    if (!computed_) compute_();
    if (q_.qbegin == 0) return 0; // no Given's rotations
    else {
      if (Q_.get() == 0)
	Q_.reset(new sparseqr::GivensRotations<Real>(A_.spaceX(), q_, false));
      return Q_.get();
    }
  }

  sparseqr::GivensRotations<Real>* Driver::Qt() {
    if (!computed_) compute_();
    if (q_.qbegin == 0) return 0; // no Given's rotations
    else {
      if (Qt_.get() == 0)
	Qt_.reset(new sparseqr::GivensRotations<Real>(A_.spaceX(), q_, true));
      return Qt_.get();
    }
  }

  concepts::TrivExtendRestrict<Real>* Driver::restriction() {
    if (!computed_) compute_();
    if (freeDofs_.get() == 0)
      freeDofs_.reset(new concepts::DummySpace<Real>(A_.spaceX().dim()-rank_));
    if (restr_.get() == 0)
      restr_.reset(new concepts::TrivExtendRestrict<Real>(*freeDofs_,
							  A_.spaceX(), false));
    return restr_.get();
  }

  concepts::TrivExtendRestrict<Real>* Driver::extension() {
    if (!computed_) compute_();
    if (freeDofs_.get() == 0)
      freeDofs_.reset(new concepts::DummySpace<Real>(A_.spaceX().dim()-rank_));
    if (ext_.get() == 0)
      ext_.reset(new concepts::TrivExtendRestrict<Real>(A_.spaceX(),
							*freeDofs_, true));
    return ext_.get();
  }

  void Driver::operator()(const concepts::Function<Real>& fncY,
			  concepts::Function<Real>& fncX) {
    const concepts::Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const concepts::Vector<Real>*>(&fncY))) {
      concepts::Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<concepts::Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(concepts::MissingFeature("unknown function"));
  }

  void Driver::operator()(const concepts::Vector<Real>& fncY,
			  concepts::Vector<Real>& fncX) {
    throw conceptsException(concepts::MissingFeature("not implemented yet"));
//     if (!computed_) compute_();
//     DEBUGL(GivensAppl_D, "fncY = " << fncY);
//     concepts::Vector<Real> tmp(fncY.space());
//     (*Q())(fncY, tmp);
//     DEBUGL(GivensAppl_D, "tmp = " << tmp);
//     qr_->b = tmp;
//     qr_->BackSolve(rank_, fncX);
//     DEBUGL(GivensAppl_D, "tmp = " << tmp);
//     DEBUGL(GivensAppl_D, "fncX = " << fncX);
  }

  void Driver::compute_() {
    prow_.resize(A_.spaceX().dim());
    pcol_.resize(A_.spaceY().dim());
    const uint n = A_.spaceX().dim();
    concepts::Array<sparseqr::Packed> pmatrix(n);
    concepts::Array<uint> nnz(n, 0);
    DEBUGL(compute_D, "size of nnz = " << nnz.size());

    // find number of non zeros per row of C_ (has to be QR factored)
    // and store it in nnz
    concepts::HashedSparseMatrix<Real>::Value** matrix =
      A_.m()->Matrix();
    const int hashBits = A_.m()->HashBits();
    const int p = 1<<hashBits;
    for (uint r = 0; r < n; ++r) {
      for (int c_mod = 0; c_mod < p; ++c_mod){
	for (concepts::HashedSparseMatrix<Real>::Value* v =
	       matrix[(r<<hashBits) + c_mod];
	     v != NULL; v = v->lnk){
	  if (v->val != 0.0){
	    ++nnz[r];
	  }  
	}
      }
    }
    DEBUGL(compute_D, "nnz = " << nnz);

    // allocate space in PENTRY
    for (uint r = 0; r < n; ++r) {
      pmatrix[r].init(nnz[r]);
    }
    // From now on, nnz is used as a counter: how many entries in pmatrix
    // for the given index in nnz / pmatrix are already given.
    // At the end of the next loops, it has the same content as now,
    // ie. can given as input 'numrow' to the routine sparseqr()
    nnz.zeros();
    for (uint r = 0; r < n; ++r) {
//       DEBUGL(compute_D, "r = " << r << ", pmatrix[r] = " << pmatrix[r]
// 	     << ", nnz[r] = " << nnz[r]);
      for (int c_mod = 0; c_mod < p; ++c_mod){
	for (concepts::HashedSparseMatrix<Real>::Value* v =
	       matrix[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk){
	  if (v->val != 0.0){
// 	    DEBUGL(compute_D, "r = " << r << ", v->idx = " << v->idx
// 		   << ", v->val = " << v->val);
	    uint& i = nnz[r];
	    pmatrix[r].col[i] = v->idx;
	    pmatrix[r].value[i] = v->val;
	    ++i;
	  }  
	}
      }
//       DEBUGL(compute_D, "r = " << r << ", pmatrix[r] = " << pmatrix[r]
// 	     << ", nnz[r] = " << nnz[r]);
    }
    for (uint r = 0; r < n; ++r) {
      pmatrix[r].Sort();
    }
    DEBUGL(compute_D, "pmatrix = " << pmatrix);

    for (uint j = 0; j < prow_.size(); ++j) prow_[j] = j;
    for (uint j = 0; j < pcol_.size(); ++j) pcol_[j] = j;

    qr_.reset(new QR((sparseqr::Packed*)pmatrix, n,
		    A_.spaceY().dim(), 0, &q_, (int*)prow_, (int*)pcol_));

    DEBUGL(compute_D, "nr = " << qr_->nr << ", nc = " << qr_->nc);
    rank_ = qr_->sparseqr();
    DEBUGL(compute_D, "rank = " << rank_);
    DEBUGL(compute_D, "fillin = " << qr_->fillin << ", flops = " << qr_->flops
	   << ", ngivens = " << qr_->ngivens);
    DEBUGL(compute_D, "q = " << q_);
    DEBUGL(compute_D, "prow = " << prow_);
    DEBUGL(compute_D, "pcol = " << pcol_);

    concepts::Array<bool> used(qr_->nr, false);
    for (uint j = 0; j < (uint)rank_; ++j) used[prow_[j]] = true;
    DEBUGL(compute_D, "used = " << used);
    uint k = 0;
    for (uint j = rank_; j < prow_.size(); ++j) {
      while (used[k]) ++k;
      prow_[j] = k; used[k++] = true;
    }
    DEBUGL(compute_D, "prow = " << prow_);

    computed_ = true;
  }

  std::ostream& Driver::info(std::ostream& os) const {
    return os << "Driver";
  }

} // namespace sparseqr
