/*----------------------------------------------------------------------------
Sparse QR solver for linear systems of equations or least squares problems
Copyright (C) 1996, Thomas H. Robey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Inquiries can be directed to trobey@arc.unm.edu or

Thomas H. Robey
925 Madison NE
Albuquerque, NM  87110
USA
----------------------------------------------------------------------------*/

#include "sparseqr.hh"
#include "fill.hh"

namespace sparseqr {

  int Fill::CalcFill(Smatrix *rptr1,Smatrix *rptr2) {
    int fill=0;
    while (rptr1!=NULL && rptr2!=NULL) {
      if (rptr1!=NULL || rptr2!=NULL) {
	if (rptr1->col!=rptr2->col) {
	  fill++;
	  if (rptr1->col<rptr2->col)
	    rptr1=rptr1->right;
	  else
	    rptr2=rptr2->right;
	}
	else {
	  rptr1=rptr1->right;
	  rptr2=rptr2->right;
	}
      }
      if (rptr1==NULL) {
	while (rptr2!=NULL) {
	  fill++;
	  rptr2=rptr2->right;
	}
      }
      if (rptr2==NULL) {
	while (rptr1!=NULL) {
	  fill++;
	  rptr1=rptr1->right;
	}
      }
    }
    return(fill);
  };

  void Fill::IndexSort(QR *qr,int *sortindex,int nlist) {
    int k=0,sup,employee=0;
    if (nlist==1)
      return;
    int level=1;
    while (1<<level<=nlist)
      level++;
    int end=nlist;
    while (end>1) {
      if (level>1)
	level--;
      /* Promotion phase */
      else {
	end--;
	k=sortindex[end];
	sortindex[end]=sortindex[0];
      }
      for (int i=1<<(level-1);(i<1<<level && i<<1<=end);i++) {
	sup=i;
	/* Hiring phase */
	if (end==nlist)
	  k=sortindex[sup-1];
	while (sup<<1<=end) {
	  employee=sup<<1;
	  /* Find best employee */
	  int i1=sortindex[employee-1];
	  int n1=qr->nrow[row1[i1]]+qr->nrow[row2[i1]];
	  if (employee<end) {
	    int i2=sortindex[employee];
	    int n2=qr->nrow[row1[i2]]+qr->nrow[row2[i2]];
	    if (pairfill[i1]>pairfill[i2] || (pairfill[i1]==pairfill[i2] &&
					      (n1>n2 || (n1==n2 && (row1[i1]>row1[i2] || (row1[i1]==row1[i2] &&
											  row2[i1]>row2[i2])))))) {
	      employee++;
	      i1=i2;
	      n1=n2;
	    }
	  }
	  /* Promote employee */
	  int nsup=qr->nrow[row1[k]]+qr->nrow[row2[k]];
	  if (pairfill[k]>pairfill[i1] || (pairfill[k]==pairfill[i1] &&
					   (nsup>n1 || (nsup==n1 && (row1[k]>row1[i1] || (row1[k]==row1[i1] &&
											  row2[k]>row2[i1])))))) {
	    sortindex[sup-1]=sortindex[employee-1];
	    sup=employee;
	  }
	  else {
	    employee=sup;
	    break;
	  }
	}
	/* Place person at their level */
	if (level==1 || sup!=i)
	  sortindex[employee-1]=k;
      }
    }
    sortindex[0]=k;
    return;
  };

  void Fill::Init(QR *qr,int n) {
    nlist1=0;
    for (int i=0;i<n-1;i++) {
      for (int j=i+1;j<n;j++) {
	pairfill[nlist1]=CalcFill(qr->srow[rindex[i]],qr->srow[rindex[j]]);
	if (rindex[i]<rindex[j]) {
	  row1[nlist1]=rindex[i];
	  row2[nlist1]=rindex[j];
	}
	else {
	  row1[nlist1]=rindex[j];
	  row2[nlist1]=rindex[i];
	}
	nlist1++;
      }
    }
  };

  Fill::Fill(Smatrix *cptr,int k) {
    sortindex1=new int[k*(k-1)/2];
    row1=new int[k*(k-1)/2];
    row2=new int[k*(k-1)/2];
    pairfill=new int[k*(k-1)/2];
    rindex=new int[k];
    assert(sortindex1!=NULL && row1!=NULL && row2!=NULL && pairfill!=NULL && 
	   rindex!=NULL);
    for (int i=0;i<k*(k-1)/2;i++)
      sortindex1[i]=i;
    if (k>3) {
      sortindex2=new int[k-2];
      assert(sortindex2!=NULL);
    }
    else
      sortindex2=NULL;
    for (int j=0;cptr!=NULL;cptr=cptr->down,j++)
      rindex[j]=cptr->row;
  };

  Fill::~Fill() {
    delete[] sortindex1;
    delete[] sortindex2;
    delete[] row1;
    delete[] row2;
    delete[] pairfill;
    delete[] rindex;
  };

  void Fill::Merge(QR *qr) {
    int i,j,l,m=0,i1=0,i2=sortindex2[0],n1=0,n2;
    n2=qr->nrow[row1[i2]]+qr->nrow[row2[i2]];
    for (l=nlist1-2;l>=0;l--) {
      i1=sortindex1[l];
      if (i1!=-1) {
	n1=qr->nrow[row1[i1]]+qr->nrow[row2[i2]];
	m++;
	if (m+l==nlist2*(nlist2-1)/2 && (pairfill[i1]>pairfill[i2] || 
					 (pairfill[i1]==pairfill[i2] && (n1>n2 || (n1==n2 &&
										   (row1[i1]>row1[i2] || (row1[i1]==row1[i2] && row2[i1]>row2[i2])))))))
	  break;
      }
    }
    l++;
    m=l;
    i=l;
    if (m<nlist2*(nlist2-1)/2) {
      while (sortindex1[i]==-1)
	i++;
      i1=sortindex1[i];
      n1=qr->nrow[row1[i1]]+qr->nrow[row2[i1]];
    }
    j=0;
    i2=sortindex2[0];
    n2=qr->nrow[row1[i2]]+qr->nrow[row2[i2]];
    nlist1=(nlist2+1)*nlist2/2;
    while (l<nlist1) {
      if (j==nlist2 || (m<nlist2*(nlist2-1)/2 && (pairfill[i1]>pairfill[i2] ||
						  (pairfill[i1]==pairfill[i2] && (n1>n2 || (n1==n2 &&
											    (row1[i1]>row1[i2] || (row1[i1]==row1[i2] && row2[i1]>row2[i2])))))))) {
	if (l!=i)
	  sortindex1[l]=i1;
	l++;
	m++;
	if (m<nlist2*(nlist2-1)/2) {
	  i++;
	  while (sortindex1[i]==-1)
	    i++;
	  i1=sortindex1[i];
	  n1=qr->nrow[row1[i1]]+qr->nrow[row2[i1]];
	}
      }
      else {
	if (m<nlist2*(nlist2-1)/2 && l==i) {
	  i++;
	  while (sortindex1[i]!=-1)
	    i++;
	  while (i>l) {
	    sortindex1[i]=sortindex1[i-1];
	    i--;
	  }
	  i++;
	}
	sortindex1[l]=i2;
	l++;
	j++;
	if (j<nlist2) {
	  i2=sortindex2[j];
	  n2=qr->nrow[row1[i2]]+qr->nrow[row2[i2]];
	}
      }
    }
    return;
  };

  void Fill::Update(int p,int z) {
    int k=0;
    sortindex1[nlist1-1]=-1;
    int n=0;
    for (int i=nlist1-2;i>=0 && n<2*nlist2;i--) {
      int j=sortindex1[i];
      if (row1[j]==z || row2[j]==z) {
	sortindex1[i]=-1;
	n++;
      }
      else {
	if (row1[j]==p || row2[j]==p) {
	  sortindex2[k]=j;
	  k++;
	  sortindex1[i]=-1;
	  n++;
	}
      }
    }
    return;
  };

}
