// alpha strategy

#include "alpha.hh"
#include "timestepping/vectors.hh"
#include "operator/compositions.hh"

// debugging
#include "basics/debug.hh"

#define Alpha_D 0

namespace timestepping {

  // ***************************************************************** Alpha **

  Alpha::Alpha(concepts::SolverFabric<Real>& fabric,
               concepts::Operator<Real>& D2,
               concepts::Operator<Real>& D0,
               timestepping::TimeVector& trhs, 
               const concepts::Vector<Real>& Y0,
               const concepts::Vector<Real>& Z0,
               Real dt, Real alpha) :
    TimeStepStrategy(fabric, new concepts::LiCo<Real>(D2, D0, 1.0,
      0.25*(1.+alpha)*(1.-alpha)*(1.-alpha)*dt*dt), Y0, Z0, dt),
    D0_(D0), trhs_(trhs), Z_(Z0), A_(Y0.space()), alpha_(alpha),
    beta_(0.25*(1.-alpha)*(1.-alpha)), gamma_(0.5-alpha) {
      if(alpha>0 || alpha<-1./3.) {
        conceptsException(concepts::MissingFeature("alpha not in [-1/3, 0]"));
      }
    }
  
  void Alpha::next() {
    t_ += dt_;

    // contribution to rhs from old correctors
    D0_(sol_, rhs_);
    rhs_ *= alpha_;
    trhs_.time(t_ + alpha_*dt_);
    rhs_ += trhs_;

    // predictor in position (sol_)
    concepts::Vector<Real> temp(Z_);
    temp *= dt_;
    sol_ += temp;

    temp = A_;
    temp *= dt_*dt_*(0.5-beta_);
    sol_ += temp;

    // predictor in velocity (Z_)
    temp = A_;
    temp *= dt_*(1-gamma_);
    Z_ += temp;

    // contribution to rhs from predictors
    D0_(sol_, temp);
    temp *= -1.-alpha_;
    rhs_ += temp;
    
    // solve for acceleration
    A_ *= 0;
    (*solver_)(rhs_, A_);

    // corrector in velocity
    temp = A_;
    temp *= dt_*gamma_;
    Z_ += temp;

    // corrector in position
    temp = A_;
    temp *= dt_*dt_*beta_;
    sol_ += temp;
  }

  std::ostream& Alpha::info(std::ostream& os) const {
    return os << "Alpha(alpha = " << alpha_ << ", dt = " << dt_ << ')';
  }

}
