// Runge Kutta 2 strategy

#ifndef rungekutta2_hh
#define rungekutta2_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "strategy.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeVector;

  // ************************************************* Runge Kutta 2nd order **

  /** Timestep strategy for the explicit 2nd order RungeKutta algorithm 
      to solve first order problems in time.
      \f[
      [ D_1 \partial_t + D_0 ] y(x,t) = f(x,t)
      \f]
      The scheme has one free parameter \c alpha and has convergence order 2.
      Please notice that this algorithm is not absolutely stable and may 
      not converge if the stability condition is not fullfilled.
      @see H. R. Schwarz <em>Numerische Mathematik</em>, Teubner, 1997
      @author Manuel Walser, 2002
  */
  class RungeKutta2 : public TimeStepStrategy {
  public:
    /** Constructor
	@param fabric Solver fabric for solving the occuring systems
	@param D1 Space opeartor D<sub>1</sub>
	@param D0 Space opeartor D<sub>0</sub>
	@param trhs Timedependent external driver  f(x,t)
	@param Y0 Initial condition y(x,0)
	@param dt Time step size
	@param alpha Parameter of the Runge Kutta 2 scheme
    */
    RungeKutta2(concepts::SolverFabric<Real>& fabric,
		concepts::Operator<Real>& D1,
		concepts::Operator<Real>& D0,
		timestepping::TimeVector& trhs,
		const concepts::Vector<Real>& Y0,
		Real dt,
		Real alpha = 0.5);
    virtual ~RungeKutta2();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_;
    /// External driver function
    timestepping::TimeVector &trhs_;
    /// Store the latest timestep
    concepts::Vector<Real> Yn1_;    // Yn1 = Y(t_{n-1})
    /// Parameters of the scheme
    Real alpha_;
  };

}

#endif // rungekutta2_hh

