// TVD-Runge-Kutta strategy

#include "tvd_rk2.hh"
#include "operator/compositions.hh"

// debugging
#include "basics/debug.hh"

#define TvdRK2_D 1

namespace timestepping {

  // ****************************************************** TVD Runge Kutta **

  // *************************************************************** TvdRK2 **

  TvdRK2::TvdRK2(concepts::SolverFabric<Real>& fabric,
		 concepts::Operator<Real>& D1,
		 concepts::Operator<Real>& D0,
		 TimeVector& trhs, 
		 const concepts::Vector<Real>& Y0,
		 Real dt):
    TimeStepStrategy(fabric, &D1, Y0, Y0, dt),
    D1_(D1), D0_(D0), trhs_(trhs), Yn1_(Y0) {}

  TvdRK2::~TvdRK2() {}

  void TvdRK2::next() {
    Yn1_ = sol_;
    Yn1_.time(t_);

    TimeVector temp1(Yn1_);

    // build dt*L_h(Yn1_)
    D0_(Yn1_, temp1);
    trhs_.time(t_);
    temp1 -= trhs_;
    temp1 *= -dt_;
    (*solver_)(temp1, rhs_);

    t_ += dt_;
    TimeVector uStar(Yn1_);
    uStar.time(t_);
    uStar += rhs_;
        
    // build dt*L_h(uStar)
    D0_(uStar, temp1);
    trhs_.time(t_);
    temp1 -= trhs_;
    temp1 *= -0.5*dt_;
    (*solver_)(temp1, rhs_);

    // Add up solution
    sol_ += uStar;
    sol_ *= 0.5;
    sol_ += rhs_;
  }

  std::ostream& TvdRK2::info(std::ostream& os) const {
    return os << "TvdRK2(dt = " << dt_ << ')';
  }

  //******************************************************** LimitingTvdRK2 **
  LimitingTvdRK2::LimitingTvdRK2(concepts::SolverFabric<Real>& fabric,
				 concepts::Operator<Real>& D1,
				 concepts::Operator<Real>& D0,
				 concepts::Operator<Real>& limiter,
				 TimeVector& trhs,
				 const concepts::Vector<Real>& Y0,
				 Real dt) :
    TimeStepStrategy(fabric, &D1, Y0, Y0, dt),
    D1_(D1), D0_(D0), limiter_(limiter), trhs_(trhs), Yn1_(Y0) {}

  LimitingTvdRK2::~LimitingTvdRK2() {}

  void LimitingTvdRK2::next() {
    Yn1_ = sol_;
    
    TimeVector temp1(Yn1_);

    limiter_(Yn1_, sol_);

    // build dt*L_h(Yn1_)
    D0_(Yn1_, temp1);
    trhs_.time(t_);
    temp1 -= trhs_;
    temp1 *= -dt_;
    (*solver_)(temp1, rhs_);

    t_ += dt_;
    TimeVector uStar(Yn1_);
    uStar.time(t_);
    uStar += rhs_;

    // build dt*L_h(uStar)
    D0_(uStar, temp1);
    trhs_.time(t_);
    temp1 -= trhs_;
    temp1 *= -0.5*dt_;
    (*solver_)(temp1, rhs_);

    limiter_(uStar, temp1);
    
    // Add up solution
    sol_ += uStar;
    sol_ *= 0.5;
    sol_ += rhs_;
  }

  std::ostream& LimitingTvdRK2::info(std::ostream& os) const {
    return os << "LimitingTvdRK2(dt = " << dt_ << ')';
  }

}
