// TVD-Runge-Kutta strategy

#ifndef tvd_rk2_hh
#define tvd rk2_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "timestepping/strategy.hh"
#include "timestepping/vectors.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  // class TimeVector;

  // ****************************************************** TVD Runge Kutta **
  /** Timestep stategy for an explicit two stage TVD Runge Kutta scheme to
      solve problems in time.
  */
  class TvdRK2 : public TimeStepStrategy {
  public:
    /** Constructor
       @param D1 Space opeartor D<sub>1</sub>
       @param D0 Space opeartor D<sub>0</sub>
       @param trhs Timedependent external driver  f(x,t)
       @param Y0 Initial condition y(x,0)
       @param dt Time step size
     */
    TvdRK2(concepts::SolverFabric<Real>& fabric,
	   concepts::Operator<Real>& D1,
	   concepts::Operator<Real>& D0,
	   TimeVector& trhs,
	   const concepts::Vector<Real>& Y0,
	   Real dt);
    virtual ~TvdRK2();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_;
    /// External driver function
    TimeVector &trhs_;
    /// Store the latest timestep
    TimeVector Yn1_;    // Yn1 = Y(t_{n-1})
  };

  /** Timestep stategy for an explicit two stage TVD Runge Kutta scheme to
      solve problems in time.
  */
  class LimitingTvdRK2 : public TimeStepStrategy {
  public:
    /** Constructor
	@param D1 Space opeartor D<sub>1</sub>
	@param D0 Space opeartor D<sub>0</sub>
	@param trhs Timedependent external driver  f(x,t)
	@param Y0 Initial condition y(x,0)
	@param dt Time step size
    */
    LimitingTvdRK2(concepts::SolverFabric<Real>& fabric,
		   concepts::Operator<Real>& D1,
		   concepts::Operator<Real>& D0,
		   concepts::Operator<Real>& limiter,
		   TimeVector& trhs,
		   const concepts::Vector<Real>& Y0,
		   Real dt);
    virtual ~LimitingTvdRK2();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_, &limiter_;
    /// External driver function
    TimeVector &trhs_;
    /// Store the latest timestep
    TimeVector Yn1_;    // Yn1 = Y(t_{n-1})
  };
}

#endif // tvd_rk2
