/* Array which does enlarge on request, zero out and more.
 * In a sense a very minimalistic implementation of STLs vector.
 * It's mainly intended to be an exception safe way to have an array
 * of something. Especially suitable for vectors and matrices.
 */

#ifndef Array_hh
#define Array_hh

#include <cstring>
#include <memory>
#include <iostream>

#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

// debugging
#define ArrayConstr_D 0
#define ArrayDestr_D 0

namespace concepts {

  // **************************************** Output operator for auto_ptr's **

  /** Output operator for auto_ptr's
   */
  template<class F>
  std::ostream& operator<<(std::ostream& os, std::auto_ptr<F>& a) {
    return os << "std::auto_ptr output";
  }


  // ***************************************************************** Array **

  /** An array of objects. It does enlarge on request, zero out and more.
      In a sense a very minimalistic implementation of STLs vector.
      It's mainly intended to be an exception safe way to have an array
      of something. Especially suitable for vectors and matrices.

      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class Array {
  public:
    /// Constructor
    Array(const uint sz = 0) : data_(sz ? new F[sz] : 0),
			       size_(sz), n_(sz) {
      DEBUGL(ArrayConstr_D, "start: size = " << sz << ", data = " << data_);
    }
    /** Constructor.

	@param sz Size of the array
	@param dft Default value. This is used in the first allocation and
	           not in subsequent allocations.
    */
    Array(const uint sz, const F& dft) : data_(sz ? new F[sz] : 0),
					 size_(sz), n_(sz) {
      F* d = data_; for(uint i = size_; i--;) *d++ = dft;
    }
    /// Copy constructor
    Array(const Array<F>& a) : data_(a.n_ ? new F[a.n_] : 0),
			       size_(a.n_), n_(a.n_) {
      std::memcpy(data_, a.data_, n_*sizeof(F));
    }
    ~Array() {
      DEBUGL(ArrayDestr_D, "start: size = " << size_ << ", n = " << n_
	     << ", data = " << data_);
      delete[] data_;
      DEBUGL(ArrayDestr_D, "done.");
    }

    /** Resizes the array. A reallocation occurs only if the requested space
	is not available, ie. making the array smaller does not result in a
	reallocation. If a reallocation occurs, the data is not preserved.
    */
    void resize(const uint sz);
    /** Resizes the array. Basically, it does the same as resize. But, if
	a reallocation occurs, the data is preserved.
    */
    void resizePreserve(const uint sz);
    /// Fills the memory with zeros
    void zeros() { std::memset(data_, 0, n_*sizeof(F)); }

    /// Returns a pointer to the array
    operator F*() { return data_; }
    /// Returns a pointer to the array
    operator const F*() const { return data_; }

    /// Index operator
    const F& operator[] (const int i) const { 
      conceptsAssert(i < (int)size_, Assertion());
      return data_[i];
    }
    /// Index operator
    F& operator[] (const int i) { 
      conceptsAssert(i < (int)size_, Assertion());
      return data_[i];
    }

    /// Returns the requested size of the array
    uint size() const { return n_; }

    /// Returns the size of the allocated memory
    uint cursize() const { return size_;}

    /// Returns the memory usage in bytes
    int memory() const { return size_*sizeof(F); }
    std::ostream& info(std::ostream& os) const;
  protected:
    /// Data
    F* data_;
    /// Current real size of the array
    uint size_;
    /// Requested size of the array
    uint n_;
  };

  template<class F>
  void Array<F>::resize(const uint sz) {
    if (sz > size_) {
      delete[] data_;
      data_ = new F[size_ = sz];
    }
    n_ = sz;
  }

  template<class F>
  void Array<F>::resizePreserve(const uint sz) {
    if (sz > size_) {
      F* data = new F[sz];
      std::memcpy(data, data_, n_*sizeof(F));
      delete[] data_;
      data_ = data;
      size_ = sz;
    }
    n_ = sz;
  }

  template<class F>
  std::ostream& Array<F>::info(std::ostream& os) const {
    os << "Array<F>(" << n_ << ", [";
    F* d = data_;
    for (uint i = n_; i--;)
      os << *d++ << ((i == 0) ? "" : ", ");
    return os << "])";
  }

  template<class F>
  std::ostream& operator<<(std::ostream& os, const Array<F>& o) {
#ifdef DEBUG
    os << std::flush;
#endif
    return o.info(os);
  }

} // namespace concepts

#endif // Array_hh
