/* The boundary class is used the describe an element on the boundary
 */

#ifndef boundary_hh
#define boundary_hh

#include <memory>
#include <sstream>
#include "basics/outputOperator.hh"
#include "basics/vectorsMatricesForward.hh"

namespace concepts {

  // forward declaration
  class Formula;

  // ************************************************************** Boundary **

  /** Class to describe an element of the boundary.
      A boundary has a type (Neumann, Dirichlet) and
      a function associated.

      Boundary conditions of type Robin are not yet implemented.

      It is possible to use this class in the following way:
      <pre>Boundary(Boundary::NEUMANN, ParsedFormula("(x)"));</pre>
      since the \c ParsedFormula is cloned (ie. it is copied) before
      the temporary object is destroyed.

      @author Philipp Frauenfelder, 2000
  */
  class Boundary : public OutputOperator {
  public:
    /// The different boundary condition types
    enum boundaryTypes { FREE = 0, DIRICHLET, NEUMANN, MAX_TYPE };

    /** Default constructor.
	Initializes the type to FREE.
    */
    Boundary();

    /** Constructor. The type of the boundary condition must be one of
	FREE, DIRICHLET or NEUMANN. In this constructor, no formula can
	be given. It is always evaluated to 0 by default.
    */
    Boundary(const enum boundaryTypes type);

    /** Constructor. The type of the boundary condition must be one of
	FREE, DIRICHLET or NEUMANN.

	@param type Boundary type
	@param frm Formula (can be a temporary object since it is copied)
    */
    Boundary(const enum boundaryTypes type, const Formula& frm);

    /** Constructor. The type of the boundary condition must be one of
 	FREE, DIRICHLET or NEUMANN. The formula can be given as a string which
	will be used to create a \c ParsedFormula.

 	@param type Boundary type
 	@param frm Formula as a string
    */
    Boundary(const enum boundaryTypes type, const char* frm);

    /// Copy constructor
    Boundary(const Boundary& bnd);

    virtual ~Boundary();

    /// Returns the type of the boundary condition
    enum boundaryTypes type() const { return type_; }

    /// Return the type of the boundary condition as output string
    std::string typeStr() const;

    /// Assignment operator
    virtual Boundary& operator=(const Boundary& bnd);

    bool isNull() { return frm_.get()==0; }

//     /// Application operator. Returns the formula.
//     const Formula& operator()() const { return *frm_; }
    /** Application operator. Calculates the value of the boundary
	function at a specific point.
    */
    Real operator()(const Real x, const Real t = 0.0) const;

    /** Application operator. Calculates the value of the boundary
	function at a specific point.
    */
    Real operator()(const Real2d& x, const Real t = 0.0) const;

    /** Application operator. Calculates the value of the boundary
	function at a specific point.
    */
    Real operator()(const Real3d& x, const Real t = 0.0) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Type of the boundary
    enum boundaryTypes type_;
    /// The function of the boundary condition
    std::auto_ptr<const Formula> frm_;
  };

} // namespace concepts

#endif // boundary_hh
