/* iterator over array of coefficients
 */

#ifndef coeffIterator_hh
#define coeffIterator_hh

#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/debug.hh"

// debugging
#define CoeffItIncrOp_D 0

namespace concepts {

  // ********************************************************* CoeffIterator **

  /** Iterator for an array of coefficients.

      The coefficients can be scalars, 2x2 or 3x3 matrices.
      The indices \c i and \c j chose an entry from the matrix and
      the post and prefix increment operators always return this entry from
      the matrix.

      It is of no use, if your coefficients are scalars and arranged in an
      array.

      @param F Return type (type of the coefficients)
      @param Coeff Type of the coefficient data structure.
             Usefull here are the same type as \c F (for scalars) and
	     \c MapReal2d or \c MapReal3d (for matrices)
      @author Philipp Frauenfelder, 2002
  */
  template<typename F, typename Coeff>
  class CoeffIterator {
  public:
    /// Constructor
    CoeffIterator(const Coeff* data, uint i = 0, uint j = 0) :
      data_(data), i_(i), j_(j) {}
    /// Prefix increment operator
    inline const F* operator++();
    /// Postfix increment operator
    inline const F* operator++(int);
    /// Returns the current value
    inline const F* data();
  private:
    /// Data
    const Coeff* data_;
    const uint i_;
    const uint j_;
  };

  template<typename F, typename Coeff>
  const F* CoeffIterator<F, Coeff>::operator++() {
    ++data_;
    DEBUGL(CoeffItIncrOp_D, '(' << i_ << ", " << j << ") is "
	   << data_->operator()(i_, j_) << " at "
	   << &(data_->operator()(i_, j_)));
    return &(data_->operator()(i_, j_));
  }

  template<typename F, typename Coeff>
  const F* CoeffIterator<F, Coeff>::operator++(int) {
    const F* tmp = &(data_->operator()(i_, j_));
    ++data_;
    return tmp;
  }

  template<typename F, typename Coeff>
  const F* CoeffIterator<F, Coeff>::data() {
    return &(data_->operator()(i_, j_));
  }

  /** Iterator for an array of coefficients.

      The coefficients can be scalars, 2x2 or 3x3 matrices.
      The indices \c i and \c j chose an entry from the matrix and
      the post and prefix increment operators always return this entry from
      the matrix.

      It is of no use, if your coefficients are scalars and arranged in an
      array.

      @param F Return type (type of the coefficients)
      @param Coeff Type of the coefficient data structure.
             Usefull here are the same type as \c F (for scalars) and
	     \c MapReal2d or \c MapReal3d (for matrices)
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class CoeffIterator<F, F> {
  public:
    CoeffIterator(const F* data) : data_(data) {}
    const F* operator++() { return ++data_; }
    inline const F* operator++(int) { return data_++; }
  private:
    const F* data_;
  };

} // namespace

#endif // coeffIterator_hh
