/* -*- c++ -*-
 * parsing an input file, holding parameters, producing output
 */

#ifndef inputOutput_hh
#define inputOutput_hh

#include <iostream>
#include <string>
#include <cstring>
#include <map>

#include "basics/exceptions.hh"
#include "basics/outputOperator.hh"

namespace concepts {

  // ****************************************************** MissingParameter **

  /** Indicates that a requested parameter is not present. Therefore,
      no value can be returned (there are no default values).

      @author Philipp Frauenfelder, 2001.
  */
  class MissingParameter : public MissingFeature {
  public:
    /** Constructor.
	@param parameter The name of the missing parameter
    */
    MissingParameter(const std::string& parameter) throw();
  protected:
    virtual std::ostream& info(std::ostream& os) const throw();
  };

  // ******************************************************* InOutParameters **

  struct ltstr {
    bool operator()(const char* s1, const char* s2) const {
      return std::strcmp(s1, s2) < 0;
    }
  };

  /** Holds parameters in hashes. This class has four hashes to hold
      integer, double, bool and string parameters and four hashes to
      hold arrays of integers, doubles, bools and strings. It does
      therefore store (name, value) pairs in the hashes and (name,
      number, value) triples in the arrays.

      The main usage of this class is by inputParser which reads an
      input file and enters the parameters into such an object.

      If a parameter already exists and is added a second time, the
      value is simply overwritten. The same is true for the entries in
      the arrays. If an array is created a second time, nothing happens.

      The output operator writes the contents of the whole class to the
      output stream in the format of the input file. The output is not
      in the order of the input file.

      @see inputParser

      @author Philipp Frauenfelder, 2000
  */
  class InOutParameters : public OutputOperator {
  public:
    /// Constructor
    InOutParameters() {}
    virtual ~InOutParameters();

    /// Adds a double to the hash of doubles
    void addDouble(const char* name, const double value);
    /// Adds a string to the hash of strings
    void addString(const char* name, const char* value);
    /// Adds an int to the hash of ints
    void addInt(const char* name, const int value);
    /// Adds a bool to the hash of bools
    void addBool(const char* name, const int value);

    /// Returns a double from the hash of doubles
    double getDouble(const char* name) const;
    /// Returns a string from the hash of strings
    std::string getString(const char* name) const;
    /// Returns an int from the hash of ints
    int getInt(const char* name) const;
    /// Returns a bool from the hash of bools
    bool getBool(const char* name) const;

    /// Creates an empty array for doubles if necessary
    void addArrayDouble(const char* array);
    /// Creates an empty array for strings if necessary
    void addArrayString(const char* array);
    /// Creates an empty array for ints if necessary
    void addArrayInt(const char* array);
    /// Creates an emtpy array for bools if necessary
    void addArrayBool(const char* array);

    /** Adds a double to an array.

	@param array Name of the array
	@param number Number of the entry
	@param value Value of the entry

	@pre The array must exist
    */
    void addArrayDouble(const char* array, const int number, 
			const double value);
    /** Adds a string to an array.

	@param array Name of the array
	@param number Number of the entry
	@param value Value of the entry

	@pre The array must exist
    */
    void addArrayString(const char* array, const int number, 
			const char* value);
    /** Adds an int to an array.

	@param array Name of the array
	@param number Number of the entry
	@param value Value of the entry

	@pre The array must exist
    */
    void addArrayInt(const char* array, const int number, const int value);
    /** Adds a bool to an array.

	@param array Name of the array
	@param number Number of the entry
	@param value Value of the entry

	@pre The array must exist
    */
    void addArrayBool(const char* array, const int number, const int value);

    /** Returns a double from an array.

	@param array Name of the array
	@param number Number of the entry in the array
    */
    double getArrayDouble(const char* array, const int number) const;
    /** Returns a string from an array.

	@param array Name of the array
	@param number Number of the entry in the array
    */
    std::string getArrayString(const char* array, const int number) const;
    /** Returns a int from an array.

	@param array Name of the array
	@param number Number of the entry in the array
    */
    int getArrayInt(const char* array, const int number) const;
    /** Returns a bool from an array.

	@param array Name of the array
	@param number Number of the entry in the array
    */
    bool getArrayBool(const char* array, const int number) const;

    /** Returns a reference to the requested map.
	@throws MissingParameter
    */
    const std::map<int, double>& getMapDouble(const char* array) const;
    /** Returns a reference to the requested map.
	@throws MissingParameter
    */
    const std::map<int, std::string>& getMapString(const char* array) const;
    /** Returns a reference to the requested map.
	@throws MissingParameter
    */
    const std::map<int, int>& getMapInt(const char* array) const;
    /** Returns a reference to the requested map.
	@throws MissingParameter
    */
    const std::map<int, bool>& getMapBool(const char* array) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Hash of doubles
    std::map<const char*, double, ltstr> double_;
    /// Hash of strings
    std::map<const char*, std::string, ltstr> string_;
    /// Hash of ints
    std::map<const char*, int, ltstr> int_;
    /// Hash of bools
    std::map<const char*, bool, ltstr> bool_;

    /// Hash of arrays of doubles
    std::map<const char*, std::map<int, double>*, ltstr> doubleArrays_;
    /// Hash of arrays of strings
    std::map<const char*, std::map<int, std::string>*, ltstr> stringArrays_;
    /// Hash of arrays of ints
    std::map<const char*, std::map<int, int>*, ltstr> intArrays_;
    /// Hash of arrays of bools
    std::map<const char*, std::map<int, bool>*, ltstr> boolArrays_;
  };

  // *********************************************************** InputParser **

  /** Parses an input file and extracts parameters. The parameters are
      stored in a separate object in hashes and arrays. Another instance
      of the same class can be used to store output information.

      The output operator writes the date and time in a comment and then
      content of the input parameters, then "end" and then the content
      of the output parameters (eg. convergence data). This can be used
      to generate "unfified" result files where the input data is also
      present. Advice: put the name of the executable and the command
      line options in a comment on the first line of the output file.

      The format of the input file is the following. Comments are C/C++
      style. In an input line, only C-style comments are
      allowed, the // form is only allowed in otherwise empty lines.
      An example of an input file: <pre>// Here, it begins
string title "Title of the problem"
double eps   1e-16
bool   debug true
array string bc {
  0 "(0)"
  1 "(x)"
  2 "(y)"
}
int    iter  10
array string bc {
  3 "(x*y)"
}
end
This is ignored</pre>
      "end" marks the end of the input file, everything after it will be
      ignored. The possible types are string, double, int and bool (also
      as arrays -- arrays are not allowed to be mixed). An array may be
      continued (like shown in the example).

      The parser is written using a combination of bison(1) and flex(1)

      @author Philipp Frauenfelder, 2000
  */
  class InputParser : public OutputOperator {
  public:
    /** Default constructor. Enters the string parameters title, author and
	comment into the input field with value "(empty)". These may be
	overriden by the input file.
	@param quiet If set to true, the output of the current time, the
	machine name etc. in the output of this class is suppressed.
    */
    InputParser(bool quiet = false);

    virtual ~InputParser();

    /// Parses the input file
    void parse();
    /// Parses the input file
    void parse(std::string file);

    /// Returns the input data
    InOutParameters& inputParameters() { return input_; }

    /// Returns the output data
    InOutParameters& outputParameters() { return output_; }

    /// The buffer representing the input file (for the parser)
    typedef struct yy_buffer_state *YY_BUFFER_STATE;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// The current input buffer (for the parser)
    YY_BUFFER_STATE current_buffer_;

    /// Input data
    InOutParameters input_;

    /// Output data
    InOutParameters output_;

    /// Suppress output of current time, machine name etc. in the output
    bool quiet_;
  };

} // namespace

#endif // inputOutput_hh
