/* Scanners and Containers
 */

#ifndef scannerConnectors_hh
#define scannerConnectors_hh

#include "basics/typedefs.hh"
#include "basics/outputOperator.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"

// debugging Joiner
#define JoinerConstr_D 0
#define JoinerDestr_D 0

namespace concepts {

  // ****************************************************************** Scan **

  /** An abstract class for scanning a mesh (a set of cells) or a space
      (a set of elements).
      @param T Element of the set to be scanned
  */
  template <class T> 
  class Scan {
  public:
    virtual ~Scan() {};

    /// Returns true as long as the end of the scanned set is not reached yet
    operator int() { return !eos(); }

    /// Returns true if the end of the scanned set is reached
    virtual bool eos() const = 0;
    /// Returns the next element in the scanned set
    virtual T& operator++(int) = 0;
    /// Returns a clone of the scanner
    virtual Scan* clone() const = 0;
  };

  // ************************************************************* Joiner **

  template<class T, unsigned nlnk>
  class Joiner;

  template<class T, unsigned nlnk>
  std::ostream& operator<<(std::ostream& os,
			   const Joiner<T, nlnk>& j);

  /** Joiner class with multiple successors. This can be a node in a tree or
      in a linked list.

      Adding a new node at the beginning of the list is as easy as
      <pre>Joiner<int, 1>* list;
int a = 2;
list = new Joiner<int, 1>(a, list);</pre>
      It is <b>not possible to use auto_ptr</b> and Joiner together in
      a construction like <pre>auto_ptr<Joiner<int, 1> > list;</pre>
      though. The assignment operator = of auto_ptr releases
      (ie. deletes) the old object when a new object is assigned. In
      the statement above, the list on the right hand side would be
      deleted...

      @param T Class to be stored
      @param nlnk Number of successors (ie. links)
  */
  template<class T, unsigned nlnk>
  class Joiner {
    friend std::ostream& operator<< <>(std::ostream& os,
				       const Joiner<T, nlnk>& j);
  public:
    /** Constructor. Fills the container and sets the successors to default
	of NULL. If a successor is given as argument its filled in as the
	last successor.
	@param val Content of the container.
	@param lnk Initialization value for the last successor.
    */
    inline Joiner(const T& val, Joiner* lnk = 0);

    /** Empty destructor. To delete the list/tree do it by
	hand or use destructor (only if T pointer).
    */
    inline ~Joiner() {}

    /** Static function to delete the list/tree. This
	Function is only valid if T is a pointer type
	allocated with new().
	@param j Root of the list/tree. It is overwritten with 0
    */
    static void destructor(Joiner<T, nlnk>*& j);

    /** Index operator for the container.
	@return The ith successor.
    */
    inline Joiner*& operator[](uint i) {
      conceptsAssert(i < nlnk, Assertion());
      return lnk_[i]; }

    /// Returns the content of the container
    inline T& value() { return val_; }

  private:
    /// Array of the successors
    Joiner<T, nlnk>* lnk_[nlnk];

    /// The content of the container
    T val_;

    /// Copy constructor
    Joiner(const Joiner&);
    void operator =(const Joiner&);

    /// Recursive destructor
    static void destructor_(Joiner<T, nlnk>* j);
  };
  
  template <class T, unsigned nlnk>
  Joiner<T, nlnk>::Joiner(const T& val, Joiner<T, nlnk>* lnk) :
    val_(val) {
    DEBUGL(JoinerConstr_D, "start.");
    for(uint i = 0; i < nlnk - 1; ++i)
      lnk_[i] = 0;
    lnk_[nlnk - 1] = lnk;
    DEBUGL(JoinerConstr_D, "done.");
  }

  template <class T, unsigned nlnk>
  void Joiner<T, nlnk>::destructor(Joiner<T, nlnk>*& j) {
    DEBUGL(JoinerDestr_D, "start.");
    if (nlnk > 1) {
      if (j) {destructor_(j);  j = 0;}
    }
    else {
      while (j) {
	Joiner<T, 1>* foo = j;
	j = j->lnk_[0];
	delete foo->val_;
	delete foo;
      }
    }
    DEBUGL(JoinerDestr_D, "done.");    
  }

  template <class T, unsigned nlnk>
  void Joiner<T, nlnk>::destructor_(Joiner<T, nlnk>* j) {
    for(uint i = 0; i < nlnk; i++) {
      if (j->lnk_[i]) destructor_(j->lnk_[i]);
    }
    delete j->val_;
    delete j;
  }

  template<class T, unsigned nlnk>
  std::ostream& operator<<(std::ostream& os, const Joiner<T, nlnk>& j) {
    os << "J(" << j.val_ << " -> [" << std::flush;
    for(uint i = 0; i < nlnk-1; ++i)
      if (j.lnk_[i])
	os << *j.lnk_[i] << ", " << std::flush;
    if (j.lnk_[nlnk-1])
      os << *j.lnk_[nlnk-1] << std::flush;
    return os << "])" << std::flush;
  }

  // ************************************************************* PListScan **

  /** Scanner for a list of pointers.
   */
  template <class T>
  class PListScan : public Scan<T> {
    /// Pointer to the head of the list
    Joiner<T*, 1>* head_;

    /// Pointer to the current element
    Joiner<T*, 1>* current_;

    /// Is true if the end of the list is reached
    bool eos_;
  public:
    /** Constructor.
	@param cnr Reference to the list
    */
    PListScan(Joiner<T*, 1>& cnr) :
      head_(&cnr), current_(&cnr), eos_(head_ ? false : true) {}

    bool eos() const { return eos_; }

    PListScan<T>* clone() const { return new PListScan<T>(*this); }

    T& operator++(int);
  };

  template <class T> 
  T& PListScan<T>::operator++(int) {

    T* tmp = 0;
    if (!eos_) {
      tmp = current_->value();
      current_ = (*current_)[0];
      if ((eos_ = current_ == 0)) current_ = head_;
    }
    return *tmp;
  }

  // ************************************************************** ListScan **

  /** Scanner for a list
   */
  template <class T>
  class ListScan : public Scan<T> {
    /// Pointer to the head of the list
    Joiner<T, 1>* head_;

    /// Pointer to the current element
    Joiner<T, 1>* current_;

    /// Is true if the end of the list is reached
    bool eos_;
  public:
    /** Constructor.
	@param cnr Reference to the list
    */
    ListScan(Joiner<T, 1>& cnr) :
      head_(&cnr), current_(&cnr), eos_(head_ ? false : true) {}

    bool eos() const { return eos_; }

    ListScan<T>* clone() const { return new ListScan<T>(*this); }

    T& operator++(int);
  };

  template <class T> 
  T& ListScan<T>::operator++(int) {

    T& tmp = current_->value();

    if (!eos_) {
      current_ = (*current_)[0];
      if ((eos_ = current_ == 0)) current_ = head_;
    }

    return tmp;
  }

} // namespace concepts

#endif //  scannerConnectors_hh
