#ifndef blockOperator_hh
#define blockOperator_hh

#include <string>
#include <memory>
#include "basics/typedefs.hh"
#include "operator/compositions.hh"
#include "vectorial/vectorial.hh"

namespace vectorial {

  // ********************************************************* BlockOperator **

  /** Operator defined by 4 operators arranged in 2x2 blocks:
      \f[ \left(\begin{array}{lr}A&B\\C&D\end{array}\right) \f].
      BlockOperator is an Operator on a vectorial::Space of vectorial
      dimension 2, consisting of, say, spc1 and spc2. A is an operator
      spc1->spc1, B is spc2->spc1, C is spc1->spc2 and D is spc2->spc2. This
      is useful for example in mixed FEM for the Stokes problem, where spc1
      is the space of velocities and spc2 is the space of pressures. spc1
      is then again a vectorial::Space.
  */
  template<class F>
  class BlockOperator : public concepts::Operator<F> {
  public:
    /** Constructor.
        @param A Upper left sub-operator
        @param B Upper right sub-operator
        @param C Lower left sub-operator
        @param D Lower right sub-operator
        @param spc Entire vectorial space
    */
    BlockOperator(Space<F>& spc,
                  concepts::Operator<F>* A = 0, concepts::Operator<F>* B = 0,
                  concepts::Operator<F>* C = 0, concepts::Operator<F>* D = 0) :
      A_(A), B_(B), C_(C), D_(D), spc_(spc) , tempX_(0), tempY_(0) {
      if (A!=0 && B!=0) {
        tempX_.reset(new concepts::Vector<F>(*(spc.get(0))));
      }
      if (C!=0 && D!=0) {
        tempY_.reset(new concepts::Vector<F>(*(spc.get(1))));
      }
      conceptsAssert(!A || &(A->spaceX()) == spc.get(0),concepts::Assertion());
      conceptsAssert(!A || &(A->spaceY()) == spc.get(0),concepts::Assertion());
      conceptsAssert(!B || &(B->spaceX()) == spc.get(0),concepts::Assertion());
      conceptsAssert(!B || &(B->spaceY()) == spc.get(1),concepts::Assertion());
      conceptsAssert(!C || &(C->spaceX()) == spc.get(1),concepts::Assertion());
      conceptsAssert(!C || &(C->spaceY()) == spc.get(0),concepts::Assertion());
      conceptsAssert(!D || &(D->spaceX()) == spc.get(1),concepts::Assertion());
      conceptsAssert(!D || &(D->spaceY()) == spc.get(1),concepts::Assertion());
    }

    /// Application operator for functions
    virtual void operator()(const concepts::Function<F>& fncY,
                            concepts::Function<F>& fncX);
    /// Application operator for vectors
    void operator()(const concepts::Vector<F>& fncY,
                    concepts::Vector<F>& fncX);
    /// Image space
    virtual const concepts::Space<F>& spaceX() const { return spc_; }
    /// Source space
    virtual const concepts::Space<F>& spaceY() const { return spc_; }

  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Upper left submatrix
    concepts::Operator<F>* A_;
    /// Upper right submatrix
    concepts::Operator<F>* B_;
    /// Lower left submatrix
    concepts::Operator<F>* C_;
    /// Lower right submatrix
    concepts::Operator<F>* D_;
    /// Entire space
    Space<F>& spc_;
    /// temporary upper vector
    std::auto_ptr<concepts::Vector<F> > tempX_;
    /// temporary lower vector
    std::auto_ptr<concepts::Vector<F> > tempY_;
  };

} // namespace vectorial

#endif // blockOperator_hh
