#ifndef bramblePasciak_hh
#define bramblePasciak_hh

#include <string>

#include "basics/typedefs.hh"
#include "operator.hh"
#include "vectorial.hh"

namespace vectorial {

  using concepts::Real;

  // ******************************************************** BramblePasciak **

  /** Bramble-Pasciak preconditioned solver for generalized saddle point
      problems.
      \f[ \left(\begin{array}{lr}A&B\\B^{\top}&-C\end{array}\right)
      \left(\begin{array}{c}U\\P\end{array}\right) = 
      \left(\begin{array}{c}L_u\\L_p\end{array}\right) \f]
      A has to be symmetric and positive definite, C has to be symmetric
      and positive semidefinite, and B has to satisfy
      the discrete inf-sup-condition.

      The main idea is to transform the
      equations and to provide an inner product in which the transformed
      matrix is symmetric and positive definite. The transformed system is
      then solved by conjugate gradients in this inner product. See [1] for
      more details.

      The class is an operator on a
      vectorial::Space consisting of two spaces, say spc1 and spc2. A is
      spc1->spc1, B is spc2->spc1 and C is spc2->spc2. The entire vectorial
      space has to be provided at construction time.

      Constructing an object of this class does not solve the given system.
      Use the application operator to solve the system. If you want to
      specify a starting vector for the cg iterations, set \c fncX
      before calling the application operator to this starting value.
      \c fncX also holds the result after the solve.

      The application operator throws \c NoConvergence if the desired
      residual \c maxeps is not reached within the given number of
      iterations \c maxit.
      @see J. H. Bramble, J. E. Pasciak: A Preconditioning Technique for Indefinite Systems Resulting from Mixed Approximations of Elliptic Problems. Mathematics of Computation, Vol. 50, Nr. 181, <a href="http://links.jstor.org/sici?sici=0025-5718%28198801%2950%3A181%3C1%3AAPTFIS%3E2.0.CO%3B2-K">pp. 1-17</a>, 1988.
      @see J. H. Bramble, J. E. Pasciak: Corrigenda for A Preconditioning Technique for Indefinite Systems Resulting from Mixed Approximations of Elliptic Problems. Mathematics of Computation, Vol. 51, Nr. 183, <a href="http://links.jstor.org/sici?sici=0025-5718%28198807%2951%3A183%3C387%3AC%3E2.0.CO%3B2-W">pp. 387-388</a>, 1988.
  */
  class BramblePasciak : public concepts::Operator<Real> {
  public:
    /** Constructor.
        @param A Upper left submatrix
        @param B Upper right submatrix
        @param Bt Lower left submatrix
        @param C Lower right submatrix
        @param W Preconditioner for A. Must approximate A<sup>-1</sup> and satisfy c0 (A U,U) < (W<sup>-1</sup> U,U) <= c1 (A U,U) for all U with constants 0 < c0 < c1 < 1.
        @param spc Entire vectorial space
        @param maxeps Maximal residual
        @param maxit Maximal number of iterations
        @param relres Relative residual
    */
    BramblePasciak(concepts::Operator<Real>& A, concepts::Operator<Real>& B,
                   concepts::Operator<Real>& Bt, concepts::Operator<Real>& C,
                   concepts::Operator<Real>& W, Space<Real>& spc,
                   Real maxeps, int maxit = 0, uint relres = false) :
      A_(A), B_(B), Bt_(Bt), C_(C), W_(W), spc_(spc), maxeps_(maxeps*maxeps), 
      maxit_(maxit), eps_(1.0), it_(0), relres_(relres) {
      conceptsAssert(spc.vdim() == 2, concepts::Assertion());
      conceptsAssert(&(A.spaceX()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(A.spaceY()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(W.spaceX()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(W.spaceY()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(B.spaceX()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(B.spaceY()) == spc.get(1), concepts::Assertion());
      conceptsAssert(&(Bt.spaceX()) == spc.get(1), concepts::Assertion());
      conceptsAssert(&(Bt.spaceY()) == spc.get(0), concepts::Assertion());
      conceptsAssert(&(C.spaceX()) == spc.get(1), concepts::Assertion());
      conceptsAssert(&(C.spaceY()) == spc.get(1), concepts::Assertion());
    }

    virtual void operator()(const concepts::Function<Real>& fncY,
                            concepts::Function<Real>& fncX);
    void operator()(const concepts::Vector<Real>& fncY,
                    concepts::Vector<Real>& fncX);

    /// Image space
    virtual const concepts::Space<Real>& spaceX() const { return spc_; }
    /// Source space
    virtual const concepts::Space<Real>& spaceY() const { return spc_; }

    /** Returns the number of iterations. Calling this method makes
        only sense after a linear system has been solved.
    */
    uint iterations() const { return it_; }

    /** Returns the residual. Calling this method makes
        only sense after a linear system has been solved.
    */
    Real epsilon() const { return eps_; }
  protected:
    std::ostream& info(std::ostream& os) const;
  private:
    /// Upper left submatrix
    concepts::Operator<Real>& A_;
    /// Upper right submatrix
    concepts::Operator<Real>& B_;
    /// Lower left submatrix
    concepts::Operator<Real>& Bt_;
    /// Lower right submatrix
    concepts::Operator<Real>& C_;
    /// Preconditioner for A
    concepts::Operator<Real>& W_;
    /// Entire space
    Space<Real>& spc_;
    /// Convergence criterion
    Real maxeps_;
    /// Maximal number of iterations until abortion
    uint maxit_;
    /// Current residual
    Real eps_;
    /// Number of iterations
    uint it_;
    /// false: absolute residual, true: relative residual
    bool relres_;
  };

} // namespace vectorial

#endif // bramblePasciak_hh
