/* base class for vectorial things
 */

#ifndef vectorialThings_hh
#define vectorialThings_hh

#include <memory>
#include "basics/cloneable.hh"
#include "toolbox/array.hh"
#include "space/tmatrix.hh"
#include "space/element.hh"
#include "space/space.hh"
#include "operator/bilinearForm.hh"
#include "toolbox/scannerConnectors.hh"
#include "vectorial/graphics.hh"

// debugging
#include "basics/debug.hh"

#define VectorialConstr_D 0
#define VectorialPut_D 0

#define TMatrixOffsetConstr_D 0

#define TMatrixConstr_D 0

#define ElementConstr_D 0

#define BilinearFormConstr_D 0

#define LinearFormConstr_D 0

#define SpaceConstr_D 0

namespace concepts {
  // forward declaration
  template<typename F>
  class ElementGraphics;
}

namespace vectorial {

  using concepts::Real;

  // ************************************************************* Vectorial **

  /** Base class for most classes which do vector valued problems. This
      class is mainly responsible for the memory handling.
      @author Kersten Schmidt, 2002
  */
  template<class F>
  class Vectorial {
  public:
    /** Constructor
	@param vdim Number of components
	@param arrayWidth Storage to be allocated
    */
    Vectorial(uint vdim, uint arrayWidth) : 
      vdim_(vdim), idx_(0), vdata_(arrayWidth, 0) {
      DEBUGL(VectorialConstr_D,"vdim = " << vdim_ << ", arraywidth = "
	     << arrayWidth);
    }
    virtual ~Vectorial();
    /// Add a component
    virtual void put(F& vdata, const int a = 0, const int b = 0);
    /// Get a component
    virtual const F* get(const int a, const int b = 0) const
    { return vdata_[a]; }
    /// Returns number of components
    uint vdim() const { return vdim_; }
  protected:
    /// Number of components
    uint vdim_;
    /// Index of the last added component
    uint idx_;
    /// Storage
    concepts::Array<F*> vdata_;
  private:
    /// Private copy constructor, forbids copying
    Vectorial(const Vectorial<F>& c);
  }; 

  template<class F>
  Vectorial<F>::~Vectorial() {}

  template<class F>
  void Vectorial<F>::put(F& vdata, int, int) {
    vdata_[idx_++] = &vdata;
    DEBUGL(VectorialConstr_D,"idx = " << idx_);
  }

  // ********************************************************* TMatrixOffset **

  template<class F>
  class TMatrix;

  /** T matrix with an offset. Used as a part of \c TMatrix.
      @author Kersten Schmidt, 2002
  */
  template<class F>
  class TMatrixOffset : public concepts::TMatrixBase<F> {
    friend class TMatrix<F>;
  public:
    TMatrixOffset(const concepts::TMatrixBase<F> &T, int offsetRow,
		  int offsetColumn) :
      concepts::TMatrixBase<F>(0),
      offsetRow_(offsetRow), offsetColumn_(offsetColumn), T_(T) {
      DEBUGL(TMatrixOffsetConstr_D, "offsetRow = " << 
	     offsetRow_ << ", offsetColumn = " << offsetColumn_ << ", T = "
	     << T_);
    }
    virtual ~TMatrixOffset();
    virtual void operator ()(const concepts::ElementMatrix<F>& A,
			     concepts::ElementMatrix<F>& B) const
    { return T_(A,B); }

    virtual uint index(unsigned int i) const { return T_.index(i); }

    inline uint m() const { return T_.m(); }

    inline uint n() const { return T_.n(); }

    virtual void usedIdx(concepts::TColumn<bool>& c) const {}
    virtual void extract(const concepts::Vector<F>& solution,
			 concepts::Array<F>& coeff) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    //private:
    uint offsetRow_, offsetColumn_;
    const concepts::TMatrixBase<F> &T_;
  };

  // *************************************************************** TMatrix **

  /** Vector valued T matrix
      @author Kersten Schmidt, 2002
  */
  template<class F>
  class TMatrix : public concepts::TMatrixBase<F>,
		  public Vectorial<vectorial::TMatrixOffset<F> > {
  public:
    TMatrix(uint vdim, uint arrayWidth = 0) :
      concepts::TMatrixBase<F>(0),
      Vectorial<vectorial::TMatrixOffset<F> >(vdim,vdim),
      offsetRow_(0), offsetColumn_(0) {
      DEBUGL(TMatrixConstr_D, "done.");
    }
    virtual ~TMatrix();

    virtual void operator()(const concepts::ElementMatrix<F>& A, 
			    concepts::ElementMatrix<F>& B) const;
    /** insert a T-Matrix T (of type TMatrixOffset) of one vectorial dimension
	to the all v-dimensionial T-Matrix 

	@param dim Number of global degrees of freedom (size of the subspace)
    */
    void put(const concepts::TMatrixBase<F> &T, const int dim);
    
    virtual uint index(unsigned int i) const;

    // not needed in Vectorial problems
    virtual void usedIdx(concepts::TColumn<bool>& c) const {}
    /** Extracts a part of \c solution according to this T matrix
	into \c coeff. The data in \c coeff are scalar: depending on
	the chosen component in Graphics, this routine extracts the
	right part from \c solution.
	@param solution Solution vector
	@param coeff Extracted coefficients in local numbering
    */	
    virtual void extract(const concepts::Vector<F>& solution,
			 concepts::Array<F>& coeff) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    uint offsetRow_, offsetColumn_;
  };

  // *************************************************************** Element **

  // forward declaration
  template<class F>
  class Graphics;

  /** Vector valued element
      @author Kersten Schmidt, 2002
  */
  template<class F>
  class Element : public concepts::Element<F>,
		  public Vectorial<concepts::Element<F> > {
  public:
    Element(uint vdim, uint arrayWidth=0) :
      concepts::Element<F>(), Vectorial<concepts::Element<F> >(vdim, vdim),
      T_(new TMatrix<F>(vdim)) {
      DEBUGL(ElementConstr_D, "done.");
    }
    virtual ~Element();

    /** Insert element \c em into the vector valued element.
	This also creates the T matrix \c T_ of this element.
	@param dim Number of global degrees of freedom (size of the subspace)
    */
    virtual void put(concepts::Element<F>& em, const int dim, const int b = 0);

    virtual vectorial::TMatrix<F>& T() const { return *T_; }
    virtual const Graphics<F>* graphics() const;
    static Graphics<F>* vecGraphics();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// T matrix for the vector valued element
    std::auto_ptr<vectorial::TMatrix<F> > T_;
    static std::auto_ptr<Graphics<F> > graphics_;
  };

  // ********************************************************** BilinearForm **

  /** Vector valued bilinear form.
      @author Kersten Schmidt, 2002
  */
  template<class F>
  class BilinearForm : public concepts::BilinearForm<F>,
		       public Vectorial<concepts::BilinearForm<F> > {
  public:
    /** Constructor.
        @param vdim1 vectorial dimension of first argument of bilinear form
        @param vdim2 vectorial dimension of second argument. Defaults to vdim1
    */
    BilinearForm(const uint vdim1, const uint vdim2 = 0) : 
      Vectorial<concepts::BilinearForm<F> >(vdim1,
                                            vdim1*(vdim2 ? vdim2 : vdim1)),
      vdim2_(vdim2 ? vdim2 : vdim1), deepCopies_(0) {
      DEBUGL(BilinearFormConstr_D, "done.");
    }
    /// Copy constructor. This copy constructor performs a deep copy of \c b
    BilinearForm(const BilinearForm& b);
    virtual ~BilinearForm();
    virtual BilinearForm* clone() const;

    /** insert a Bilinearform bf of one vectorial dimension to the all 
	v-dimensionial Bilinearform

	@param i  row number of v-dimensions
	@param j  column number of v-dimensions
    */
    virtual void put(concepts::BilinearForm<F> &bf,const int i,const int j);
    /** Acts similarly to the normal put() command,
	but stores \c bf in \c deepCopies_ (without cloning \c bf first).
	At cleanup time, \c bf is destroyed by this class.
    */
    void putStore(concepts::BilinearForm<F>* bf, const int i, const int j);

    /** Stores \c data in \c sharedData_ (this includes destruction of
	\c sharedData_ at cleanup).
    */
    void putData(concepts::Cloneable* data) { sharedData_.reset(data); }

    virtual void operator ()(const concepts::Element<F>& elmX,
                             const concepts::Element<F>& elmY,
                             concepts::ElementMatrix<F>& em);
    virtual void operator ()(const concepts::Element<F>& elmV,
                             const concepts::Element<F>& elmU,
                             concepts::ElementMatrix<F>& em,
                             const concepts::ElementPair<F>& elmp);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// vectorial dimension of second argument of bilinear form
    uint vdim2_;
    /// Stores the deep copies of the component bilinear forms, if there are
    /// any
    concepts::Array<std::auto_ptr<concepts::BilinearForm<F> > > deepCopies_;
    /// Stores the data which is shared between the components of
    /// the bilinear forms
    std::auto_ptr<concepts::Cloneable> sharedData_;
  };

} // namespace vectorial

namespace concepts {

  // ****************************************************************** Scan **

  template<class F>
  class Scan<vectorial::Element<F> > :
    public concepts::Scan<concepts::Element<F> > {
  public:  
    virtual vectorial::Element<F>& operator++(int) = 0;
  };

} // namespace concepts

namespace vectorial {

  // ***************************************************************** Space **

  /** Vector valued space
      @author Kersten Schmidt, 2002
  */
  template<class F>
  class Space : public Vectorial<concepts::Space<F> >,
		public concepts::Space<F> {
  public:
    Space(const uint vdim, const uint arrayWidth=0) : 
      Vectorial<concepts::Space<F> >(vdim,vdim), dim_(0), nelm_(0), elm_(0) {
      DEBUGL(SpaceConstr_D, "done.");
    }
    virtual ~Space();

    virtual uint dim() const { return dim_;}
    virtual uint nelm() const { return nelm_;}
    typedef concepts::Scan<vectorial::Element<F> > Scanner;
    virtual Scanner* scan() const;

    /** insert a Space sp of one vectorial dimension to the all 
	v-dimensionial Space
    */
    virtual void put(concepts::Space<F> &spc,const int i=0,const int j=0);

    /// Rebuilds the vectorial space from the scalar spaces
    void rebuild();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    uint dim_, nelm_;
    concepts::Joiner<Element<F>*, 1>* elm_;
    void buildElm_(const concepts::Space<F>& spc, uint idx);
  };

  // *********************************************************** ElementPair **
  
  /** Vector valued element pair
      @author Christoph Winkelmann, 2003
  */
  template<class F>
  class ElementPair 
    : public concepts::ElementPair<F>,
      public Vectorial<concepts::ElementPair<F> > {
  public:
    ElementPair(const Element<F>& elm1, const Element<F>& elm2) 
      : Vectorial<concepts::ElementPair<F> >(elm2.vdim(),
                                             elm1.vdim()*elm2.vdim()), 
        elm1_(elm1), elm2_(elm2) { }
    virtual ~ElementPair() { }
    virtual void put(concepts::ElementPair<F>& ep,
                     const int iV, const int iU);
    /// Returns reference to the first element.
    virtual const concepts::Element<F>& elm1() const;
    /// Returns reference to the second element.
    virtual const concepts::Element<F>& elm2() const;
  private:
    const Element<F>& elm1_;
    const Element<F>& elm2_;
  };

} // namespace vectorial

#endif // vectorialThings_hh
